#include "analysis.h"


















#define GENE_INFO_FNAME "data/ALL/gene_info/hg19_refSeq_genes.txt"

#define SNP_FNAME "data/ALL/snp/hg19_snp137Common.txt"






#define PEAK_WINDOW 200


























void do_analysis( void )
{

  extern char minor_mode;
  
  switch ( minor_mode ) {
  
  case 'r' :
    do_repeat_analysis();
    break;
    
  case 'c' :
    do_cell_type_specific();
    break;
    
  case 's' :
    do_snp_analysis();
    break;
    
  case 'x' :
    do_compare_spec();
    break;
    
  default :
    error( 5000, "Minor mode needed for analysis run.\n" );
    break;
  
  }

}

















/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/














void do_repeat_analysis( void )
{

  FILE *open_text_file_reading( char a[], char b[] );
  FILE *open_text_file_writing( char a[], char b[] );
  
  extern char seq_align_file[ BUFFER ], input_file[ BUFFER ], region_file[ BUFFER ];
  
  strcpy( input_file, seq_align_file );
  
  REPEAT_FH fh = { .in = NULL, .rep_fh = NULL, .info_fh = NULL, .out_fh = NULL,
    .hist_fh = NULL };
  
  fh.in = (FILE *)open_text_file_reading( input_file, "" );
  fh.rep_fh = (FILE *)open_text_file_reading( region_file, "" );
  fh.info_fh = (FILE *)open_text_file_reading( GENE_INFO_FNAME, "" );
  fh.out_fh = (FILE *)open_text_file_writing( input_file, "Repeat.txt" );
  fh.hist_fh = (FILE *)open_text_file_writing( input_file, "RepeatHist.txt" );
  fh.total_fh = (FILE *)open_text_file_writing( input_file, "RepeatTotal.txt" );
  
  do_repeat_analysis_w_fhs( &fh );
  
  close_file( fh.in );
  close_file( fh.rep_fh );
  close_file( fh.info_fh );
  close_file( fh.out_fh );
  close_file( fh.hist_fh );
  close_file( fh.total_fh );

}
















void do_repeat_analysis_w_fhs( REPEAT_FH *fh )
{

  CHROM_LIST *get_chrom_list_from_bed_fh( CHROM_LIST *chrom_header, FILE *fh );
  CHROM_LIST *get_gene_info_list_fh( CHROM_LIST *info_header, FILE *fh );
  TREE *make_and_set_tree( TREE *tree );
  void do_repeat_origin_w_chrom_list( CHROM_LIST *chrom_header, FILE *rep_fh,
    FILE *hist_fh, FILE *total_fh, FILE *out_fh );

  TREE *tree = NULL;

  /*  In ../tree/create_tree.c:  */
  tree = (TREE *)make_and_set_tree( tree );
  
  CHROM_LIST *chrom_header = NULL;
  CHROM_LIST *info_header = NULL;
  CHROM_LIST *rep_header = NULL;
  
  /*  In list.c:  */
  chrom_header = (CHROM_LIST *)get_chrom_list_from_bed_fh( chrom_header, fh->in );
  rep_header = (CHROM_LIST *)get_chrom_list_from_bed_fh( rep_header, fh->rep_fh );
  info_header = (CHROM_LIST *)get_gene_info_list_fh( info_header, fh->info_fh );
  
  CHROM_LIST *chrom = chrom_header;
  
  int total[ BUFFER ], hits[ BUFFER ], n = 0, euth_n = 0;
  
  NODE_LIST *node = tree->spec_header;
  
  while ( node != NULL ) {
    if ( ! strcmp( node->my_node->label, "Eutheria" ))
      euth_n = n;
    ++n;
    node = node->up;
  }
  
  int j;
  
  for ( j = 0; j < BUFFER; j++ ) {
    total[ j ] = hits[ j ] = 0;
  }
  
  while ( chrom != NULL ) {
  
    CHROM_LIST *rep_chrom = rep_header;
    
    while ( rep_chrom != NULL ) {
      if ( ! strcmp( rep_chrom->chr, chrom->chr ))
        break;
      rep_chrom = rep_chrom->next;
    }
    
    printf( "chr = %s\n", chrom->chr );
  
    SITE_LIST *site = chrom->site_header;
    
    while ( site != NULL ) {
    
      j = 0;
      
      node = tree->spec_header;
      
      while ( node != NULL  ) {
        if ( ! strcmp( site->stats->node, node->my_node->label ))
          break;
        ++j;
        node = node->up;
      }
    
      if ( 1 ) {
      
        ++(total[ j ]);
        
        SITE_LIST *rep = rep_chrom->site_header;
        
        while ( rep != NULL ) {
        
          if (( rep->site->align_start <= site->site->align_start ) &&
            ( rep->site->align_end >= site->site->align_start )) 
          {
            ++(hits[ j ]);
            break;
          }
          
          rep = rep->next;
        
        }
          
      }

      site = site->next;
    
    }
    
    chrom = chrom->next;
  
  }
  
  int k = 0;

  fprintf( fh->hist_fh, "%9s\t%15s\t%9s\t%9s\t%s\n", 
    "\"Index\"", "\"Node\"", "\"Sites\"", "\"Total\"", 
    "\"Frac\"" );
    
  node = tree->spec_header;
  
  for ( k = 0; k < n; k++ ) {
    if ( k > euth_n ) {
      hits[ euth_n ] += hits[ k ];
      total[ euth_n ] += total[ k ];
    }
  }
  
  for ( k = 0; k <= euth_n; k++ )  {
    fprintf( fh->hist_fh, "%9d\t%15s\t%9d\t%9d\t%.2f\n",
      k, node->my_node->label, hits[ k ], total[ k ], 
        (float)hits[ k ] * 100.0 / (float)total[ k ] );
    node = node->up;
  }
  
  fflush( fh->hist_fh );
  
  /*  In repeat.c:  */
  do_repeat_origin_w_chrom_list( chrom_header, fh->rep_fh, fh->hist_fh, fh->total_fh, 
    fh->out_fh );
  
  /*  In this file, below:  */
  do_repeat_total_overlap_w_chrom_lists( chrom_header, info_header, rep_header );

}















void do_repeat_total_overlap_w_chrom_lists( CHROM_LIST *chrom_header,
  CHROM_LIST *info_header, CHROM_LIST *rep_header )
{

  long rep_x = 0, total_x = 0;
  long chip_x = 0, total_chip = 0;
  
  get_repeat_coverage_window( &rep_x, &total_x, &chip_x, &total_chip, info_header, 
    rep_header, 50000 );
  
  printf( "%10ld   %10ld   (%.2f%%)\n", rep_x, total_x, 
    (float)rep_x * 100.0 / (float)total_x );
    
  float pr = (float)rep_x / (float)total_x;
  
  PVAL pval = get_repeat_total_pval( chrom_header, rep_header, pr );
  
  printf( "\n\n" );
  printf( "Pval :  %.2e\n\n", (double)pval );

}












void get_repeat_coverage_window( long *rep_x, long *total_x, long *chip_x,
  long *total_chip, CHROM_LIST *info_header, CHROM_LIST *rep_header, int window )
{

  CHROM_LIST *get_chrom_lister_w_str( CHROM_LIST *chrom_header, char chr[] );

  CHROM_LIST *rep_chrom = rep_header;
  
  while ( rep_chrom != NULL ) {
  
    SITE_LIST *rep_site = rep_chrom->site_header;
  
    CHROM_LIST *info_chrom = NULL; 
    /*  In list.c:  */   
    info_chrom = (CHROM_LIST *)get_chrom_lister_w_str( info_header, rep_chrom->chr );
    
    if ( info_chrom == NULL ) {
      rep_chrom = rep_chrom->next;
      continue;
    }
    
    while ( rep_site != NULL ) {
    
      do_repeat_coverage_window_single_site( rep_x, total_x, chip_x, total_chip,
        rep_site, info_chrom, window );
    
      rep_site = rep_site->next;
    
    }
    
    printf( "%s\n", rep_chrom->chr );
    
    rep_chrom = rep_chrom->next;
  
  }

  CHROM_LIST *info_chrom = info_header;
  
  while ( info_chrom != NULL ) {
  
    SITE_LIST *info_site = info_chrom->site_header;
    
    while ( info_site != NULL ) {
    
      *total_x += window;
      
      info_site = info_site->next;
    
    }
    
    info_chrom = info_chrom->next;
  
  }

}










void do_repeat_coverage_window_single_site( long *rep_x, long *total_x,
  long *chip_x, long *total_chip, SITE_LIST *rep_site, CHROM_LIST *info_chrom, 
  int window )
{

  long dist = 10000000000000;
  
  SITE_LIST *info_site = info_chrom->site_header;
  
  while ( info_site != NULL ) {
  
    long my_dist = info_site->site->loc->start - rep_site->site->align_start;
    
    if ( my_dist < 0 )
      my_dist *= -1;
    
    if ( my_dist <= window ) {
    
      long end, start;
      
      start = MAX( rep_site->site->align_start, info_site->site->loc->start-window );
      end = MIN( rep_site->site->align_end, info_site->site->loc->start + window );

      
      if ( info_site->site->loc->strand == '-' ) {
        if ( info_site->site->loc->start >= rep_site->site->align_end )
          start = end;
      }
      
      if ( info_site->site->loc->strand == '+' ) {
        if ( info_site->site->loc->start <= rep_site->site->align_start )
          start = end;
      }
      
      
      long y = end - start + 1;

      *rep_x += y;
      return;

    }
      
    info_site = info_site->next;
      
  }
  
}
















PVAL get_repeat_total_pval( CHROM_LIST *chrom_header, CHROM_LIST *rep_header, 
  float pr )
{

  CHROM_LIST *get_chrom_lister_w_str( CHROM_LIST *chrom_header, char chr[] );
  PVAL cumulative_binom_prob( STAT p, int k, int n );
  
  int total = 0, hits = 0;

  CHROM_LIST *chrom = chrom_header;
  
  while ( chrom != NULL ) {
  
    SITE_LIST *site = chrom->site_header;
  
    CHROM_LIST *rep_chrom = NULL;    
    /*  In list.c:  */   
    rep_chrom = (CHROM_LIST *)get_chrom_lister_w_str( rep_header, chrom->chr );
    
    while ( site != NULL ) {
    
      if ( site_lister_is_recent( site )) {
      
        ++total;
        
        SITE_LIST *rep = rep_chrom->site_header;
        
        while ( rep != NULL ) {
        
          if (( rep->site->align_start <= site->site->align_start ) &&
            ( rep->site->align_end >= site->site->align_start )) 
          {
            ++hits;
            break;
          }
          
          rep = rep->next;
        
        }
      
      }
      
      site = site->next;
    
    }
    
    chrom = chrom->next;
  
  }
  
  printf( "rep-sites:   %10d   %10d   %.2f%%\n", hits, total, 
    (float)(hits * 100.0) / (float)total );
    
  
  
  /*  In ../model/sig.c:  */
  return cumulative_binom_prob( pr, hits, total );

}













int site_lister_is_recent( SITE_LIST *site )
{

  char node[ BUFFER ];
  
  strcpy( node, site->stats->node );
  
  if ( ! strcmp( node, "hg19" ))
    return 1;
   
  if ( ! strcmp( node, "Homininae2" ))
    return 1;
  
  if ( ! strcmp( node, "Homininae" ))
    return 1;
  
   
  if ( ! strcmp( node, "Hominidae" ))
    return 1;
  
  /*
  if ( ! strcmp( node, "Catarrhini" ))
    return 1;
    
  if ( ! strcmp( node, "Simiiformes" ))
    return 1;
  */
  return 0;

}
















/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/















void do_cell_type_specific( void )
{

  extern MOTIF major_motif;

  do_my_cell_type_origin( &major_motif );

}















void do_my_cell_type_origin( MOTIF *mf )
{

  FILE *open_text_file_writing( char a[], char b[] );
  
  char fname[ BUFFER ];
  
  get_cell_type_origin_fname( fname, mf );
  
  FILE *out = NULL, *dist_out = NULL, *score_out = NULL;
  out = (FILE *)open_text_file_writing( fname, "" );
  dist_out = (FILE *)open_text_file_writing( fname, "tss_dist.txt" );
  score_out = (FILE *)open_text_file_writing( fname, "score_dist.txt" );
  
  /*  In cell_type.c:  */
  do_cell_type_origin_w_output_fh( mf, out, dist_out, score_out );
  
  close_file( out );
  close_file( dist_out );
  close_file( score_out );

}












void get_cell_type_origin_fname( char fname[], MOTIF *mf )
{
  sprintf( fname, "data/CTCF/hg19_cellType.%s.txt", mf->cons );
}










/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
















void do_compare_spec( void )
{

  extern char major_spec[ BUFFER ], minor_spec[ BUFFER ];
  
  strcpy( major_spec, "hg19" );
  strcpy( minor_spec, "mm9" );
  
  do_my_compare_spec( major_spec, minor_spec );

}










void do_my_compare_spec( char major_spec[], char minor_spec[] )
{

  FILE *open_text_file_reading( char a[], char b[] );
  FILE *open_binary_file_reading( char a[], char b[] );
  FILE *open_text_file_writing( char a[], char b[] );
  
  extern char seq_align_file[ BUFFER ], input_file[ BUFFER ];
  extern char region_file[ BUFFER ], minor_region_file[ BUFFER ];
  extern int reg_start, reg_end;
  extern MOTIF major_motif;
  
  char fname[ BUFFER ], fname2[ BUFFER ];
  
//  sprintf( fname, "%s.%s.BooBED_SimpleAnc.txt", seq_align_file,
  //  major_motif.cons );
  
  sprintf( fname, "%s.%s.BooBED_All.txt", seq_align_file,
    major_motif.cons );
  
  sprintf( fname2, "%s.%s", seq_align_file, major_motif.cons );
  
  FILE *seq_fh = NULL, *in = NULL, *spec_fh = NULL;
  FILE *reg_fh = NULL, *minor_reg_fh = NULL, *minor_seq_fh = NULL, *out = NULL, *peak_out = NULL;

  seq_fh = (FILE *)open_text_file_reading( seq_align_file, "" );
  in = (FILE *)open_text_file_reading( fname, "" );
  spec_fh = (FILE *)open_binary_file_reading( fname2, "x" );
  reg_fh = (FILE *)open_text_file_reading( region_file, "" );
  minor_reg_fh = (FILE *)open_text_file_reading( minor_region_file, "" );
  minor_seq_fh = (FILE *)open_text_file_reading( minor_region_file, "seq.txt" );
  out = (FILE *)open_text_file_writing( input_file, "mm9_compare.txt" );
  peak_out = (FILE *)open_text_file_writing( input_file, "peak_state.txt" );

  do_compare_spec_w_fhs( major_spec, minor_spec, in, seq_fh, spec_fh, reg_fh, 
    minor_reg_fh, minor_seq_fh, out, peak_out );
  
  close_file( in );
  close_file( seq_fh );
  close_file( spec_fh );
  close_file( reg_fh );
  close_file( minor_reg_fh );
  close_file( minor_seq_fh );
  close_file( out );
  
}













void do_compare_spec_w_fhs( char major_spec[], char minor_spec[], FILE *in,
  FILE *seq_fh, FILE *spec_fh, FILE *reg_fh, FILE *minor_reg_fh, 
  FILE *minor_seq_fh, FILE *out, FILE *peak_out )
{

  MOTIF_LIST *get_motif_list_from_file( MOTIF_LIST *mf_header, MOTIF *mf );
  MOTIF_LIST *get_empty_align_seq_mf_list( MOTIF_LIST *align_header, FILE *fh );
  CHROM_LIST *get_chrom_list_from_bed_fh( CHROM_LIST *chrom_header, FILE *fh );
  TREE *make_and_set_tree( TREE *tree );
  
  extern MOTIF major_motif;
  extern int score_thresh;
  
  MOTIF_LIST *mf_header = NULL, *spec_header = NULL;
  CHROM_LIST *origin_header = NULL;
  TREE *tree = NULL;
  CHROM_LIST *minor_header = NULL;
  
  /*  In ../format/format.c:  */
  mf_header = (MOTIF_LIST *)get_motif_list_from_file( mf_header, &major_motif );
  
  /*  In list.c:  */
  origin_header = (CHROM_LIST *)get_chrom_list_from_bed_fh( origin_header, in );
  
  /*  In ../format/ft.c:  */  
  spec_header = (MOTIF_LIST *)get_empty_align_seq_mf_list( spec_header,
    spec_fh );

  /*  In ../tree/create_tree.c:  */
  tree = (TREE *)make_and_set_tree( tree );
  
  /*  In list.c:  */
  minor_header = (CHROM_LIST *)get_chrom_list_from_bed_fh( minor_header, 
    minor_reg_fh );
    
  minor_header = (CHROM_LIST *)filter_peaks_without_mf( minor_header,
    mf_header, minor_seq_fh );
  
  minor_header = (CHROM_LIST *)filter_low_peaks_chrom_list( minor_header,
    score_thresh );
    
  /*  In this file, below:  */
  do_compare_spec_w_mf_list( origin_header, mf_header, minor_header, spec_header, 
    tree, major_spec, minor_spec, seq_fh, reg_fh, out, peak_out );

}












CHROM_LIST *filter_low_peaks_chrom_list( CHROM_LIST *chrom_header, int score )
{

  CHROM_LIST *chrom = chrom_header;
  
  while ( chrom != NULL ) {
  
    SITE_LIST *header = chrom->site_header;
  
    chrom->site_header = (SITE_LIST *)filter_site_list_low_peaks( header, score );
    
    chrom = chrom->next;
  
  }
  
  return chrom_header;

}











SITE_LIST *filter_site_list_low_peaks( SITE_LIST *site_header, int score_thresh )
{

  SITE_LIST *my_header = NULL;
  SITE_LIST *site = site_header, *prev = NULL;
  
  while ( site != NULL ) {
  
    int score = get_score_peak_entry( site->stats->node );

    if ( score < score_thresh ) {
      
      SITE_LIST *next = site->next;
  
      kill_site_lister( site );
      
      if ( prev != NULL ) 
        prev->next = next;
        
      site = next;
      
    }
    else {
    
      if ( prev == NULL )
        my_header = site;
    
      prev = site;
      site = site->next;
  
    }
    
  }
  
  return my_header;

}









int get_score_peak_entry( char label[] )
{

  int i;
  
  for ( i = 0; i < strlen( label ); i++ ) {
    if ( label[ i ] == '.' )
      return atoi( label + i + 1 );
  }
  
  return 0;

}













CHROM_LIST *filter_peaks_without_mf( CHROM_LIST *chrom_header, 
  MOTIF_LIST *mf_header, FILE *seq_fh )
{

  CHROM_LIST *chrom = chrom_header;
  
  while ( chrom != NULL ) {
  
    chrom->site_header = (SITE_LIST *)filter_seq_list_without_mf( chrom,
      mf_header, seq_fh );
      
    chrom = chrom->next;
  
  }
  
  return chrom_header;

}













SITE_LIST *filter_seq_list_without_mf( CHROM_LIST *chrom, MOTIF_LIST *mf_header,
  FILE *seq_fh )
{

  SITE_LIST *site_header = NULL;
  SITE_LIST *site = chrom->site_header, *prev = NULL;
  
  while ( site != NULL ) {
  
    char seq[ LINE_BUFFER ];
    
    printf( "%s   %d    %d\n", chrom->chr, (int)site->site->align_start,
      (int)site->site->align_end );
    
    get_fasta_seq_check_region( seq, site->site->loc, seq_fh );
    
    if ( motif_in_sequence( seq, mf_header )) {
      if ( site_header == NULL )
        site_header = site;
      prev = site;
      site = site->next;
      continue;
    }
    
    SITE_LIST *next = site->next;
    
    kill_site_lister( site );
    
    if ( prev != NULL ) 
      prev->next = next;
      
    site = next;      
  
  }
  
  return site_header;

}













void get_fasta_seq_check_region( char seq[], SEQ_INFO *info, FILE *seq_fh )
{

  char get_nucl_from_bp( MF bp );

  char line[ LINE_BUFFER ], c;
  
  fgets( line, LINE_BUFFER, seq_fh );
  
  if ( line[ 0 ] != '>' )
    error( 3000, "Expected angle char.\n" );
    
  printf( "%s\n", line );
    
  int i = 0, rand_char = 0;  
    
  while (( c = fgetc( seq_fh )) != EOF ) {
  
    if ( c == '>' ) {
      ungetc( c, seq_fh );
      break;
    }
  
    if ( isspace( c ))
      continue;
      
    seq[ i++ ] = tolower( c );
  
  }
  
  seq[ i ] = '\0';
  
}









int motif_in_sequence( char seq[], MOTIF_LIST *mf_header )
{

  int site_in_motif_list( char site[], MOTIF_LIST *mf_header );
  
  extern int reg_start, reg_end;
  extern MOTIF major_motif;
  
  int length = strlen( seq );
  int start = reg_start + (length / 2);
  int end = start + reg_end - reg_start;
  
  char site[ BUFFER ];
  int len = 0;
  
  int i;
  
  for ( i = start; i < end; i++ ) {
  
    if ( seq[ i ] == 'n' ) {
      len = 0;
      continue;
    }
    
    if ( len < major_motif.length ) {
      site[ len++ ] = tolower( seq[ i ] );
      site[ len ] = '\0';
    }
    else {
      int j;
      for ( j = 0; j < major_motif.length; j++ )
        site[ j ] = site[ j + 1 ];
      site[ len - 1 ] = tolower( seq[ i ] );
      site[ len ] = '\0';
    }
    
    if (( len == major_motif.length ) && ( site_in_motif_list( site, mf_header )))
      return 1;
  
  }
  
  printf( "\t\t\tyo yo :  %d\n", (int)strlen( seq ) );
  
  return 0;

}









void do_compare_spec_w_mf_list( CHROM_LIST *origin_header, MOTIF_LIST *mf_header, 
  CHROM_LIST *minor_header, MOTIF_LIST *spec_header, TREE *tree, char major_spec[], 
  char minor_spec[], FILE *seq_fh, FILE *reg_fh, FILE *out, FILE *peak_out )
{

  COMP_HIST my_hist;
  
  initialize_comp_hist( &my_hist );
  
  char line[ LINE_BUFFER ];
  
  FPOS fpos = 0;
  
  int seq = 0;

  while ( fgets( line, LINE_BUFFER, reg_fh ) != NULL ) {
    fpos = get_compare_spec_single_gene( &my_hist, origin_header, mf_header, 
      minor_header, spec_header, tree, major_spec, minor_spec, fpos, line, seq_fh, peak_out );
    printf( "seq = %d\n", ++seq );
  }
  
  print_comp_hist( &my_hist, out );
  print_comp_hist( &my_hist, stdout );
      
}






void initialize_comp_hist( COMP_HIST *my_hist )
{

  int i, j;
  
  for ( i = 0; i < PEAK_BUFFER; i++ ) {
  
    for ( j = 0; j < SITE_BUFFER; j++ ) {
    
      my_hist->type_mx[ i ][ j ].total = 0;
      my_hist->type_mx[ i ][ j ].present = 0;
      my_hist->type_mx[ i ][ j ].overlap = 0;
      my_hist->type_mx[ i ][ j ].aligned = 0;
    
    }
  
  }

}













FPOS get_compare_spec_single_gene( COMP_HIST *my_hist, CHROM_LIST *origin_header,
  MOTIF_LIST *mf_header, CHROM_LIST *minor_header, MOTIF_LIST *spec_header, 
  TREE *tree, char major_spec[], char minor_spec[], FPOS fpos, char line[], 
  FILE *seq_fh , FILE *peak_out)
{

  int get_gencoord_entry_from_bed_line( GENCOORD_ENTRY *gene_entry, char line[] );
  MAF_LIST *get_alignment_maf_list_single_gene( MAF_LIST *maf_header, 
    GENCOORD_ENTRY *gene_entry, MOTIF_LIST *align_header, MOTIF_LIST *mf_header, 
    FPOS *fpos, FILE *seq_fh );
    
  extern int reg_start, reg_end;

  GENCOORD_ENTRY gene_entry;
  MAF_LIST *maf_header = NULL;
  
  get_gencoord_entry_from_bed_line( &gene_entry, line );
  
  strcpy( gene_entry.spec, major_spec );
  
  maf_header = (MAF_LIST *)get_alignment_maf_list_single_gene( maf_header,
    &gene_entry, spec_header, mf_header, &fpos, seq_fh );
    
  int center = (gene_entry.start + gene_entry.end) / 2;  
    
  gene_entry.start = center + reg_start;
  gene_entry.end = center + reg_end;
  
  SITE_LIST *site = NULL;
  site = (SITE_LIST *)get_gencoord_entry_first_site_origin( origin_header,
    &gene_entry );
    
  if ( site != NULL )
    get_compare_spec_single_gene_maf_list( my_hist, maf_header, origin_header, 
      &gene_entry, mf_header, minor_header, site, spec_header, tree, major_spec, 
      minor_spec, peak_out );
    
  kill_maf_list( maf_header );
  
  return fpos;
  
}









SITE_LIST *get_gencoord_entry_first_site_origin( CHROM_LIST *origin_header,
  GENCOORD_ENTRY *gene_entry )
{

  CHROM_LIST *chrom = origin_header;
  
  while ( chrom != NULL ) {
    if ( ! strcmp( chrom->chr, gene_entry->chr ))
      break;
    chrom = chrom->next;
  }
  
  if ( chrom == NULL )
    return NULL;
    
  SITE_LIST *site = chrom->site_header;
    
  while ( site != NULL ) {
    
    if (( site->site->align_start <= gene_entry->end ) &&
      ( site->site->align_end > gene_entry->start ))
    {
      return site;
    }
    
    if ( site->site->align_start >= gene_entry->end )
      return NULL;
      
    site = site->next;
    
  }
  
  return NULL;
  
}










void get_compare_spec_single_gene_maf_list( COMP_HIST *my_hist, 
  MAF_LIST *maf_header, CHROM_LIST *origin_header, GENCOORD_ENTRY *gene_entry, 
  MOTIF_LIST *mf_header, CHROM_LIST *minor_header, SITE_LIST *first_site,
  MOTIF_LIST *spec_header, TREE *tree, char major_spec[], char minor_spec[] , FILE *peak_out)
{

  int peak = get_peak_type_from_chrom_lists( maf_header, gene_entry, minor_header, 
    major_spec, minor_spec );
  
  // pring peak annoation into file
  if (peak == NOTHING)
    fprintf(peak_out, "%s\t%lld\t%lld\tUnmapped\n",gene_entry->chr,gene_entry->start,gene_entry->end);
  else if (peak == PEAK_SHARED)
    fprintf(peak_out, "%s\t%lld\t%lld\tShared\n",gene_entry->chr,gene_entry->start,gene_entry->end);
  else
    fprintf(peak_out, "%s\t%lld\t%lld\tLineage_specific\n",gene_entry->chr,gene_entry->start,gene_entry->end);

  printf( "\t\t\t\tPeak = %d\n", peak );
  
  int site = get_site_type_from_first_site( first_site, gene_entry, tree );
  
  SITE_TYPE *elem = &(my_hist->type_mx[ peak ][ site ]);
  
  int present = 0;
  update_compare_spec_site_type( elem, maf_header, mf_header, gene_entry,
    major_spec, minor_spec, &present );
    
  if ( peak == NOTHING ) {
    elem = &(my_hist->type_mx[ PEAK_LS ][ site ]);
    update_compare_spec_site_type( elem, maf_header, mf_header, gene_entry,
      major_spec, minor_spec, &present  );
  }
  /* 
  // # output more stringent ground truth region
  if (present > 0)
    if (peak == PEAK_SHARED)
        fprintf(peak_out, "%s\t%lld\t%lld\tShared\n",gene_entry->chr,gene_entry->start,gene_entry->end);
    else
        fprintf(peak_out, "%s\t%lld\t%lld\tLineage_specific\n",gene_entry->chr,gene_entry->start,gene_entry->end);
  // added by Yang
  */

  if ( peak == PEAK_SHARED ) {
  
    printf( "Human:  %10s\t%12d\t%12d\t%10s\t\t\t%12d\t%12d\n", 
      first_site->site->loc->chr, 
      (int)first_site->site->loc->start, 
      (int)first_site->site->loc->end, 
      first_site->stats->node,
      (int)first_site->site->align_start,
      (int)first_site->site->align_end );
  
  }

}












int get_peak_type_from_chrom_lists( MAF_LIST *maf_header, GENCOORD_ENTRY *entry,
  CHROM_LIST *minor_header, char major_spec[], char minor_spec[] )
{

  POS get_col_nucl_offset( char seq[], POS x );
  int char_is_nucl( char c );
  
  MAF_LIST *major_maf = NULL, *minor_maf = NULL;
  major_maf = (MAF_LIST *)get_maf_lister_w_spec( maf_header, major_spec );
  minor_maf = (MAF_LIST *)get_maf_lister_w_spec( maf_header, minor_spec );
  
  if ( minor_maf->my_maf->seq[ 0 ] == '\0' )
    return NOTHING;
  
  printf( "Human :  %s  %d  %d\n", major_maf->my_maf->entry->chr,
    (int)major_maf->my_maf->entry->start,
    (int)major_maf->my_maf->entry->end );
    
  printf( "Mouse :  %s  %d  %d\n", minor_maf->my_maf->entry->chr,
    (int)minor_maf->my_maf->entry->start,
    (int)minor_maf->my_maf->entry->end );
    
  CHROM_LIST *chrom = minor_header;
  
  while ( chrom != NULL ) {
    if ( ! strcmp( chrom->chr, minor_maf->my_maf->entry->chr ))
      break;
    chrom = chrom->next;
  }
  
  if ( chrom == NULL )
    return PEAK_LS;
    
  int hg_center = (entry->start + entry->end) / 2;
  
  /*  In ../main/nucl.c:  */
  int hg_align_center = get_col_nucl_offset( major_maf->my_maf->seq, 
    hg_center - major_maf->my_maf->entry->start );
    
  int mm_center = minor_maf->my_maf->entry->start;
  int strand_inc = 1;
  
  if ( minor_maf->my_maf->entry->strand == '-' ) {
    mm_center = minor_maf->my_maf->entry->end;  
    strand_inc = -1;
  }
    
  int i;
  for ( i = 0; i < hg_align_center; i++ ) 
    mm_center += ( char_is_nucl( minor_maf->my_maf->seq[ i ] )) ? (strand_inc)
      : (0);
    
  SITE_LIST *site = chrom->site_header;
  
  while ( site != NULL ) {
  
    int site_center = (site->site->align_start + site->site->align_end) / 2;
    
    if (( site_center + PEAK_WINDOW > mm_center )
      &&
      ( site_center - PEAK_WINDOW <= mm_center ))
    {
      printf( "\t\t\t\t\t\tShared!!!\n" );
      return PEAK_SHARED;
    }
    
    if ( site_center - PEAK_WINDOW > mm_center )
      return PEAK_LS;
      
    site = site->next;
  
  }
  
  return PEAK_LS;

}













MAF_LIST *get_maf_lister_w_spec( MAF_LIST *maf_header, char spec[] )
{

  MAF_LIST *maf = maf_header;
  
  while ( maf != NULL ) {
  
    if ( ! strcmp( maf->my_maf->entry->spec, spec ))
      return maf;
      
    maf = maf->next;
  
  }
  
  printf( "hello:  %s\n", spec );
  
  return NULL;

}











int get_site_type_from_first_site( SITE_LIST *first_site, 
  GENCOORD_ENTRY *gene_entry, TREE *tree )
{

  SITE_LIST *site = first_site;
  
  while ( site != NULL ) {
  
    if ( site->site->align_start > gene_entry->end )
      return SITE_LS;
      
    if ( ! site_lister_is_lineage_specific( site ))
      return SITE_ANC;
      
    site = site->next;
  
  }
  
  return SITE_LS;

}









int site_lister_is_lineage_specific( SITE_LIST *site )
{

  char *labels[ BUFFER ] = 
    { "hg19", "Homininae2", "Homininae", "Hominidae",
      "Catarrhini", "Simiiformes", "Haplorrhini", "Primates", "Node_9" };
      
  int n = 9;
  
  int i;
  
  for ( i = 0; i < n; i++ ) {
    if ( ! strcmp( labels[ i ], site->stats->node ))
      return 1;
  }
  
  return 0;

}








void update_compare_spec_site_type( SITE_TYPE *elem, MAF_LIST *maf_header, 
  MOTIF_LIST *mf_header, GENCOORD_ENTRY *gene_entry, char major_spec[], 
  char minor_spec[] , int *present)
{

  int major_total, minor_total;
  int major_start[ BUFFER ], major_end[ BUFFER ];
  int minor_start[ BUFFER ], minor_end[ BUFFER ];
  
  int pos_start = 0, pos_end = 0;

  MAF_LIST *major_maf = NULL, *minor_maf = NULL;
  major_maf = (MAF_LIST *)get_maf_lister_w_spec( maf_header, major_spec );
  minor_maf = (MAF_LIST *)get_maf_lister_w_spec( maf_header, minor_spec );
  
  ++(elem->total);
  
  int start_offset = gene_entry->start - major_maf->my_maf->entry->start;
  int end_offset = gene_entry->end - major_maf->my_maf->entry->start;
   
  if (( minor_maf == NULL ) || ( strlen( minor_maf->my_maf->seq ) == 0 ))
    return;
  
  major_total = get_site_array_from_seq( major_start, major_end, 
    major_maf->my_maf->seq, mf_header, &pos_start, &pos_end,
    start_offset, end_offset );
    
  minor_total = get_site_array_from_seq( minor_start, minor_end, 
    minor_maf->my_maf->seq, mf_header, &pos_start, &pos_end,
    start_offset, end_offset );

  if ( minor_total == 0 ) 
    return;
    
  (*present)++;
  ++(elem->present);
  
  int i, j;
  
  for ( i = 0; i < major_total; i++ ) {
    for ( j = 0; j < minor_total; j++ ) {
      if ( major_start[ i ] == minor_start[ j ] ) {
        if ( major_end[ i ] == minor_end[ j ] ) {
          ++(elem->overlap);
          ++(elem->aligned);
          return;
        }
      }
    }
  }
  
  for ( i = 0; i < major_total; i++ ) {
    for ( j = 0; j < minor_total; j++ ) {
      if ( major_start[ i ] <= minor_end[ j ] ) {
        if ( major_end[ i ] >= minor_start[ j ] ) {
          ++(elem->overlap);
          break;
        }
      }
    }
  }
  
  for ( i = 0; i < minor_total; i++ ) {
  
    printf( "Mouse (%d)\t\t%12d\t%12d\n",
      i,
      minor_start[ i ],
      minor_end[ i ] );
  
  }
    
}








int get_site_array_from_seq( int start[], int end[], char seq[],
  MOTIF_LIST *mf_header, int *pos_start, int *pos_end, int start_offset,
  int end_offset )
{

  int site_in_motif_list( char site[], MOTIF_LIST *mf_header );

  int total = 0;
  int total_nucls = 0;
  
  char site[ BUFFER ];
  int pos[ BUFFER ];
  int len = 0;
  
  int my_start = 0, my_end = strlen( seq );
  
  int set_pos = 1;
  
  if ( *pos_start != *pos_end ) {
    my_start = *pos_start;
    my_end = *pos_end;
    set_pos = 0;
  }

  int i, j;
  
  for ( i = my_start; i < my_end; i++ ) {
  
    char c = tolower( seq[ i ] );
    
    switch ( c ) {
    
    case '-' :
      continue;
      
    case '.' :
      len = 0;
      continue;
      
    case 'n' :
    
      if ( set_pos ) {
      
        ++total_nucls;
        
        if ( total_nucls == start_offset )
          *pos_start = i;
      
        if (( total_nucls > start_offset ) && ( total_nucls <= end_offset ))
          *pos_end = i;
        
      }
      
      len = 0;
      
      continue;
      
    default :
      
      if ( len == WORD_LENGTH ) {
        for ( j = 0; j < WORD_LENGTH; j++ ) {
          site[ j ] = site[ j + 1 ];
          pos[ j ] = pos[ j + 1 ];
        }
        site[ WORD_LENGTH - 1 ] = c;
        pos[ WORD_LENGTH - 1 ] = i;
      }
      else {
        pos[ len ] = i;
        site[ len++ ] = c;
        site[ len ] = '\0';
      }
      
      if ( set_pos ) {
      
        ++total_nucls;
        
        if ( total_nucls == start_offset )
          *pos_start = i;
      
        if (( total_nucls > start_offset ) && ( total_nucls <= end_offset ))
          *pos_end = i;
        
      }
      
      break;
      
    }   
    
    if (( len == WORD_LENGTH ) && ( site_in_motif_list( site, mf_header ))) {
      start[ total ] = pos[ 0 ];
      end[ total ] = i;
      ++total;
    }
  
  }
  
  return total;

}













void print_comp_hist( COMP_HIST *my_hist, FILE *out )
{

  fprintf( out, "%12s\t%16s\t%16s\t%16s\t%16s\n",
    "",
    "Shared pk",
    "",
    "LS peak",
    "" );
    
  fprintf( out, "%12s\t%16s\t%16s\t%16s\t%16s\n",
    "",
    "Anc site",
    "LS site",
    "Anc site",
    "LS site" );
    
  int i, j;
  
  int total[ PEAK_BUFFER ] = { 0, 0 };
  
  for ( i = 0; i < PEAK_BUFFER; i++ ) {
    for ( j = 0; j < SITE_BUFFER; j++ ) 
      total[ i ] += my_hist->type_mx[ i ][ j ].total;
  }
  
  fprintf( out, "%12s\t", "Total" );
  
  for ( i = 0; i < PEAK_BUFFER; i++ ) {
    for ( j = 0; j < SITE_BUFFER; j++ )
      fprintf( out, "%7d  %7.2f\t", my_hist->type_mx[ i ][ j ].total,
        (float)(my_hist->type_mx[ i ][ j ].total * 100.0) / (float)total[ i ] );
  }
  
  fprintf( out, "\n" );

  fprintf( out, "%12s\t", "Present" );
  
  for ( i = 0; i < PEAK_BUFFER; i++ ) {
    for ( j = 0; j < SITE_BUFFER; j++ )
      fprintf( out, "%7d  %7.2f\t", my_hist->type_mx[ i ][ j ].present,
        (float)(my_hist->type_mx[ i ][ j ].present * 100.0) / (float)total[ i ] );
  }

  fprintf( out, "\n" );

  fprintf( out, "%12s\t", "Overlap" );
  
  for ( i = 0; i < PEAK_BUFFER; i++ ) {
    for ( j = 0; j < SITE_BUFFER; j++ )
      fprintf( out, "%7d  %7.2f\t", my_hist->type_mx[ i ][ j ].overlap,
        (float)(my_hist->type_mx[ i ][ j ].overlap * 100.0) / (float)total[ i ] );
  }

  fprintf( out, "\n" );

  fprintf( out, "%12s\t", "Aligned" );
  
  for ( i = 0; i < PEAK_BUFFER; i++ ) {
    for ( j = 0; j < SITE_BUFFER; j++ )
      fprintf( out, "%7d  %7.2f\t", my_hist->type_mx[ i ][ j ].aligned,
        (float)(my_hist->type_mx[ i ][ j ].aligned * 100.0) / (float)total[ i ] );
  }

  fprintf( out, "\n" );

}















/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/

















void do_snp_analysis( void )
{

  extern MOTIF major_motif;

  do_my_snp_analysis( &major_motif );

}















void do_my_snp_analysis( MOTIF *mf )
{

  FILE *open_text_file_reading( char a[], char b[] );
  FILE *open_text_file_writing( char a[], char b[] );
  
  extern char input_file[ BUFFER ];
  
  char fname[ BUFFER ];
  
  sprintf( fname, "%s", input_file );
  
  FILE *in = NULL, *snp_fh = NULL, *out = NULL;
  in = (FILE *)open_text_file_reading( fname, "BooBED_All.txt" );
  snp_fh = (FILE *)open_text_file_reading( SNP_FNAME, "" );
  out = (FILE *)open_text_file_writing( fname, "SNP_hist.txt" );
  
  /*  In cell_type.c:  */
  do_snp_analysis_w_fh( mf, in, snp_fh, out );
  
  close_file( in );
  close_file( snp_fh );
  close_file( out );

}













void do_snp_analysis_w_fh( MOTIF *mf, FILE *in, FILE *snp_fh, FILE *out )
{

  CHROM_LIST *get_chrom_list_from_bed_fh( CHROM_LIST *chrom_header, FILE *fh );
  TREE *make_and_set_tree( TREE *tree );
    
  extern int score_thresh;
  
  CHROM_LIST *site_header = NULL, *snp_header = NULL;
  TREE *tree = NULL;
  
  SNP_HIST my_hist;
  
  /*  In list.c:  */
  site_header = (CHROM_LIST *)get_chrom_list_from_bed_fh( site_header, in );
  snp_header = (CHROM_LIST *)get_chrom_list_from_bed_fh( snp_header, snp_fh );
  
  /*  In ../tree/create_tree.c:  */
  tree = (TREE *)make_and_set_tree( tree );
  
  /*  In this file, below:  */
  initialize_snp_hist( &my_hist );
  
  /*  In this file, below:  */
  set_snp_hist_w_chrom_lists( &my_hist, site_header, snp_header, tree );
  
  /*  In this file, below:  */
  print_snp_hist_to_fh( &my_hist, out );
  print_snp_hist_to_fh( &my_hist, stdout );

}













void initialize_snp_hist( SNP_HIST *my_hist )
{

  my_hist->total_ct = 0;
  my_hist->total_sites = 0;
  my_hist->total_freq = 0.0;
  
  my_hist->n = SNP_BUFFER;
  
  int i;
  
  for ( i = 0; i < SNP_BUFFER; i++ ) {
    my_hist->ct[ i ] = 0;
    my_hist->sites[ i ] = 0;
    my_hist->freq[ i ] = 0.0;
  }
  
  strcpy( my_hist->nodes[ HG_INDEX ], "hg19" );
  strcpy( my_hist->nodes[ HOM_INDEX ], "Hominid" );
  strcpy( my_hist->nodes[ SIM_INDEX ], "Simian" );
  strcpy( my_hist->nodes[ ANC_INDEX ], "Ancestral" );

}













void set_snp_hist_w_chrom_lists( SNP_HIST *my_hist, CHROM_LIST *site_header, 
  CHROM_LIST *snp_header, TREE *tree )
{

  CHROM_LIST *get_chrom_lister_w_str( CHROM_LIST *chrom_header, char chr[] );

  CHROM_LIST *chrom_site = site_header;
  
  while ( chrom_site != NULL ) {
  
    CHROM_LIST *chrom_snp = NULL;
    
    /*  In list.c:  */   
    chrom_snp = (CHROM_LIST *)get_chrom_lister_w_str( snp_header, chrom_site->chr );
    
    SITE_LIST *site = chrom_site->site_header;
    SITE_LIST *snp = chrom_snp->site_header;
    
    while ( site != NULL ) {
    
      /*  In this file, below:  */
      set_snp_hist_single_binding_site( my_hist, site, &snp, tree );
    
      site = site->next;
      
    }
    
    chrom_site = chrom_site->next;
  
  }

}










void set_snp_hist_single_binding_site( SNP_HIST *my_hist, SITE_LIST *site,
  SITE_LIST **snp, TREE *tree )
{

  int index = get_branch_snp_index( site->stats->node, tree );
  
  ++(my_hist->total_sites);
  
  if ( index >= 0 )
    ++(my_hist->sites[ index ]);

  while (( *snp != NULL ) && ( (*snp)->site->align_end < site->site->align_start ))
    *snp = (*snp)->next;
    
  if (( *snp == NULL ) || ( (*snp)->site->align_start >= site->site->align_end ))
    return;
    
  ++(my_hist->total_ct);
    
  if ( index >= 0 )
    ++(my_hist->ct[ index ]);  

}












int get_branch_snp_index( char node[], TREE *tree )
{

  NODE_LIST *get_node_from_label( NODE_LIST *node_header, char label[] );
  int node_downstream( NODE_LIST *node, NODE_LIST *leaf );
  
  NODE_LIST *hg = NULL, *hom = NULL, *sim = NULL, *anc = NULL;
  NODE_LIST *x = NULL;
  
  hg = (NODE_LIST *)get_node_from_label( tree->spec_header, "hg19" );
  hom = (NODE_LIST *)get_node_from_label( tree->spec_header, "Hominidae" );
  sim = (NODE_LIST *)get_node_from_label( tree->spec_header, "Simiiformes" );
  anc = (NODE_LIST *)get_node_from_label( tree->spec_header, "Eutheria" );
  
  x = (NODE_LIST *)get_node_from_label( tree->spec_header, node );
  
  if ( node_downstream( hg, x ))
    return HG_INDEX;
    
  if ( node_downstream( hom, x ))
    return HOM_INDEX;
    
  if ( node_downstream( sim, x ))
    return SIM_INDEX;
    
  if ( node_downstream( x, anc ))
    return ANC_INDEX;
  
  return -1;

}
















void print_snp_hist_to_fh( SNP_HIST *my_hist, FILE *out )
{

  fprintf( out, "%9s\t%12s\t%9s\t%9s\t%9s\n",
    "\"INDEX\"",
    "\"NODE\"",
    "\"CT\"",
    "\"SITES\"",
    "\"FREQ\"" );
    
  fprintf( out, "%9d\t%12s\t%9d\t%9d\t%9.3f\n",
    0,
    "All",
    my_hist->total_ct,
    my_hist->total_sites,
    (float)(my_hist->total_ct * 100.0) / (float)my_hist->total_sites );
    
  int i;
  
  for ( i = 0; i < my_hist->n; i++ ) {
    fprintf( out, "%9d\t%12s\t%9d\t%9d\t%9.3f\n",
      i + 1,
      my_hist->nodes[ i ],
      my_hist->ct[ i ],
      my_hist->sites[ i ],
      (float)(my_hist->ct[ i ] * 100.0) / (float)my_hist->sites[ i ] );
  }

}















