#include "../ext/struct.h"


















typedef struct repeat_fh_struct {
  FILE *in;
  FILE *rep_fh;
  FILE *info_fh;
  FILE *out_fh;
  FILE *hist_fh;
  FILE *total_fh;
} REPEAT_FH;





#define PEAK_BUFFER 3
#define SITE_BUFFER 2

#define PEAK_SHARED 0
#define PEAK_LS 1
#define NOTHING 2

#define SITE_ANC 0
#define SITE_LS 1




typedef struct site_type_struct {
  int total;
  int present;
  int overlap;
  int aligned;
} SITE_TYPE;



typedef struct comp_hist_struct {
  SITE_TYPE type_mx[ PEAK_BUFFER ][ SITE_BUFFER ];
} COMP_HIST;






typedef struct snp_hist_struct {
  int ct[ SHORT_BUFFER ];
  int sites[ SHORT_BUFFER ];
  float freq[ SHORT_BUFFER ];
  char nodes[ SHORT_BUFFER ][ SHORT_BUFFER ];
  int total_ct;
  int total_sites;
  float total_freq;
  int n;
} SNP_HIST;





#define SNP_BUFFER 4
#define HG_INDEX 0
#define HOM_INDEX 1
#define SIM_INDEX 2
#define ANC_INDEX 3





void do_analysis( void );



/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/



void do_repeat_analysis( void );

void do_repeat_analysis_w_fhs( REPEAT_FH *fh );

void do_repeat_total_overlap_w_chrom_lists( CHROM_LIST *chrom_header,
  CHROM_LIST *info_header, CHROM_LIST *rep_header );

void get_repeat_coverage_window( long *rep_x, long *total_x, long *chip_x,
  long *total_chip, CHROM_LIST *info_header, CHROM_LIST *rep_header, int window );

void do_repeat_coverage_window_single_site( long *rep_x, long *total_x,
  long *chip_x, long *total_chip, SITE_LIST *rep_site, CHROM_LIST *info_chrom, 
  int window );

PVAL get_repeat_total_pval( CHROM_LIST *chrom_header, CHROM_LIST *rep_header, 
  float pr );

int site_lister_is_recent( SITE_LIST *site );



/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/



void do_cell_type_specific( void );

void do_my_cell_type_origin( MOTIF *mf );

void get_cell_type_origin_fname( char fname[], MOTIF *mf );



/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/



void do_compare_spec( void );

void do_my_compare_spec( char major_spec[], char minor_spec[]);

void do_compare_spec_w_fhs( char major_spec[], char minor_spec[], FILE *in,
  FILE *seq_fh, FILE *spec_fh, FILE *reg_fh, FILE *minor_reg_fh, 
  FILE *minor_seq_fh, FILE *out, FILE *peak_out );

CHROM_LIST *filter_low_peaks_chrom_list( CHROM_LIST *chrom_header, int score );

SITE_LIST *filter_site_list_low_peaks( SITE_LIST *site_header, int score );

int get_score_peak_entry( char label[] );

CHROM_LIST *filter_peaks_without_mf( CHROM_LIST *chrom_header, 
  MOTIF_LIST *mf_header, FILE *seq_fh );

SITE_LIST *filter_seq_list_without_mf( CHROM_LIST *chrom, MOTIF_LIST *mf_header,
  FILE *seq_fh );

void get_fasta_seq_check_region( char seq[], SEQ_INFO *info, FILE *seq_fh );

int motif_in_sequence( char seq[], MOTIF_LIST *mf_header );

void do_compare_spec_w_mf_list( CHROM_LIST *origin_header, MOTIF_LIST *mf_header, 
  CHROM_LIST *minor_header, MOTIF_LIST *spec_header, TREE *tree, char major_spec[], 
  char minor_spec[], FILE *seq_fh, FILE *reg_fh, FILE *out, FILE *peak_out );

void initialize_comp_hist( COMP_HIST *my_hist );

FPOS get_compare_spec_single_gene( COMP_HIST *my_hist, CHROM_LIST *origin_header,
  MOTIF_LIST *mf_header, CHROM_LIST *minor_header, MOTIF_LIST *spec_header, 
  TREE *tree, char major_spec[], char minor_spec[], FPOS fpos, char line[], 
  FILE *in, FILE *peak_out );

SITE_LIST *get_gencoord_entry_first_site_origin( CHROM_LIST *origin_header,
  GENCOORD_ENTRY *gene_entry );

void get_compare_spec_single_gene_maf_list( COMP_HIST *my_hist, 
  MAF_LIST *maf_header, CHROM_LIST *origin_header, GENCOORD_ENTRY *gene_entry, 
  MOTIF_LIST *mf_header, CHROM_LIST *minor_header, SITE_LIST *site,
  MOTIF_LIST *spec_header, TREE *tree, char major_spec[], char minor_spec[], FILE *peak_out );

int get_peak_type_from_chrom_lists( MAF_LIST *maf_header, GENCOORD_ENTRY *entry,
  CHROM_LIST *minor_header, char major_spec[], char minor_spec[] );

MAF_LIST *get_maf_lister_w_spec( MAF_LIST *maf_header, char spec[] );

int get_site_type_from_first_site( SITE_LIST *first_site, 
  GENCOORD_ENTRY *gene_entry, TREE *tree );

int site_lister_is_lineage_specific( SITE_LIST *site );

void update_compare_spec_site_type( SITE_TYPE *elem, MAF_LIST *maf_header, 
  MOTIF_LIST *mf_header, GENCOORD_ENTRY *gene_entry, char major_spec[], 
  char minor_spec[], int *present );

int get_site_array_from_seq( int start[], int end[], char seq[],
  MOTIF_LIST *mf_header, int *pos_start, int *pos_end, int start_offset,
  int end_offset );

void print_comp_hist( COMP_HIST *my_hist, FILE *out );



/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/



void do_snp_analysis( void );

void do_my_snp_analysis( MOTIF *mf );

void do_snp_analysis_w_fh( MOTIF *mf, FILE *in, FILE *snp_fh, FILE *out );

void initialize_snp_hist( SNP_HIST *my_hist );

void set_snp_hist_w_chrom_lists( SNP_HIST *my_hist, CHROM_LIST *site_header, 
  CHROM_LIST *snp_header, TREE *tree );

void set_snp_hist_single_binding_site( SNP_HIST *my_hist, SITE_LIST *site,
  SITE_LIST **snp, TREE *tree );

int get_branch_snp_index( char node[], TREE *tree );

void print_snp_hist_to_fh( SNP_HIST *my_hist, FILE *out );








