#include "cell_type.h"












#define BIG 3000000000

#define CELL_T_BUFFER 4
#define CELL_T_ARRAY { "Broad_GM12878", "Broad_H1hESC", "UW_HAC", "UW_HRE" }







#define GENE_INFO_FNAME "data/ALL/gene_info/hg19_refSeq_genes.txt"
#define DIST_BUFFER 6
#define DIST_ARRAY { 5000, 20000, 50000, 100000, 200000, 1000000000 }
#define LARGE 100000000



#define SCORE_BUFFER 7
#define SCORE_ARRAY { 300, 500, 700, 900, 1100, 1300, 1500 }













void do_cell_type_origin_w_output_fh( MOTIF *mf, FILE *out, FILE *dist_out,
  FILE *score_out )
{

  SEQ_LIST *seq_header[ CELL_T_BUFFER ];

  char *cell_array[ CELL_T_BUFFER ] = CELL_T_ARRAY;
  
  int i;
  
  for ( i = 0; i < CELL_T_BUFFER; i++ )
    seq_header[ i ] = (SEQ_LIST *)do_single_cell_type_origin( cell_array[ i ],
      mf );
      
  do_cell_type_origin_analysis( seq_header, CELL_T_BUFFER, out );
  
//  do_cell_type_origin_tss_dist( seq_header, CELL_T_BUFFER, dist_out );
  
  do_cell_type_origin_score_dist( seq_header, score_out );

}














SEQ_LIST *do_single_cell_type_origin( char *cell_type, MOTIF *mf )
{

  FILE *open_text_file_reading( char a[], char b[] );

  extern char chip_type[ BUFFER ];
  
  char fname[ BUFFER ];
  char chip_holder[ BUFFER ];
  
  strcpy( chip_holder, chip_type );
  strcpy( chip_type, cell_type );
  
  get_origin_bed_fname( fname, chip_type, mf );
  
  FILE *fh = NULL;
  fh = (FILE *)open_text_file_reading( fname, "" );
  
  SEQ_LIST *seq_header = NULL;
  seq_header = (SEQ_LIST *)get_single_cell_type_origin_w_fh( seq_header, 
    cell_type, fh );
    
  strcpy( chip_type, chip_holder );  
    
  close_file( fh );
  
  return seq_header;

}











void get_origin_bed_fname( char fname[], char cell_type[], MOTIF *mf )
{
  sprintf( fname, "data/CTCF/hg19_multiz_%s.CTCF.txt.%s.BooBED_All.txt",
    cell_type, mf->cons );
}









SEQ_LIST *get_single_cell_type_origin_w_fh( SEQ_LIST *seq_header,
  char *cell_type, FILE *fh )
{

  char line[ LINE_BUFFER ];
  
  while ( fgets( line, LINE_BUFFER, fh ) != NULL )
    seq_header = (SEQ_LIST *)get_single_cell_type_origin_single_line( seq_header,
      line );
      
  return seq_header;

}













SEQ_LIST *get_single_cell_type_origin_single_line( SEQ_LIST *seq_header,
  char line[] )
{

  SEQ_LIST *create_seq_lister( SEQ_LIST *x );
  SEQ_LIST *put_seq_lister_into_complex_list( SEQ_LIST *seq_header, SEQ_LIST *seq );
  
  SEQ_LIST *seq = NULL;
  seq = (SEQ_LIST *)create_seq_lister( seq );
  
  /*  In this file, below:  */
  set_seq_lister_bed_file_line( seq, line );
  
  /*  In compare.c:  */
  return put_seq_lister_into_complex_list( seq_header, seq );

}








SEQ_LIST *put_seq_lister_into_complex_list( SEQ_LIST *seq_header, SEQ_LIST *seq )
{

  if ( seq_header == NULL )
    return seq;
    
  SEQ_LIST *curr = seq_header, *prev = NULL;
  
  while (( curr != NULL ) && ( strcmp( curr->my_seq->chr, seq->my_seq->chr ))) {
    prev = curr;
    curr = curr->down;
  }
  
  if ( curr == NULL ) {
    prev->down = seq;
    return seq_header;
  }
  
  while ( curr->next != NULL )
    curr = curr->next;
    
  curr->next = seq;
  
  return seq_header;

}












void set_seq_lister_bed_file_line( SEQ_LIST *seq, char line[] )
{

  int start, end;

  sscanf( line, "%s%d%d%s", seq->my_seq->chr, &start, &end, seq->my_seq->gene );
  
  seq->my_seq->start = start;
  seq->my_seq->end = end;
  
  printf( "%s\t%12d\t%12d\t%s\n", seq->my_seq->chr, (int)seq->my_seq->start,
    (int)seq->my_seq->end, seq->my_seq->gene );
    
  seq->yes = 0;

}














void do_cell_type_origin_analysis( SEQ_LIST *seq_header[], int buffer, FILE *out )
{

  TREE *make_and_set_tree( TREE *tree );
  void initialize_origin_hist( ORIGIN_HIST *my_hist, TREE *tree, char spec[] );

  extern char major_spec[ BUFFER ];
  
  TREE *tree = NULL;
  ORIGIN_HIST my_hist[ CELL_T_BUFFER ];
  
  /*  In ../tree/create_tree.c:  */
  tree = (TREE *)make_and_set_tree( tree );
  
  /*  In ../origin/data.c:  */
  set_tree_ancestor_branches( tree );
  
  int i;
  
  for ( i = 0; i < CELL_T_BUFFER; i++ )
    /*  In origin.c:  */
    initialize_origin_hist( my_hist + i, tree, major_spec );
  
  /*  In this file, below:  */
  get_origin_hist_cell_type_origin( my_hist, seq_header, buffer );
  
  /*  In this file, below:  */
  shorten_origin_hist_cell( my_hist, tree, buffer );
  
  /*  In this file, below:  */
  print_cell_type_origin_hist_array( my_hist, buffer, out );

}













void shorten_origin_hist_cell( ORIGIN_HIST my_hist[], TREE *tree, int buffer )
{

  int i, j;
  
  for ( i = 0; i < buffer; i++ ) {

    int bool = 0;
    int index = -1;

    for ( j = 0; j < my_hist[ i ].n; j++ ) {
      
      if ( bool )
        my_hist[ i ].ct[ index ] += my_hist[ i ].ct[ j ];
      
      if ( ! strcmp( "Eutheria", my_hist[ i ].node[ j ] )) {
        bool = 1;
        index = j;
      }
      
    }
    
  }

}









void get_origin_hist_cell_type_origin( ORIGIN_HIST my_hist[], SEQ_LIST *seq_header[],
  int buffer )
{

  set_site_complex_seq_list_intersection( my_hist, seq_header, buffer );
  
  get_my_origin_hist_cell_type_origin( my_hist, seq_header, buffer );

}












void set_site_complex_seq_list_intersection( ORIGIN_HIST my_hist[],
  SEQ_LIST *seq_header[], int buffer )
{

  SEQ_LIST *seq[ CELL_T_BUFFER ];

  seq[ 0 ] = seq_header[ 0 ];
  
  int i;
  
  while ( seq[ 0 ] != NULL ) {
  
    for ( i = 1; i < buffer; i++ ) {
    
      seq[ i ] = seq_header[ i ];
      
      while (( seq[ i ] != NULL ) && 
        ( strcmp( seq[ i ]->my_seq->chr, seq[ 0 ]->my_seq->chr )))
      {
        seq[ i ] = seq[ i ]->down;
      }
      
    }
    
    set_site_complex_seq_list_single_chr( my_hist, seq, buffer );
    
    seq[ 0 ] = seq[ 0 ]->down;
  
  }

}











void set_site_complex_seq_list_single_chr( ORIGIN_HIST my_hist[],
  SEQ_LIST *seq_head[], int buffer )
{

  SEQ_LIST *seq[ CELL_T_BUFFER ];

  int i, j;
  
  for ( i = 0; i < buffer; i++ )
    seq[ i ] = seq_head[ i ];
    
  while ( seq_list_array_is_null( seq, buffer ) == 0 ) {
  
    int lowest[ CELL_T_BUFFER ];
  
    get_site_lowest_array_complex_seq_list( lowest, seq, buffer );
    
    int total = 0, index;
    for ( i = 0; i < buffer; i++ ) {
      total += lowest[ i ];
      if ( lowest[ i ] )
        index = i;
    }
    
    int x = -1;
    
    for ( i = my_hist[ total - 1 ].n - 1; i >= 0; i-- ) {
      for ( j = 0; j < buffer; j++ ) {
        if (( x == -1 ) && ( lowest[ j ] ) && ( ! strcmp( seq[ j ]->my_seq->gene, 
          my_hist[ total - 1 ].node[ i ] )))
        {
          x = i;
        }
      }
    }
    
    ++(my_hist[ total - 1 ].ct[ x ]);
    ++(my_hist[ total - 1 ].total_ct);
    
    step_seq_list_array_w_int_array( seq, lowest, buffer );
    
  }

}












int seq_list_array_is_null( SEQ_LIST *seq[], int buffer )
{

  int i;
  
  for ( i = 0; i < buffer; i++ ) {
    if ( seq[ i ] != NULL )
      return 0;
  }
  
  return 1;

}








void get_site_lowest_array_complex_seq_list( int lowest[], SEQ_LIST *seq[],
  int buffer )
{

  long int start = BIG;
  
  int i;
  
  for ( i = 0; i < buffer; i++ ) {
    if (( seq[ i ] != NULL ) && ( seq[ i ]->my_seq->start < start ))
      start = seq[ i ]->my_seq->start;
  }
  
  for ( i = 0; i < buffer; i++ ) {
    if (( seq[ i ] != NULL ) && ( seq[ i ]->my_seq->start == start ))
      lowest[ i ] = 1;
    else
      lowest[ i ] = 0;
  }
  
  printf( "start = %d\n", (int)start );

}













void step_seq_list_array_w_int_array( SEQ_LIST *seq[], int lowest[], int buffer )
{

  int i;
  
  for ( i = 0; i < buffer; i++ ) {
    if ( lowest[ i ] )
      seq[ i ] = seq[ i ]->next;
  }

}















void get_my_origin_hist_cell_type_origin( ORIGIN_HIST my_hist[], 
  SEQ_LIST *seq_header[], int buffer )
{

  int i, j;

  /*
  my_hist->total_ct = 0;

  int i, j;
  
  for ( i = 0; i < buffer; i++ ) {
  
    SEQ_LIST *my_seq = seq_header[ i ];
    
    while ( my_seq != NULL ) {
    
      SEQ_LIST *seq = my_seq;
      
      while ( seq != NULL ) {
      
        if ( seq->yes == 0 ) {
        
          for ( j = 0; j < my_hist->n; j++ ) {
            if ( ! strcmp( seq->my_seq->gene, my_hist->node[ j ] )) {
              ++(my_hist->ct[ j ]);
              ++(my_hist->total_ct);
            }
          }
        
        }
      
        seq = seq->next;
      
      }
      
      my_seq = my_seq->down;
  
    }
    
  }
  */
  
  for ( i = 0; i < CELL_T_BUFFER; i++ ) {
    for ( j = 0; j < my_hist[ i ].n; j++ ) {
      
      printf( "%15s :  %8d   %.1f%%\n",
        my_hist[ i ].node[ j ],
        my_hist[ i ].ct[ j ],
        (float)(my_hist[ i ].ct[ j ] * 100.0) / (float)(my_hist[ i ].total_ct) );
      
      if ( ! strcmp( my_hist[ i ].node[ j ], "Eutheria" ))
        break;
    
    }
  }

}












void print_cell_type_origin_hist_array( ORIGIN_HIST my_hist[], int buffer, 
  FILE *out )
{

  fprintf( out, "#  Cell type origin histogram\n" );
  fprintf( out, "#  \n" );
  fprintf( out, "#  CellSh :  Number of cell types with binding site in peak\n" );
  fprintf( out, "#  X :       Node index\n" );
  fprintf( out, "#  Node :    Node (branch) label\n" );
  fprintf( out, "#  Ct :      Number of binding sites with origins at node\n" );
  fprintf( out, "#  Frac :    Fraction of binding sites with origins at node\n" );
  fprintf( out, "#            (out of all binding sites in CellSh group\n" );
  fprintf( out, "#  \n" );
  
  fprintf( out, "%10s\t%10s\t%15s\t%10s\t%10s\n",
    "\"CellSh\"",
    "\"X\"",
    "\"Node\"",
    "\"Ct\"",
    "\"Frac\"" );
    
  int i, j;
  
  for ( i = 0; i < buffer; i++ ) {
  
    for ( j = 0; j < my_hist[ i ].n; j++ ) {
    
      fprintf( out, "%10d\t%10d\t%15s\t%10d\t%10.1f\n",
        i + 1,
        j,
        my_hist[ i ].node[ j ],
        my_hist[ i ].ct[ j ],
        (float)(my_hist[ i ].ct[ j ] * 100.0) / (float)(my_hist[ i ].total_ct) );
        
      if ( ! strcmp( my_hist[ i ].node[ j ], "Eutheria" ))
        break;
    
    }
  
  }

}











/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/












void do_cell_type_origin_tss_dist( SEQ_LIST *seq_header[], int buffer, FILE *out )
{

  FILE *open_text_file_reading( char a[], char b[] );

  char info_fname[ BUFFER ];

  sprintf( info_fname, GENE_INFO_FNAME );
  
  FILE *info_fh = NULL;
  info_fh = (FILE *)open_text_file_reading( info_fname, "" );  
  
  do_cell_type_origin_tss_dist_w_fhs( seq_header, buffer, info_fh, out );
  
  close_file( info_fh );

}











void do_cell_type_origin_tss_dist_w_fhs( SEQ_LIST *seq_header[], int buffer, 
  FILE *info_fh, FILE *out )
{

  CHROM_LIST *get_gene_info_list_fh( CHROM_LIST *info_header, FILE *fh );
  
  CHROM_LIST *gene_header = NULL;
  /*  In origin.c:  */
  gene_header = (CHROM_LIST *)get_gene_info_list_fh( gene_header, info_fh );
  
  long dist_array[ DIST_BUFFER ] = DIST_ARRAY;
  long dist_count[ CELL_T_BUFFER ][ DIST_BUFFER ];
  long dist_total[ CELL_T_BUFFER ];
  
  int i, j;
  
  for ( i = 0; i < CELL_T_BUFFER; i++ ) {
    for ( j = 0; j < DIST_BUFFER; j++ )
      dist_count[ i ][ j ] = 0;
    dist_total[ i ] = 0;
  }
  
  do_cell_type_origin_tss_dist_w_lists( seq_header, gene_header, dist_array,
    dist_total, dist_count );
      
  fprintf( out, "\"CellSh\"\t\"Dist\"\t\"Ct\"\t\"Frac\"\n" );
    
  for ( i = 0; i < CELL_T_BUFFER; i++ ) {
    for ( j = 0; j < DIST_BUFFER; j++ ) 
      fprintf( out, "%d\t%6ld\t%10ld\t%.2f\n", i, dist_array[ j ],
        dist_count[ i ][ j ],
        (float)(dist_count[ i ][ j ] * 100.0) / 
        (float)(dist_total[ i ]) );
  }
      
}






void do_cell_type_origin_tss_dist_w_lists( SEQ_LIST *seq_header[], 
  CHROM_LIST *gene_header, long dist_array[], long dist_total[],
  long dist_count[ CELL_T_BUFFER ][ DIST_BUFFER ] )
{

  CHROM_LIST *get_chrom_lister_w_str( CHROM_LIST *chrom_header, char chr[] );

  SEQ_LIST *seq[ CELL_T_BUFFER ];

  seq[ 0 ] = seq_header[ 0 ];
  
  int i;
  
  while ( seq[ 0 ] != NULL ) {
  
    for ( i = 1; i < CELL_T_BUFFER; i++ ) {
    
      seq[ i ] = seq_header[ i ];
      
      while (( seq[ i ] != NULL ) && 
        ( strcmp( seq[ i ]->my_seq->chr, seq[ 0 ]->my_seq->chr )))
      {
        seq[ i ] = seq[ i ]->down;
      }
      
    }
    
    /*  In origin.c:  */
    CHROM_LIST *chrom = (CHROM_LIST *)get_chrom_lister_w_str( gene_header, 
      seq[ 0 ]->my_seq->chr );
      
    do_cell_type_origin_tss_dist_single_chrom( seq, chrom, dist_array, dist_total,
      dist_count );
    
    seq[ 0 ] = seq[ 0 ]->down;
  
  }
  
}






















void do_cell_type_origin_tss_dist_single_chrom( SEQ_LIST *seq_head[], 
  CHROM_LIST *chrom, long dist_array[], long dist_total[],
  long dist_count[ CELL_T_BUFFER ][ DIST_BUFFER ] )
{

  SEQ_LIST *seq[ CELL_T_BUFFER ];

  int i, j, k;
  
  for ( i = 0; i < CELL_T_BUFFER; i++ )
    seq[ i ] = seq_head[ i ];
    
  while ( seq_list_array_is_null( seq, CELL_T_BUFFER ) == 0 ) {
  
    int lowest[ CELL_T_BUFFER ];
  
    get_site_lowest_array_complex_seq_list( lowest, seq, CELL_T_BUFFER );
    
    int total = 0, index;
    for ( i = 0; i < CELL_T_BUFFER; i++ ) {
      total += lowest[ i ];
      if ( lowest[ i ] )
        index = i;
    }
    
    int x = -1;
    
//    for ( i = my_hist[ total - 1 ].n - 1; i >= 0; i-- ) {
      for ( j = 0; j < CELL_T_BUFFER; j++ ) {
        if (( x == -1 ) && ( lowest[ j ] )) {
        
          long dist = LARGE;
          
          SITE_LIST *info_site = chrom->site_header;
          
          while ( info_site != NULL ) {
          
            long my_dist = info_site->site->loc->start - seq[ j ]->my_seq->start;
            
            if ( my_dist < 0 )
              my_dist *= -1;
              
            if ( my_dist < dist )
              dist = my_dist;
              
            info_site = info_site->next;
          
          }
          
          if ( dist <= dist_array[ 0 ] )
            ++(dist_count[ total - 1 ][ 0 ]);
        
          else {
            for ( k = 1; k < DIST_BUFFER; k++ ) {
              if (( dist <= dist_array[ k ] ) && ( dist > dist_array[ k - 1 ] ))
              ++(dist_count[ total - 1 ][ k ]);
            }
          }
          
          ++(dist_total[ total - 1 ]);
          
        }
      }
//  }  
      
  step_seq_list_array_w_int_array( seq, lowest, CELL_T_BUFFER );
    
  }

}


















/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/








/*
#define SCORE_BUFFER 7
#define SCORE_ARRAY { 300, 500, 700, 900, 1100, 1300, 1500 }
*/




void do_cell_type_origin_score_dist( SEQ_LIST *seq_header[], FILE *out )
{

  long score_array[ SCORE_BUFFER ] = SCORE_ARRAY;
  long score_dist[ CELL_T_BUFFER ][ SCORE_BUFFER ];
  
  int i, j;
  for ( i = 0; i < CELL_T_BUFFER; i++ ) {
    for ( j = 0; j < SCORE_BUFFER; j++ ) 
      score_dist[ i ][ j ] = 0;
  }
  
  for ( i = 0; i < CELL_T_BUFFER; i++ ) {
  
    SEQ_LIST *site_chrom = seq_header[ i ];
  
    while ( site_chrom != NULL ) {
      SEQ_LIST *site = site_chrom;
      while ( site != NULL ) {
        site->my_seq->score = 0;
        site = site->next;
      }
      site_chrom = site_chrom->down;
    }
  }
  
  get_cell_type_origin_score_dist_array( score_array, score_dist, seq_header );
      
  fprintf( out, "\"CellSh\"\t\"Score\"\t\"Ct\"\t\"Frac\"\n" );
    
  for ( i = 0; i < CELL_T_BUFFER; i++ ) {
    for ( j = 0; j < SCORE_BUFFER; j++ ) {
      fprintf( out, "%d\t%5ld\t%10ld\t%.1f\n", i,
        score_array[ j ],
        score_dist[ i ][ j ],
        (float)(score_dist[ i ][ j ] * 100.0) / 
        (float)(score_dist[ i ][ 0 ]) );
      printf( "%5ld\t%10ld\t%.1f\n", score_array[ j ],
        score_dist[ i ][ j ],
        (float)(score_dist[ i ][ j ] * 100.0) / 
        (float)(score_dist[ i ][ 0 ]) );
    }
    printf( "\n\n" );
  }

}










/*
#define CELL_T_BUFFER 4
#define CELL_T_ARRAY { "Broad_GM12878", "Broad_H1hESC", "UW_HAC", "UW_HRE" }
*/


void get_cell_type_origin_score_dist_array( long score_array[], 
  long score_dist[ CELL_T_BUFFER ][ SCORE_BUFFER ], SEQ_LIST *seq_header[] )
{

  char *cell_t[ CELL_T_BUFFER ] = CELL_T_ARRAY;

  int i;
  
  for ( i = 0; i < CELL_T_BUFFER; i++ )
    get_cell_type_origin_score_single_cell_type( seq_header[ i ], cell_t[ i ] );
  
  get_my_cell_type_origin_score_dist_array( score_array, score_dist, seq_header );
  
}
  
  
  
    













void get_cell_type_origin_score_single_cell_type( SEQ_LIST *seq_header,
  char *cell_t )
{

  FILE *open_text_file_reading( char a[], char b[] );
  
  extern char major_spec[ BUFFER ];
  
  char fname[ BUFFER ];
  
  sprintf( fname, "data/CTCF/chip_genCoord_%s.hg19.CTCF.txt", cell_t );
    
  FILE *in = NULL;
  in = (FILE *)open_text_file_reading( fname, "" );
  
  get_cell_type_origin_score_single_cell_type_w_fh( seq_header, in );
  
  close_file( in );

}













void get_cell_type_origin_score_single_cell_type_w_fh( SEQ_LIST *seq_header,
  FILE *in )
{

  SEQ_LIST *site_chrom = seq_header;
  
  char line[ LINE_BUFFER ];
  
  while ( fgets( line, LINE_BUFFER, in ) != NULL )
    get_cell_type_origin_score_single_cell_single_line( seq_header, &site_chrom,
      line );

}











void get_cell_type_origin_score_single_cell_single_line( SEQ_LIST *seq_header,
  SEQ_LIST **site_chrom, char line[] )
{

  char chr[ BUFFER ], entry[ BUFFER ];
  int start, end;
  
  sscanf( line, "%s%d%d%s", chr, &start, &end, entry );

  if (( *site_chrom == NULL ) || ( strcmp( chr, (*site_chrom)->my_seq->chr ))) {
    *site_chrom = seq_header;
    while ( *site_chrom != NULL ) {
      if ( ! strcmp( chr, (*site_chrom)->my_seq->chr ))
        break;
      *site_chrom = (*site_chrom)->down;
    }
  }
  
  if ( *site_chrom == NULL )
    return;
    
  SEQ_LIST *site = *site_chrom;
  
  while ( site != NULL ) {
  
    if ( site->my_seq->start > end )
      return;
      
    if (( site->my_seq->start >= start ) && ( site->my_seq->end <= end )) {
      int i, score = 0;
      for ( i = 0; i < strlen( entry ); i++ ) {
        if ( entry[ i ] == '.' ) {
          score = atoi( entry + i + 1 );
          break;
        }
      }
      
      printf( "sc = %d\n", score );
      
      if ( score >= 300 ) 
        site->my_seq->score = score;
      
    }
    
    site = site->next;
  
  }

}












void get_my_cell_type_origin_score_dist_array( long score_array[], 
  long score_dist[ CELL_T_BUFFER ][ SCORE_BUFFER ], SEQ_LIST *seq_header[] )
{

  CHROM_LIST *get_chrom_lister_w_str( CHROM_LIST *chrom_header, char chr[] );

  SEQ_LIST *seq[ CELL_T_BUFFER ];

  seq[ 0 ] = seq_header[ 0 ];
  
  int i;
  
  while ( seq[ 0 ] != NULL ) {
  
    for ( i = 1; i < CELL_T_BUFFER; i++ ) {
    
      seq[ i ] = seq_header[ i ];
      
      while (( seq[ i ] != NULL ) && 
        ( strcmp( seq[ i ]->my_seq->chr, seq[ 0 ]->my_seq->chr )))
      {
        seq[ i ] = seq[ i ]->down;
      }
      
    }
    
    get_cell_type_score_dist_single_chrom( seq, score_array, score_dist );
    
    seq[ 0 ] = seq[ 0 ]->down;
  
  }

}










void get_cell_type_score_dist_single_chrom( SEQ_LIST *seq_head[], long score_array[],
  long score_dist[ CELL_T_BUFFER ][ SCORE_BUFFER ] )
{
  SEQ_LIST *seq[ CELL_T_BUFFER ];

  int i, j, k;
  
  for ( i = 0; i < CELL_T_BUFFER; i++ )
    seq[ i ] = seq_head[ i ];
    
  while ( seq_list_array_is_null( seq, CELL_T_BUFFER ) == 0 ) {
  
    int lowest[ CELL_T_BUFFER ];
  
    get_site_lowest_array_complex_seq_list( lowest, seq, CELL_T_BUFFER );
    
    int total = 0, index;
    for ( i = 0; i < CELL_T_BUFFER; i++ ) {
      total += lowest[ i ];
      if ( lowest[ i ] )
        index = i;
    }
    
    float score = 0.0;
    
    for ( j = 0; j < CELL_T_BUFFER; j++ ) {
      if ( lowest[ j ] ) {
//        if ( score > seq[ j ]->my_seq->score )
          score += seq[ j ]->my_seq->score;
      }
    }
    
    score /= (float)total;
    
    for ( k = 0; k < SCORE_BUFFER; k++ ) {
      if ( score >= score_array[ k ] )
        ++(score_dist[ total - 1 ][ k ]);
    }
      
    step_seq_list_array_w_int_array( seq, lowest, CELL_T_BUFFER );
    
  }

}















