#include "list.h"
















#define REG_WIDTH 10














CHROM_LIST *get_chrom_list_sites_bed_fh( CHROM_LIST *chrom_header, FILE *in )
{

  char line[ LINE_BUFFER ];
  
  CHROM_LIST *chrom = NULL;
  SITE_LIST *site = NULL;
  
  char old_chr[ BUFFER ] = "", old_node[ BUFFER ] = "";
  int old_start = 0, old_end = 0;
  int same = 0;
  
  while ( fgets( line, LINE_BUFFER, in ) != NULL )
    chrom_header = (CHROM_LIST *)get_chrom_list_sites_single_line( chrom_header,
      line, &chrom, &site, old_chr, &old_start, &old_end, old_node, &same );
      
  if (( same == 0 ) && ( old_node[ 0 ] != '\0' ))
    chrom_header = (CHROM_LIST *)get_chrom_list_sites_single_entry( chrom_header, 
      old_chr, old_start, old_end, old_node, &chrom, &site );
        
  return chrom_header;

}










CHROM_LIST *get_chrom_list_sites_single_line( CHROM_LIST *chrom_header,
  char line[], CHROM_LIST **chrom, SITE_LIST **site, char old_chr[],
  int *old_start, int *old_end, char old_node[ BUFFER ], int *same )
{

  char chr[ BUFFER ], node[ BUFFER ];
  int start, end;
  
  sscanf( line, "%s%d%d%s", chr, &start, &end, node );
  
  if ( line[ 0 ] == '#' )
    return chrom_header;

  if ( site_in_same_region( chr, start, end, old_chr, *old_start, *old_end )) {
  
    strcpy( old_chr, chr );
    strcpy( old_node, node );
    *old_start = start;
    *old_end = end;
    *same = 1;
  
  }
  
  else {
    if (( *same == 0 ) && ( old_node[ 0 ] != '\0' ))
      chrom_header = (CHROM_LIST *)get_chrom_list_sites_single_entry( chrom_header, 
        old_chr, *old_start, *old_end, old_node, chrom, site );
        
    strcpy( old_chr, chr );
    strcpy( old_node, node );
    *old_start = start;
    *old_end = end;
    
    *same = 0;
    
  }
  
  return chrom_header;

}












int site_in_same_region( char chr[], int start, int end, char old_chr[], 
  int old_start, int old_end )
{

  if ( strcmp( chr, old_chr ))
    return 0;
    
  return (( start <= old_end ) && ( end >= old_start )) ? (1) : (0);

}










CHROM_LIST *get_chrom_list_sites_single_entry( CHROM_LIST *chrom_header, char chr[],
  int start, int end, char node[], CHROM_LIST **chrom, SITE_LIST **site )
{

  SITE_LIST *create_site_lister( SITE_LIST *x );

  if (( *chrom == NULL ) || ( strcmp( (*chrom)->chr, chr )))
    return get_chrom_list_new_chrom( chrom_header, chr, start, end, node, chrom,
      site );
      
  SITE_LIST *new_site = NULL;
  new_site = (SITE_LIST *)create_site_lister( new_site );
  
  set_site_lister_from_info( new_site, start, end, node );
  
  (*site)->next = new_site;
  *site = new_site;
  
  return chrom_header;  

}











CHROM_LIST *get_chrom_list_new_chrom( CHROM_LIST *chrom_header, char chr[], 
  int start, int end, char node[], CHROM_LIST **chrom, SITE_LIST **site )
{

  CHROM_LIST *create_chrom_lister( CHROM_LIST *x );
  SITE_LIST *create_site_lister( SITE_LIST *x );
  
  CHROM_LIST *new_chrom = NULL;
  SITE_LIST *new_site = NULL;
  new_chrom = (CHROM_LIST *)create_chrom_lister( new_chrom );
  new_site = (SITE_LIST *)create_site_lister( new_site );
  
  strcpy( new_chrom->chr, chr );
  
  set_site_lister_from_info( new_site, start, end, node );
  
  new_chrom->site_header = new_site;
  
  if ( *chrom == NULL ) {
    chrom_header = new_chrom;
    *chrom = new_chrom;
  }
  else {
    (*chrom)->next = new_chrom;
    *chrom = new_chrom;
  }
  
  *site = new_site;
  
  return chrom_header;

}













void set_site_lister_from_info( SITE_LIST *site, int start, int end, char node[] )
{
  site->site->align_start = start;
  site->site->align_end = end;
  strcpy( site->stats->node, node );
}














/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/













CHROM_LIST *get_chrom_list_from_bed_fh( CHROM_LIST *chrom_header, FILE *fh )
{

  CHROM_LIST *chrom = NULL;
  SITE_LIST *site = NULL;
  
  char line[ LINE_BUFFER ];
  
  while ( fgets( line, LINE_BUFFER, fh ) != NULL )
    chrom_header = (CHROM_LIST *)get_chrom_list_single_line( chrom_header,
      &chrom, &site, line );
      
  return chrom_header;

}













CHROM_LIST *get_chrom_list_single_line( CHROM_LIST *chrom_header, CHROM_LIST **chrom,
  SITE_LIST **site, char line[] )
{

  if ( line[ 0 ] == '#' )
    return chrom_header;

  char chr[ BUFFER ], node[ BUFFER ];
  int start, end;
  
  sscanf( line, "%s%d%d%s", chr, &start, &end, node );
  
  return get_and_put_site_lister_chrom_list( chrom_header, chrom, site, chr,
    start, end, node );
  
}










CHROM_LIST *get_and_put_site_lister_chrom_list( CHROM_LIST *chrom_header,
  CHROM_LIST **chrom, SITE_LIST **site, char chr[], int start, int end, 
  char node[] )
{

  /*  In this file, above:  */
  return get_chrom_list_sites_single_entry( chrom_header, chr, start, end, node, 
    chrom, site );

}













/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/















CHROM_LIST *get_gene_info_list_fh( CHROM_LIST *info_header, FILE *fh )
{

  char line[ LINE_BUFFER ];
  
  while ( fgets( line, LINE_BUFFER, fh ) != NULL )
    info_header = (CHROM_LIST *)get_gene_info_single_line( info_header, line );
    
  return info_header;

}












CHROM_LIST *get_gene_info_single_line( CHROM_LIST *info_header, char line[] )
{

  CHROM_LIST *create_chrom_lister( CHROM_LIST *x );
  SITE_LIST *create_site_lister( SITE_LIST *x );
  
  if ( line[ 0 ] == '#' )
    return info_header;
  
  SITE_LIST *site = NULL;
  site = (SITE_LIST *)create_site_lister( site );

  char chr[ BUFFER ], strand[ BUFFER ];
  int start, end;
  
  sscanf( line, "%*s%s%s%d%d", chr, strand, &start, &end );
  
  set_site_lister_gene_info( site, start, end, strand[ 0 ] );
  
  CHROM_LIST *curr_chr = NULL;
  curr_chr = (CHROM_LIST *)get_chrom_lister_w_str( info_header, chr );
  
  if ( curr_chr == NULL ) {
  
    CHROM_LIST *new_chrom = NULL;
    new_chrom = (CHROM_LIST *)create_chrom_lister( new_chrom );
    
    strcpy( new_chrom->chr, chr );
  
    new_chrom->site_header = site;
    
    return put_chrom_lister_end_of_list( info_header, new_chrom );
  
  }
  
  SITE_LIST *curr = curr_chr->site_header;
  
  while ( curr->next != NULL )
    curr = curr->next;
    
  curr->next = site;
  
  return info_header;

}












void set_site_lister_gene_info( SITE_LIST *site, int start, int end, char strand )
{

  site->site->loc->start = start;
  site->site->loc->end = end;
  site->site->loc->strand = strand;

  if ( strand == '-' ) {
    site->site->loc->start = end;
    site->site->loc->end = start;
  }

}








CHROM_LIST *get_chrom_lister_w_str( CHROM_LIST *chrom_header, char chr[] )
{

  CHROM_LIST *curr = chrom_header;
  
  while ( curr != NULL ) {
    if ( ! strcmp( curr->chr, chr ))
      return curr;
    curr = curr->next;
  }
  
  printf( "here\n" );
  
  return NULL;

}











CHROM_LIST *put_chrom_lister_end_of_list( CHROM_LIST *chrom_header, 
  CHROM_LIST *new_chrom )
{

  if ( chrom_header == NULL )
    return new_chrom;
    
  CHROM_LIST *chrom = chrom_header;
  
  while ( chrom->next != NULL )
    chrom = chrom->next;
    
  chrom->next = new_chrom;
  
  return chrom_header;

}









