#include "repeat.h"















#define REPT_NODE_BUFF 20


#define TOTAL 0
#define SITES 1

#define MIN_FRAC 1e-3









void do_repeat_origin_w_chrom_list( CHROM_LIST *chrom_header, FILE *rep_fh,
  FILE *hist_fh, FILE *total_fh, FILE *out_fh )
{

  REPEAT_FAM_LIST *repeat_fam_array[ REPT_NODE_BUFF ], *total_repeat = NULL;

  initialize_repeat_fam_array( repeat_fam_array, REPT_NODE_BUFF );
  
  set_position( 0, rep_fh );
  
  get_repeat_origin_array( repeat_fam_array, &total_repeat, chrom_header, rep_fh );
  
  print_repeat_origin_array( repeat_fam_array, total_repeat, chrom_header, 
    hist_fh, total_fh, out_fh );
  
}
















void initialize_repeat_fam_array( REPEAT_FAM_LIST *repeat_fam_array[], int n )
{

  int i;
  
  for ( i = 0; i < n; i++ )
    repeat_fam_array[ i ] = NULL;

}









void get_repeat_origin_array( REPEAT_FAM_LIST *repeat_fam_array[],
  REPEAT_FAM_LIST **total_repeat, CHROM_LIST *chrom_header, FILE *in )
{

  TREE *make_and_set_tree( TREE *tree );
  
  TREE *tree = NULL;
  /*  In ../tree/create_tree.c:  */
  tree = (TREE *)make_and_set_tree( tree );

  char line[ LINE_BUFFER ];
  
  CHROM_LIST *chrom = chrom_header;
  
  while ( fgets( line, LINE_BUFFER, in ) != NULL )
    get_repeat_origin_array_single_line( repeat_fam_array, total_repeat, line, 
      chrom_header, &chrom, tree, in );

}










void get_repeat_origin_array_single_line( REPEAT_FAM_LIST *repeat_fam_array[],
  REPEAT_FAM_LIST **total_repeat, char line[], CHROM_LIST *chrom_header, 
  CHROM_LIST **chrom, TREE *tree, FILE *in )
{

  if ( line[ 0 ] == '#' )
    return;
    
  char chr[ BUFFER ], repeat[ BUFFER ], rep_family[ BUFFER ];
  int start, end;
  
  sscanf( line, "%s%d%d%s%*s%s", chr, &start, &end, repeat, rep_family );
  
  *total_repeat = (REPEAT_FAM_LIST *)update_repeat_fam_list( *total_repeat,
    repeat, rep_family, end - start + 1, SITES );
    
  SITE_LIST *site = NULL;  
    
  if (( site = (SITE_LIST *)repeat_in_chrom_list( chrom_header, chrom, chr, start, 
    end )) != NULL )
  { 
    update_repeat_fam_list_cell( repeat_fam_array, *chrom, site, repeat, rep_family, 
      tree );
  }

}






SITE_LIST *repeat_in_chrom_list( CHROM_LIST *chrom_header, CHROM_LIST **chrom, 
  char chr[], int start, int end )
{

  if (( *chrom == NULL ) || ( strcmp( (*chrom)->chr, chr ))) {
    *chrom = chrom_header;
    while ( *chrom != NULL ) {
      if ( ! strcmp( (*chrom)->chr, chr ))
        break;
      *chrom = (*chrom)->next;
    }
  }
  
  if ( *chrom == NULL )
    return NULL;
  
  SITE_LIST *site = (*chrom)->site_header;
  
  SITE_LIST *my_site = NULL;
  
  while ( site != NULL ) {
  
    if (( site->stats->new_site == 0 ) && ( site->site->align_start >= start )) {
      if (( my_site == NULL ) && ( site->site->align_start <= end )) {
        my_site = site;
        site->stats->new_site = 1;
      }
    }
    
    if ( site->site->align_start > end )
      break;
    
    site = site->next;
  
  }
  
  return my_site;

}














void update_repeat_fam_list_cell( REPEAT_FAM_LIST *repeat_fam_array[], 
  CHROM_LIST *chrom, SITE_LIST *site, char repeat[], char rep_family[], TREE *tree )
{

  int i = 0;
  
  NODE_LIST *node = tree->spec_header;
  
  while ( node != NULL ) {
  
    if ( ! strcmp( node->my_node->label, site->stats->node )) {
      if ( ! strcmp( node->my_node->label, "Hominidae" )) {
        printf( "%s\t%d\t%d\t%s\n", chrom->chr, site->site->align_start, 
          site->site->align_end, repeat );
      }
      break;
    }
      
//   if ( ! strcmp( node->my_node->label, "Eutheria" ))
  //    break;
      
    node = node->up;
    
    ++i;
  
  }
  
  repeat_fam_array[ i ] = 
    (REPEAT_FAM_LIST *)update_repeat_fam_list( repeat_fam_array[ i ], repeat, 
      rep_family, 1, SITES );

}











REPEAT_FAM_LIST *update_repeat_fam_list( REPEAT_FAM_LIST *repeat_header,
  char repeat[], char rep_family[], int x, int type )
{

  REPEAT_FAM_LIST *rep = repeat_header;
  
  while ( rep != NULL ) {

    if ( ! strcmp( rep->family, rep_family )) {

      switch ( type ) {
      case TOTAL :
        rep->total += x;
        break;
      case SITES :
        rep->sites += x;
        break;
      default :
        error( 100, "What type?\n" );
        break;
      }
      
      return update_repeat_list( repeat_header, rep, repeat, x, type );

    }
    
    rep = rep->next;

  }
  
  return add_new_repeat_fam_lister( repeat_header, repeat, rep_family, x, type );

}










REPEAT_FAM_LIST *add_new_repeat_fam_lister( REPEAT_FAM_LIST *repeat_header, 
  char repeat[], char rep_family[], int x, int type )
{

  REPEAT_FAM_LIST *create_repeat_fam_lister( REPEAT_FAM_LIST *x );
  REPEAT_LIST *create_repeat_lister( REPEAT_LIST *x );
  
  REPEAT_FAM_LIST *new_fam = NULL;
  REPEAT_LIST *new = NULL;
  new_fam = (REPEAT_FAM_LIST *)create_repeat_fam_lister( new_fam );
  new = (REPEAT_LIST *)create_repeat_lister( new );
  
  new_fam->repeat_header = new;
  
  strcpy( new_fam->family, rep_family );
  strcpy( new->repeat, repeat );
  
  switch ( type ) {
  
  case TOTAL :
    new_fam->total = new->total = x;
    break;
    
  case SITES :
    new_fam->sites = new->sites = x;
    break;
    
  default :
    error( 102, "What type of type?\n" );
    break;
    
  }
  
  if ( repeat_header == NULL )
    return new_fam;
    
  REPEAT_FAM_LIST *curr = repeat_header;
  
  while ( curr->next != NULL )
    curr = curr->next;
    
  curr->next = new_fam;
  
  return repeat_header;

}









REPEAT_FAM_LIST *update_repeat_list( REPEAT_FAM_LIST *repeat_header,
  REPEAT_FAM_LIST *my_rep_fam, char repeat[], int x, int type )
{

  REPEAT_LIST *rep = my_rep_fam->repeat_header;
  
  while ( rep != NULL ) {
  
    if ( ! strcmp( rep->repeat, repeat )) {
    
    switch ( type ) {
    
    case TOTAL :
      rep->total += x;
      break;
      
    case SITES :
      rep->sites += x;
      break;
      
    default :
      error( 101, "What type again?\n" );
      break;
    
    }
      
    return repeat_header;
    
    }
    
    rep = rep->next;
  
  }
  
  return add_new_repeat_lister( repeat_header, my_rep_fam, repeat, x, type );

}









REPEAT_FAM_LIST *add_new_repeat_lister( REPEAT_FAM_LIST *repeat_header, 
  REPEAT_FAM_LIST *my_rep_fam, char repeat[], int x, int type )
{

  REPEAT_LIST *create_repeat_lister( REPEAT_LIST *x );

  REPEAT_LIST *new = NULL;
  new = (REPEAT_LIST *)create_repeat_lister( new );
  
  strcpy( new->repeat, repeat );

  REPEAT_LIST *rep = my_rep_fam->repeat_header;
  
  while ( rep->next != NULL ) 
    rep = rep->next;
    
  rep->next = new;
  
  switch ( type ) {
  
  case TOTAL :
    new->total = x;
    break;
    
  case SITES :
    new->sites = x;
    break;
    
  default :
    error( 103, "oh my\n" );
    break;
  
  }
  
  return repeat_header;

}













void print_repeat_origin_array( REPEAT_FAM_LIST *repeat_fam_array[], 
  REPEAT_FAM_LIST *total_repeat, CHROM_LIST *chrom_header, FILE *hist_fh, 
  FILE *total_fh, FILE *out )
{

  TREE *make_and_set_tree( TREE *tree );
  
  TREE *tree = NULL;
  /*  In ../tree/create_tree.c:  */
  tree = (TREE *)make_and_set_tree( tree );
  
  /*  In this file, below:  */
  print_repeat_origin_fracs( repeat_fam_array, chrom_header, tree, hist_fh );
  
  /*  In this file, below:  */
  print_repeat_origin_fams( repeat_fam_array, total_repeat, chrom_header,
    tree, out );

  /*  In this file, below:  */
  print_repeat_total_fracs( total_repeat, total_fh );
  
}











void print_repeat_origin_fracs( REPEAT_FAM_LIST *repeat_fam_array[],
  CHROM_LIST *chrom_header, TREE *tree, FILE *hist_fh )
{

  int i = 0;

  NODE_LIST *node = tree->spec_header;
  
  while ( node != NULL ) {
  
    int sites = get_numb_of_repeat_sites( repeat_fam_array[ i ] );
    int total = get_numb_of_total_sites( chrom_header, node->my_node->label );
    
    /*
    fprintf( hist_fh, "%15s\t%9d\t%9d\t%.2f\n",
      node->my_node->label, sites, total, 
        (float)sites * 100.0 / (float)total );
    */  
    ++i;
    
    if ( ! strcmp( node->my_node->label, "Eutheria" ))
      break;
    
    node = node->up;
  
  }

}









int get_numb_of_new_sites( CHROM_LIST *chrom_header, char node[] )
{

  int x = 0;
  
  CHROM_LIST *chrom = chrom_header;
  
  while ( chrom != NULL ) {
  
    SITE_LIST *site = chrom->site_header;
    
    while ( site != NULL ) {
    
      if (( ! strcmp( site->stats->node, node )) && ( site->stats->new_site > 0 ))
        ++x;
        
      site = site->next;
    
    }
    
    chrom = chrom->next;
  
  }
  
  return x;

}







int get_numb_of_repeat_sites( REPEAT_FAM_LIST *repeat_fam )
{

  int x = 0;
  
  REPEAT_FAM_LIST *fam = repeat_fam;
  
  while ( fam != NULL ) {
  
    x += fam->sites;
    
    fam = fam->next;
  
  }
  
  return x;

}









int get_numb_of_total_sites( CHROM_LIST *chrom_header, char node[] )
{

  int x = 0;
  
  CHROM_LIST *chrom = chrom_header;
  
  while ( chrom != NULL ) {
  
    SITE_LIST *site = chrom->site_header;
    
    while ( site != NULL ) {
    
      if ( ! strcmp( site->stats->node, node )) {
        ++x;
      }
        
      site = site->next;
    
    }
    
    chrom = chrom->next;
  
  }
  
  return x;

}












void print_repeat_total_fracs( REPEAT_FAM_LIST *total_repeat, FILE *total_fh )
{

  REPEAT_FAM_LIST *total = total_repeat;
  
  fprintf( total_fh, "%8s\t%8s\n", "\"Fam\"", "\"Frac\"" );
  
  while ( total != NULL ) {
    fprintf( total_fh, "%8s\t%8.4f\n", total->family, 
      (float)total->sites / (float)total->total );
    total = total->next;
  }

}








void print_repeat_origin_fams( REPEAT_FAM_LIST *repeat_fam_array[], 
  REPEAT_FAM_LIST *total_repeat, CHROM_LIST *chrom_header, TREE *tree, FILE *out )
{

  set_totals( repeat_fam_array, total_repeat );

  NODE_LIST *node = tree->spec_header;
  
  int i = 0;
  
  while ( node != NULL ) {
  
    REPEAT_FAM_LIST *fam_header = NULL;
    
    REPEAT_FAM_LIST *fam = repeat_fam_array[ i ];
    
    while ( fam != NULL ) {
    
      REPEAT_FAM_LIST *next = fam->next;
      
      fam_header = (REPEAT_FAM_LIST *)set_and_put_repeat_fam( fam_header, fam,
        total_repeat );
      
      fam = next;
    
    }
    
    fprintf( out, "%s\n", node->my_node->label );
    
    fam = fam_header;
    
    while ( fam != NULL ) {
      fprintf( out, "\t%10s\t%8d\t%8d\t%8.3f\t%10.2e\t%.3fx\n",
        fam->family, fam->sites, fam->total, fam->exp * fam->total, 
        (double)fam->pval, fam->fold );
      fam = fam->next;
    }
    
    print_sep( out );
    
    if ( ! strcmp( node->my_node->label, "Eutheria" ))
      break;
    
    node = node->up;
    
    ++i;
  
  }

}









void set_totals( REPEAT_FAM_LIST *repeat_fam_array[], REPEAT_FAM_LIST *total_repeat )
{

  int i;
  
  for ( i = 0; i < REPT_NODE_BUFF; i++ )
    set_total_repeat_fam( repeat_fam_array[ i ] );
    
  set_total_repeat_fam( total_repeat );

}











void set_total_repeat_fam( REPEAT_FAM_LIST *repeat_fam )
{

  int x = 0;
  
  REPEAT_FAM_LIST *rep = repeat_fam;
  
  while ( rep != NULL ) {
  
    x += rep->sites;
    
    rep = rep->next;
  
  }
  
  rep = repeat_fam;
  
  while ( rep != NULL ) {
  
    rep->total = x;
    
    rep = rep->next;
  
  }

}












REPEAT_FAM_LIST *set_and_put_repeat_fam( REPEAT_FAM_LIST *fam_header, 
  REPEAT_FAM_LIST *fam, REPEAT_FAM_LIST *total_repeat )
{

  PVAL cumulative_binom_prob( STAT p, int k, int n );

  REPEAT_FAM_LIST *total = total_repeat;
  
  while ( total != NULL ) {
    if ( ! strcmp( total->family, fam->family ))
      break;
    total = total->next;
  }
  
  fam->next = NULL;
  
  fam->exp = (float)total->sites / (float)total->total;
    
  fam->fold = ((float)fam->sites / (float)fam->total) / fam->exp;

  if ( fam->exp > MIN_FRAC )
    /*  In ../model/sig.c:  */  
    fam->pval = cumulative_binom_prob( fam->exp, fam->sites, fam->total );
  else
    fam->pval = 1.0;
  
  if ( fam_header == NULL )
    return fam;
    
  if ( fam_header->pval > fam->pval ) {
    fam->next = fam_header;
    return fam;
  }
  
  REPEAT_FAM_LIST *curr = fam_header, *prev = NULL;
  
  while (( curr != NULL ) && ( curr->pval <= fam->pval )) {
    prev = curr;
    curr = curr->next;
  }
  
  if ( prev != NULL )
    prev->next = fam;
    
  fam->next = curr;
  
  return fam_header;

}









































