#include "sig.h"




















#define MAXIT 100
#define EPS 3.0e-7
#define FPMIN 1.0e-30



















PVAL cumulative_binom_prob( STAT p, int k, int n )
{
  return betai( (PVAL)k, (PVAL)(n - k + 1.0), (PVAL)p );
}

















PVAL betai( PVAL a, PVAL b, PVAL x )
{

  if ( x < 0.0 || x > 1.0 )
    return 1.0;
  
  PVAL bt;
    
  if ( x == 0.0 || x == 1.0 )
    bt = 0.0;
  else
    bt = exp( gammln( a + b ) - gammln( a ) - gammln( b ) + (a * log( x ))
      + (b * log( 1.0 - x )) );
      
  if ( x < (a + 1.0) / (a + b + 2.0) )
    return bt * betacf( a, b, x ) / a;

  return 1.0 - (bt * betacf( b, a, 1.0 - x ) / b);

}














PVAL betacf( PVAL a, PVAL b, PVAL x )
{

  int m, m2;
  PVAL aa, c, d, del, h, qab, qam, qap;
  
  qab = a + b;
  qap = a + 1.0;
  qam = a - 1.0;
  c = 1.0;
  d = 1.0 - (qab * x / qap);
  
  if ( fabs( d ) < FPMIN )
    d = FPMIN;
    
  d = 1.0 / d;
  h = d;
  
  for ( m = 1; m <= MAXIT; m++ ) {
  
    m2 = 2 * m;
    aa = (m * (b - m) * x) / ((qam + m2) * (a + m2));
    d = 1.0 + (aa * d);

    if ( fabs( d ) < FPMIN )
      d = FPMIN;

    c = 1.0 + (aa / c);

    if ( fabs( c ) < FPMIN )
      c = FPMIN;

    d = 1.0 / d;
    h *= d * c;
    aa = (-1.0 * (a + m) * (qab + m) * x) / ((a + m2) * (qap + m2));

    d = 1.0 + (aa * d);

    if ( fabs( d ) < FPMIN )
      d = FPMIN;

    c = 1.0 + (aa / c);

    if ( fabs( c ) < FPMIN )
      c = FPMIN;

    d = 1.0 / d;
    del = d * c;
    h *= del;

    if ( fabs( del - 1.0 ) < EPS )
      break;
  
  }
  
  if ( m > MAXIT )
    error( 6300, "Maximum iterations for P-value generation.\n" );
    
  return h;

}










PVAL gammln( PVAL xx )
{

  PVAL x, y, tmp, ser;
  
  static PVAL cof[ 6 ] = { 76.18009172947146, -86.50532032941677, 24.01409824083091,
    -1.231739572450155, 0.1208650973866179e-2, -0.5395239384953e-5 };

  int j;
  
  y = x = xx;
  tmp = x + 5.5;
  tmp -= (x + 0.5) * log( tmp );
  ser = 1.000000000190015;
  
  for ( j = 0; j <= 5; j++ )
    ser += cof[ j ] / ++y;
    
  return (-1.0 * tmp) + log( 2.5066282746310005 * ser / x );

}


















































