#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <math.h>
#include <time.h>















#define BUFFER 256
#define LINE_BUFFER 20000
#define SEQ_BUFFER 20000
#define SPEC_BUFFER 10
#define SHORT_BUFFER 25
#define MX_BUFFER 200
#define CONS_BUFFER 40
#define MY_NODE_BUFFER 25
#define SHORT_SEQ_BUFFER 700


#define IMPOSSIBLE -1
#define IMPOSSIBLE_PVAL -1.0

#define PVAL_THRESH 1e-10

#define INCREM 1
#define DECREM -1

#define MY_MILLION 1
#define MILLION 1000000.0

#define A_NUCL 0
#define C_NUCL 1
#define G_NUCL 2
#define T_NUCL 3

#define WORD_LENGTH 8

/* change MOTIF_LENGTH in to 30*/
#define MOTIF_LENGTH 30

#define BITS_PER_NUCL 2
#define NUMB_OF_NUCL 4
#define WORD_BITS (WORD_LENGTH * BITS_PER_NUCL)
#define NUMB_OF_WORDS 65536


#define START_INDEX 0
#define END_INDEX 1


#define BD 2
#define BIRTH 0
#define DEATH 1

#define MM 2
#define MN 0
#define MZ 1


#define DEFAULT_SCORE_THRESH 300


#define DBUFF 2


#define BRANCH_BUFFER 10
#define OCCUR_BUFFER_LARGE 550
#define OCCUR_BUFFER_MEDIUM 150
#define OCCUR_BUFFER_SMALL 80

#define OCCUR_BUFFER (OCCUR_BUFFER_LARGE)

#define NUMB_OF_SPEC 100
#define NUMB_OF_NODES 110
#define BITS_PER_CT 8

#define MATCH_BUFFER 2
#define MATCH 0
#define NON_MATCH 1


#define CROSS_BUFFER 2


#define STRAND_BUFFER 2
#define PLUS_STRAND 0
#define MINUS_STRAND 1


#define MAJOR 0
#define MINOR 1


#define INT_TYPE 0
#define DOUBLE_TYPE 1
#define FLOAT_TYPE 2
#define CHAR_TYPE 3



#define PP 2
#define PRESENT 0
#define ABSENT 1









#define DEFAULT_TREE_FILE "data/tree_file.nh"









/*  Previous:
#define MAIN_REG_START -500
#define MAIN_REG_END 250
#define MAIN_REG_MODE '0'
*/

#define MAIN_REG_START -400
#define MAIN_REG_END 400
#define MAIN_REG_MODE '0'

#define MAIN_REG_LENGTH (MAIN_REG_END - MAIN_REG_START)

#define REG_RADIUS 400

#define MAX_ALIGNMENT_LENGTH 4000




/*  Called in ../main/fname.c:  */
#define CHIP_TYPE_BUFF 18
#define CHIP_TYPE { "HAIB_GM12878", "SYDH_GM12878", "HAIB_H1", "UTA_GM12878", "Uchicago_K562", "SYDH_PBDE", "UW_HRE", "UW_HAC", "Broad_H1hESC", "Broad_GM12878", "PSU_CH12", "GSM_HL1", "PSU_Erythrobl", "SING_E14", "GSM_V65", "LICR_Kidney", "LICR_Cbellum", "SYDH_CH12" }




#define BROY_GOING -1
#define BROY_CONVERGED 0
#define BROY_MAX_ITER 1
#define BROY_DIVERGED 2
#define BROY_NAN 3
#define BROY_NEG_PROB 4

#define OUTCOME_GOING -1
#define OUTCOME_CONVERGED 0
#define OUTCOME_MAX_ITER 1
#define OUTCOME_DIVERGED 2
#define OUTCOME_NAN 3
#define OUTCOME_NEG_PROB 4



#define CLASS_NONE 0
#define CLASS_GC 1
#define CLASS_NON_SIG 2
#define CLASS_MF_BIRTH 3
#define CLASS_MF_DEATH 4



#define NONE_EXISTS -1
#define DISTAL 0
#define CORE_PROM_UP 1
#define CORE_PROM_DOWN 2
#define PROX_PROM_UP 3
#define PROX_PROM_DOWN 4
#define UPSTREAM 5
#define FIVE_UTR 6
#define THREE_UTR 7
#define CODING_SEQ 8
#define INTRON 9






#define MIN(a,b) (((a) > (b)) ? (b) : (a))
#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#define ABS(a) (((a) > 0) ? (a) : ((a) * -1))









typedef long FPOS;
typedef long long COORD;
typedef short POS;
typedef long COUNT;
typedef long SEQ;            //  static:  don't change
typedef long INDEX;          //  static:  don't change
typedef unsigned long MF;    //  static:  don't change
typedef unsigned char DD;

typedef double TIME;
typedef long double STAT;
typedef long double PVAL;






/*  In ../main/stream.c:  */
int get( void *x, size_t size, FILE *fh );
void put( void *x, size_t size, FILE *fh );
int get_array( void *x, size_t size, int n, FILE *fh );
void put_array( void *x, size_t size, int n, FILE *fh );







/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/








typedef struct ft_struct {
  SEQ seq;
  POS pos[ MOTIF_LENGTH ];
  char site[ MOTIF_LENGTH + 1 ];
} FT;





typedef struct ft_ind_struct {
  FPOS fpos;
  COUNT n;
  COUNT x;
} FT_IND;





typedef struct seq_info_struct {
  FPOS fpos;
  COUNT n;
  COUNT x;
  char gene[ SHORT_BUFFER ];
  char chr[ SHORT_BUFFER ];
  COORD start;
  COORD end;
  char strand;
  int score;
} SEQ_INFO;






typedef struct spec_struct {
  char spec[ SPEC_BUFFER ];
  INDEX numb;
} SPEC;





typedef struct spec_cell_struct {
  SPEC *x;
  INDEX n;
} SPEC_CELL;





typedef struct fh_struct {
  FILE *text;
  FILE *seq;
  FILE *ct;
  FILE *spec;
  FILE *ft_ind;
  FILE *ft;
  FILE *nn_ind;
  FILE *nn;
  SPEC_CELL *spec_cell;
  INDEX spec_index;
  char reg;
} FH;









typedef struct ft_list_struct {
  FT *ft;
  struct ft_list_struct *next;
} FT_LIST;






typedef struct ft_array_struct {
  FPOS *fpos_array;
  int n;
  int total;
  FT_LIST *ft_array[ NUMB_OF_WORDS ];
  FT_LIST *ft_end[ NUMB_OF_WORDS ];
  FT_LIST *empty_header;
  FT_LIST *empty_last;
} FT_ARRAY;




/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/










typedef struct motif_struct {
  MF my_motif;
  char cons[ BUFFER ];
  char length;
  FT ft;
} MOTIF;







typedef struct rep_struct {
  int occur;
  int seq_occur;
  int n;
  float exp;
  STAT pr;
  STAT pr_seq;
  STAT z;
  STAT pscore;
  char spec[ SHORT_BUFFER ];
  unsigned use : 1;
} REP;







typedef struct motif_stats_struct {
  STAT zval;
  STAT pscore;
  char spec[ SHORT_BUFFER ];
  REP rep_array[ NUMB_OF_SPEC ];
  int n;
  int a;
  int b;
  int ab;
  float exp;
  PVAL pval;
  int total_ct;
  int strand_ct[ 2 ];
  float frac;
  float pr;
  char type;
} MOTIF_STATS;







typedef struct motif_list_struct {
  MOTIF *my_mf;
  MOTIF_STATS *my_stats;
  char spec[ BUFFER ];
  struct motif_list_struct *next;
} MOTIF_LIST;






typedef struct site_info_struct {
  SEQ_INFO *loc;
  char seq[ SHORT_BUFFER ];
  char gene[ SHORT_BUFFER ];
  SEQ seq_numb;
  int align_start;
  int align_end;
} SITE_INFO;







typedef struct site_stats_struct {
  int occur_in_seq;
  char node[ SHORT_BUFFER ];
  STAT pr;
  char unique;
  char new_site;
} SITE_STATS;




typedef struct site_list_struct {
  SITE_INFO *site;
  SITE_STATS *stats;
  struct site_list_struct *next;
  struct site_list_struct *prev;
} SITE_LIST;






typedef struct chrom_list_struct {
  char chr[ SHORT_BUFFER ];
  SITE_LIST *site_header;
  struct chrom_list_struct *next;
} CHROM_LIST;





typedef struct origin_struct {
  MOTIF *mf;
  struct result_list_struct *res;
  SITE_LIST *site_header[ STRAND_BUFFER ];
  int total_seqs;
  int n;
  int n_strand[ 2 ];
  float frac;
  char spec[ BUFFER ];
  int anc_bd;
} ORIGIN;








typedef struct origin_list_struct {
  ORIGIN *my_origin;
  struct origin_list_struct *next;
  struct origin_list_struct *prev;
} ORIGIN_LIST;






typedef struct pos_list_struct {
  POS start;
  POS end;
  POS pos[ CONS_BUFFER ];
  struct pos_list_struct *next;
} POS_LIST;





typedef struct seq_list_struct {
  SEQ numb;
  SEQ_INFO *my_seq;
  POS_LIST *pos_header;
  POS_LIST *nn_header;
  int length[ NUMB_OF_SPEC ];
  int numb_of_spec;
  char sequence[ SHORT_SEQ_BUFFER ];
  char minor_seq[ SHORT_SEQ_BUFFER ];
  char use;
  char yes;
  char tfbs;
  struct seq_list_struct *next;
  struct seq_list_struct *down;
  struct seq_list_struct *across;
} SEQ_LIST;








typedef struct occur_cell_struct {
  SEQ_LIST *x[ BUFFER ];
  char spec[ BUFFER ][ SPEC_BUFFER ];
  INDEX n;
} OCCUR_CELL;







typedef struct ct_struct {
  short ct;
} CT;






typedef struct use_struct {
  char use;
} USE;






typedef struct use_cell_struct {
  USE *x;
  INDEX n;
} USE_CELL;




typedef struct oc_struct {
  COUNT x;
  STAT freq;
} OC;







typedef struct base_freq_struct {
  OC mono[ NUMB_OF_NUCL ];
  OC di[ NUMB_OF_NUCL ][ NUMB_OF_NUCL ];
  COUNT total;
  COUNT seq_count;
} BASE_FREQ;





typedef struct base_freq_cell_struct {
  BASE_FREQ x[ NUMB_OF_SPEC ];
  char spec[ NUMB_OF_SPEC ][ SHORT_BUFFER ];
  INDEX n;
} BASE_FREQ_CELL;




typedef struct leaf_data_struct {
  CT ct_array[ NUMB_OF_SPEC ];
  CT ch_array[ NUMB_OF_SPEC ];
  USE use_array[ NUMB_OF_SPEC ];
  SEQ seq_numb;
  COUNT n;
  int changes;
  unsigned done : 1;
} LEAF_DATA;





typedef struct leaf_data_list_struct {
  LEAF_DATA *data;
  struct leaf_data_list_struct *next;
  struct leaf_data_list_struct *prev;
} LEAF_DATA_LIST;







typedef struct pdata_struct {
  DD ct_array[ NUMB_OF_SPEC ];
  char use_array[ NUMB_OF_SPEC ];
  char gene[ SHORT_BUFFER ];
  COUNT n;
  INDEX numb_of_spec;
  SEQ seq_numb;
  COUNT changes;
  STAT best_lhood;
  unsigned best_lhood_nan : 1;
  unsigned done : 1;
} PDATA;







typedef struct pdata_list_struct {
  PDATA *my_pdata;
  struct pdata_list_struct *next;
  struct pdata_list_struct *prev;
} PDATA_LIST;







typedef struct go_term_struct {
  char id[ SHORT_BUFFER ];
  char term[ BUFFER ];
  int index;
} GO_TERM;






typedef struct go_term_list_struct {
  GO_TERM *my_go;
  struct go_term_list_struct *next;
  struct go_term_list_struct *down;
} GO_TERM_LIST;






typedef struct go_gene_struct {
  int numb;
  char acc[ SHORT_BUFFER ];
  struct str_list_struct *uniprot_header;
  GO_TERM_LIST *go_header;
} GO_GENE;







typedef struct go_gene_list_struct {
  GO_GENE *my_gene;
  struct go_gene_list_struct *next;
} GO_GENE_LIST;






typedef struct go_stats_struct {
  struct memb_list_struct *memb;
  GO_TERM_LIST *go_term;
  PVAL pval;
  int a;
  float exp;
  int aa;
  int b;
  int n;
} GO_STATS;






typedef struct go_stats_list_struct {
  GO_STATS *my_stats;
  struct go_stats_list_struct *next;
} GO_STATS_LIST;





typedef struct go_struct {
  struct clus_list_struct *clus;
  PVAL pval;
  GO_STATS_LIST *stats_header;
} GO;






typedef struct go_list_struct {
  GO *my_go;
  struct go_list_struct *next;
  struct go_list_struct *prev;
  struct go_list_struct *clus_next;
  struct go_list_struct *clus_prev;
} GO_LIST;







typedef struct cross_info {
  char chr[ BUFFER ];
  char gene[ BUFFER ];
  char name[ BUFFER ];
  char spec[ BUFFER ];
  COORD center;
  char strand;
  int dist;
  char location;
} CROSS_INFO;




typedef struct cross_struct {
  CROSS_INFO info[ CROSS_BUFFER ];
  int score;
  char exists[ CROSS_BUFFER ];
  char near_gene[ CROSS_BUFFER ];
  struct cross_list_struct *alt;
} CROSS;




typedef struct cross_list_struct {
  CROSS *my_cross;
  struct cross_list_struct *next;
  struct cross_list_struct *prev;
} CROSS_LIST;




typedef struct chip_pk_struct {
  char tf[ BUFFER ];
  char chr[ BUFFER ];
  COORD start;
  COORD end;
  int score;
} CHIP_PK;






typedef struct chip_pk_list_struct {
  CHIP_PK *my_chip;
  struct chip_pk_list_struct *next;
} CHIP_PK_LIST;





typedef struct chip_gene_struct {
  int numb;
  char acc[ SHORT_BUFFER ];
  char chr[ SHORT_BUFFER ];
  COORD start;
  COORD end;
  char strand;
  CHIP_PK_LIST *chip_header;
} CHIP_GENE;








typedef struct chip_gene_list_struct {
  CHIP_GENE *my_gene;
  struct chip_gene_list_struct *next;
  struct chip_gene_list_struct *prev;
  struct chip_gene_list_struct *chr_next;
  struct chip_gene_list_struct *chr_prev;
  struct chip_gene_list_struct *coord_next;
  struct chip_gene_list_struct *coord_prev;
} CHIP_GENE_LIST;







typedef struct chip_stats_struct {
  struct memb_list_struct *memb;
  char tf[ SHORT_BUFFER ];
  PVAL pval;
  int a;
  int aa;
  int b;
  int n;
  float exp;
  char node[ SHORT_BUFFER ];
} CHIP_STATS;






typedef struct chip_stats_list {
  CHIP_STATS *my_stats;
  CHIP_STATS *p_stats;
  struct chip_stats_list *next;
} CHIP_STATS_LIST;






typedef struct chip_struct {
  struct clus_list_struct *clus;
  char node[ SHORT_BUFFER ];
  PVAL pval;
  CHIP_STATS_LIST *chip_header;
} CHIP;







typedef struct chip_list_struct {
  CHIP *my_chip;
  struct chip_list_struct *next;
} CHIP_LIST;




typedef union elem_struct {
  int d;
  double f;
} ELEM;





typedef struct list_struct {
  ELEM elem;
  char type;
  struct list_struct *next;
} LIST;







typedef struct mx_struct {
  STAT mx[ MX_BUFFER ][ MX_BUFFER ];
  INDEX i;
  INDEX j;
} MX;







typedef struct vec_struct {
  STAT v[ MX_BUFFER ];
  INDEX i;
} VEC;







typedef struct repeat_list_struct {
  char repeat[ BUFFER ];
  int total;
  int sites;
  PVAL pval;
  float fold;
  float exp;
  struct repeat_list_struct *next;
} REPEAT_LIST;








typedef struct repeat_fam_list_struct {
  char family[ BUFFER ];
  int total;
  int sites;
  PVAL pval;
  float fold;
  float exp;
  REPEAT_LIST *repeat_header;
  struct repeat_fam_list_struct *next;
} REPEAT_FAM_LIST;






/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/








typedef struct str_list_struct {
  char str[ BUFFER ];
  struct str_list_struct *next;
} STR_LIST;





typedef struct fpos_list_struct {
  char c;
  char str[ SHORT_BUFFER ];
  FPOS fpos;
  struct fpos_list_struct *next;
} FPOS_LIST;




typedef struct count_cell_struct {
  COUNT *x;
  INDEX n;
} COUNT_CELL;






typedef struct dinucl_struct {
  char spec[ SHORT_BUFFER ];
  int dinucl[ NUMB_OF_NUCL ][ NUMB_OF_NUCL ];
  char c;
} DINUCL;






typedef struct dinucl_list_struct {
  DINUCL *my_d;
  struct dinucl_list_struct *next;
} DINUCL_LIST;







typedef struct gencoord_entry_struct {
  char spec[ BUFFER ];
  char gene[ BUFFER ];
  char chr[ BUFFER ];
  COORD start;
  COORD end;
  COORD chr_length;
  char strand;
  char info[ BUFFER ];
} GENCOORD_ENTRY;







typedef struct maf_struct {
  GENCOORD_ENTRY *entry;
  GENCOORD_ENTRY *curr;
  char seq[ SEQ_BUFFER ];
  int node_q;
  int parent_q;
  int score;
  unsigned use : 1;
} MAF;







typedef struct maf_list_struct {
  MAF *my_maf;
  struct maf_list_struct *next;
  struct maf_list_struct *same;
} MAF_LIST;








typedef struct gene_stats_struct {
  int node_q;
  int parent_q;
  int a;
} GENE_STATS;








typedef struct gene_struct {
  GENCOORD_ENTRY *gen_coord;
  GENE_STATS *stats;
  unsigned use : 1;
} GENE;








typedef struct gene_list_struct {
  GENE *my_gene;
  struct gene_list_struct *next;
} GENE_LIST;








typedef struct region_struct {
  char reg;
  COORD a;
  COORD b;
} REGION;







typedef struct region_list_struct {
  REGION *my_region;
  FH *my_fh;
  struct region_list_struct *next;
} REGION_LIST;








/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/










typedef struct node_info_struct {
  STAT x[ OCCUR_BUFFER ];
  STAT mean;
  INDEX max;
  INDEX q;
  unsigned use : 1;
  unsigned change : 1;
} NODE_INFO;






typedef struct node_info_cell_struct {
  NODE_INFO *x;
  INDEX n;
} NODE_INFO_CELL;





typedef struct node_trans_struct {

  STAT p[ OCCUR_BUFFER ][ OCCUR_BUFFER ];
  STAT u;
  STAT v;
  STAT n;

  STAT p_d[ BD ][ OCCUR_BUFFER ][ OCCUR_BUFFER ];
  STAT u_d[ BD ];
  STAT v_d[ BD ];

  STAT p_dd[ BD ][ BD ][ OCCUR_BUFFER ][ OCCUR_BUFFER ];
  STAT u_dd[ BD ][ BD ];
  STAT v_dd[ BD ][ BD ];
  
  int part;

} NODE_TRANS;





typedef struct node_param_struct {
  STAT a[ OCCUR_BUFFER ][ OCCUR_BUFFER ];
  STAT b[ OCCUR_BUFFER ][ OCCUR_BUFFER ];
} NODE_PARAM;










typedef struct node_struct {
  int numb;
  char label[ SHORT_BUFFER ];
  char spec[ SHORT_BUFFER ];
  char seq[ LINE_BUFFER ];
  TIME t;
  unsigned leaf : 1;
} NODE;








typedef struct node_list_struct {
  NODE *my_node;
  NODE_TRANS *my_trans;
  NODE_INFO *my_info;
  NODE_INFO *my_d[ DBUFF ];
  NODE_INFO *my_dd;
  NODE_INFO *prior;
  NODE_PARAM *param;
  struct tab_struct *tab;
  struct node_list_struct *up;
  struct node_list_struct *down[ BRANCH_BUFFER ];
  struct node_list_struct *next;
  struct node_list_struct *prev;
} NODE_LIST;









typedef struct branch_struct {
  NODE_LIST *node;
  STAT a;
  STAT b;
  unsigned used : 1;
} BRANCH;






typedef struct branch_list_struct {
  BRANCH *my_branch;
  struct branch_list_struct *next;
  struct branch_list_struct *prev;
} BRANCH_LIST;








typedef struct rel_stats_struct {
  char node[ SHORT_BUFFER ];
  COUNT ct[ PP ][ PP ];
  PVAL pval;
} REL_STATS;







typedef struct rel_stats_list_struct {
  REL_STATS *s;
  struct rel_stats_list_struct *next;
} REL_STATS_LIST;







typedef struct rel_struct {
  MOTIF *a_mf;
  MOTIF *b_mf;
  REL_STATS_LIST *s_header;
  PVAL pval;
} REL;









typedef struct rel_list_struct {
  REL *my_rel;
  struct rel_list_struct *next;
} REL_LIST;







/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/








typedef struct tab_struct {
  STAT birth;
  STAT death;
  STAT siten;
  STAT lhood;
  COUNT n;
  TIME t;
} TAB;







typedef struct tab_cell_struct {
  TAB *tab[ NUMB_OF_NODES ];
  INDEX n;
} TAB_CELL;




typedef struct params_struct {
  STAT a;
  STAT b;
} PARAMS;





typedef struct part_struct {
  PARAMS *pp;
  LIST *branch_header;
  int part;
} PART;





typedef struct part_list_struct {
  PART *my_part;
  struct part_list_struct *next;
} PART_LIST;





typedef struct pt_cell_struct {
  PARAMS *pp[ BUFFER ];
  LIST *branch[ BUFFER ];
  INDEX n;
} PT_CELL;





typedef struct part_stats_struct {
  STAT lhood;
  STAT old_lhood;
  STAT best_lhood;
  PVAL pval;
  COUNT n;
  int dof_numer;
  int dof_denom;
  COUNT max_occur;
  char node_label[ BUFFER ];
  char best_label[ BUFFER ];
  char beta_label[ BUFFER ];
  char outcome;
  unsigned conv : 1;
} MODEL_STATS;





typedef struct model_struct {
  INDEX m;
  INDEX part_n;
  MOTIF *mf;
  REGION *my_region;
  PART_LIST *part_header;
  MODEL_STATS *stats;
  MODEL_STATS *null_stats;
  struct model_struct *next;
} MODEL;






typedef struct model_list_struct {
  MODEL *my_model;
  struct model_list_struct *next;
  struct model_list_struct *down;
} MODEL_LIST;









/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/









typedef struct memb_struct {
  MOTIF *mf;
  char node[ SHORT_BUFFER ];
  char offset;
  PVAL pval;
  char class;
  MODEL_LIST *model;
  unsigned used : 1;
} MEMB;





typedef struct memb_list_struct {
  MEMB *my_memb;
  struct memb_list_struct *next;
} MEMB_LIST;




typedef struct clus_struct {
  char cons[ SHORT_BUFFER ];
  char node[ SHORT_BUFFER ];
  char label[ SHORT_BUFFER ];
  MEMB_LIST *memb_header;
  struct clus_list_struct *rc;
  struct clus_list_struct *rc_other;
  struct clus_list_struct *tf;
  PVAL pval;
  char class;
  char min_offset;
  char max_offset;
  char reg;
  int numb;
  unsigned dupl : 1;
} CLUS;








typedef struct clus_list_struct {
  CLUS *my_clus;
  struct clus_list_struct *next;
  struct clus_list_struct *down;
} CLUS_LIST;






/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/












typedef struct tree_struct {
  NODE_LIST *root;
  NODE_LIST *spec_header;
  NODE_LIST *inner_nodes;
  BRANCH_LIST *branch_header;
  BRANCH_LIST *death_header;
  BRANCH_LIST *low_birth_header;
  BRANCH_LIST *low_death_header;
} TREE;







typedef struct result_struct {
  MODEL_LIST *model_header;
  MODEL_LIST *my_model;
  MODEL_LIST *null_model;
  PVAL pval;
  int m;
  MOTIF *mf;
  char label[ BUFFER ];
  unsigned use : 1;
} RESULT;






typedef struct result_list_struct {
  RESULT *my_result;
  struct result_list_struct *next;
  struct result_list_struct *down;
} RESULT_LIST;







typedef struct tree_data_struct {
  MODEL_LIST *model;
  OCCUR_CELL *occur;
  NODE_LIST *root;
  LEAF_DATA_LIST *leaf;
} TREE_DATA;










typedef struct data_struct {
  TREE *tree;
  OCCUR_CELL *occur;
  PDATA_LIST *pdata;
  USE_CELL *use_cell;
  MOTIF *mf;
  REGION *reg;
  SPEC_CELL *spec_cell;
  MODEL_LIST *new_model;
  MODEL_LIST *old_model;
} DATA;











typedef struct data_list_struct {
  DATA *data;
  DATA *rc_data;
  struct data_list_struct *next;
} DATA_LIST;







/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/










typedef struct broy_struct {

  VEC *x;
  VEC *old_x;
  VEC *s;
  VEC *f;
  
  MX *j;
  MX *b;
  
  TREE_DATA *d;
  
  INDEX n;
  
  STAT lhood;
  STAT old_lhood;
  
  int major_iter;
  int loop_iter;
  int total_iter;
  
  char conv;
  char outcome;

} BROY;










typedef struct em_struct {
  
  PT_CELL *x;
  PT_CELL *old_x;
  
  DATA *data;
  
  STAT lhood;
  STAT old_lhood;
  
  int part_n;
  
  int iter;
  char conv;
  char outcome;
  
} EM_INFO;








#define STAMP_HIST_BUFFER 200



typedef struct stamp_hist_struct {
  int x[ STAMP_HIST_BUFFER ];
  PVAL pval[ STAMP_HIST_BUFFER ];
  int match[ STAMP_HIST_BUFFER ];
  int total[ STAMP_HIST_BUFFER ];
  int hg_match[ STAMP_HIST_BUFFER ];
  int hg_total[ STAMP_HIST_BUFFER ];
  int euth_match[ STAMP_HIST_BUFFER ];
  int euth_total[ STAMP_HIST_BUFFER ];
  int fish_match[ STAMP_HIST_BUFFER ];
  int fish_total[ STAMP_HIST_BUFFER ];
} STAMP_HIST;






typedef struct sum_struct {
  char spec[ SHORT_BUFFER ];
  int numb_of_seqs;
  float t;
  int clus_n;
  int memb_n;
  int memb_gc_n;
  int memb_gc_only;
  int memb_at_n;
  int memb_mf_n;
  int clus_array[ SHORT_BUFFER ];
} SUM;








typedef struct sum_list_struct {
  SUM *my_sum;
  struct sum_list_struct *next;
} SUM_LIST;







typedef struct evol_stats_struct {
  int numb_of_seqs;
  int n;
  char length;
  int sub[ BUFFER ];
  int indel[ BUFFER ];
  TIME total_t;
} EVOL_STATS;





typedef struct evol_struct {
  MOTIF *mf;
  EVOL_STATS *s;
} EVOL;







typedef struct evol_list_struct {
  EVOL *my_evol;
  struct evol_list_struct *next;
} EVOL_LIST;








typedef struct origin_hist_struct {
  char node[ MY_NODE_BUFFER ][ BUFFER ];
  float time[ MY_NODE_BUFFER ];
  int ct[ MY_NODE_BUFFER ];
  int new_site[ MY_NODE_BUFFER ];
  int overlap[ MY_NODE_BUFFER ];
  int total_seqs;
  int total_ct;
  int seq_ct;
  int n;
  int euth;
} ORIGIN_HIST;







typedef struct rel_hist_struct {
  char node[ MY_NODE_BUFFER ][ BUFFER ];
  float time[ MY_NODE_BUFFER ];
  int ct[ MY_NODE_BUFFER ][ MY_NODE_BUFFER ];
  int seq_ct[ MY_NODE_BUFFER ][ MY_NODE_BUFFER ];
  float exp[ MY_NODE_BUFFER ][ MY_NODE_BUFFER ];
  PVAL pval;
  int total_seqs;
  int n;
  int x;
  int major_index[ STRAND_BUFFER ];
  int minor_index[ STRAND_BUFFER ];
} REL_HIST;







typedef struct orel_stats_struct {
  int n;
  int a;
  int b;
  int ab;
  float exp;
  PVAL pval;
} OREL_STATS;




typedef struct rel_origin_struct {
  ORIGIN_HIST *major_hist[ STRAND_BUFFER ];
  ORIGIN_HIST *minor_hist[ STRAND_BUFFER ];
  REL_HIST *rel_hist[ STRAND_BUFFER ][ STRAND_BUFFER ];
  OREL_STATS *rel_stats[ STRAND_BUFFER ][ STRAND_BUFFER ];
  REL_HIST *best_hist;
  REL_HIST *rev_hist;
  REL_HIST *all_hist;
  OREL_STATS *best_stats;
  OREL_STATS *rev_stats;
  MOTIF *major_mf;
  MOTIF *minor_mf;
} REL_ORIGIN;





typedef struct rel_origin_list_struct {
  REL_ORIGIN *my_rel;
  struct rel_origin_list_struct *next;
} REL_ORIGIN_LIST;





typedef struct model_fh_struct {
  FILE *sum_fh;
  FILE *model_fh;
  FILE *all_fh;
} MODEL_FH;





typedef struct pred_fh_struct {
  FILE *text_fh;
  FILE *res_fh;
  FILE *sum_fh;
  FILE *null_fh;
} PRED_FH;






typedef struct go_fh_struct {
  FILE *out;
  FILE *hist_fh;
  FILE *sum_fh;
  FILE *primate_fh;
  FILE *rodent_fh;
  FILE *euth_fh;
  FILE *laura_fh;
  FILE *mammal_fh;
  FILE *fish_fh;
} GO_FH;








typedef struct mf_fh_struct {
  FILE *align_fh;
  FILE *col_fh;
  FILE *reg_fh;
} MOTIF_FH;







/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/















char mode;

char directory[ BUFFER ];
char seq_align_file[ BUFFER ];
char region_file[ BUFFER ], minor_region_file[ BUFFER ];
char tree_file[ BUFFER ];
char motif_file[ BUFFER ];




char minor_mode;






int reg_start, reg_end;






STAT branch_factor;








char combine_mode;

char input_file[ BUFFER ], output_file[ BUFFER ];
char info_file[ BUFFER ], info_out[ BUFFER ];

char major_spec[ BUFFER ], minor_spec[ BUFFER ];
char phylo_spec[ BUFFER ];

char major_tf[ BUFFER ], minor_tf[ BUFFER ];

char chip_type[ BUFFER ];
char chip_type_minor[ BUFFER ];

REGION main_reg;

MOTIF major_motif, minor_motif;

char major_reg, minor_reg;

int min_rank, max_rank;

int mismatch_thresh;

int numb_of_seqs_ext;



int score_thresh;




int max_occur;

int newton;





char command[ LINE_BUFFER ];






int init;






