#include "param.h"

















#define DEFAULT_REG '\0'













void set_parameters( int argc, char *argv[] )
{

  set_default_params();
  
  set_command( argc, argv );
  
  set_my_parameters( argc, argv );

}

















void set_default_params( void )
{

  extern char mode, minor_mode, combine_mode;
  extern char input_file[ BUFFER ], output_file[ BUFFER ];
  extern char tree_file[ BUFFER ];
  extern char info_file[ BUFFER ], info_out[ BUFFER ];
  extern char motif_file[ BUFFER ];
  extern char region_file[ BUFFER ], minor_region_file[ BUFFER ];

  extern char major_spec[ BUFFER ], minor_spec[ BUFFER ], phylo_spec[ BUFFER ];
  extern char major_tf[ BUFFER ], minor_tf[ BUFFER ];
  extern char major_reg, minor_reg;
  
  extern REGION main_reg;
  
  extern MOTIF major_motif, minor_motif;
  
  extern int max_occur;
  extern int init;
  extern int newton;
  extern int score_thresh;
  extern int min_rank, max_rank;
  extern int mismatch_thresh;
  
  mode = minor_mode = '\0';
  combine_mode = 0;
  
  strcpy( tree_file, DEFAULT_TREE_FILE );
  
  input_file[ 0 ] = output_file[ 0 ] = '\0';
  info_file[ 0 ] = info_out[ 0 ] = '\0';
  region_file[ 0 ] = minor_region_file[ 0 ] = '\0';
  major_spec[ 0 ] = minor_spec[ 0 ] = phylo_spec[ 0 ] = '\0';
  major_tf[ 0 ] = minor_tf[ 0 ] = '\0';
  
  motif_file[ 0 ] = '\0';
  
  major_reg = DEFAULT_REG;
  minor_reg = '\0';
  
  major_motif.cons[ 0 ] = '\0';
  major_motif.length = 0;
  
  minor_motif.cons[ 0 ] = '\0';
  minor_motif.length = 0;
  
  max_occur = 0;
  
  init = 0;
  
  newton = 0;
  
  score_thresh = DEFAULT_SCORE_THRESH;
  
  min_rank = -1;
  max_rank = -1;
  
  mismatch_thresh = 1;
  
  set_main_reg( &main_reg );

}














void set_command( int argc, char *argv[] )
{

  extern char command[ LINE_BUFFER ];
  
  char str[ BUFFER ];
  
  int i;
  
  for ( i = 0; i < argc; i++ ) {
  
    sprintf( str, "%s ", argv[ i ] );
    
    strcat( command, str );
  
  }
  
}









void set_main_reg( REGION *main_reg )
{

  /*  Defined in '../ext/struct.h':  */
  main_reg->a = MAIN_REG_START;
  main_reg->b = MAIN_REG_END;
  main_reg->reg = MAIN_REG_MODE;

}











void set_my_parameters( int argc, char *argv[] )
{

  extern char input_file[ BUFFER ], seq_align_file[ BUFFER ];
  extern MOTIF major_motif;
  
  char cons[ BUFFER ];
  
  int i;
  
  for ( i = 1; i < argc; i += 2 )
    set_single_param( argv[ i ], argv[ i + 1 ] );
    
  upper( cons, major_motif.cons );
    
  sprintf( input_file, "%s.%s", seq_align_file, cons );

}














void set_single_param( char *x, char *arg )
{

  switch ( *(x + 1) ) {
  
  case 'm' :
    set_mode( arg );
    break;
    
  case 'M' :
    set_minor_mode( arg );
    break;
    
  case 'a' :
    set_major_motif( arg );
    break;
    
  case 'd' :
    set_directory( arg );
    break;
    
  case 's' :
    set_seq_align_file( arg );
    break;
    
  case 'r' :
    set_region_file( arg );
    break;
    
  case 'R' :
    set_minor_region_file( arg );
    break;
    
  case 't' :
    set_tree_file( arg );
    break;
    
  case 'x' :
    set_major_spec( arg );
    break;
    
  case 'k' :
    set_score_thresh( arg );
    break;
    
  case 'b' :
    set_mismatch_thresh( arg );
    break;
    
  case 'q' :
    set_motif_file( arg );
    break;
    
  case 'w' :
    set_region_window( arg );
    break;
    
  default :
    error( 100, "Unrecognized parameter type specified.\n" );
    break;
  
  }

}













void set_mode( char *x )
{

  extern char mode;
  
  mode = *x;

}










void set_minor_mode( char *x )
{

  extern char minor_mode;
  
  minor_mode = *x;

}








void set_major_motif( char *x )
{

  extern MOTIF major_motif;
  
  strcpy( major_motif.cons, x );
  
  major_motif.length = strlen( x );

}











void set_motif_file( char *x )
{

  extern char motif_file[ BUFFER ];
  
  strcpy( motif_file, x );

}








void set_major_spec( char *x )
{

  extern char major_spec[ BUFFER ];
  
  strcpy( major_spec, x );

}











void set_directory( char *x )
{

  extern char directory[ BUFFER ];
  
  strcpy( directory, x );

}




    
    
    
    
    
void set_seq_align_file( char *x )
{

  extern char seq_align_file[ BUFFER ];
  
  strcpy( seq_align_file, x );

}
    








void set_region_file( char *x )
{

  extern char region_file[ BUFFER ];
  
  strcpy( region_file, x );

}







void set_minor_region_file( char *x )
{

  extern char minor_region_file[ BUFFER ];
  
  strcpy( minor_region_file, x );

}








void set_tree_file( char *x )
{

  extern char tree_file[ BUFFER ];
  
  strcpy( tree_file, x );

}








void set_score_thresh( char *x )
{

  extern int score_thresh;
  
  score_thresh = atoi( x );

}













void set_mismatch_thresh( char *x )
{

  extern int mismatch_thresh;
  
  mismatch_thresh = atoi( x );

}










void set_region_window( char *x )
{

  extern int reg_start, reg_end;
  
  char start[ BUFFER ], end[ BUFFER ];
  
  int i;
  
  for ( i = 0; i < strlen( x ); i++ ) {
  
    if ( *(x + i) == ':' ) {
      start[ i ] = '\0';
      strcpy( end, x + i + 1 );
      continue;
    }
    
    start[ i ] = *(x + i);
    
  }
  
  reg_start = atoi( start );
  reg_end = atoi( end );

}










