#include "deriv.h"














/*  For p[ij] derivative checking :  */
#define CHECK 0
#define DIFF_RATIO_THRESH 3e-3
#define MAX_DIFF_THRESH 1e-8























void set_tree_derivs( NODE_LIST *root, MODEL_LIST *model )
{

  if ( root == NULL )
    return;
    
  set_node_derivs( root, model );
  
  int i;
  
  for ( i = 0; i < BRANCH_BUFFER; i++ )
    set_tree_derivs( root->down[ i ], model );

}














void set_node_derivs( NODE_LIST *node, MODEL_LIST *model )
{
  set_node_single_p_deriv( node, model );
}













void set_node_single_p_deriv( NODE_LIST *node, MODEL_LIST *model )
{

  PARAMS *get_params_node_by_numb( NODE_LIST *node, MODEL_LIST *model );
  
  PARAMS *pp = NULL;
  /*  In lhood.c:  */
  pp = (PARAMS *)get_params_node_by_numb( node, model );
  
  /*  In this file, below:  */
  node->my_trans->part = get_part_numb_from_model( node, model );
  
  /*  In this file, below:  */
  set_node_u_deriv( node->my_trans->u_d, pp, node->my_node->t );
  set_node_v_deriv( node->my_trans->v_d, pp, node->my_node->t );
  
  /*  In this file, below:  */
  set_node_u_double_deriv( node->my_trans->u_dd, pp, node->my_node->t );
  set_node_v_double_deriv( node->my_trans->v_dd, pp, node->my_node->t );
  
  /*  In this file, below:  */
  set_p_deriv( node, pp, model->my_model->stats->max_occur );
  
  if ( CHECK ) 
    /*  In this file, below:  */
    check_p_derivs( node, pp, model->my_model->stats->max_occur );

}












int get_part_numb_from_model( NODE_LIST *node, MODEL_LIST *model )
{

  int node_in_part( NODE_LIST *node, PART_LIST *part );

  PART_LIST *part = model->my_model->part_header;
  
  while ( part != NULL ) {
  
    /*  In lhood.c:  */
    if ( node_in_part( node, part ))
      return part->my_part->part;
      
    part = part->next;
  
  }
  
  error( 5500, "Could not find node in a partition: Looking for numb.\n" );

}












void set_node_u_deriv( STAT u[], PARAMS *pp, TIME t )
{

  STAT a = pp->a;
  STAT b = pp->b;
  STAT denom = (a + b) * (a + b);
  
  STAT x = exp( -1.0 * (a + b) * MILLION * t );

  u[ BIRTH ] = (b - (b * (a + b) * t * MILLION * x) - (b * x)) / denom;

  u[ DEATH ] = ((-1.0 * a) - (b * (a + b) * t * MILLION * x) + (a * x)) / denom;

}















void set_node_v_deriv( STAT v[], PARAMS *pp, TIME t )
{

  STAT a = pp->a;
  STAT b = pp->b;
  STAT denom = (a + b) * (a + b);
  
  STAT x = exp( -1.0 * (a + b) * MILLION * t );

  v[ BIRTH ] = (b + (a * (a + b) * MILLION * t * x) - (b * x)) / denom;

  v[ DEATH ] = ((-1.0 * a) + (a * (a + b) * MILLION * t * x) + (a * x)) / denom;

}












void set_node_u_double_deriv( STAT u[ BD ][ BD ], PARAMS *pp, TIME t )
{

  STAT a = pp->a;
  STAT b = pp->b;
  
  STAT c = a + b;
  
  STAT d = b - a;
  
  STAT tt = MILLION * t;
  
  STAT x = exp( -1.0 * c * tt );

  u[ BIRTH ][ BIRTH ] = (-2.0 * b) + (2.0 * b * tt * c * x) + (2.0 * b * x);

  u[ BIRTH ][ DEATH ] = (a - b) + ((b - a) * tt * c * x) + ((b - a) * x);

  u[ DEATH ][ BIRTH ] = (a - b) + ((b - a) * tt * c * x) + ((b - a) * x);

  u[ DEATH ][ DEATH ] = (2.0 * a) - (2.0 * a * tt * c * x) - (2.0 * a * x);

  STAT denom = c * c * c;
  
  STAT add = (b * tt * tt * x) / c;
  
  int i, j;
  
  for ( i = 0; i < BD; i++ ) {
    for ( j = 0; j < BD; j++ ) {
      u[ i ][ j ] /= denom;
      u[ i ][ j ] += add;
    }
  }

}














void set_node_v_double_deriv( STAT v[ BD ][ BD ], PARAMS *pp, TIME t )
{

  STAT a = pp->a;
  STAT b = pp->b;
  
  STAT c = a + b;
  
  STAT d = b - a;
  
  STAT tt = MILLION * t;
  
  STAT x = exp( -1.0 * c * tt );

  v[ BIRTH ][ BIRTH ] = (-2.0 * b) + (2.0 * b * tt * c * x) + (2.0 * b * x);

  v[ BIRTH ][ DEATH ] = (a - b) + ((b - a) * tt * c * x) + ((b - a) * x);

  v[ DEATH ][ BIRTH ] = (a - b) + ((b - a) * tt * c * x) + ((b - a) * x);

  v[ DEATH ][ DEATH ] = (2.0 * a) - (2.0 * a * tt * c * x) - (2.0 * a * x);

  STAT denom = c * c * c;
  
  STAT add = (-1.0 * a * tt * tt * x) / c;
  
  int i, j;
  
  for ( i = 0; i < BD; i++ ) {
    for ( j = 0; j < BD; j++ ) {
      v[ i ][ j ] /= denom;
      v[ i ][ j ] += add;
    }
  }

}

















/*=============================================================================*/
/*=============================================================================*/
/*=============================================================================*/
/*=============================================================================*/
/*=============================================================================*/














void set_p_deriv( NODE_LIST *node, PARAMS *pp, COUNT n )
{

  set_p_single_deriv( node, pp, n );
  
  set_p_second_derivs( node, pp, n );
  
  /*
  int i, j;
  
  for ( i = 0; i < OCCUR_BUFFER; i++ ) {
  
    for ( j = 0; j < OCCUR_BUFFER; j++ )
      printf( "%3d  %3d    %15.5e    %15.5e\n", i, j,
        (float)node->my_trans->p_dd[ 0 ][ 1 ][ i ][ j ],
        (float)node->my_trans->p_dd[ 1 ][ 0 ][ i ][ j ] );
  
  }
  */
  
}














void set_p_single_deriv( NODE_LIST *node, PARAMS *pp, COUNT n )
{

  int end = MIN( n + 3, OCCUR_BUFFER );

  int i;
  
  for ( i = 0; i < end; i++ ) 
    set_p_single_deriv_single_i_val( node, pp, i, end );
    
}










void set_p_single_deriv_single_i_val( NODE_LIST *node, PARAMS *pp, int i, int end )
{

  int j, bd;
  
  STAT uu[ OCCUR_BUFFER ], vv[ OCCUR_BUFFER ];
  STAT uu_d[ OCCUR_BUFFER ], vv_d[ OCCUR_BUFFER ];
  
  get_two_binom_arrays( uu, vv, node, i );
  
  for ( bd = 0; bd < DBUFF; bd++ ) {
    
    get_two_deriv_binom_arrays( uu_d, vv_d, node, bd, i );
    
    for ( j = 0; j < end; j++ ) 
      node->my_trans->p_d[ bd ][ i ][ j ] = 
        get_p_deriv_elem( node, i, j, bd, uu, vv, uu_d, vv_d );  
    
  }
  
}

















void set_p_second_derivs( NODE_LIST *node, PARAMS *pp, int n )
{

  int end = MIN( n + 3, OCCUR_BUFFER );

  int i, j, bd_i, bd_j;
  
  STAT uu[ OCCUR_BUFFER ], vv[ OCCUR_BUFFER ];
  STAT uu_d[ OCCUR_BUFFER ], vv_d[ OCCUR_BUFFER ];
  STAT uu_f[ OCCUR_BUFFER ], vv_f[ OCCUR_BUFFER ];
  STAT uu_dd[ OCCUR_BUFFER ], vv_dd[ OCCUR_BUFFER ];
  
  for ( i = 0; i < end; i++ ) {
  
    get_two_binom_arrays( uu, vv, node, i );
   
    for ( bd_i = 0; bd_i < DBUFF; bd_i++ ) {

      get_two_deriv_binom_arrays( uu_d, vv_d, node, bd_i, i );

      for ( bd_j = 0; bd_j < DBUFF; bd_j++ ) {
    
        get_two_deriv_binom_arrays( uu_f, vv_f, node, bd_j, i );   
      
        get_two_dd_binom_arrays( uu_dd, vv_dd, node, bd_i, bd_j, i );
      
        for ( j = 0; j < end; j++ ) 
          node->my_trans->p_dd[ bd_i ][ bd_j ][ i ][ j ] = 
            get_p_double_deriv_elem( node, i, j, bd_i, bd_j, uu, vv, uu_d, vv_d,
              uu_f, vv_f, uu_dd, vv_dd );  
      
      }

    }
  }

}










void get_two_binom_arrays( STAT uu[], STAT vv[], NODE_LIST *node, int i )
{ 
  get_binom_array( uu, node->my_trans->u, i, OCCUR_BUFFER );
  get_binom_array( vv, node->my_trans->v, node->my_trans->n - i, OCCUR_BUFFER );
}









void get_two_deriv_binom_arrays( STAT uu_d[], STAT vv_d[], NODE_LIST *node,
  int bd, int i )
{

  get_deriv_binom_array( uu_d, node->my_trans->u, node->my_trans->u_d[ bd ],
    i, OCCUR_BUFFER );

  get_deriv_binom_array( vv_d, node->my_trans->v, node->my_trans->v_d[ bd ],
    node->my_trans->n - i, OCCUR_BUFFER );

}  














void get_two_dd_binom_arrays( STAT uu_dd[], STAT vv_dd[], NODE_LIST *node,
  int bd_i, int bd_j, int i )
{

  get_dd_binom_array( uu_dd, node->my_trans->u, node->my_trans->u_d[ bd_i ],
    node->my_trans->u_d[ bd_j ], node->my_trans->u_dd[ bd_i ][ bd_j ],
    i, OCCUR_BUFFER );

  get_dd_binom_array( vv_dd, node->my_trans->v, node->my_trans->v_d[ bd_i ],
    node->my_trans->v_d[ bd_j ], node->my_trans->v_dd[ bd_i ][ bd_j ],
    node->my_trans->n - i, OCCUR_BUFFER );

}














STAT get_p_deriv_elem( NODE_LIST *node, int i, int j, int bd, STAT uu[], STAT vv[],
  STAT uu_d[], STAT vv_d[] )
{

  int k, end = MIN( i, j );
  
  STAT p = 0.0;
  
  for ( k = 0; k <= end; k++ )
    p += (vv[ j - k ] * uu_d[ k ]) + (vv_d[ j - k ] * uu[ k ]);

  return p;

}















void get_binom_array( STAT array[], STAT pr, COUNT n, COUNT max_k )
{

  STAT x = pow( 1.0 - pr, n );
  
  STAT step = pr / (1.0 - pr);
  STAT numer = n;
  STAT denom = 1.0;
  
  int end = MIN( max_k - 1, n );
  
  int k;
  
  for ( k = 0; k <= end; k++ ) {
  
    array[ k ] = x;
    
    x *= step * (numer--) / (denom++);
    
  }
  
  fill_zero_array( array, k, max_k );

}













void fill_zero_array( STAT array[], COUNT start, COUNT max )
{

  int k;
  
  for ( k = start; k < max; k++ )
    array[ k ] = 0.0;

}












void get_deriv_binom_array( STAT d_array[], STAT pr, STAT deriv, COUNT n,
  COUNT max_k )
{

  if ( n == 0 )
    /*  In this file, above:  */
    fill_zero_array( d_array, 0, max_k );
    
  else 
    /*  In this file, below:  */
    get_my_deriv_binom_array( d_array, pr, deriv, n, max_k );
    
}













void get_my_deriv_binom_array( STAT d_array[], STAT pr, STAT deriv, COUNT n,
  COUNT max_k )
{

  STAT x = pow( 1.0 - pr, n - 1.0 ) / pr;
  
  STAT step = pr / (1.0 - pr);
  STAT numer = n;
  STAT denom = 1.0;
  
  int end = MIN( n, max_k - 1 );
  
  int k;
  
  for ( k = 0; k <= end; k++ ) {
  
    d_array[ k ] = x * ((STAT)k - ((STAT)n * pr)) * deriv;
    
    x *= step * (numer--) / (denom++);
    
  }
  
  fill_zero_array( d_array, k, max_k );

}















void get_dd_binom_array( STAT array[], STAT pr, STAT da, STAT db, STAT dd, int n, 
  int max_k )
{

  if ( n == 0 )
    /*  In this file, above:  */
    fill_zero_array( array, 0, max_k );
    
  else 
    /*  In this file, below:  */
    get_my_dd_binom_array( array, pr, da, db, dd, n, max_k );  

}
















void get_my_dd_binom_array( STAT array[], STAT pr, STAT da, STAT db, STAT dd,
  int n, int max_k )
{

  STAT x = pow( 1.0 - pr, n - 2.0 ) / (pr * pr);
  
  STAT y = pr * (1.0 - pr);
  
  STAT step = pr / (1.0 - pr);
  
  STAT numer = n;
  STAT denom = 1.0;
  
  int end = MIN( n, max_k - 1 );
  
  int k;
  
  for ( k = 0; k <= end; k++ ) {
    
    STAT a = x * (k - (n * pr)) * y * dd;
    
    STAT b = (k - 1) * (1.0 - pr) * (k - (n * pr));
    
    b -= (n - k - 1) * pr * (k - (n * pr));
    b -= n * pr * (1.0 - pr);
    
    array[ k ] = a + (x * b * da * db);
      
    x *= step * (numer--) / (denom++);
    
  }
  
  fill_zero_array( array, k, max_k );

}












STAT get_p_double_deriv_elem( NODE_LIST *node, int i, int j, int bd_i, int bd_j, 
  STAT uu[], STAT vv[], STAT uu_d[], STAT vv_d[], STAT uu_f[], STAT vv_f[],
  STAT uu_dd[], STAT vv_dd[] )
{

  STAT p = 0.0;

  int k, end = MIN( i, j );
  
  for ( k = 0; k <= j; k++ ) {
  
    int a = j - k;
    
    p += (vv_dd[ a ] * uu[ k ]) + (uu_dd[ k ] * vv[ a ]);
    p += (uu_d[ k ] * vv_f[ a ]) + (vv_d[ a ] * uu_f[ k ]);
    
  }

  return p;

}

















/*=============================================================================*/
/*=============================================================================*/
/*=============================================================================*/
/*=============================================================================*/
/*=============================================================================*/



















void check_p_derivs( NODE_LIST *node, PARAMS *pp, int max )
{

  check_p_single_deriv( node, pp, max );
  
  check_p_double_deriv( node, pp, max );
  
}













void check_p_single_deriv( NODE_LIST *node, PARAMS *pp, int max )
{

  int bd;
  
  for ( bd = 0; bd < BD; bd++ )
    check_p_single_deriv_single_param( node, pp, bd, max );

}











void check_p_single_deriv_single_param( NODE_LIST *node, PARAMS *pp, int bd,
  int max )
{

  int flag = 0;
  int finished;

  STAT p_deriv;

  int end = 2 * max;

  int i, j;
  
  for ( i = 0; i < end; i++ ) {
  
    for ( j = 0; j < end; j++ ) 
      flag = check_p_single_deriv_ij( node, pp, bd, i, j, flag );
    
    if ( flag )
      error( 5188, "P-derivatives are different.\n" );

  }
  
}











int check_p_single_deriv_ij( NODE_LIST *node, PARAMS *pp, int bd, int i, int j,
  int flag )
{

//  STAT get_p_numerical_deriv( NODE_LIST *node, PARAMS *pp, int bd, int i, int j,
  //  int *finished );
    
  int finished;
  
  STAT p_deriv = 0.0;
  
  /*  In num_d.c:  */
//  STAT p_deriv = get_p_numerical_deriv( node, pp, bd, i, j, &finished );
      
  printf( "%3d    %3d      %15.5e    %15.5e\n", i, j, 
    (double)node->my_trans->p_d[ bd ][ i ][ j ], (double)p_deriv );   
      
  if ( ! finished )
    return flag;
        
  if ( ! same_val( p_deriv, node->my_trans->p_d[ bd ][ i ][ j ] ))
    return 1;
  
  return flag;
  
}













int same_val( STAT a, STAT b )
{

  if (( fabs( a ) < MAX_DIFF_THRESH ) && ( fabs( b ) < MAX_DIFF_THRESH ))
    return 1;

  STAT diff = fabs( a - b );
  
  if ( diff >= fabs( a ) * DIFF_RATIO_THRESH ) 
    return 0;
    
  if ( diff >= fabs( b ) * DIFF_RATIO_THRESH )
    return 0;
    
  return 1;

}









void check_p_double_deriv( NODE_LIST *node, PARAMS *pp, int max )
{

  int bd_i, bd_j;
  
  for ( bd_i = 0; bd_i < BD; bd_i++ ) {
  
    for ( bd_j = 0; bd_j < BD; bd_j++ )
      check_p_double_deriv_single_pp( node, pp, bd_i, bd_j, max );
  
  }

}













void check_p_double_deriv_single_pp( NODE_LIST *node, PARAMS *pp, int bd_i,
  int bd_j, int max )
{

  int flag = 0;
  int finished;

  STAT p_deriv;

  int end = 2 * max;

  int i, j;
  
  for ( i = 0; i < end; i++ ) {
  
    for ( j = 0; j < end; j++ ) 
      ;
//      flag = check_p_double_deriv_ij( node, pp, bd_i, bd_j, i, j, flag );
    
    if ( flag )
      error( 5188, "P-derivatives are different.\n" );

  }
  
}











int check_p_double_deriv_ij( NODE_LIST *node, PARAMS *pp, int bd_i, int bd_j,
  int i, int j, int flag )
{

//  STAT get_p_numerical_double_deriv( NODE_LIST *node, PARAMS *pp, int bd_i, int bd_j,
  //  int i, int j, int *finished );
    
  int finished;
  
  STAT p_deriv = 0.0;
  
  /*  In num_d.c:  */
//  STAT p_deriv = get_p_numerical_double_deriv( node, pp, bd_i, bd_j, i, j, 
  //  &finished );
      
  printf( "double :  %3d    %3d      %15.5e    %15.5e   (%.1f)\n", i, j, 
    (double)node->my_trans->p_dd[ bd_i ][ bd_j ][ i ][ j ], (double)p_deriv,
    (float)node->my_node->t );   
      
  if ( ! finished )
    return flag;  
        
  if ( ! same_val( p_deriv, node->my_trans->p_dd[ bd_i ][ bd_j ][ i ][ j ] )) {
    printf( "\n\n\nHere!!!!\n\n" );
    return 1;
  }
  
  return flag;
  
  


}







