#include "hess_mx.h"

















#define OUTPUT 0














void get_hessian_and_grad( MX *mx, VEC *f, TREE_DATA *d )
{

  void set_trans_probs_tree_w_model( NODE_LIST *root, MODEL_LIST *model );
  void set_tree_derivs( NODE_LIST *root, MODEL_LIST *model );
  
  /*  In lhood.c:  */
  set_trans_probs_tree_w_model( d->root, d->model );
  
  /*  In deriv.c:  */
  set_tree_derivs( d->root, d->model );
  
  printf( "Creating new hessian matrix and gradient vector...\n\n" );
  
  /*  In this file, below:  */
  get_my_hessian_and_grad( mx, f, d );
  
  printf( "Hessian matrix J :\n" );
  print_mx( mx );
  printf( "Gradient vector F :\n" );
  print_vec( f );
  
}
















void get_my_hessian_and_grad( MX *mx, VEC *f, TREE_DATA *d )
{

  void sym_mx( MX *mx );

  int n = d->model->my_model->part_n;
  
  int i, j, pi, bi, pj, bj;
  
  for ( pi = 0; pi < n; pi++ ) {  
    for ( bi = 0; bi < BD; bi++ ) 
      get_hessian_and_grad_single_row( mx, f, d, pi, bi, n );
  }
  
  /*  In mx.c:  */
  sym_mx( mx );  

}

















void get_hessian_and_grad_single_row( MX *mx, VEC *f, TREE_DATA *d, int pi,
  int bi, int n )
{

  int i = (BD * pi) + bi;
  
  int pj, bj, j;
  
  for ( pj = 0; pj < n; pj++ ) {
    for ( bj = 0; bj < BD; bj++ ) {
    
      j = (BD * pj) + bj;
      
      if ( i > j )
        continue;
        
      /*  In this file, below:  */  
      mx->mx[ i ][ j ] = get_hessian_elem_and_grad( d, f, i, j, pi, bi, pj, bj );
      
      if ( OUTPUT )
        printf( "j[ %d ][ %d ] = %15.5e\n", i, j, (double)mx->mx[ i ][ j ] );
      
    }
  }

}















STAT get_hessian_elem_and_grad( TREE_DATA *d, VEC *f, int i, int j, int pi, int bi, 
  int pj, int bj )
{

  TREE_DATA *create_tree_data( TREE_DATA *x );
  void copy_tree_data( TREE_DATA *d, TREE_DATA *a );
  
  TREE_DATA *data = NULL;
  data = (TREE_DATA *)create_tree_data( data );
  
  /*  In model_set.c:  */  
  copy_tree_data( data, d );
  
  /*  In this file, below:  */
  STAT x = get_my_hessian_elem_and_grad( data, f, i, j, pi, bi, pj, bj );
  
  kill_all_tree_data( data );
  
  return x;

}
















STAT get_my_hessian_elem_and_grad( TREE_DATA *data, VEC *f, int i, int j, int pi, 
  int bi, int pj, int bj )
{

  if ( i == 0 )
    f->v[ j ] = 0.0;

  STAT x = 0.0;
  
  while ( data->occur->x[ 0 ] != NULL )
    x += get_hessian_elem_and_grad_single_seq_and_step( data, f, i, j, pi, bi, 
      pj, bj );

  return x;

}















STAT get_hessian_elem_and_grad_single_seq_and_step( TREE_DATA *data, VEC *f,
  int i, int j, int pi, int bi, int pj, int bj )
{

  void step_occur_cell( OCCUR_CELL *occur );

  /*  In this file, below:  */
  STAT x = get_hessian_elem_and_grad_single_seq( data, f, i, j, pi, bi, pj, bj );
  
  /*  In lhood.c:  */
  step_occur_cell( data->occur );
  
  return x;

}















STAT get_hessian_elem_and_grad_single_seq( TREE_DATA *data, VEC *f, int i, int j,
  int pi, int bi, int pj, int bj )
{

  int numb_of_spec_in_align( OCCUR_CELL *occur );
  void clear_tree_all( NODE_LIST *root );
  void set_leaf_occurs( NODE_LIST *root, OCCUR_CELL *occur );

  /*  In lhood.c:  */
  if ( numb_of_spec_in_align( data->occur ) < 2 )
    return 0.0;
    
  /*  In set.c:  */
  clear_tree_all( data->root );
  
  /*  In set.c:  */
  set_leaf_occurs( data->root, data->occur );
    
  /*  In this file, below:  */
  return get_my_hessian_elem_and_grad_single_seq( data, f, i, j, pi, bi, pj, bj );

}















STAT get_my_hessian_elem_and_grad_single_seq( TREE_DATA *data, VEC *f, int i, 
  int j, int pi, int bi, int pj, int bj )
{

  NODE_LIST *get_spec_header( NODE_LIST *node );
  STAT get_array_numerical_sum( STAT array[], int mean, int max );
  STAT get_node_info_lhood( NODE_INFO *x, NODE_INFO *prior );
  
  NODE_LIST *root = data->root;
  
  /*  In ../tree/get.c:  */
  NODE_LIST *node = (NODE_LIST *)get_spec_header( root );
  
  while ( node != NULL )
    /*  In this file, below:  */
    get_hessian_elem_ct_d_dd_single_node_and_step( &node, pi, bi, pj, bj );
    
  /*  In lhood.c:  */
  STAT a = get_node_info_lhood( root->my_info, root->prior );
  STAT b = get_node_info_lhood( root->my_dd, root->prior );
  STAT c1 = get_node_info_lhood( root->my_d[ 0 ], root->prior );
  STAT c2 = get_node_info_lhood( root->my_d[ 1 ], root->prior );
  
  if ( i == 0 )
    f->v[ j ] += c2 / a;
    
  return ((a * b) - (c1 * c2)) / (a * a);

}















void get_hessian_elem_ct_d_dd_single_node_and_step( NODE_LIST **node, int pi, 
  int bi, int pj, int bj )
{
  get_hessian_elem_ct_d_dd_single_node( *node, pi, bi, pj, bj );
  *node = (*node)->next;
}















void get_hessian_elem_ct_d_dd_single_node( NODE_LIST *node, int pi, int bi, 
  int pj, int bj )
{

  if (( node->my_info->use == 0 ) || ( node->my_node->leaf ))
    return;
    
  get_my_hessian_elem_ct_d_dd_single_node( node, pi, bi, pj, bj );

}















void get_my_hessian_elem_ct_d_dd_single_node( NODE_LIST *node, int pi, int bi, 
  int pj, int bj )
{

  STAT get_node_lhood_single_ct( NODE_LIST *node, int i );
  STAT get_deriv_single_ct( NODE_LIST *node, int d_index, int i, int p, int b );
  int ct_done( STAT x, STAT total, STAT *last );

  STAT ct_total = 0.0, ct_last = 0.0;
  STAT dd_total = 0.0, dd_last = 0.0;
  
  int i;
  
  for ( i = 0; i < OCCUR_BUFFER; i++ ) {
  
    /*  In lhood.c:  */
    node->my_info->x[ i ] = get_node_lhood_single_ct( node, i );
  
    /*  In grad_vec.c:  */
    node->my_d[ 0 ]->x[ i ] = get_deriv_single_ct( node, 0, i, pi, bi );
    
    if (( pi == pj ) && ( bi == bj ))
      node->my_d[ 1 ]->x[ i ] = node->my_d[ 0 ]->x[ i ];
    else
      /*  In grad_vec.c:  */
      node->my_d[ 1 ]->x[ i ] = get_deriv_single_ct( node, 1, i, pj, bj );

    /*  In this file, below:  */
    node->my_dd->x[ i ] = get_dd_single_ct( node, i, pi, bi, pj, bj );
    
    STAT cx = node->my_info->x[ i ] * node->prior->x[ i ];
    STAT dd = node->my_dd->x[ i ];
    
    ct_total += cx;
    dd_total += dd;
    
    /*  In lhood.c:  */
    if (( ct_done( cx, ct_total, &ct_last )) && ( ct_done( dd, dd_total, &dd_last )))
      break;
      
    node->my_info->max = i;
    
  }

}











STAT get_dd_single_ct( NODE_LIST *node, int i, int pi, int bi, int pj, int bj )
{

  STAT x = node->my_info->x[ i ];

  if ( x == 0.0 )
    return 0.0;
    
  STAT da = node->my_d[ 0 ]->x[ i ];
  STAT db = node->my_d[ 1 ]->x[ i ];
    
  STAT dd = get_dd_term_single_ct( node, i, pi, bi, pj, bj );
  
  return ( x * dd ) + ( da * db / x );

}















STAT get_dd_term_single_ct( NODE_LIST *node, int i, int pi, int bi, int pj,
  int bj )
{

  STAT sum = 0.0;
  
  int k;
  
  for ( k = 0; k < BRANCH_BUFFER; k++ ) {
  
    NODE_LIST *d = node->down[ k ];
    
    if ( d == NULL )
      break;
      
    sum += get_dd_term_single_daughter_node( d, i, pi, bi, pj, bj );
  
  }
  
  return sum;

}












STAT get_dd_term_single_daughter_node( NODE_LIST *d, int i, int pi, int bi,
  int pj, int bj )
{

  STAT get_node_lhood_single_ct_pair( NODE_LIST *d, int i, int j );
  STAT get_deriv_single_daughter_numer( NODE_LIST *d, int d_index, int i, int j, 
    int p, int b );

  if ( d->my_info->use == 0 )
    return 0.0;
    
  STAT a = 0.0, b = 0.0, c[ DBUFF ] = { 0.0, 0.0 };

  int j;
  
  for ( j = 0; j <= d->my_info->max; j++ ) {
  
    /*  In lhood.c:  */
    a += get_node_lhood_single_ct_pair( d, i, j );
    
    /*  In this file, below:  */
    b += get_second_deriv_dd_single_ct( d, i, j, pi, bi, pj, bj );
    
    /*  In grad_vec.c:  */
    c[ 0 ] += get_deriv_single_daughter_numer( d, 0, i, j, pi, bi );
    c[ 1 ] += get_deriv_single_daughter_numer( d, 1, i, j, pj, bj );
    
  }
  
  /*  In this file, below:  */
  return get_second_deriv_term( a, b, c );

}













/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/













void get_hessian_mx( MX *mx, TREE_DATA *d )
{

  void set_trans_probs_tree_w_model( NODE_LIST *root, MODEL_LIST *model );
  void set_tree_derivs( NODE_LIST *root, MODEL_LIST *model );
  
  /*  In lhood.c:  */
  set_trans_probs_tree_w_model( d->root, d->model );
  
  /*  In deriv.c:  */
  set_tree_derivs( d->root, d->model );
  
  printf( "Creating new hessian matrix...\n\n" );
  
  /*  In this file, below:  */
  get_my_hessian_mx( mx, d );
  
  printf( "Hessian matrix J :\n" );
  print_mx( mx );
  
}













void get_my_hessian_mx( MX *mx, TREE_DATA *d )
{
  
  void sym_mx( MX *mx );

  int n = d->model->my_model->part_n;
  
  int i, j, pi, bi, pj, bj;
  
  for ( pi = 0; pi < n; pi++ ) {  
    for ( bi = 0; bi < BD; bi++ ) 
      get_hessian_mx_single_row( mx, d, pi, bi, n );
  }
  
  /*  In mx.c:  */
  sym_mx( mx );  
  
}











void get_hessian_mx_single_row( MX *mx, TREE_DATA *d, int pi, int bi, int n )
{

  int i = (BD * pi) + bi;
  
  int pj, bj, j;
  
  for ( pj = 0; pj < n; pj++ ) {
    for ( bj = 0; bj < BD; bj++ ) {
    
      j = (BD * pj) + bj;
      
      if ( i > j )
        continue;
        
      /*  In this file, below:  */  
      mx->mx[ i ][ j ] = get_hessian_elem( d, pi, bi, pj, bj );     
      
      printf( "j[ %d ][ %d ] = %15.5e\n", i, j, (double)mx->mx[ i ][ j ] );
 
    
    }
  }

}








STAT get_hessian_elem( TREE_DATA *d, int pi, int bi, int pj, int bj )
{

  TREE_DATA *create_tree_data( TREE_DATA *x );
  void copy_tree_data( TREE_DATA *d, TREE_DATA *a );
  
  TREE_DATA *data = NULL;
  data = (TREE_DATA *)create_tree_data( data );
  
  /*  In model_set.c:  */  
  copy_tree_data( data, d );
  
  /*  In this file, below:  */
  STAT x = get_my_hessian_elem( data, pi, bi, pj, bj );
  
  kill_all_tree_data( data );
  
  return x;

}
















STAT get_my_hessian_elem( TREE_DATA *data, int pi, int bi, int pj, int bj )
{

  STAT x = 0.0;
  
  while ( data->occur->x[ 0 ] != NULL )
    x += get_hessian_elem_single_seq_and_step( data, pi, bi, pj, bj );

  return x;

}













STAT get_hessian_elem_single_seq_and_step( TREE_DATA *data, int pi, int bi,
  int pj, int bj )
{

  void step_occur_cell( OCCUR_CELL *occur );

  /*  In this file, below:  */
  STAT x = get_hessian_elem_single_seq( data, pi, bi, pj, bj );
  
  /*  In lhood.c:  */
  step_occur_cell( data->occur );
  
  return x;

}















STAT get_hessian_elem_single_seq( TREE_DATA *data, int pi, int bi, int pj, int bj )
{

  int numb_of_spec_in_align( OCCUR_CELL *occur );
  void clear_tree_all( NODE_LIST *root );
  void set_leaf_occurs( NODE_LIST *root, OCCUR_CELL *occur );

  /*  In lhood.c:  */
  if ( numb_of_spec_in_align( data->occur ) < 2 )
    return 0.0;
    
  /*  In set.c:  */
  clear_tree_all( data->root );
  
  /*  In set.c:  */
  set_leaf_occurs( data->root, data->occur );
    
  /*  In this file, below:  */
  return get_my_hessian_elem_single_seq( data, pi, bi, pj, bj );

}















STAT get_my_hessian_elem_single_seq( TREE_DATA *data, int pi, int bi, int pj,
  int bj )
{

  NODE_LIST *get_spec_header( NODE_LIST *node );
  STAT get_array_numerical_sum( STAT array[], int mean, int max );
  STAT get_node_info_lhood( NODE_INFO *x, NODE_INFO *prior );
  
  NODE_LIST *root = data->root;
  
  /*  In ../tree/get.c:  */
  NODE_LIST *node = (NODE_LIST *)get_spec_header( root );
  
  while ( node != NULL )
    /*  In this file, below:  */
    get_hessian_elem_single_node_and_step( &node, pi, bi, pj, bj );

  int center = (int)root->my_info->mean, n = OCCUR_BUFFER;
  
  STAT a = get_node_info_lhood( root->my_info, root->prior );
  STAT b = get_node_info_lhood( root->my_dd, root->prior );
  STAT c1 = get_node_info_lhood( root->my_d[ 0 ], root->prior );
  STAT c2 = get_node_info_lhood( root->my_d[ 1 ], root->prior );

  /*  In grad_vec.c:  
  STAT a = get_array_numerical_sum( root->my_info->x, center, n );
  STAT b = get_array_numerical_sum( root->my_dd->x, center, n );
  STAT c1 = get_array_numerical_sum( root->my_d[ 0 ]->x, center, n );
  STAT c2 = get_array_numerical_sum( root->my_d[ 1 ]->x, center, n );
  */
  return ((a * b) - (c1 * c2)) / (a * a);

}















void get_hessian_elem_single_node_and_step( NODE_LIST **node, int pi, int bi,
  int pj, int bj )
{
  get_hessian_elem_single_node( *node, pi, bi, pj, bj );
  *node = (*node)->next;
}













void get_hessian_elem_single_node( NODE_LIST *node, int pi, int bi, int pj, int bj )
{

  void set_ct_single_node( NODE_LIST *node );
  void set_derivative_vals_single_node( NODE_LIST *node, int d_index, int p, int b );

  if (( node->my_info->use == 0 ) || ( node->my_node->leaf ))
    return;
    
  /*  In grad_vec.c:  */  
  set_ct_single_node( node );
  
  /*  In grad_vec.c:  */
  set_derivative_vals_single_node( node, 0, pi, bi ); 
  
  if (( pi == pj ) && ( bi == bj ))
    *(node->my_d[ 1 ]) = *(node->my_d[ 0 ]);
  else
    /*  In grad_vec.c:  */ 
    set_derivative_vals_single_node( node, 1, pj, bj );
    
  /*  In this file, below:  */
  set_second_deriv_vals_single_node( node, pi, bi, pj, bj );

}
















/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/




















void set_second_deriv_vals_single_node( NODE_LIST *node, int pi, int bi, int pj, 
  int bj )
{

  int center = (int)node->my_info->mean;
  
  STAT total = 0.0;
  
  set_second_deriv_vals_single_node_single_dir( node, center, &total, pi, bi, pj,
    bj, INCREM );
    
  set_second_deriv_vals_single_node_single_dir( node, center - 1, &total, pi, bi,
    pj, bj, DECREM );

}
















void set_second_deriv_vals_single_node_single_dir( NODE_LIST *node, int start,
  STAT *total, int pi, int bi, int pj, int bj, int dir )
{

  int ct_done( STAT x, STAT total, STAT *last );

  STAT my_total = *total;
  
  STAT last = 0.0;
  
  int i;
  
  for ( i = start; i < OCCUR_BUFFER && i >= 0; i += dir ) {
  
    /*  In this file, below:  */
    node->my_dd->x[ i ] = get_second_deriv_single_ct( node, i, pi, bi, pj, bj );
    
    my_total += node->my_dd->x[ i ];
    
  }
  
  *total = my_total;

}














STAT get_second_deriv_single_ct( NODE_LIST *node, int i, int pi, int bi, int pj,
  int bj )
{

  STAT x = node->my_info->x[ i ];

  if ( x == 0.0 )
    return 0.0;
    
  STAT deriv_a = node->my_d[ 0 ]->x[ i ];
  STAT deriv_b = node->my_d[ 1 ]->x[ i ];
    
  STAT dd = get_second_deriv_term_single_ct( node, i, pi, bi, pj, bj );
  
  return ( x * dd ) + ( deriv_a * deriv_b / x );

}












STAT get_second_deriv_term_single_ct( NODE_LIST *node, int i, int pi, int bi, 
  int pj, int bj )
{

  STAT sum = 0.0;
  
  int k;
  
  for ( k = 0; k < BRANCH_BUFFER; k++ ) {
  
    NODE_LIST *d = node->down[ k ];
    
    if ( d == NULL )
      break;
      
    if ( d->my_info->use == 0 )
      continue;
      
    sum += get_second_deriv_term_single_daughter_node( d, i, pi, bi, pj, bj );
  
  }
  
  return sum;

}















STAT get_second_deriv_term_single_daughter_node( NODE_LIST *d, int i, int pi, 
  int bi, int pj, int bj )
{

  int center = (int)d->my_info->mean;
  
  STAT a = 0.0, b = 0.0, c[ DBUFF ] = { 0.0, 0.0 };
  
  get_second_deriv_term_single_daughter_single_dir( d, &a, &b, c, center, i, 
    pi, bi, pj, bj, INCREM );

  return get_second_deriv_term_single_daughter_single_dir( d, &a, &b, c,
    center - 1, i, pi, bi, pj, bj, DECREM );

}














STAT get_second_deriv_term_single_daughter_single_dir( NODE_LIST *d, STAT *a,
  STAT *b, STAT c[], int start, int i, int pi, int bi, int pj, int bj, int dir )
{

  int ct_done( STAT x, STAT total, STAT *last );
  
  STAT last = 0.0, total = 0.0;
  
  STAT my_a = *a, my_b = *b, my_c[ DBUFF ] = { c[ 0 ], c[ 1 ] };

  int j;
  
  for ( j = start; j < OCCUR_BUFFER && j >= 0; j += dir ) {
  
    
    if ( d->my_info->x[ j ] == 0.0 ) {
      if ( *a != 0.0 )
        break;
      continue;
    }
  
    /*  In this file, below:  */
    STAT curr_a = get_second_deriv_s_single_ct( d, i, j );
    STAT curr_b = get_second_deriv_dd_single_ct( d, i, j, pi, bi, pj, bj );
    
    STAT curr_c[ DBUFF ];
    
    /*  In this file, below:  */
    curr_c[ 0 ] = get_second_deriv_d_single_ct( d, 0, i, j, pi, bi );
    curr_c[ 1 ] = get_second_deriv_d_single_ct( d, 1, i, j, pj, bj );
    
    my_a += curr_a;
    my_b += curr_b;
    my_c[ 0 ] += curr_c[ 0 ];
    my_c[ 1 ] += curr_c[ 1 ];
    
    /*  In this file, below:  */
    STAT curr = get_second_deriv_term( curr_a, curr_b, curr_c );    
    total = get_second_deriv_term( my_a, my_b, my_c );
    
    /*  In grad_vec.c:  */
    if ( ct_done( curr, total, &last ))
      break;
    
  }
  
  *a = my_a;
  *b = my_b;
  c[ 0 ] = my_c[ 0 ];
  c[ 1 ] = my_c[ 1 ];
  
  /*  In this file, below:  */
  return get_second_deriv_term( *a, *b, c );

}













STAT get_second_deriv_term( STAT a, STAT b, STAT c[] )
{
  return (b / a) - ((c[ 0 ] * c[ 1 ]) / (a * a));
}









STAT get_second_deriv_s_single_ct( NODE_LIST *d, int i, int j )
{
  return d->my_trans->p[ i ][ j ] * d->my_info->x[ j ];
}












STAT get_second_deriv_d_single_ct( NODE_LIST *d, int d_index, int i, int j, 
  int p, int b )
{

  STAT get_deriv_single_daughter_numer( NODE_LIST *d, int d_index, int i, int j, 
    int p, int b );

  /*  In grad_vec.c:  */
  return get_deriv_single_daughter_numer( d, d_index, i, j, p, b );

}









STAT get_second_deriv_dd_single_ct( NODE_LIST *d, int i, int j, int pi, int bi, 
  int pj, int bj )
{

  STAT x = d->my_trans->p[ i ][ j ] * d->my_dd->x[ j ];

  if ( d->my_trans->part == pi )
    x += d->my_trans->p_d[ bi ][ i ][ j ] * d->my_d[ 1 ]->x[ j ];
    
  if ( d->my_trans->part == pj )
    x += d->my_trans->p_d[ bj ][ i ][ j ] * d->my_d[ 0 ]->x[ j ];
        
  if (( d->my_trans->part == pi ) && ( pi == pj ))
    x += d->my_trans->p_dd[ bi ][ bj ][ i ][ j ] * d->my_info->x[ j ];

  return x;
  
}


















