#include "mx.h"

#define THIS_FILE "mx.c"














#define CHECK_INV 1
#define STAT_LOW 1e-12














void matrix_mult( MX *c, MX *a, MX *b )
{

  int i, j;
  
  for ( i = 0; i < a->i; i++ ) {
  
    for ( j = 0; j < b->j; j++ ) 
      c->mx[ i ][ j ] = get_matrix_mult_elem( a, b, i, j );
  
  }

}















STAT get_matrix_mult_elem( MX *a, MX *b, int i, int j )
{

  STAT elem = 0.0;

  int k;
  
  for ( k = 0; k < a->j; k++ )
    elem += a->mx[ i ][ k ] * b->mx[ k ][ j ];
    
  return elem;

}













void mx_times_vec( VEC *x, MX *a, VEC *b )
{

  int i;
  
  for ( i = 0; i < a->i; i++ )
    x->v[ i ] = get_mx_times_vec_elem( a, b, i );

}














STAT get_mx_times_vec_elem( MX *a, VEC *b, int i )
{

  STAT elem = 0.0;
  
  int k;
  
  for ( k = 0; k < b->i; k++ )
    elem += a->mx[ i ][ k ] * b->v[ k ];
    
  return elem;

}











void vec_mult_mx( VEC *c, VEC *s, MX *a )
{

  int i, j;
  
  for ( i = 0; i < c->i; i++ )
    c->v[ i ] = get_vec_mult_mx_elem( s, a, i );

}












STAT get_vec_mult_mx_elem( VEC *s, MX *a, int i )
{

  STAT elem = 0.0;
  
  int j;
  
  for ( j = 0; j < a->i; j++ )
    elem += s->v[ j ] * a->mx[ j ][ i ];
    
  return elem;

}









void vec_mult_scalar( VEC *v, STAT c )
{

  int i;
  
  for ( i = 0; i < v->i; i++ )
    v->v[ i ] *= c;

}













void add_vec( VEC *c, VEC *a, VEC *b )
{

  int i;
  
  for ( i = 0; i < a->i; i++ )
    c->v[ i ] = a->v[ i ] + b->v[ i ];

}











void subtract_vec( VEC *c, VEC *a, VEC *b )
{

  int i;
  
  for ( i = 0; i < a->i; i++ )
    c->v[ i ] = a->v[ i ] - b->v[ i ];

}












void increm_mx( MX *a, MX *b )
{

  int i, j;
  
  for ( i = 0; i < a->i; i++ ) {
  
    for ( j = 0; j < a->j; j++ )
      a->mx[ i ][ j ] += b->mx[ i ][ j ];
  
  }

}









STAT dot_prod( VEC *a, VEC *b )
{

  STAT x = 0.0;
  
  int i;
  
  for ( i = 0; i < a->i; i++ )
    x += a->v[ i ] * b->v[ i ];
    
  return x;

}













void outer_prod( MX *c, VEC *a, VEC *b )
{

  int i, j;
  
  for ( i = 0; i < a->i; i++ ) {
  
    for ( j = 0; j < b->i; j++ )
      c->mx[ i ][ j ] = a->v[ i ] * b->v[ j ];
  
  }

}









void sym_mx( MX *mx )
{

  int i, j;
  
  for ( i = 0; i < mx->i; i++ ) {
    for ( j = i + 1; j < mx->j; j++ )
      mx->mx[ j ][ i ] = mx->mx[ i ][ j ];
  }

}













void get_pos_x_vec( VEC *x, VEC *s, VEC *old_x )
{

  STAT step = 0.7, c = 1.0;
  
  while ( vec_is_neg( x )) {
  
    vec_mult_scalar( s, step );
    
    add_vec( x, s, old_x );
    
    c *= step;
    
  }
  
  vec_mult_scalar( s, 1.0 / c );

}








/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/














void get_mx_inv( MX *inv, MX *a )
{

  VEC v, sol;
  
  initialize_vec( &v, a->i );
  initialize_vec( &sol, a->i );
  
  get_my_mx_inv( inv, a, &v, &sol );
  
  if ( CHECK_INV )
    check_inv( inv, a );

}















void get_my_mx_inv( MX *inv, MX *a, VEC *v, VEC *sol )
{

  int col;
  
  for ( col = 0; col < a->i; col++ )
    get_mx_inv_single_col( inv, a, v, sol, col );

}












void get_mx_inv_single_col( MX *inv, MX *a, VEC *v, VEC *sol, int col )
{

  get_ident_vec_for_inv( v, col );
  
  system_of_eq( sol, v, a );
  
  put_col_into_mx( inv, sol, col );

}











void get_ident_vec_for_inv( VEC *v, int col )
{

  int i;
  
  for ( i = 0; i < v->i; i++ )
    v->v[ i ] = 0.0;
    
  v->v[ col ] = 1.0;

}











void system_of_eq( VEC *sol, VEC *v, MX *a )
{

  MX L, U;
  
  initialize_mx( &L, a->i, a->j );
  initialize_mx( &U, a->i, a->j );
  
  do_system_of_eq( sol, v, a, &L, &U );

}














void do_system_of_eq( VEC *sol, VEC *v, MX *a, MX *L, MX *U )
{

  set_diag_l( L );
  
  do_lu_decomp( L, U, a );
  
  do_my_system_of_eq( sol, L, U, v );

}












void set_diag_l( MX *L )
{

  int i;
  
  for ( i = 0; i < L->i; i++ )
    L->mx[ i ][ i ] = 1.0;

}











void do_lu_decomp( MX *L, MX *U, MX *a )
{

  int i, j;
  
  for ( j = 0; j < a->i; j++ ) {
  
    for ( i = 0; i <= j; i++ )
      U->mx[ i ][ j ] = get_u_elem( a, L, U, i, j );
      
    for ( i = j + 1; i < a->i; i++ )
      L->mx[ i ][ j ] = get_l_elem( a, L, U, i, j ); 
  
  }
  
}














STAT get_u_elem( MX *a, MX *L, MX *U, int i, int j )
{

  STAT elem = a->mx[ i ][ j ];
  
  int k;
  
  for ( k = 0; k < i; k++ )
    elem -= L->mx[ i ][ k ] * U->mx[ k ][ j ];
    
  return elem;

}











STAT get_l_elem( MX *a, MX *L, MX *U, int i, int j )
{

  STAT elem = a->mx[ i ][ j ];
  
  int k;
  
  for ( k = 0; k < j; k++ )
    elem -= L->mx[ i ][ k ] * U->mx[ k ][ j ];
    
  return elem / U->mx[ j ][ j ];

}















void do_my_system_of_eq( VEC *sol, MX *L, MX *U, VEC *v )
{

  VEC y;
  
  initialize_vec( &y, sol->i );
  
  do_system_of_eq_y_vec( sol, &y, L, U, v );

}













void do_system_of_eq_y_vec( VEC *sol, VEC *y, MX *L, MX *U, VEC *v )
{

  find_y_system_of_eq( y, L, v );
  
  find_sol_vector_system_of_eq( sol, U, y );

}













void find_y_system_of_eq( VEC *y, MX *L, VEC *v )
{

  y->v[ 0 ] = v->v[ 0 ] / L->mx[ 0 ][ 0 ];
  
  int j, k;
  
  for ( k = 1; k < y->i; k++ ) {
  
    y->v[ k ] = v->v[ k ];
    
    for ( j = 0; j < k; j++ )
      y->v[ k ] -= L->mx[ k ][ j ] * y->v[ j ];
      
    y->v[ k ] /= L->mx[ k ][ k ];
  
  }

}















void find_sol_vector_system_of_eq( VEC *sol, MX *U, VEC *y )
{

  int n = y->i - 1;

  sol->v[ n ] = y->v[ n ] / U->mx[ n ][ n ];

  int j, k;
  
  for ( k = n - 1; k >= 0; k-- ) {
  
    sol->v[ k ] = y->v[ k ];
  
    for ( j = n; j > k; j-- ) 
      sol->v[ k ] -= U->mx[ k ][ j ] * sol->v[ j ];
    
    sol->v[ k ] = sol->v[ k ] / U->mx[ k ][ k ];
  
  }

}













void put_col_into_mx( MX *inv, VEC *sol, int col )
{

  int i;
  
  for ( i = 0; i < inv->i; i++ )
    inv->mx[ i ][ col ] = sol->v[ i ];

}








/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/











void print_mx( MX *mx )
{
  print_mx_fh( mx, stdout );
}










void print_mx_fh( MX *mx, FILE *fh )
{

  print_sep( fh );
  
  print_my_mx( mx, fh );
  
  print_sep( fh );

}











void print_my_mx( MX *mx, FILE *fh )
{

  int i, j;
  
  for ( i = 0; i < mx->i; i++ ) {
  
    for ( j = 0; j < mx->j; j++ )     
      fprintf( fh, "%12.5e    ", (double)mx->mx[ i ][ j ] );
      
    fprintf( fh, "\n" );
  
  }

}















void print_vec( VEC *v )
{
  print_vec_fh( v, stdout );
}













void print_vec_fh( VEC *v, FILE *fh )
{

  print_sep( fh );
  
  int i;
  
  for ( i = 0; i < v->i; i++ )
    fprintf( fh, "%15.5e\n", (double)v->v[ i ] );
  
  print_sep( fh );

}








/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/















void check_mx_nan( MX *mx )
{
  if ( mx_is_nan( mx ))
    error( 5600, "Matrix contains NaN elements.\n" );
}













int mx_is_nan( MX *mx )
{

  int i, j;
  
  for ( i = 0; i < mx->i; i++ ) {
  
    for ( j = 0; j < mx->j; j++ ) {
    
      if ( stat_is_nan( mx->mx[ i ][ j ] ))
        return 1;
          
    }
  
  }
  
  return 0;

}











void check_vec_nan( VEC *v )
{
  if ( vec_is_nan( v ))
    error( 5601, "Vector contains NaN elements.\n" );
}










int vec_is_nan( VEC *v )
{

  int i;
  
  for ( i = 0; i < v->i; i++ ) {
    if ( stat_is_nan( v->v[ i ] ))
      return 1;
  }
  
  return 0;

}











int stat_is_nan( STAT x )
{
  return ( isnan( x ) || ( isinf( x ))) ? (1) : (0);
}












int vec_is_neg( VEC *v )
{

  int i;
  
  for ( i = 0; i < v->i; i++ ) {
  
    if ( v->v[ i ] < 0.0 )
      return 1;
  
  }
  
  return 0;

}










int vec_is_zero( VEC *v )
{

  int i;
  
  for ( i = 0; i < v->i; i++ ) {
  
    if ( v->v[ i ] != 0.0 )
      return 0;
  
  }
  
  return 1;

}












void check_inv( MX *inv, MX *a )
{

  check_mx_nan( inv );
  
  MX d = { .i = a->i, .j = a->j };
  
  matrix_mult( &d, inv, a );
  
  int i, j;
  
  for ( i = 0; i < d.i; i++ ) {
  
    for ( j = 0; j < d.j; j++ ) {
    
      if ( i == j ) {
        if ( fabs( 1.0 - d.mx[ i ][ j ] ) > STAT_LOW )
          error( 5602, "Failed inverse.\n" );
      }
      else {
        if ( fabs( d.mx[ i ][ j ] ) > STAT_LOW ) {
          print_mx( &d );
          error( 5603, "Failed inverse.\n" );
        }
      }
    
    }
  
  }

}








