#include "origin.h"














/*  Pyron et al, 2010  */
//#define EUTH_TIME 89.7


/*  Bininda et al-Emonds et al, 2007  */
#define EUTH_TIME 94.5




#define REMOVE_OVERLAP 1
#define REMOVE_ANC_OVERLAP 0




#define BRANCH_MIN 30.0
#define BRANCH_MIN_FACTOR 1.7






void do_origin( void )
{

  extern MOTIF major_motif;
  
  check_major_motif();
  
  do_my_origin( &major_motif );

}











void do_my_origin( MOTIF *mf )
{

  DATA_LIST *get_origin_data( DATA_LIST *data, MOTIF *mf );
  ORIGIN_LIST *get_origin_w_data_lister( ORIGIN_LIST *origin, DATA_LIST *data, 
    MOTIF *mf );
  
  DATA_LIST *data = NULL;
  ORIGIN_LIST *origin = NULL;
  
  /*  In data.c:  */
  data = (DATA_LIST *)get_origin_data( data, mf );
  
  /*  In this file, below:  */
  origin = (ORIGIN_LIST *)get_origin_w_data_lister( origin, data, mf );
  
  update_short_branches( data->data->tree );
  
  /*  In this file, below:  */
  print_origin_list( origin, data, mf );

}







void update_short_branches( TREE *tree )
{

  extern STAT branch_factor;

  NODE_LIST *node = tree->spec_header;
  
  while ( node != NULL ) {
  
    if ( node->my_node->t < BRANCH_MIN )
      node->my_node->t *= BRANCH_MIN_FACTOR;
      
    node = node->next;
  
  }

}







ORIGIN_LIST *get_origin_w_data_lister( ORIGIN_LIST *origin, DATA_LIST *data,
  MOTIF *mf )
{

  ORIGIN_LIST *create_origin_lister( ORIGIN_LIST *x );
  ORIGIN_LIST *get_set_origin_lister_w_data_lister( ORIGIN_LIST *origin, 
    DATA_LIST *data );
  
  origin = (ORIGIN_LIST *)create_origin_lister( origin );
  
  printf( "a1\n" );
  
  /*  In this file, below:  */
  set_origin_lister( origin, data, mf );
  
  /*  In origin_list.c:  */
  return get_set_origin_lister_w_data_lister( origin, data );

}










void set_origin_lister( ORIGIN_LIST *origin, DATA_LIST *data, MOTIF *mf )
{

  extern char major_spec[ BUFFER ];

  *(origin->my_origin->mf) = *mf;
  
  strcpy( origin->my_origin->spec, major_spec );  
  
  int n = 0;
  int total_seqs = get_numb_of_seqs_data( data, &n, origin->my_origin->spec );
  
  origin->my_origin->total_seqs = total_seqs;
  origin->my_origin->n = n;
  origin->my_origin->frac = (float)n / (float)total_seqs;
  
  printf( "\n" );
  printf( "TotalSeqs  = %d\n", total_seqs );
  printf( "OccurSeqs  = %d\n", n );
  printf( "Frac       = %.2f%%\n", origin->my_origin->frac * 100.0 );
  printf( "\n" );

}









int get_numb_of_seqs_data( DATA_LIST *data, int *n, char spec[] )
{

  int index = get_spec_numb_from_spec_cell( data->data->spec_cell, spec );
  
  PDATA_LIST *pdata = data->data->pdata;
  
  int total = 0;
  
  while ( pdata != NULL ) {
  
    if ( pdata->my_pdata->use_array[ 0 ] ) {
      ++total;
    
      if ( pdata->my_pdata->ct_array[ 0 ] > 0 ) 
        ++(*n);
      
    }
      
    pdata = pdata->next;
  
  }
  
  return total;
  
}












int get_spec_numb_from_spec_cell( SPEC_CELL *spec_cell, char spec[] )
{

  int i;
  
  for ( i = 0; i < spec_cell->n; i++ ) {
    if ( ! strcmp( (spec_cell->x + i)->spec, spec ))
      return i;
  }
  
  error( 13100, "Could not find species in occur_cell.\n" );
  
  return 0;

}

















void print_origin_list( ORIGIN_LIST *origin, DATA_LIST *data, MOTIF *mf )
{

  FILE *open_text_file_writing( char a[], char b[] );
  
  extern char input_file[ BUFFER ];
  
  FILE *out = NULL, *hist_fh = NULL, *hom_fh = NULL, *simple_fh = NULL;
  FILE *hg_fh = NULL, *sim_fh = NULL, *anc_fh = NULL;
  FILE *all_fh = NULL;
  
  if ( REMOVE_OVERLAP ) {
    out = (FILE *)open_text_file_writing( input_file, "BooBED_NoOverlap.txt" );
    simple_fh = (FILE *)open_text_file_writing( input_file, "BooBED_Simple.txt" );
    hist_fh = (FILE *)open_text_file_writing( input_file, "Boo_NoOverHist.txt" );
  }
  else if ( REMOVE_ANC_OVERLAP ) {
    out = (FILE *)open_text_file_writing( input_file, "BooBED_NoAncOverlap.txt" );
    simple_fh = (FILE *)open_text_file_writing( input_file, "BooBED_SimpleAnc.txt" );
    hist_fh = (FILE *)open_text_file_writing( input_file, "Boo_NoAncOverHist.txt" );
  }
  else {
    out = (FILE *)open_text_file_writing( input_file, "BooBED_out.txt" );
    hist_fh = (FILE *)open_text_file_writing( input_file, "Boo_Hist.txt" );
  }
  
  if ( REMOVE_ANC_OVERLAP == 0 ) {  
  hg_fh = (FILE *)open_text_file_writing( input_file, "BooBED_Hg.txt" );
  hom_fh = (FILE *)open_text_file_writing( input_file, "BooBED_Hominid.txt" );
  sim_fh = (FILE *)open_text_file_writing( input_file, "BooBED_Simian.txt" );
  anc_fh = (FILE *)open_text_file_writing( input_file, "BooBED_Anc.txt" );
  }

  all_fh = (FILE *)open_text_file_writing( input_file, "BooBED_All.txt" );
    
  print_origin_list_w_fhs( origin, data, mf, out, hg_fh, hom_fh, sim_fh, anc_fh,
    simple_fh, hist_fh, all_fh );
  
  close_file( out );
  
  close_file( all_fh );
  
  if ( REMOVE_ANC_OVERLAP == 0 ) {
  close_file( hg_fh );
  close_file( hom_fh );
  close_file( anc_fh );
  close_file( sim_fh );
  }
  
  close_file( hist_fh );
  
  if ( simple_fh != NULL )
    close_file( simple_fh );

}












void print_origin_list_w_fhs( ORIGIN_LIST *origin, DATA_LIST *data, MOTIF *mf,
  FILE *out, FILE *hg_fh, FILE *hom_fh, FILE *sim_fh, FILE *anc_fh, FILE *simple_fh, 
  FILE *hist_fh, FILE *all_fh )
{

  ORIGIN_HIST my_hist;
  
  initialize_origin_hist( &my_hist, data->data->tree );
  
  print_origin_list_bed_fh( origin, data, mf, &my_hist, out, hg_fh, hom_fh, sim_fh,
    anc_fh, simple_fh, all_fh );
  
  set_origin_hist_set( &my_hist );
  
  print_origin_hist_fh( &my_hist, hist_fh );
  print_origin_hist_fh( &my_hist, stdout );

}














void initialize_origin_hist( ORIGIN_HIST *my_hist, TREE *tree )
{

  NODE_LIST *get_node_from_label( NODE_LIST *node_header, char label[] );

  extern char major_spec[ BUFFER ];
  extern STAT branch_factor;

  NODE_LIST *node = NULL;
  /*  In ../tree/get.c:  */
  node = (NODE_LIST *)get_node_from_label( tree->spec_header, major_spec );
  
  STAT total_time = EUTH_TIME, total_branch = 0.0;
  
  my_hist->euth = 100;
  
  int i = 0;
  
  while ( node != NULL ) {
  
    strcpy( my_hist->node[ i ], node->my_node->label );
    
    if ( ! strcmp( node->my_node->label, "Eutheria" ))
      my_hist->euth = i;
    
    STAT branch = node->my_node->t / (float)branch_factor;
    
    my_hist->time[ i ] = branch;
    my_hist->ct[ i ] = 0;
    my_hist->overlap[ i ] = 0;
    
    if ( my_hist->euth > i )
      total_branch += branch;

    ++i;
    
    node = node->up;
  
  }  
  
  my_hist->total_seqs = 0;
  my_hist->total_ct = 0;
  my_hist->seq_ct = 0;
  my_hist->n = i;
  
  for ( i = 0; i < my_hist->n; i++ ) {
    my_hist->time[ i ] *= total_time / total_branch;
  }

}












void print_origin_list_bed_fh( ORIGIN_LIST *origin, DATA_LIST *data, MOTIF *mf, 
  ORIGIN_HIST *my_hist, FILE *out, FILE *hg_fh, FILE *hom_fh, FILE *sim_fh,
  FILE *anc_fh, FILE *simple_fh, FILE *all_fh )
{

  print_origin_list_bed_header( my_hist, out );
  print_origin_list_bed_header( my_hist, all_fh );
  
  if ( simple_fh != NULL )
    print_origin_list_bed_header( my_hist, simple_fh );

  SITE_LIST *site = origin->my_origin->site_header[ PLUS_STRAND ];
  
  while ( site != NULL )
    print_origin_list_bed_single_seq_and_step( origin, &site, data->data, my_hist, 
      out, hg_fh, hom_fh, sim_fh, anc_fh, simple_fh, all_fh );

}












void print_origin_list_bed_header( ORIGIN_HIST *my_hist, FILE *out )
{

  extern char seq_align_file[ BUFFER ];
  extern char region_file[ BUFFER ];
  extern char tree_file[ BUFFER ];
  extern char major_spec[ BUFFER ];

  fprintf( out, "#  Branch of Origin finder\n" );
  fprintf( out, "#  BED output file\n" );
  fprintf( out, "#  \n" );
  fprintf( out, "#  TFBS occurrences are given in BED format from multiz\n" );
  fprintf( out, "#  alignment file:\n" );
  fprintf( out, "#    %s\n", seq_align_file );
  fprintf( out, "#  Region file:\n" );
  fprintf( out, "#    %s\n", region_file );
  fprintf( out, "#  Tree file:\n" );
  fprintf( out, "#    %s\n", tree_file );
  fprintf( out, "#  Species:\n" );
  fprintf( out, "#    %s\n", major_spec );
  fprintf( out, "#  \n" );
  fprintf( out, "#  Branches from leaf node to root:\n" );
  
  int i;
  for ( i = 0; i < my_hist->n; i++ ) {
    char t[ BUFFER ] = "";
    if ( strcmp( my_hist->node[ i ], "Root" )) {
      if ( my_hist->time[ i ] > 1.0 )
        sprintf( t, "%.1f", (float)my_hist->time[ i ] );
      else
        sprintf( t, "%.5f", (float)my_hist->time[ i ] );
    }
    fprintf( out, "#    %20s    %s\n", my_hist->node[ i ], t );
  }
  
  fprintf( out, "#  \n" );
    
  fprintf( out, "#  First three columns are in BED format.\n" );
  fprintf( out, "#  The fourth column gives the possible branch of origin");
  fprintf( out, "#  \n" );

}















void print_origin_list_bed_single_seq_and_step( ORIGIN_LIST *origin, 
  SITE_LIST **site, DATA *data, ORIGIN_HIST *my_hist, FILE *out, FILE *hg_fh,
  FILE *hom_fh, FILE *sim_fh, FILE *anc_fh, FILE *simple_fh, FILE *all_fh )
{

  NODE_LIST *get_node_from_label( NODE_LIST *node_header, char label[] );
  int node_downstream( NODE_LIST *node, NODE_LIST *leaf );
  
  TREE *tree = data->tree;
  NODE_LIST *hg = NULL, *hom = NULL, *sim = NULL, *anc = NULL, *my_node = NULL;
  
  SITE_LIST *first_site = *site;
  
  hg = (NODE_LIST *)get_node_from_label( tree->spec_header, "hg19" );
  hom = (NODE_LIST *)get_node_from_label( tree->spec_header, "Hominidae" );  
  sim = (NODE_LIST *)get_node_from_label( tree->spec_header, "Simiiformes" );
  anc = (NODE_LIST *)get_node_from_label( tree->spec_header, "Eutheria" );
  
//  hg = hg->up;
//  hom = hom->up;
//  sim = sim->up;
  //  Don't move the ancestral node...
  
  char my_node_str[ BUFFER ];
  
              
  int n = get_numb_of_sites_w_seq_numb( first_site, 
    first_site->site->seq_numb );
    
  char branch_str[ LINE_BUFFER ] = "";
  char branches[ BUFFER ][ BUFFER ];
  
  int numb = get_branch_str_origin_file( branch_str, branches, first_site, n, 
    my_hist );
  
  ++(my_hist->total_seqs);

  if ( first_site->site->loc->n == first_site->site->loc->x )
    ++(my_hist->seq_ct);
    
  SITE_LIST *prev_site = NULL;
  SITE_LIST *curr_site = *site;
    
  int i, j = 0;
  
  for ( i = 0; i < n; i++ ) {
  
    SEQ_INFO *site_info = curr_site->site->loc;
    my_node = (NODE_LIST *)get_node_from_label( tree->spec_header, 
      curr_site->stats->node );

    /*
    if ( ! node_downstream( anc, my_node ))
      strcpy( my_node_str, (*site)->stats->node );
    else
      strcpy( my_node_str, "Eutheria" );
    */
    
    fprintf( all_fh, "%s\t%d\t%d\t%s\n",
      site_info->chr,
      (int)site_info->start,
      (int)site_info->end,
      curr_site->stats->node );
      
    curr_site = curr_site->next;
      
  }
              
  if ( REMOVE_ANC_OVERLAP ) {
  
    for ( i = 0; i < n; i++ ) {
    
      SEQ_INFO *info = (*site)->site->loc;
  
      if (( prev_site == NULL ) || ( prev_site->site->loc->end < info->start )) {

        fprintf( out, "%s\t%d\t%d\t%s\t%d\t%s\t%d:%d\n",
          info->chr,
          (int)info->start,
          (int)info->end,
          (*site)->site->seq,
          (int)(*site)->site->seq_numb,
          branch_str,
          (int)info->n,
          (int)info->x );
        
        if ( simple_fh != NULL ) {
          if ( j < numb ) {
            while ( branches[ j ][ 0 ] == '\0' )
              ++j;
            fprintf( simple_fh, "%s\t%d\t%d\t%s\n",
              info->chr,
              (int)info->start,
              (int)info->end,
              branches[ j++ ] );
          }
        }
        
      
      }

      prev_site = *site;
      *site = (*site)->next;
    
    }
    
    return;
  
  }

  for ( i = 0; i < n; i++ ) {
  
    SEQ_INFO *info = (*site)->site->loc;
  
    my_node = (NODE_LIST *)get_node_from_label( tree->spec_header, 
      (*site)->stats->node );

    if ( REMOVE_OVERLAP ) {
      if (( prev_site == NULL ) || ( prev_site->site->loc->end < info->start )) {
        fprintf( out, "%s\t%d\t%d\t%s\t%d\t%s\t%d:%d\n",
          info->chr,
          (int)info->start,
          (int)info->end,
          (*site)->site->seq,
          (int)(*site)->site->seq_numb,
          branch_str,
          (int)info->n,
          (int)info->x );
        
        if ( simple_fh != NULL ) {
          fprintf( simple_fh, "%s\t%d\t%d\t%s\n",
            info->chr,
            (int)info->start,
            (int)info->end,
            (*site)->stats->node );
        }
        
        if ( REMOVE_ANC_OVERLAP == 0 ) {
        
        if ( node_downstream( hg, my_node )) 
          fprintf( hg_fh, "%s\t%d\t%d\t%s\n",
            info->chr,
            (int)info->start,
            (int)info->end,
            (*site)->stats->node );
        
        if ( node_downstream( hom, my_node )) 
          fprintf( hom_fh, "%s\t%d\t%d\t%s\n",
            info->chr,
            (int)info->start,
            (int)info->end,
            (*site)->stats->node );
        
        if ( node_downstream( sim, my_node ))
          fprintf( sim_fh, "%s\t%d\t%d\t%s\n",
            info->chr,
            (int)info->start,
            (int)info->end,
            (*site)->stats->node );
        
        if ( ! node_downstream( anc, my_node ))        
          fprintf( anc_fh, "%s\t%d\t%d\t%s\n",
            info->chr,
            (int)info->start,
            (int)info->end,
            (*site)->stats->node );
            
        }
        
      }
    }
    else {
        fprintf( out, "%s\t%d\t%d\t%s\t%d\t%s\t%d:%d\n",
          info->chr,
          (int)info->start,
          (int)info->end,
          (*site)->site->seq,
          (int)(*site)->site->seq_numb,
          branch_str,
          (int)info->n,
          (int)info->x );    
    }
      
    prev_site = *site;  
    *site = (*site)->next;
  
  }
  
}











int get_numb_of_sites_w_seq_numb( SITE_LIST *first_site, SEQ seq_numb )
{

  int n = 0;
  
  SITE_LIST *site = first_site;
  
  while ( site != NULL ) {
  
    if ( site->site->seq_numb != seq_numb )
      break;
      
    ++n;
    
    site = site->next;
  
  }
  
  return n;

}










void print_hominid_sites_fh( SITE_LIST *first_site, TREE *tree, int n, FILE *fh )
{

  NODE_LIST *get_node_from_label( NODE_LIST *node_header, char label[] );
  int node_downstream( NODE_LIST *node, NODE_LIST *leaf );

  SITE_LIST *site = first_site, *prev_site = NULL;
  
  NODE_LIST *hom = NULL;
  hom = (NODE_LIST *)get_node_from_label( tree->spec_header, "Hominidae" );
  
  hom = hom->up;
  
  int i;
  
  for ( i = 0; i < n; i++ ) {
  
    SEQ_INFO *info = site->site->loc;
  
    NODE_LIST *x = NULL;
    x = (NODE_LIST *)get_node_from_label( tree->spec_header, site->stats->node );
  
    if (( prev_site == NULL ) || ( prev_site->site->loc->end >= info->start ))
    {
          fprintf( fh, "%s\t%d\t%d\t%s\n",
          info->chr,
          (int)info->start,
          (int)info->end,
          site->stats->node );
    }
      
    prev_site = site;
    site = site->next;
  
  }

}







int get_branch_str_origin_file( char branch_str[], char branches[ BUFFER ][ BUFFER ],
  SITE_LIST *first_site, int n, ORIGIN_HIST *my_hist )
{

  int i, j, count = 0, numb = 0;
  
  char branch[ BUFFER ] = "";
  int node_index[ BUFFER ];
  
  int curr_count[ MY_NODE_BUFFER ];
  
  for ( i = 0; i < MY_NODE_BUFFER; i++ )
    curr_count[ i ] = 0;
  
  SITE_LIST *site = first_site, *prev_site = NULL;
  
  int set = 0, overlap = 0;
  
  for ( i = 0; i < n; i++ ) {
  
    if ( strcmp( site->stats->node, branch )) {
    
      char copy[ LINE_BUFFER ] = "";
      
      if ( set ) {
        if ( strlen( branch_str ) > 0 )
          sprintf( copy, "%s,", branch_str );
        sprintf( branch_str, "%s%s:%d", copy, branch, ++count );
      }
      else {
        set = 1;
      }
        
      count = 0;
      strcpy( branch, site->stats->node );
    
    }
    else {
      ++count;
    }
    
    int overlap_bool = 0;
    
    if (( prev_site != NULL ) && 
      ( prev_site->site->loc->end > site->site->loc->start )) 
    {
      ++overlap;
    }
    
    for ( j = 0; j < my_hist->n; j++ ) {
      if ( ! strcmp( site->stats->node, my_hist->node[ j ] )) {
        ++(my_hist->ct[ j ]);
        ++(curr_count[ j ]);
        node_index[ numb ] = j;
        strcpy( branches[ numb++ ], branch );
      }
    }
    
    prev_site = site;
    site = site->next;
  
  }
  
  char copy[ LINE_BUFFER ] = "";
  
  if ( set ) {
    if ( strlen( branch_str ) > 0 )
      sprintf( copy, "%s,", branch_str );
    sprintf( branch_str, "%s%s:%d", copy, branch, ++count );
  }

  if ( REMOVE_ANC_OVERLAP ) {
  
    site = first_site;
  
    j = my_hist->n - 1;
    
    int k = numb - 1;
  
    for ( i = 0; i < overlap; i++ ) {
  
      while (( j >= 0 ) && ( curr_count[ j ] <= 1 ))
        --j;
  
      --(curr_count[ j ]);
      --(my_hist->ct[ j ]);
      
      while ( node_index[ k ] != j )
        --k;
        
      branches[ k ][ 0 ] = '\0';
    
    }
    
    return numb;
      
  }
  
  if ( REMOVE_OVERLAP ) {
  
    site = first_site;
  
    j = my_hist->n - 1;
    j = 0;
  
    for ( i = 0; i < overlap; i++ ) {
  
      while (( j < my_hist->n ) && ( curr_count[ j ] <= 1 ))
        ++j;
  
      if ( j >= my_hist->n )
        break;
      
      --(curr_count[ j ]);
      --(my_hist->ct[ j ]);
    
    }
      
  }
  
  return numb;

}










void update_origin_hist( ORIGIN_HIST *my_hist, SITE_LIST *site )
{

  int i;
  
  for ( i = 0; i < my_hist->n; i++ ) {
  
    if ( ! strcmp( site->stats->node, my_hist->node[ i ] )) {
      ++(my_hist->ct[ i ]);
      ++(my_hist->total_ct);
      return;
    }
  
  }

}










void set_origin_hist_set( ORIGIN_HIST *my_hist )
{

  int i, j = 0;
  
  my_hist->ct[ my_hist->euth - 2 ] += my_hist->ct[ my_hist->euth - 1 ];
  my_hist->time[ my_hist->euth - 2 ] += my_hist->time[ my_hist->euth - 1 ];
  
  my_hist->ct[ my_hist->euth - 1 ] = 0;
  my_hist->time[ my_hist->euth - 1 ] = 0.0;
  
  for ( i = my_hist->euth; i < my_hist->n; i++ ) {
    my_hist->ct[ my_hist->euth - 1 ] += my_hist->ct[ i ];
    my_hist->time[ my_hist->euth - 1 ] += my_hist->time[ i ];
  }
    
  strcpy( my_hist->node[ my_hist->euth - 1 ], my_hist->node[ my_hist->euth ] );

}








void print_origin_hist_fh( ORIGIN_HIST *my_hist, FILE *hist_fh )
{

  int i, total = 0;
  
  for ( i = 0; i < my_hist->euth; i++ ) 
    total += my_hist->ct[ i ];
  
  fprintf( hist_fh, "%6s\t%12s\t%9s\t%9s\t%9s\t%9s\n",
    "\"X\"",
    "\"NODE\"",
    "\"T\"",
    "\"CT\"",
    "\"Frac\"",
    "\"Rate\"" );

  for ( i = 0; i < my_hist->euth; i++ )  
    fprintf( hist_fh, "%6d\t%12s\t%9.2f\t%9d\t%9.3f\t%9.3f\n",
      i + 1,
      my_hist->node[ i ],
      (float)my_hist->time[ i ],
      my_hist->ct[ i ],
      (float)my_hist->ct[ i ] * 100.0 / (float)total,
      (float)my_hist->ct[ i ] / (float)my_hist->time[ i ] );
    
}










