#include "clus.h"





















void do_clustering( void )
{

  TREE *make_and_set_tree( TREE *tree );
  void set_tree_ancestor_branches( TREE *tree );
  RESULT_LIST *get_result_list( RESULT_LIST *res_header, TREE *tree, MOTIF *mf );
  CLUS_LIST *get_cluster_list_from_r_list( CLUS_LIST *clus_header,
    RESULT_LIST *res_header );

  TREE *tree = NULL;
  RESULT_LIST *res_header = NULL;
  CLUS_LIST *clus_header = NULL;
  
  MOTIF mf = { .length = WORD_LENGTH };
  
  /*  In ../tree/create_tree.c:  */
  tree = (TREE *)make_and_set_tree( tree );
  
  /*  In ../model/data.c:  */
  set_tree_ancestor_branches( tree );
  
  int i;
  for ( i = 0; i < WORD_LENGTH; i++ )
    mf.cons[ i ] = 'n';
    
  mf.cons[ WORD_LENGTH ] = '\0';
  
  /*  In get_rlist.c:  */
  res_header = (RESULT_LIST *)get_result_list( res_header, tree, &mf );

  /*  In cluster.c:  */
  clus_header = (CLUS_LIST *)get_cluster_list_from_r_list( clus_header,
    res_header );
    
  /*  In this file, below:  */
  print_motif_clustering( clus_header, res_header, tree, &mf );
    
//  kill_result_list( res_header );
  
}











void print_motif_clustering( CLUS_LIST *clus_header, RESULT_LIST *res_header,
  TREE *tree, MOTIF *mf )
{

  FILE *open_text_file_writing( char a[], char b[] );
  
  extern char input_file[ BUFFER ];
  extern int reg_start, reg_end;

  char fname[ BUFFER ], ext[ BUFFER ];
  
  sprintf( fname, "%s.%d:%d.cluster", input_file, reg_start, reg_end );
  
  FILE *out = NULL, *fasta_fh = NULL, *phylo_fasta = NULL, *non_fasta = NULL;
  out = (FILE *)open_text_file_writing( fname, "out" );
  fasta_fh = (FILE *)open_text_file_writing( fname, "fasta" );

  print_cluster_list_to_file( clus_header, out );
  
  print_fasta_cluster_list_w_fh( clus_header, tree, fasta_fh );
  
  close_file( out );
  close_file( fasta_fh );

}








void print_cluster_list_to_file( CLUS_LIST *clus_header, FILE *out )
{

  print_mf_cluster_lists_fh( clus_header, out );
  
  print_mf_cluster_lists( clus_header );

}












void print_mf_cluster_lists( CLUS_LIST *clus_header )
{
  print_mf_cluster_lists_fh( clus_header, stdout );
}











void print_mf_cluster_lists_fh( CLUS_LIST *clus_header, FILE *out )
{

  int i = 0;

  CLUS_LIST *clus = clus_header;
  
  while ( clus != NULL )
    print_single_mf_cluster_list_and_step( &clus, &i, out );

}












void print_single_mf_cluster_list_and_step( CLUS_LIST **clus, int *i, FILE *out )
{
  print_single_mf_cluster_list( *clus, i, out );
  *clus = (*clus)->next;
}










void print_single_mf_cluster_list( CLUS_LIST *clus, int *i, FILE *out )
{

  fprintf( out, "%5d\t\t%s\n", ++(*i),
    clus->my_clus->cons );

  print_memb_list_fh( clus->my_clus->memb_header, clus->my_clus->min_offset, 
    clus->my_clus->max_offset, out );
    
  fprintf( out, "\n" );
  
  CLUS_LIST *c = clus->down;
  
  while ( c != NULL ) {
  
    fprintf( out, "\t\t\t%25s\t%15s\t%.2e\n",
      c->my_clus->cons,
      c->my_clus->memb_header->my_memb->node,
      (double)c->my_clus->memb_header->my_memb->pval );
      
    c = c->down;
  
  }
  
  fprintf( out, "\n" );

}












void print_memb_list_fh( MEMB_LIST *memb_header, char min_offset, char max_offset,
  FILE *out )
{

  int i = 0;

  MEMB_LIST *memb = memb_header;
  
  while ( memb != NULL )
    print_single_memb_lister_and_step( &memb, ++i, min_offset, max_offset, out );

}













void print_single_memb_lister_and_step( MEMB_LIST **memb, int i, int min_offset,
  int max_offset, FILE *out )
{
  print_single_memb_lister( *memb, i, min_offset, max_offset, out );
  *memb = (*memb)->next;
}












void print_single_memb_lister( MEMB_LIST *memb, int numb, int min_offset, 
  int max_offset, FILE *out )
{

  fprintf( out, "\t\t%5d   ", numb );
  
  int i;
  
  for ( i = 0; i < memb->my_memb->offset - min_offset; i++ )
    fprintf( out, "." );
    
  fprintf( out, "%s", memb->my_memb->mf->cons );
  
  for ( i = 0; i < max_offset - memb->my_memb->offset; i++ )
    fprintf( out, "." );
    
  fprintf( out, "\t%.3e\t%s\n", (double)memb->my_memb->pval,
    memb->my_memb->node );

}
















void print_fasta_cluster_list_w_fh( CLUS_LIST *clus_header, TREE *tree,
  FILE *fasta_fh )
{

  CLUS_LIST *clus = clus_header;
  
  int i = 0;
  
  while ( clus != NULL )
    print_single_clus_fasta_list_and_step( &clus, &i, tree, fasta_fh );
      
}










void print_single_clus_fasta_list_and_step( CLUS_LIST **clus, int *numb, 
  TREE *tree, FILE *fasta_fh )
{
  print_single_clus_fasta_list( *clus, numb, tree, fasta_fh );
  *clus = (*clus)->next;
}












void print_single_clus_fasta_list( CLUS_LIST *clus, int *numb, TREE *tree,
  FILE *fasta_fh )
{

  fprintf( fasta_fh, ">Motif_%d_%.3e\n", ++(*numb), (double)clus->my_clus->pval );
  fprintf( fasta_fh, "%s\n", clus->my_clus->cons );
  
}


























