#include "cons.h"

















#define REPEAT_DENOM 2

#define MAX_MEMB_INCLUDE 7

#define MAX_WEIGHT 200













void set_clus_lister_cons_seq( CLUS_LIST *clus )
{

  int len = clus->my_clus->memb_header->my_memb->mf->length;
  int offset = clus->my_clus->min_offset;
  
  int i = 0;
  
  while ( offset < len + clus->my_clus->max_offset )
    clus->my_clus->cons[ i++ ] = get_cons_nucl_single_site( clus, offset++ );
    
  clus->my_clus->cons[ i ] = '\0';

}












char get_cons_nucl_single_site( CLUS_LIST *clus, char offset )
{

  int string_is_repeat( char str[], char length );

  int nucl_array[ NUMB_OF_NUCL ];
  
  initialize_int_array( nucl_array, NUMB_OF_NUCL );
  
  MEMB_LIST *memb = clus->my_clus->memb_header;
  
  int numb = 0;
  
  int n = 0;
  
  while (( memb != NULL ) && ( numb < MAX_MEMB_INCLUDE )) {

    n += get_pval_weight( memb->my_memb->pval );
    
    if ( memb->next != NULL ) {
      /*  In ../main/nucl.c:  */
      if ( string_is_repeat( memb->my_memb->mf->cons, 1 )) {
        memb = memb->next;
        continue;
      }
      if ( string_is_repeat( memb->my_memb->mf->cons, 2 )) {
        memb = memb->next;
        continue;
      }
    }

    update_nucl_array_single_memb_and_step( nucl_array, &memb, offset );
    
    ++numb;

  }
    
  return get_cons_nucl_single_site_w_array( nucl_array, n );

}












int get_pval_weight( PVAL p )
{

  if ( p == 0.0 )
    return MAX_WEIGHT;
    
  return -1.0 * log10( p );

}









void update_nucl_array_single_memb_and_step( int nucl_array[], MEMB_LIST **memb, 
  char offset )
{
  update_nucl_array_single_memb( nucl_array, *memb, offset );
  *memb = (*memb)->next;
}










void update_nucl_array_single_memb( int nucl_array[], MEMB_LIST *memb, char offset )
{

  MF get_bp_from_nucl( char nucl );
  int string_is_repeat( char str[], char length );

  char index = offset - memb->my_memb->offset;
  
  if (( index >= 0 ) && ( index < memb->my_memb->mf->length )) {

    /*  In ../main/nucl.c:  */
    MF bp = get_bp_from_nucl( memb->my_memb->mf->cons[ index ] );
    
//    ++(nucl_array[ bp ]);

    nucl_array[ bp ] += get_pval_weight( memb->my_memb->pval );

  }  

}













char get_cons_nucl_single_site_w_array( int ct_array[], int numb_of_memb )
{

  int nucl_array[ NUMB_OF_NUCL ];
  
  int i;
  
  for ( i = 0; i < NUMB_OF_NUCL; i++ )
    nucl_array[ i ] = i;
    
  sort_ct_array_index_array( ct_array, nucl_array, NUMB_OF_NUCL );
  
  return get_cons_nucl_sorted_array( ct_array, nucl_array, numb_of_memb );

}













void sort_ct_array_index_array( int ct_array[], int nucl_array[], int n )
{

  int i, j;
  
  int ct, nucl;
  
  for ( j = 1; j < n; j++ ) {
  
    ct = ct_array[ j ];
    nucl = nucl_array[ j ];
    
    i = j - 1;
    
    while (( i >= 0 ) && ( ct_array[ i ] < ct )) {
      ct_array[ i + 1 ] = ct_array[ i ];
      nucl_array[ i + 1 ] = nucl_array[ i ];
      --i;
    }
    
    ct_array[ i + 1 ] = ct;
    nucl_array[ i + 1 ] = nucl;
  
  }

}












char get_cons_nucl_sorted_array( int ct_array[], int nucl_array[],
  int numb_of_memb )
{

  if ( (float)ct_array[ 0 ] / (float)numb_of_memb < 0.15 )
    return tolower( get_my_cons_nucl_sorted_array( ct_array, nucl_array ));
      
  return get_my_cons_nucl_sorted_array( ct_array, nucl_array );

}











char get_my_cons_nucl_sorted_array( int ct_array[], int nucl_array[] )
{
  
  char get_nucl_from_bp( MF bp );
  char get_double_degen_char( MF a, MF b );
  
  /*  In this file, below:  */
  if ( cons_nucl_single_nucl( ct_array ))
    /*  In ../main/nucl.c:  */
    return toupper( get_nucl_from_bp( nucl_array[ 0 ] ));
    
  /*  In this file, below:  */
  if ( cons_nucl_double_nucl( ct_array ))
    /*  In ../main/nucl.c:  */
    return toupper( get_double_degen_char( nucl_array[ 0 ], nucl_array[ 1 ] ));
    
  return 'N';
  
}











int cons_nucl_single_nucl( int ct_array[] )
{

  int total = get_my_total_int_array( ct_array, NUMB_OF_NUCL );
  
  if ( (float)ct_array[ 0 ] / (float)total <= 0.5 )
    return 0;
    
  return ( ct_array[ 0 ] >= ct_array[ 1 ] * 2 ) ? (1) : (0);

}









int cons_nucl_double_nucl( int ct_array[] )
{

  int total = get_my_total_int_array( ct_array, NUMB_OF_NUCL );
  
  return ( (float)(ct_array[ 0 ] + ct_array[ 1 ]) / (float)total >= 0.75 ) ? 
    (1) : (0);

}








int get_my_total_int_array( int array[], int n )
{

  int total = 0, i;
  
  for ( i = 0; i < n; i++ )
    total += array[ i ];
    
  return total;

}






/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/













CLUS_LIST *get_swapped_sites_clus_list( CLUS_LIST *swap_header,
  CLUS_LIST *clus_header )
{

  CLUS_LIST *clus = clus_header;
  
  while ( clus != NULL )
    swap_header = (CLUS_LIST *)get_swapped_sites_single_clus_and_step( swap_header,
      &clus );
      
  return swap_header;

}












CLUS_LIST *get_swapped_sites_single_clus_and_step( CLUS_LIST *swap_header,
  CLUS_LIST **clus )
{

  swap_header = (CLUS_LIST *)get_swapped_sites_single_clus( swap_header, *clus );
  
  *clus = (*clus)->next;
  
  return swap_header;

}












CLUS_LIST *get_swapped_sites_single_clus( CLUS_LIST *swap_header, CLUS_LIST *clus )
{

  char class = clus->my_clus->class;
  
  if (( class != CLASS_MF_BIRTH ) && ( class != CLASS_MF_DEATH ))
    return swap_header;
    
  return get_my_swapped_sites_single_clus( swap_header, clus );

}













CLUS_LIST *get_my_swapped_sites_single_clus( CLUS_LIST *swap_header, 
  CLUS_LIST *clus )
{

  CLUS_LIST *create_clus_lister( CLUS_LIST *x );
  CLUS_LIST *put_clus_lister_end_of_list( CLUS_LIST *clus_header, CLUS_LIST *new );
  
  CLUS_LIST *new = NULL;
  new = (CLUS_LIST *)create_clus_lister( new );
  
  set_swapped_sites_clus_lister( new, clus );
  
  return put_clus_lister_end_of_list( swap_header, new );

}













void set_swapped_sites_clus_lister( CLUS_LIST *new, CLUS_LIST *clus )
{

  int array[ BUFFER ], len = strlen( clus->my_clus->cons ), i;
  
  for ( i = 0; i < len - 1; i++ )
    array[ i ] = get_next_index_w_array( array, i, len );
    
  if ( value_in_array( len - 1, array, len - 1 ) == 0 ) {
    int x = rand() % (len - 1);
    array[ x ] = len - 1;
    array[ len - 1 ] = x;
  }
  else {
    array[ len - 1 ] = get_next_index_w_array( array, len - 1, len );
  }
  
  for ( i = 0; i < len; i++ )
    new->my_clus->cons[ i ] = clus->my_clus->cons[ array[ i ] ];
    
  new->my_clus->cons[ len ] = '\0';
  
  new->my_clus->pval = clus->my_clus->pval;
  new->my_clus->reg = clus->my_clus->reg;
  new->my_clus->class = clus->my_clus->class;
  
  printf( "clus = %s\n", new->my_clus->cons );

}













int get_next_index_w_array( int array[], int i, int len )
{

  int x = rand() % len;
  
  while (( x == i ) || ( value_in_array( x, array, i )))
    x = rand() % len;
    
  return x;

}












int value_in_array( int x, int array[], int buffer )
{

  int i;
  
  for ( i = 0; i < buffer; i++ ) {
    if ( x == array[ i ] )
      return 1;
  }
  
  return 0;

}













CLUS_LIST *put_clus_lister_end_of_list( CLUS_LIST *clus_header, 
  CLUS_LIST *clus )
{

  if ( clus_header == NULL )
    return clus;
    
  CLUS_LIST *curr = clus_header;
  
  while ( curr->down != NULL )
    curr = curr->down;
    
  curr->down = clus;
  
  return clus_header;

}















