#include "read_model.h"
























MODEL_LIST *get_null_model_from_output( MODEL_LIST *null_model, TREE *tree,
  MOTIF *mf, FILE *fh )
{

  MODEL_LIST *create_model_lister( MODEL_LIST *x, MOTIF *mf, INDEX m );
  
  null_model = (MODEL_LIST *)create_model_lister( null_model, mf, 0 );
  
  return get_my_null_model_from_output( null_model, tree, mf, fh );

}















MODEL_LIST *get_my_null_model_from_output( MODEL_LIST *null_model, TREE *tree,
  MOTIF *mf, FILE *fh )
{

  check_null_model_header( fh );
  
  return get_null_model_from_known_output( null_model, tree, mf, fh );

}













void check_null_model_header( FILE *fh )
{

  char line[ LINE_BUFFER ], str[ BUFFER ] = "Null-model";
  
  while ( fgets( line, LINE_BUFFER, fh ) != NULL ) {
  
    if ( ! strncmp( str, line, strlen( str )))
      return;
  
  }

}












MODEL_LIST *get_null_model_from_known_output( MODEL_LIST *null_model, TREE *tree,
  MOTIF *mf, FILE *fh )
{

  read_motif_from_output( null_model, fh );
  
  read_model_stats_from_output( null_model->my_model->stats, fh );
  read_model_stats_from_output( null_model->my_model->null_stats, fh );
  
  read_part_list_from_output( null_model, tree, fh );
  
  return null_model;

}  













void read_motif_from_output( MODEL_LIST *model, FILE *fh )
{

  int strs_match( char a[], char b[] );

  char line[ LINE_BUFFER ], str[ BUFFER ], cons[ BUFFER ];
  int d;
  
  while ( fgets( line, LINE_BUFFER, fh ) != NULL ) {
  
    if ( line[ 0 ] == '>' ) {
      check_motif_entry( model, line, fh );
      break;
    }
    
  }
  
}











void read_model_stats_from_output( MODEL_STATS *s, FILE *fh )
{

  char line[ LINE_BUFFER ];

  while ( fgets( line, LINE_BUFFER, fh ) != NULL ) {
  
    if ( line[ 0 ] == '.' )
      return;
      
    read_model_stats_output_line( s, line );
    
  }

}











void read_model_stats_output_line( MODEL_STATS *s, char line[] )
{

  char entry[ BUFFER ], x[ BUFFER ];
  
  sscanf( line, "%s%s", entry, x );
  
  double f;
  
  if ( ! strcmp( entry, "Lhood:" )) 
    s->lhood = atof( x );

  else if ( ! strcmp( entry, "Old-Lhood:" )) 
    s->old_lhood = atof( x );

  else if ( ! strcmp( entry, "Best-Lhood:" )) 
    s->best_lhood = atof( x );

  else if ( ! strcmp( entry, "Pval:" )) 
    s->pval = atof( x );

  else if ( ! strcmp( entry, "DOF-numer:" )) 
    s->dof_numer = atoi( x );
    
  else if ( ! strcmp( entry, "DOF-denom:" ))
    s->dof_denom = atoi( x );
    
  else if ( ! strcmp( entry, "Node:" ))
    strcpy( s->node_label, x );
    
  else if ( ! strcmp( entry, "Best-node:" ))
    strcpy( s->best_label, x );

}











void read_part_list_from_output( MODEL_LIST *model, TREE *tree, FILE *fh )
{

  PART_LIST *part = model->my_model->part_header;

  char line[ LINE_BUFFER ];
  
  int part_n = 0;
  
  while ( fgets( line, LINE_BUFFER, fh ) != NULL ) {
  
    if ( line[ 0 ] == '.' )
      break;
      
    if ( ! strncmp( line, "Part-N", 6 ))
      continue;
      
    read_part_lister_from_line_and_step( model->my_model->part_header, &part, tree, 
      line );
      
    ++part_n;
  
  }
  
  model->my_model->part_n = part_n;

}










void read_part_lister_from_line_and_step( PART_LIST *part_header, PART_LIST **part, 
  TREE *tree, char line[] )
{
  read_part_lister_from_line( part_header, *part, tree, line );
  *part = (*part)->next;
}











void read_part_lister_from_line( PART_LIST *part_header, PART_LIST *part, 
  TREE *tree, char line[] )
{

  int d;
  char label[ BUFFER ];
  double a, b;
  
  sscanf( line, "%d%s%lf%lf", &d, label, &a, &b );
  
  part->my_part->pp->a = (STAT)a;
  part->my_part->pp->b = (STAT)b;
  part->my_part->part = d;
  
  if ( ! strcmp( label, "1.000000e-25" ))
    strcpy( label, "hg19" );
  
  part->my_part->branch_header = (LIST *)get_list_from_output_str( label, tree );
  
  remove_list_elem_part_list( part_header, tree, label, d );
  
}












LIST *get_list_from_output_str( char label[], TREE *tree )
{

  if ( ! strcmp( label, "All" ))
    return get_all_list_from_output_str( tree );
  
  return get_lister_single_output_node( label, tree );
  
}








LIST *get_all_list_from_output_str( TREE *tree )
{

  LIST *create_lister( LIST *x, char type );
  
  NODE_LIST *node = tree->spec_header;
  
  LIST *header = NULL, *last = NULL;
  
  while ( node != NULL ) {
  
    LIST *new = NULL;
    new = (LIST *)create_lister( new, INT_TYPE );
    
    new->elem.d = node->my_node->numb;
    
    if ( header == NULL ) {
      header = new;
      last = new;
    }
    else {
      last->next = new;
      last = new;
    }
    
    node = node->next;
  
  }
  
  return header;

}  









LIST *get_lister_single_output_node( char label[], TREE *tree )
{

  LIST *create_lister( LIST *x, char type );

  NODE_LIST *node = tree->spec_header;
  
  int d = node->my_node->numb;
  
  while ( node != NULL ) {
  
    if ( ! strcmp( node->my_node->label, label )) {
    
      LIST *new = NULL;
      new = (LIST *)create_lister( new, INT_TYPE );
      
      new->elem.d = node->my_node->numb;
      
      return new;
    
    }
    
    node = node->next;
  
  }
  
      LIST *new = NULL;
      new = (LIST *)create_lister( new, INT_TYPE );
      
      new->elem.d = d;
      
      return new;

  /*
  printf( "Label in output file : '%s'\n", label );
  
  error( 7210, "Cannot find node label in tree.\n" );
  */
  return NULL;

}










void remove_list_elem_part_list( PART_LIST *part_header, TREE *tree, char label[], 
  int n )
{

  PART_LIST *part = part_header;

  int i;
  
  for ( i = 1; i < n; i++ )
    remove_list_elem_single_part_lister_and_step( &part, tree, label );

}










void remove_list_elem_single_part_lister_and_step( PART_LIST **part, TREE *tree,
  char label[] )
{
  remove_list_elem_single_part_lister( *part, tree, label );
  *part = (*part)->next;
}











void remove_list_elem_single_part_lister( PART_LIST *part, TREE *tree, 
  char label[] )
{

  int numb = -1;

  NODE_LIST *node = tree->spec_header;
  
  while ( node != NULL ) {
  
    if ( ! strcmp( node->my_node->label, label )) {
      numb = node->my_node->numb;
      break;
    }
    
    node = node->next;
  
  }
  
  /*
  if ( numb < 0 )
    error( 7211, "Could not find label in tree.\n" );
  */
  
  if ( numb < 0 )
    numb = 1;
    
  remove_list_single_part_lister_w_int( part, numb );

}











void remove_list_single_part_lister_w_int( PART_LIST *part, int numb )
{

  LIST *x = part->my_part->branch_header;
  
  if ( x->elem.d == numb ) {
    LIST *header = x->next;
    kill_lister( x );
    part->my_part->branch_header = header;
    return;
  }
  
  LIST *prev = NULL;
  
  while ( x != NULL ) {
  
    if ( x->elem.d == numb ) {

      if ( prev != NULL )
        prev->next = x->next;

      kill_lister( x );

      return;

    }
    
    prev = x;
    x = x->next;
  
  }

}












void check_motif_entry( MODEL_LIST *model, char line[], FILE *fh )
{

  char str[ BUFFER ], cons[ BUFFER ];
  int d;
  
  sscanf( line, "%*s%s%s%d", str, cons, &d );
  
  if ( strcmp( str, "Motif:" ))
    error( 7200, "Expected 'Motif:' string.\n" );
    
  if ( isdigit( cons[ 0 ] )) {
    fgets( line, LINE_BUFFER, fh );
    fgets( line, LINE_BUFFER, fh );
    fgets( line, LINE_BUFFER, fh );
    return;
  }
  
  /*  In ../main/nucl.c:  */
  if ( ! strs_match( cons, model->my_model->mf->cons ))
    error( 7201, "Unmatching motifs.\n" );
      
  if ( d != model->my_model->mf->my_motif )
    error( 7202, "Unmatching stacks.\n" );
        
  fgets( line, LINE_BUFFER, fh );

  sscanf( line, "%s%d", str, &d );
  
  if ( strcmp( str, "M:" ))
    error( 7203, "Expected 'M:' string.\n" );
    
  if ( model->my_model->m != d )
    error( 7204, "Unmatching M-values.\n" );
    
  fgets( line, LINE_BUFFER, fh );
  
  sscanf( line, "%s%d", str, &d );
  
  if ( strcmp( str, "Part-N:" ))
    error( 7205, "Expected 'Part-N:' string.\n" );
    
  if ( d != model->my_model->part_n )
    error( 7206, "Unmatching part-N values.\n" );
    
  fgets( line, LINE_BUFFER, fh );
  
  if ( line[ 0 ] != '.' )
    error( 7207, "Expected dot separator.\n" );

}















/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/















MODEL_LIST *get_all_model_list_from_output( MODEL_LIST *model_header, TREE *tree,
  MOTIF *mf, FILE *fh )
{

  MODEL_LIST *model = NULL;

  int still_going = 1;
  
  while ( still_going ) 
    model_header = (MODEL_LIST *)get_and_put_model_down_from_output( model_header,
      &model, tree, mf, &still_going, fh );
      
  return model_header;

}











MODEL_LIST *get_and_put_model_down_from_output( MODEL_LIST *model_header,
  MODEL_LIST **model, TREE *tree, MOTIF *mf, int *still_going, FILE *fh )
{

  char line[ LINE_BUFFER ];
  
  while ( fgets( line, LINE_BUFFER, fh ) != NULL ) {
  
    if ( ! strncmp( "Total-Models", line, 12 ))
      return put_model_down_end_list( model_header, model );
          
    if ( line[ 0 ] == '>' ) {
    
      int d;
      
      sscanf( line, "%*s%*s%*s%d", &d );
      
      if ( mf->my_motif != d ) {
        *still_going = 0;
        return put_model_down_end_list( model_header, model );
      }
      else {
        model_header = (MODEL_LIST *)update_model_down_from_output( model_header,
          model, tree, mf, fh );
      }
      
    }
  
  }

  *still_going = 0;
  
  return put_model_down_end_list( model_header, model );

}








MODEL_LIST *put_model_down_end_list( MODEL_LIST *model_header, MODEL_LIST **model )
{

  if ( model_header == NULL ) {
    model_header = *model;
    *model = NULL;
    return model_header;
  }
  
  MODEL_LIST *m = model_header;
  
  while ( m->next != NULL )
    m = m->next;
    
  m->next = *model;
  
  *model = NULL;
  
  return model_header;

}









MODEL_LIST *update_model_down_from_output( MODEL_LIST *model_header,
  MODEL_LIST **model, TREE *tree, MOTIF *mf, FILE *fh )
{

  MODEL_LIST *create_model_lister( MODEL_LIST *x, MOTIF *mf, INDEX mm );

  char line[ BUFFER ];
  int mval;
  
  fgets( line, BUFFER, fh );
  
  sscanf( line, "%*s%d", &mval );
  
  MODEL_LIST *m = NULL;
  m = (MODEL_LIST *)create_model_lister( m, mf, mval );
  
  fgets( line, BUFFER, fh );
  fgets( line, BUFFER, fh );
  
  if ( line[ 0 ] != '.' )
    error( 7220, "Expected dot separator.\n" );
    
  /*  In this file, above:  */  
  read_model_stats_from_output( m->my_model->stats, fh );
  read_model_stats_from_output( m->my_model->null_stats, fh );
  read_part_list_from_output( m, tree, fh );
  
  /*  In this file, below:  */
  *model = (MODEL_LIST *)put_model_down_end_of_down( *model, m );

}












MODEL_LIST *put_model_down_end_of_down( MODEL_LIST *model, MODEL_LIST *m )
{

  if ( model == NULL )
    return m;
    
  if ( model->my_model->stats->pval > m->my_model->stats->pval ) {
    m->next = model;
    return m;
  }
    
  MODEL_LIST *curr = model, *prev = NULL;
  
  while (( curr != NULL ) && 
    ( curr->my_model->stats->pval <= m->my_model->stats->pval ))
  {
    prev = curr;
    curr = curr->down;
  }
  
  if ( prev != NULL )
    prev->down = m;
    
  m->down = curr;
    
  return model;

}














