#include "ft.h"












#define READ_ONLY 0 


















void do_binary_ft( MOTIF_LIST *mf_header, TREE *tree, FILE *seq_fh, FILE *reg_fh, 
  FILE *ft_fh, FILE *ind_fh, FILE *mf_fh, FILE *fpos_fh, FILE *spec_fh, 
  FILE *ss_fh )
{

  MOTIF_LIST *align_header = NULL;
  align_header = (MOTIF_LIST *)get_empty_align_seq_mf_list( align_header,
    spec_fh );
    
  do_my_binary_ft( align_header, mf_header, tree, seq_fh, reg_fh, ft_fh, ind_fh,
    mf_fh, fpos_fh, ss_fh );

}










MOTIF_LIST *get_empty_align_seq_mf_list( MOTIF_LIST *align_header, FILE *fh )
{

  SPEC_CELL *create_spec_cell( SPEC_CELL *x, INDEX n );
  void set_position( FPOS fpos, FILE *fh );

  set_position( 0, fh );
  
  INDEX n;
  get( &n, sizeof(INDEX), fh );
  
  SPEC_CELL *spec_cell = NULL;
  spec_cell = (SPEC_CELL *)create_spec_cell( spec_cell, n );
  
  get( spec_cell->x, sizeof(SPEC) * n, fh );
  
  int i;
  for ( i = 0; i < n; i++ )
    printf( "%2d  %s\n", i, (spec_cell->x + i)->spec );
    
  align_header = (MOTIF_LIST *)get_empty_align_seq_mf_list_w_spec_cell( align_header,
    spec_cell );
    
  kill_spec_cell( spec_cell );
  
  return align_header;

}












MOTIF_LIST *get_empty_align_seq_mf_list_w_spec_cell( MOTIF_LIST *align_header,
  SPEC_CELL *spec_cell )
{

  MOTIF_LIST *create_motif_lister( MOTIF_LIST *x );

  MOTIF_LIST *prev = NULL;

  int i;
  
  for ( i = 0; i < spec_cell->n; i++ ) {
  
    MOTIF_LIST *align = NULL;
    align = (MOTIF_LIST *)create_motif_lister( align );
    
    strcpy( align->spec, (spec_cell->x + i)->spec );
  
    if ( align_header == NULL ) {
      align_header = align;
      prev = align;
    }
    else {
      prev->next = align;
      prev = align;
    }
    
  }
  
  return align_header;

}













/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/















void do_my_binary_ft( MOTIF_LIST *align_header, MOTIF_LIST *mf_header, TREE *tree,
  FILE *seq_fh, FILE *reg_fh, FILE *ft_fh, FILE *ind_fh, FILE *mf_fh, 
  FILE *fpos_fh, FILE *ss_fh )
{

  void set_position( FPOS fpos, FILE *fh );
  
  printf( "Counting motif frequencies." );
  printf( "..\n" );
  
  int use_array[ NUMB_OF_WORDS ], kmer_array[ NUMB_OF_WORDS ];
  
  mf_header = (MOTIF_LIST *)get_motif_freq_list( mf_header, use_array, kmer_array,
    align_header, tree, seq_fh, reg_fh, ft_fh, ind_fh );
    
  printf( "Done with motif frequency list.\n" );
  printf( "\n" ); 
  
  if ( READ_ONLY ) {
    print_mf_list_stdout( mf_header );
    return;
  }
        
  print_mf_binary_file( mf_header, use_array, tree, mf_fh, fpos_fh );
  
  printf( "Tabulating motif occurrences." );
  printf( "..\n" );
  
  do_binary_ft_count_tfbs( align_header, mf_header, use_array, kmer_array, tree, 
    seq_fh, reg_fh, ft_fh, ind_fh );
    
  printf( "Writing motif occurrences." );
  printf( "..\n" );
    
  do_binary_ft_put_tfbs( align_header, mf_header, use_array, kmer_array, tree, 
    seq_fh, reg_fh, fpos_fh, ft_fh, ind_fh, ss_fh );
    
}












void print_mf_list_stdout( MOTIF_LIST *mf_header )
{

  void get_string_from_stack( char cons[], MF stack, char length );
  
  extern MOTIF major_motif;

  MOTIF_LIST *mf = mf_header;
  
  int i = 0;
  
  int x = -1;
  
  while ( mf != NULL ) {
  
    char cons[ BUFFER ];
    
    get_string_from_stack( cons, mf->my_mf->my_motif, WORD_LENGTH );
    
    printf( "%6d :  %s    %d\n", ++i, cons, (int)mf->my_mf->my_motif );
    
    if ( major_motif.length == WORD_LENGTH ) {
      if ( ! strcmp( cons, major_motif.cons ))
        x = i;
    }
    
    mf = mf->next;
  
  }
  
  if ( x >= 0 )
    printf( "\n\n\nMotif rank :  %d\n", x );

}








MOTIF_LIST *get_motif_freq_list( MOTIF_LIST *mf_header, int use_array[],
  int kmer_array[], MOTIF_LIST *align_header, TREE *tree, FILE *seq_fh, 
  FILE *reg_fh, FILE *ft_fh, FILE *ind_fh )
{

  int ct_array[ NUMB_OF_WORDS ];
  COUNT total_ct = 0;
  
  initialize_int_array( ct_array, NUMB_OF_WORDS );
  
  get_motif_freq_count_array( ct_array, &total_ct, align_header, tree, seq_fh, 
    reg_fh, ft_fh );
  
  return get_motif_freq_list_from_array( mf_header, use_array, kmer_array, ct_array, 
    total_ct );

}














void initialize_int_array( int array[], int n )
{

  int i;
  
  for ( i = 0; i < n; i++ ) 
    array[ i ] = 0;

}












void get_motif_freq_count_array( int ct_array[], COUNT *total_ct, 
  MOTIF_LIST *align_header, TREE *tree, FILE *seq_fh, FILE *reg_fh, FILE *ft_fh )
{

  int get_numb_leaves_from_tree( TREE *tree );
  void set_position( FPOS fpos, FILE *fh );
  
  /*  In format.c:  */
  int n = get_numb_leaves_from_tree( tree );
  
  char line[ LINE_BUFFER ];
  
  set_position( 0, seq_fh );
  set_position( 0, reg_fh );
  
  FPOS fpos = 0;

  while ( fgets( line, LINE_BUFFER, reg_fh ) != NULL )   
    /*  In this file, below:  */
    fpos = get_binary_motif_freq_array_block( ct_array, total_ct, n, align_header, 
      tree, fpos, line, seq_fh );
      
}












FPOS get_binary_motif_freq_array_block( int ct_array[], COUNT *total_ct, int n,
  MOTIF_LIST *align_header, TREE *tree, FPOS fpos, char line[], FILE *seq_fh )
{

  extern char major_spec[ BUFFER ];

  GENCOORD_ENTRY gene_entry;
  MAF_LIST *maf_header = NULL;
  
  get_gencoord_entry_from_bed_line( &gene_entry, line );
  
  strcpy( gene_entry.spec, major_spec );
  
  maf_header = (MAF_LIST *)get_alignment_maf_list_single_gene( maf_header,
    &gene_entry, align_header, NULL, &fpos, seq_fh );
    
  *total_ct += update_count_motif_freq_array( ct_array, maf_header );
  
  kill_maf_list( maf_header );
  
  return fpos;

}











COUNT update_count_motif_freq_array( int ct_array[], MAF_LIST *maf_header )
{

  extern char major_spec[ BUFFER ];
  
  MAF_LIST *maf = NULL;
  maf = (MAF_LIST *)get_maf_lister_w_spec_str( maf_header, major_spec );
  
  if ( maf == NULL )
    return 0;
    
  return update_count_motif_freq_array_seq( ct_array, maf->my_maf->seq );

}












MAF_LIST *get_maf_lister_w_spec_str( MAF_LIST *maf_header, char spec[] )
{

  MAF_LIST *maf = maf_header;
  
  while ( maf != NULL ) {
  
    if ( ! strcmp( maf->my_maf->entry->spec, spec ))
      return maf;
      
    maf = maf->next;
  
  }
  
  return NULL;

}











COUNT update_count_motif_freq_array_seq( int ct_array[], char seq[] )
{

  MF get_bp_from_nucl( char nucl );
  
  int i, j, n = 0, end = strlen( seq );
  
  if ( end == 0 )
    return 0;
  
  char len = 0;
  MF stack = 0;
  
  int mf_len = WORD_LENGTH;
  
  for ( i = 0; i < end; i++ ) {
  
    char c = tolower( seq[ i ] );
    
    if (( c == '.' ) || ( c == 'n' )) {
      stack = 0;
      len = 0;
      continue;
    }
    
    if ( c == '-' )
      continue;
      
    stack = stack << BITS_PER_NUCL;
    stack = stack % NUMB_OF_WORDS;
      
    /*  In ../main/nucl.c:  */
    stack += get_bp_from_nucl( c );

    if ( len < mf_len ) 
      ++len;
    
    if ( len == WORD_LENGTH ) {
      ++(ct_array[ stack ]);
      ++n;
    }
      
  }
  
  return n;

}














MOTIF_LIST *get_motif_freq_list_from_array( MOTIF_LIST *mf_header, int use_array[],
  int kmer_array[], int ct_array[], COUNT total_ct )
{

  int mf_count_array[ NUMB_OF_WORDS ], rank_array[ NUMB_OF_WORDS ];
  
  int i;
  for ( i = 0; i < NUMB_OF_WORDS; i++ )
    rank_array[ i ] = i;
    
  copy_int_array( mf_count_array, ct_array, NUMB_OF_WORDS );
    
  bubble_sort( rank_array, mf_count_array, NUMB_OF_WORDS );
  
  return get_motif_freq_list_from_sorted_array( mf_header, use_array, kmer_array, 
    mf_count_array, rank_array, ct_array );

}











void copy_int_array( int copy_array[], int ct_array[], COUNT n )
{
  int i;
  for ( i = 0; i < n; i++ )
    copy_array[ i ] = ct_array[ i ];
}











void bubble_sort( int rank_array[], int ct_array[], COUNT n )
{

  int swapped = 1;
  int j = 0;
  int tmp;
  
  while ( swapped ) {
  
    int i;
    ++j;
    
    swapped = 0;
    
    for ( i = 0; i < n - j; i++ ) {

      if ( ct_array[ i ] < ct_array[ i + 1 ] ) {

        tmp = ct_array[ i ];
        ct_array[ i ] = ct_array[ i + 1 ];
        ct_array[ i + 1 ] = tmp;
        
        tmp = rank_array[ i ];
        rank_array[ i ] = rank_array[ i + 1 ];
        rank_array[ i + 1 ] = tmp;
        
        swapped = 1;

      }

    }
  
  }

}













MOTIF_LIST *get_motif_freq_list_from_sorted_array( MOTIF_LIST *mf_header, 
  int use_array[], int kmer_array[], int mf_array[], int rank_array[], 
  int ct_array[] )
{

  extern int top_rank;
  
  initialize_int_array( use_array, NUMB_OF_WORDS );
  
  get_word_use_array_from_rank_array( use_array, rank_array, top_rank );

  return get_my_motif_freq_list_from_sorted_array( mf_header, use_array, 
    rank_array, kmer_array, top_rank );

}















void get_word_use_array_from_rank_array( int use_array[], int rank_array[],
  int top_rank )
{

  int i;
  
  for ( i = 0; i < top_rank; i++ ) 
    update_word_use_array_from_single_mf( use_array, rank_array[ i ] );

}












void update_word_use_array_from_single_mf( int use_array[], int mf )
{

  int get_motif_stack_mismatches( MOTIF *mf, MF stack );
  void get_string_from_stack( char cons[], MF stack, char length );

  extern int mismatch_thresh;

  MF x;
  MF stack = (MF)mf;
  
  MOTIF my_mf = { .my_motif = stack, .length = WORD_LENGTH };
  
  /*  In ../main/nucl.c:  */
  get_string_from_stack( my_mf.cons, stack, WORD_LENGTH );
  
  int total = 0;
  
  for ( x = 0; x < NUMB_OF_WORDS; x++ ) {
    /*  In ../main/nucl.c:  */
    if ( get_motif_stack_mismatches( &my_mf, x ) <= mismatch_thresh ) {
      if ( ! use_array[ x ] )
        ++total;
      use_array[ x ] = 1; 
    }
  }
  
}















MOTIF_LIST *get_my_motif_freq_list_from_sorted_array( MOTIF_LIST *mf_header, 
  int use_array[], int rank_array[], int kmer_array[], int top_rank )
{

  MOTIF_LIST *last_mf = NULL;
  
  int x = 0;
  
  MF i, j = 0;
  
  for ( i = 0; i < top_rank; i++ )
    mf_header = (MOTIF_LIST *)make_and_put_motif_top_list_single_mf( mf_header,
      &last_mf, rank_array[ i ] );
  
  for ( i = 0; i < NUMB_OF_WORDS; i++ ) {
    if ( use_array[ i ] )
      kmer_array[ j++ ] = i;
  }
  
  kmer_array[ j ] = -1;
      
  return mf_header;

}













MOTIF_LIST *make_and_put_motif_top_list_single_mf( MOTIF_LIST *mf_header,
  MOTIF_LIST **last_mf, MF mf )
{

  MOTIF_LIST *create_motif_lister( MOTIF_LIST *x );
  void get_string_from_stack( char cons[], MF stack, char length );

  MOTIF_LIST *new_mf = NULL;
  new_mf = (MOTIF_LIST *)create_motif_lister( new_mf );
  
  new_mf->my_mf->my_motif = (MF)mf;
  new_mf->my_mf->length = WORD_LENGTH;
  
  /*  In ../main/nucl.c:  */
  get_string_from_stack( new_mf->my_mf->cons, mf, WORD_LENGTH );
  
  if ( mf_header == NULL ) {
    *last_mf = new_mf;
    return new_mf;
  }  
  
  (*last_mf)->next = new_mf;
  *last_mf = new_mf;
  
  return mf_header;

}













void print_mf_binary_file( MOTIF_LIST *mf_header, int use_array[], TREE *tree, 
  FILE *mf_fh, FILE *fpos_fh )
{

  int get_numb_leaves_from_tree( TREE *tree );
  void get_string_from_stack( char cons[], MF stack, char length );
  
  /*  In format.c:  */
  int n = get_numb_leaves_from_tree( tree );
  
  MF x;
  
  FPOS fpos = 0;
  
  for ( x = 0; x < NUMB_OF_WORDS; x++ ) {
    if ( use_array[ x ] ) {
      put( &fpos, sizeof(FPOS), fpos_fh );
      fpos += sizeof(FT_IND) * n;
    }
    else {
      FPOS ff = -1;
      put( &ff, sizeof(FPOS), fpos_fh );
    }
  }
  
  fflush( fpos_fh );
  
  MOTIF_LIST *mf = mf_header;

  int i = 0;
  
  while ( mf != NULL ) {
    put( &(mf->my_mf->my_motif), sizeof(MF), mf_fh );
    char cons[ BUFFER ];
    get_string_from_stack( cons, mf->my_mf->my_motif, WORD_LENGTH );
    printf( "%4d  %s\n", ++i, cons );
    mf = mf->next;
  }  
  
  fflush( mf_fh );

}














/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
















void do_binary_ft_count_tfbs( MOTIF_LIST *align_header, MOTIF_LIST *mf_header,
  int use_array[], int kmer_array[], TREE *tree, FILE *seq_fh, FILE *reg_fh, 
  FILE *ft_fh, FILE *ind_fh )
{

  int get_numb_leaves_from_tree( TREE *tree );
  void set_position( FPOS fpos, FILE *fh );
  
  /*  In format.c:  */
  int n = get_numb_leaves_from_tree( tree );

  char line[ LINE_BUFFER ];
  
  set_position( 0, ind_fh );  

  int loop;
  
  for ( loop = 0; loop < NUMB_OF_LOOPS; loop++ ) {
  
    long ct[ NUMB_OF_KMERS ][ NUMB_OF_SPEC ], i, j;
    
    for ( i = 0; i < n; i++ ) {
      for ( j = 0; j < NUMB_OF_KMERS; j++ )
        ct[ j ][ i ] = 0;
    }
    
    set_position( 0, seq_fh );
    set_position( 0, reg_fh );
  
    FPOS fpos = 0;
    int seq_numb = 0;
    
    while ( fgets( line, LINE_BUFFER, reg_fh ) != NULL ) {
      /*  In this file, below:  */
      fpos = get_binary_ft_count_block( ct, n, align_header, mf_header, use_array,
        kmer_array, fpos, line, loop, seq_fh );
      printf( "seq_numb = %d\n", ++seq_numb );
    }
        
    /*  In this file, below:  */    
    put_binary_ft_tfbs_count_array( ct, n, use_array, loop, ft_fh, ind_fh );
    
  }
  
}









void put_binary_ft_tfbs_count_array( long ct[ NUMB_OF_KMERS ][ NUMB_OF_SPEC ], 
  int n, int use_array[], int loop, FILE *ft_fh, FILE *ind_fh )
{

  FPOS get_position( FILE *fh );
  void set_position( FPOS fpos, FILE *fh );

  set_position( 0, ft_fh );
  
  FT empty_ft = { .seq = 0 };
  FPOS fpos_array[ NUMB_OF_SPEC ];
  
  int i, j, k;

  for ( j = 0; j < MOTIF_LENGTH; j++ ) {
    empty_ft.pos[ j ] = 0;
    empty_ft.site[ j ] = '\0'; 
  }  
  
  fpos_array[ 0 ] = 0;
  
  for ( i = 0; i < NUMB_OF_KMERS; i++ ) {
  
    if ( use_array[ i ] ) {
    
      int index = i % NUMB_OF_KMERS;
  
      for ( j = 1; j < n; j++ ) 
        fpos_array[ j ] = fpos_array[ j - 1 ] + 
          (FPOS)(ct[ index ][ j - 1 ] * sizeof(FT));

      for ( j = 0; j < n; j++ ) {
     
        FT_IND ft_ind = { .fpos = fpos_array[ j ], .n = ct[ index ][ j ], .x = 0 };

        for ( k = 0; k < ft_ind.n; k++ )
          put( &empty_ft, sizeof(FT), ft_fh );
  
        put( &ft_ind, sizeof(FT_IND), ind_fh );
        
//        printf( "%6d   %2d    %ld\n", i, j, ft_ind.n );
    
        fflush( ft_fh );
        fflush( ind_fh );
    
      }
      
      fpos_array[ 0 ] = fpos_array[ n - 1 ] +
        (FPOS)(ct[ index ][ n - 1 ] * sizeof(FT));
        
    }
  
  }
  
  set_position( 0, ind_fh );
  
  for ( i = 0; i < NUMB_OF_KMERS; i++ ) {
  
    if ( use_array[ i ] ) {
    
      int index = i % NUMB_OF_KMERS;
      
      FT_IND ind_array[ NUMB_OF_SPEC ];
      
      get( ind_array, sizeof(FT_IND) * n, ind_fh );
      
      for ( j = 0; j < n; j++ )
        printf( "%6d   %2d     %9ld    %6d\n", i, j, 
          (long)ind_array[ j ].fpos, (int)ind_array[ j ].n );
      
    }
    
  }
  
}










FPOS get_binary_ft_count_block( FPOS ct[ NUMB_OF_KMERS ][ NUMB_OF_SPEC ], int n, 
  MOTIF_LIST *align_header, MOTIF_LIST *mf_header, int use_array[], 
  int kmer_array[], FPOS fpos, char line[], int loop, FILE *seq_fh )
{

  extern char major_spec[ BUFFER ];

  GENCOORD_ENTRY gene_entry;
  MAF_LIST *maf_header = NULL;
  
  get_gencoord_entry_from_bed_line( &gene_entry, line );
  
  strcpy( gene_entry.spec, major_spec );
  
  maf_header = (MAF_LIST *)get_alignment_maf_list_single_gene( maf_header,
    &gene_entry, align_header, mf_header, &fpos, seq_fh );
    
  update_count_binary_array( ct, n, maf_header, mf_header, use_array, kmer_array,
    loop );
    
  kill_maf_list( maf_header );
  
  return fpos;

}














void update_count_binary_array( FPOS ct[ NUMB_OF_KMERS ][ NUMB_OF_SPEC ], int n, 
  MAF_LIST *maf_header, MOTIF_LIST *mf_header, int use_array[], int kmer_array[],
  int loop )
{

  if ( maf_header == NULL )
    return;

  MAF_LIST *maf = maf_header;
  
  int j;
  
  for ( j = 0; j < n; j++ ) {
    get_numb_of_tfbs_from_seq_stack( ct, maf->my_maf->seq, use_array, j );
    maf = maf->next;
  }

}











void get_numb_of_tfbs_from_seq_stack( FPOS ct[ NUMB_OF_KMERS ][ NUMB_OF_SPEC ],
  char seq[], int use_array[], int spec_numb )
{

  MF get_stack_from_string( char cons[] );
  MF get_bp_from_nucl( char nucl );
    
  int i, j, n = 0, end = strlen( seq );
  
  if ( end == 0 )
    return;
  
  char site[ BUFFER ], len = 0;
  
  MF stack = 0;
  
  int mf_len = WORD_LENGTH;
  
  for ( i = 0; i < end; i++ ) {
  
    char c = tolower( seq[ i ] );
    
    switch ( c ) {
    
    case '.' :
    case 'n' :
      stack = 0;
      len = 0;
      break;
      
    case '-' :
      break;
      
    default :
    
      stack = stack << BITS_PER_NUCL;
      stack += get_bp_from_nucl( c );
      
      if ( len < mf_len ) 
        len++;
      else 
        stack = stack % NUMB_OF_WORDS;
    
      if (( len == mf_len ) && ( use_array[ stack ])) {
        ++(ct[ stack ][ spec_numb ]);
//        printf( "%d :  %ld\n", spec_numb, ct[ stack ][ spec_numb ] );
      }
        
      break;
      
    }
      
  }
  
}











int site_equals_stack( char site[], MF stack )
{

  MF get_stack_from_string( char cons[] );
  
  /*  In ../main/nucl.c:  */
  MF site_stack = get_stack_from_string( site );
  
  return ( site_stack == stack ) ? (1) : (0);

}











int get_gencoord_entry_from_bed_line( GENCOORD_ENTRY *gene_entry, char line[] )
{

  char entry[ BUFFER ], score_str[ BUFFER ];
  int start, end;
  
  int i = 0, j = 0;

  sscanf( line, "%s%d%d%s", gene_entry->chr, &start, &end, entry );
  
  gene_entry->start = start;
  gene_entry->end = end;
  
  for ( i = 0; i < strlen( entry ); i++ ) {
    score_str[ j++ ] = entry[ i ];
    if ( entry[ i ] == '.' )
      j = 0;
  }
  
  score_str[ j ] = '\0';
  
  return atoi( score_str );

}













/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/










void do_binary_ft_put_tfbs( MOTIF_LIST *align_header, MOTIF_LIST *mf_header,
  int use_array[], int kmer_array[], TREE *tree, FILE *seq_fh, FILE *reg_fh, 
  FILE *fpos_fh, FILE *ft_fh, FILE *ind_fh, FILE *ss_fh )
{

  int get_numb_leaves_from_tree( TREE *tree );
  void set_position( FPOS fpos, FILE *fh );
  
  /*  In format.c:  */
  int n = get_numb_leaves_from_tree( tree );
  
  int spec_numb;
  
  set_position( 0, fpos_fh );

  FPOS fpos_array[ NUMB_OF_WORDS ];  
  get( fpos_array, sizeof(FPOS) * NUMB_OF_WORDS, fpos_fh );
  
  for ( spec_numb = 0; spec_numb < n; spec_numb++ ) {
  
    set_position( 0, seq_fh );
    set_position( 0, reg_fh );
      
    FT_LIST *ft_array[ NUMB_OF_WORDS ], *last_ft[ NUMB_OF_WORDS ];  
    
    int i;
    for ( i = 0; i < NUMB_OF_WORDS; i++ ) {
      ft_array[ i ] = NULL;
      last_ft[ i ] = NULL;
    }

    char line[ LINE_BUFFER ];
  
    SEQ seq_numb = 0;  
    FPOS fpos = 0;
  
    SEQ_LIST *seq_header = NULL, *last_seq = NULL;

    while ( fgets( line, LINE_BUFFER, reg_fh ) != NULL ) {
  
      fpos = get_binary_ft_put_tfbs_block( fpos_array, n, align_header, mf_header, 
        ft_array, last_ft, spec_numb, &seq_header, &last_seq, use_array, fpos, line, 
        ++seq_numb, seq_fh, ft_fh );
  
      printf( "%2d   seq = %d\n", spec_numb, (int)seq_numb );
  
    }
    
    put_ft_info_list_binary( ft_array, spec_numb, use_array, fpos_array, ind_fh, 
      ft_fh );
    
    if ( spec_numb == 0 )    
      put_seq_info_list_binary( seq_header, ss_fh );
      
    for ( i = 0; i < NUMB_OF_WORDS; i++ ) {
      kill_ft_list( ft_array[ i ] );
      ft_array[ i ] = NULL;
    }
  
  }
        
}











FPOS get_binary_ft_put_tfbs_block( FPOS fpos_array[], int n, 
  MOTIF_LIST *align_header, MOTIF_LIST *mf_header, FT_LIST *ft_array[],
  FT_LIST *last_ft[], int spec_numb, SEQ_LIST **seq_header, SEQ_LIST **last_seq, 
  int use_array[], FPOS fpos, char line[], SEQ seq_numb, FILE *seq_fh, 
  FILE *ft_fh )
{

  extern char major_spec[ BUFFER ];

  GENCOORD_ENTRY gene_entry;
  MAF_LIST *maf_header = NULL;
  
  get_gencoord_entry_from_bed_line( &gene_entry, line );
  
  strcpy( gene_entry.spec, major_spec );
  
  maf_header = (MAF_LIST *)get_alignment_maf_list_single_gene( maf_header,
    &gene_entry, align_header, mf_header, &fpos, seq_fh );
  
  if ( spec_numb == 0 )  
    *seq_header = (SEQ_LIST *)get_and_put_seq_lister_region_line( *seq_header,
      last_seq, line, maf_header );

  update_ft_put_tfbs_array( ft_array, last_ft, spec_numb, use_array, maf_header, 
    mf_header, seq_numb, ft_fh );
    
  kill_maf_list( maf_header );
  
  return fpos;

}








void update_ft_put_tfbs_array( FT_LIST *ft_array[], FT_LIST *last_ft[], 
  int spec_numb, int use_array[], MAF_LIST *maf_header, MOTIF_LIST *mf_header, 
  SEQ seq_numb, FILE *ft_fh )
{

  MAF_LIST *maf = maf_header;

  int i = 0;
  while ( i++ < spec_numb ) 
    maf = maf->next;
  
  put_binary_ft_tfbs_array( ft_array, last_ft, maf->my_maf->seq, use_array,
    seq_numb );

}











void put_binary_ft_tfbs_array( FT_LIST *ft_array[], FT_LIST *last_ft[],
  char seq[], int use_array[], SEQ seq_numb )
{

  FT_LIST *create_ft_lister( FT_LIST *x );
  MF get_stack_from_string( char cons[] );
  void set_position( FPOS fpos, FILE *fh );

  FT ft = { .seq = seq_numb };
  
  int i, j = 0, k = 0, end = strlen( seq );
  
  for ( i = 0; i < MOTIF_LENGTH; i++ ) {
    ft.pos[ i ] = 0;
    ft.site[ i ] = '\0';
  }
  
  int len = 0, mf_len = WORD_LENGTH;
    
  for ( i = 0; i < end; i++ ) {
  
    char c = tolower( seq[ i ] );
    
    switch ( c ) {
    
    case '.' :
      ft.site[ 0 ] = '\0';
      len = 0;
      break;
      
    case 'n' :
      ft.site[ 0 ] = '\0';
      len = 0;
      ++k;
      break;
      
    case '-' :
      break;
      
    default :

      if ( len < mf_len ) {
        ft.site[ len ] = c;
        ft.pos[ len++ ] = k++;
        ft.site[ len ] = '\0';
      }
      else {
        for ( j = 0; j < len - 1; j++ ) {
          ft.site[ j ] = ft.site[ j + 1 ];
          ft.pos[ j ] = ft.pos[ j + 1 ];
        }
        ft.site[ len - 1 ] = c;
        ft.pos[ len - 1 ] = k++;
        ft.site[ len ] = '\0';
      }
      
      if ( len == mf_len ) {
        
        /*  In ../main/nucl.c:  */
        MF stack = get_stack_from_string( ft.site );        
  
        if ( use_array[ stack ] ) {

          FT_LIST *ft_list = NULL;
          ft_list = (FT_LIST *)create_ft_lister( ft_list );
          
          *(ft_list->ft) = ft;
  
          if ( ft_array[ stack ] == NULL ) {
            ft_array[ stack ] = ft_list;
            last_ft[ stack ] = ft_list;
          }
          else {
            last_ft[ stack ]->next = ft_list;
            last_ft[ stack ] = ft_list;
          }
          
        }
                
      }
 
      break;
      
    }
    
  }
  
}












void put_ft_info_list_binary( FT_LIST *ft_array[], int spec_numb, int use_array[],
  FPOS fpos_array[], FILE *ind_fh, FILE *ft_fh )
{

  void set_position( FPOS fpos, FILE *fh );
  void get_string_from_stack( char cons[], MF w, char length );
  
  int i, j;
  
  for ( i = 0; i < NUMB_OF_WORDS; i++ ) {
  
    char cons[ BUFFER ];
    
    get_string_from_stack( cons, i, WORD_LENGTH );
  
    printf( "%d :  %s\n", i, cons );
  
    if ( use_array[ i ] ) {
    
      FPOS fpos = fpos_array[ i ] + (sizeof(FT_IND) * spec_numb);
      FT_IND ft_ind;
      
      printf( "\t\t\t\tfpos = %ld\n", fpos );
      
      set_position( fpos, ind_fh );
      get( &ft_ind, sizeof(FT_IND), ind_fh );
      
      set_position( ft_ind.fpos, ft_fh );
      
      FT_LIST *ft = ft_array[ i ];
      
      int k = 0;
      
      while ( ft != NULL ) {
        put( ft->ft, sizeof(FT), ft_fh );
        ft = ft->next;
        ++k;
      }
      
      fflush( ft_fh );
      
      if ( k != ft_ind.n ) {
        printf( "list:  %d   Ind file: %d\n", k, (int)ft_ind.n );
        error( 2302, "kmer occurrence count not consistent.\n" );
      }
    
    }
  
  }

}











SEQ_LIST *get_and_put_seq_lister_region_line( SEQ_LIST *seq_header, 
  SEQ_LIST **last_seq, char line[], MAF_LIST *maf_header )
{

  SEQ_LIST *create_seq_lister( SEQ_LIST *x );
  COUNT get_numb_of_nucls_seq( char seq[] );
  
  extern char major_spec[ BUFFER ];
  
  SEQ_LIST *seq = NULL;
  seq = (SEQ_LIST *)create_seq_lister( seq );
  
  SEQ_INFO *info = seq->my_seq;
  
  GENCOORD_ENTRY gene_entry = { .chr = "", .start = 0, .end = 0 };  
  info->score = get_gencoord_entry_from_bed_line( &gene_entry, line );
  
  MAF_LIST *maf = maf_header;
  
  int i = 0;
  
  while ( maf != NULL ) {
    /*  In ../main/nucl.c:  */
    seq->length[ i++ ] = get_numb_of_nucls_seq( maf->my_maf->seq );
    maf = maf->next;
  }
  
  seq->numb_of_spec = i;
  
  maf = maf_header;
  
  while ( maf != NULL ) {
    if ( ! strcmp( major_spec, maf->my_maf->entry->spec ))
      break;
    maf = maf->next;
  }
  
  strcpy( info->chr, gene_entry.chr );
  info->start = gene_entry.start;
  info->end = gene_entry.end;
  info->strand = '+';
  info->n = gene_entry.end - gene_entry.start;
  /*  In ../main/nucl.c:  */
  info->x = get_numb_of_nucls_seq( maf->my_maf->seq );
  
  if ( seq_header == NULL ) {
    *last_seq = seq;
    return seq;
  }
  
  (*last_seq)->next = seq;
  *last_seq = seq;
  
  return seq_header;

}
















void put_seq_info_list_binary( SEQ_LIST *seq_header, FILE *ss_fh )
{

  SEQ_LIST *seq = seq_header;
  
  while ( seq != NULL ) {

    put( seq->my_seq, sizeof(SEQ_INFO), ss_fh );
    
    int i;
    
    for ( i = 0; i < seq->numb_of_spec; i++ )
      put( seq->length + i, sizeof(int), ss_fh );

    seq = seq->next;

  }

}








/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/








MAF_LIST *get_alignment_maf_list_single_gene( MAF_LIST *maf_header, 
  GENCOORD_ENTRY *gene_entry, MOTIF_LIST *align_header, MOTIF_LIST *mf_header, 
  FPOS *fpos, FILE *seq_fh )
{

  void set_position( FPOS fpos, FILE *fh );

  set_position( *fpos, seq_fh );
  
  maf_header = (MAF_LIST *)make_and_set_maf_list_all_spec_cell( maf_header,
    align_header ); 
    
  return get_my_alignment_maf_list_single_gene( maf_header, gene_entry, 
    align_header, mf_header, fpos, seq_fh );

}












MAF_LIST *make_and_set_maf_list_all_spec_cell( MAF_LIST *maf_header, 
  MOTIF_LIST *mf_header )
{

  MOTIF_LIST *spec = mf_header;
  
  while ( spec != NULL ) 
    maf_header = (MAF_LIST *)make_and_set_single_maf_spec_and_step( maf_header,
      &spec );
      
  return maf_header;

}










MAF_LIST *make_and_set_single_maf_spec_and_step( MAF_LIST *maf_header, 
  MOTIF_LIST **spec )
{

  maf_header = (MAF_LIST *)make_and_set_single_maf_spec( maf_header, *spec );
  
  *spec = (*spec)->next;
  
  return maf_header;

}










MAF_LIST *make_and_set_single_maf_spec( MAF_LIST *maf_header, MOTIF_LIST *spec )
{

  MAF_LIST *create_maf_lister( MAF_LIST *mm );
  
  MAF_LIST *maf = NULL;
  maf = (MAF_LIST *)create_maf_lister( maf );
  
  strcpy( maf->my_maf->entry->spec, spec->spec );
  
  maf->my_maf->seq[ 0 ] = '\0';
  
  return put_single_maf_spec_into_list( maf_header, maf );

}












MAF_LIST *put_single_maf_spec_into_list( MAF_LIST *maf_header, MAF_LIST *maf )
{

  if ( maf_header == NULL )
    return maf;
    
  MAF_LIST *curr = maf_header;
  
  while ( curr->next != NULL )
    curr = curr->next;
    
  curr->next = maf;
  
  return maf_header;

}











MAF_LIST *get_my_alignment_maf_list_single_gene( MAF_LIST *maf_header, 
  GENCOORD_ENTRY *gene_entry, MOTIF_LIST *align_header, MOTIF_LIST *mf_header, 
  FPOS *fpos, FILE *seq_fh )
{

  FPOS get_position( FILE *ff );
  void set_position( FPOS fpos, FILE *fh );

  char line[ LINE_BUFFER ];
  int after_found = 0;
  int same_chr = 0;
  
  FPOS my_fpos;
  
  my_fpos = *fpos;
  
  while ( fgets( line, LINE_BUFFER, seq_fh ) != NULL ) {
  
    maf_header = (MAF_LIST *)get_alignment_maf_single_multiz_line( maf_header,
      line, gene_entry, align_header, mf_header, fpos, &my_fpos, &after_found,
      &same_chr, seq_fh );
  
    if (( after_found < 0 ) || ( same_chr < 0 ))
      return maf_header;
      
  }
  
  if ( same_chr == 0 ) {
    
    *fpos = 0;

    set_position( 0, seq_fh );
    
    while ( fgets( line, LINE_BUFFER, seq_fh ) != NULL ) {
    
      maf_header = (MAF_LIST *)get_alignment_maf_single_multiz_line( maf_header,
        line, gene_entry, align_header, mf_header, fpos, &my_fpos, &after_found,
        &same_chr, seq_fh );
  
      if (( after_found < 0 ) || ( same_chr < 0 ))
        return maf_header;
      
    }
  
  }
  
  return maf_header;

}













MAF_LIST *get_alignment_maf_single_multiz_line( MAF_LIST *maf_header, char line[], 
  GENCOORD_ENTRY *gene_entry, MOTIF_LIST *align_header, MOTIF_LIST *mf_header, 
  FPOS *fpos, FPOS *my_fpos, int *after_found, int *same_chr, FILE *seq_fh )
{

  FPOS get_position( FILE *fh );

  extern char major_spec[ BUFFER ];
  
  char spec[ BUFFER ];

  if ( line[ 0 ] != 's' )
    return maf_header;
    
  get_spec_from_maf_line( spec, line );
  
  if ( strcmp( spec, major_spec )) 
    return maf_header;
    
  *my_fpos = get_position( seq_fh );

  GENCOORD_ENTRY maf_entry;
  get_gencoord_entry_from_maf_line( &maf_entry, line );
  
  *same_chr = set_same_chr( gene_entry, &maf_entry, *same_chr );
  
  if ( *same_chr < 0 )
    return maf_header;
    
  if ( ! gene_gencoord_matches_maf_gencoord( gene_entry, &maf_entry )) {
    if ( *after_found == 1 )
      *after_found = -1;
    return maf_header;
  }
  
  *after_found = 1;
  
  *fpos = *my_fpos;
  
  return get_alignment_maf_first_major_species_line( maf_header, 
    line, &maf_entry, gene_entry, after_found, fpos, seq_fh );
  
}











void get_spec_from_maf_line( char spec[], char line[] )
{

  char s[ BUFFER ];
  
  sscanf( line, "%*s%s", s );
  
  char i, end = strlen( s );
  
  for ( i = 0; i < end; i++ ) {
    if ( s[ i ] == '.' ) {
      spec[ i ] = '\0';
      break;
    }
    spec[ i ] = s[ i ];
  }

}








void get_gencoord_entry_from_maf_line( GENCOORD_ENTRY *entry, char line[] )
{

  char x[ BUFFER ], start[ BUFFER ], length[ BUFFER ], strand[ BUFFER ];
  int chr_length;
  
  if (( line[ 0 ] == '\n' ) || ( line[ 0 ] == '\0' ))
    return;
    
  sscanf( line, "%*s%s%s%s%s%d",
    x,
    start,
    length,
    strand,
    &chr_length );
    
  register char i = 0, j = 0;
  
  while ( x[ i ] != '.' ) {
    entry->spec[ i ] = x[ i ];
    ++i;
  }
    
  entry->spec[ i ] = '\0';
  
  strcpy( entry->chr, x + i + 1 );
  
  entry->start = (COORD)atoll( start );
  entry->end = entry->start + (COORD)atoll( length );
  entry->strand = strand[ 0 ];
  
}











int set_same_chr( GENCOORD_ENTRY *gene_entry, GENCOORD_ENTRY *maf_entry, 
  int same_chr )
{

  if (( ! strcmp( maf_entry->spec, gene_entry->spec )) &&
    ( ! strcmp( maf_entry->chr, gene_entry->chr )))
  return 1;
  
  if ( same_chr != 1 )
    return same_chr;
    
  if (( ! strcmp( maf_entry->spec, gene_entry->spec )) &&
    ( strcmp( maf_entry->chr, gene_entry->chr )))
  return -1;
  
  return same_chr;

}












int gene_gencoord_matches_maf_gencoord( GENCOORD_ENTRY *gene_entry, 
  GENCOORD_ENTRY *maf_entry )
{

  if ( strcmp( gene_entry->spec, maf_entry->spec ))
    return 0;
    
  if ( gene_entry->start > maf_entry->end )
    return 0;
    
  if ( gene_entry->end < maf_entry->start )
    return 0;
    
  if ( strcmp( gene_entry->chr, maf_entry->chr ))
    return 0;
    
  return 1;

}















MAF_LIST *get_alignment_maf_first_major_species_line( MAF_LIST *maf_header, 
  char line[], GENCOORD_ENTRY *maf_entry, GENCOORD_ENTRY *gene_entry, 
  int *after_found, FPOS *fpos, FILE *seq_fh )
{

  char my_line[ LINE_BUFFER ];
  
  int still_going = 1;
  
  maf_header = (MAF_LIST *)get_alignment_maf_single_matching_line( maf_header,
      line, line, maf_entry, gene_entry, after_found, fpos, &still_going, seq_fh );
  
  while (( still_going ) && ( fgets( my_line, LINE_BUFFER, seq_fh ) != NULL ))   
    maf_header = (MAF_LIST *)get_alignment_maf_single_matching_line( maf_header,
      my_line, line, maf_entry, gene_entry, after_found, fpos, &still_going, 
      seq_fh );
      
  return maf_header;

}










MAF_LIST *get_alignment_maf_single_matching_line( MAF_LIST *maf_header,
  char my_line[], char line[], GENCOORD_ENTRY *maf_entry, 
  GENCOORD_ENTRY *gene_entry, int *after_found, FPOS *fpos, int *still_going, 
  FILE *seq_fh )
{

  FPOS get_position( FILE *fh );
  
  extern char major_spec[ BUFFER ];
  
  if ( my_line[ 0 ] != 's' ) {
    if ( my_line[ 0 ] == 'a' )
      *still_going = 0;
    return maf_header;
  }
  
  char spec[ BUFFER ];
  
  get_spec_from_maf_line( spec, my_line );

//  printf( "spec = %s\n", spec );
  
//  printf( "line :  %s\n", my_line );
  
  if ( ! strcmp( spec, major_spec )) { 

    GENCOORD_ENTRY curr_maf_entry;
    get_gencoord_entry_from_maf_line( &curr_maf_entry, my_line );
  
    if ( ! gene_gencoord_matches_maf_gencoord( gene_entry, &curr_maf_entry )) {
      *still_going = 0;
      return maf_header;
    }

  }
    
  MAF_LIST *maf = maf_header;
  
  while ( maf != NULL ) {
  
//    printf( "   my_spec = %s\n", maf->my_maf->entry->spec );
  
    if ( ! strcmp( maf->my_maf->entry->spec, spec )) {
      int len = strlen( maf_header->my_maf->seq );
      if ( ! strcmp( maf->my_maf->entry->spec, major_spec ))
        len = -1;
      update_alignment_maf_single_maf_line( maf, len, my_line );
      return maf_header;
    }
    
    maf = maf->next;
  
  }
  
  error( 2201, "Species not found.\n" );
      
  return NULL;
    
}








void update_alignment_maf_single_maf_line( MAF_LIST *maf, int len, char line[] )
{

  int seq_len = strlen( maf->my_maf->seq );
  
  char new_seq[ LINE_BUFFER ], n_seq[ LINE_BUFFER ];
  
  sscanf( line, "%*s%*s%*d%*d%*s%*d%s", new_seq );
  
  int n = len - (strlen( new_seq ) + seq_len);
  
  if ( len == -1 )
    n = 0;
  
  int i;
  
  for ( i = 0; i < n; i++ )
    n_seq[ i ] = '.';
    
  n_seq[ n ] = '\0';
  
  strcat( maf->my_maf->seq, n_seq );
  strcat( maf->my_maf->seq, new_seq );
    
}









