#include "nucl.h"

















MF get_bp_from_nucl( char nucl )
{

  switch ( tolower( nucl )) {
  
  case 'a' :
    return A_NUCL;

  case 'c' :
    return C_NUCL;
    
  case 'g' :
    return G_NUCL;
    
  case 't' :
    return T_NUCL;
    
  default :
    printf( "char = %c\n", nucl );
    error( 702, "Unrecognized nucleotide type.\n" );
    break;
  
  }

}











char get_nucl_from_bp( MF bp )
{

  switch ( bp ) {
  
  case A_NUCL :
    return 'a';
    
  case C_NUCL :
    return 'c';
    
  case G_NUCL :
    return 'g';
    
  case T_NUCL :
    return 't';
    
  default :
    error( 703, "Unrecognized basepair.\n" );
    break;
  
  }

}









MF get_stack_from_string( char cons[] )
{

  int i, end = strlen( cons );
  
  MF stack = 0;
  
  for ( i = 0; i < end; i++ ) {
    stack = stack << BITS_PER_NUCL;
    stack += get_bp_from_nucl( cons[ i ] );
  }
  
  return stack;

}














void get_string_from_stack( char cons[], MF stack, char length )
{

  int i;
  
  for ( i = length - 1; i >= 0; i-- ) {
    cons[ i ] = get_nucl_from_bp( stack % NUMB_OF_NUCL );
    stack = stack >> BITS_PER_NUCL;
  }
  
  cons[ length ] = '\0';

}













int motif_matches_stack( MOTIF *mf, MF stack )
{

  MF ww = stack;

  char i;
  
  for ( i = mf->length - 1; i >= 0; i-- ) {
  
    if ( ! nucl_matches_bp_degen( mf->cons[ i ], ww % NUMB_OF_NUCL ))
      return 0;
      
    ww = ww >> BITS_PER_NUCL;
  
  }
  
  return 1;

}









int get_motif_stack_mismatches( MOTIF *mf, MF stack )
{

  MF ww = stack;

  int i, mm = 0;
  
  for ( i = mf->length - 1; i >= 0; i-- ) {
    if ( ! nucl_matches_bp_degen( mf->cons[ i ], ww % NUMB_OF_NUCL ))
      ++mm;
    ww = ww >> BITS_PER_NUCL;
  }
  
  return mm;

}











int get_motif_match_in_seq_offset( int *start, int *end, char seq[], MOTIF *mf, 
  char mm_max, int match_index )
{

  int i, seq_end = strlen( seq ) - mf->length + 1;
  int matches = 0;
  
  for ( i = 0; i < seq_end; i++ ) {
  
    if ( ! char_is_nucl( seq[ i ] ))
      continue;

    int j, len = 0, mm = 0, mf_end = strlen( seq );
    
    char site[ BUFFER ];
    
    for ( j = 0; j < mf_end; j++ ) {
    
      if ( seq[ i + j ] == '\0' )
        break;
        
      if ( char_is_nucl( seq[ i + j ] ))
        site[ len++ ] = tolower( seq[ i + j ] );
    
      if ( len == mf->length ) {
        site[ len ] = '\0';
        break;
      }
      
    }
    
    if ( len == mf->length ) {
    
      char rc_site[ BUFFER ];
      
      get_rev_complement_seq( rc_site, site );
      
      MF site_stack = get_stack_from_string( site );
      MF rc_stack = get_stack_from_string( rc_site );

      if ( get_motif_stack_mismatches( mf, site_stack ) <= mm_max ) {
        if ( matches == match_index ) {
          *start = i;
          *end = i + j;
          return 1;
        }
        ++matches;
      }

      else if ( get_motif_stack_mismatches( mf, rc_stack ) <= mm_max ) {
        if ( matches == match_index ) {
          *start = i;
          *end = i + j;
          return 1;
        }
        ++matches;
      }

    }
    
  }
  
  return 0;

}









MF numb_of_mfs_from_length( char length )
{

  MF x = 1;
  
  char i = 0;
  
  while ( i++ < length )
    x *= NUMB_OF_NUCL;
    
  return x;

}










char complement_nucl( char nucl )
{

  switch ( tolower( nucl )) {
  
  case 'a' :
    return 'T';

  case 'c' :
    return 'G';
    
  case 'g' :
    return 'C';
    
  case 't' :
    return 'A';

  case 'w' :
    return 'W';
    
  case 's' :
    return 'S';
    
  case 'm' :
    return 'K';
    
  case 'k' :
    return 'M';  
    
  case 'r' :
    return 'Y';
    
  case 'y' :
    return 'R';
    
  case '-' :
    return '-';
    
  case 'n' :
    return 'N';
    
  default :
    printf( "c = '%c'\n", nucl );
    error( 700, "Unrecognized nucleotide type.\n" );
    break;
  
  }

}









MF complement_bp( MF bp )
{

  switch ( bp ) {
  
  case A_NUCL :
    return T_NUCL;
    
  case C_NUCL :
    return G_NUCL;
    
  case G_NUCL :
    return C_NUCL;
    
  case T_NUCL :
    return A_NUCL;
    
  default :
    error( 740, "Unrecognized bp.\n" );
    break;
  
  }

}









void get_rev_complement_seq( char rev_compl[], char seq[] )
{

  int begin = strlen( seq ) - 1;
  
  register int i, j = 0;
  
  for ( i = begin; i >= 0; --i )
    /*  In this file, above:  */
    rev_compl[ j++ ] = complement_nucl( seq[ i ] );
    
  rev_compl[ j ] = '\0';
  
}









int seq_is_empty( char seq[] )
{

  int end = strlen( seq );

  register int i;
  
  for ( i = 0; i < end; i++ ) {
    
    switch ( tolower( seq[ i ] )) {
      
    case 'a' :
    case 'c' :
    case 'g' :
    case 't' :
      return 0;
      
    default :
      break;  
      
    }
    
  }
  
  return 1;

}













void get_nucl_seq_from_align( char seq[], char align[] )
{

  int i, j = 0, end = strlen( align );
  
  for ( i = 0; i < end; i++ ) {
  
    if ( char_is_nucl( align[ i ] ))
      seq[ j++ ] = tolower( align[ i ] );
  
  }
  
  seq[ j ] = '\0';

}















void upper( char a[], char b[] )
{

  int i;
  
  for ( i = 0; i < strlen( b ); i++ )
    a[ i ] = toupper( b[ i ] );
    
  a[ strlen( b ) ] = '\0';

}













void lower( char a[], char b[] )
{

  int i;
  
  for ( i = 0; i < strlen( b ); i++ )
    a[ i ] = tolower( b[ i ] );
    
  a[ i ] = '\0';

}












char get_numb_lowercase( char cons[] )
{

  char len = 0;

  int i;
  
  for ( i = 0; i < strlen( cons ); i++ ) {
    if ( islower( cons[ i ] ))
      ++len;
  }
  
  return len;

}













void get_repeat_char_string( char str[], char c, char length )
{

  int i;
  
  for ( i = 0; i < length; i++ )
    str[ i ] = c;
    
  str[ length ] = '\0';

}










int string_is_repeat( char str[], char length )
{

  char rep[ BUFFER ];

  int i, j;
  
  for ( i = 0; i < length; i++ ) {
    rep[ i ] = str[ i ];
  }
  
  rep[ length ] = '\0';
  
  for ( i = length; i < strlen( str ) - length + 1; i += length ) {
  
    for ( j = 0; j < length; j++ ) {
      if ( tolower( str[ i + j ] ) != tolower( rep[ j ] ))
        return 0;
    }
  
  }
  
  return 1;

}









int motif_is_full_degen( MOTIF *mf )
{

  int i;
  
  for ( i = 0; i < mf->length; i++ ) {
  
    char c = mf->cons[ i ];
    
    if ( tolower( c ) != 'n' )
      return 0;
  
  }
  
  return 1;

}









/*================================================================================*/
/*================================================================================*/
/*================================================================================*/
/*================================================================================*/
/*================================================================================*/

















COUNT get_numb_of_nucls_seq( char seq[] )
{

  int i, end = strlen( seq );
  
  COUNT nucl_count = 0;
  
  for ( i = 0; i < end; i++ ) 
    nucl_count += char_is_nucl( seq[ i ] );
    
  return nucl_count;

}














COUNT get_numb_of_degen_nucls_seq( char seq[] )
{

  int i, end = strlen( seq );
  
  COUNT nucl_count = 0;
  
  for ( i = 0; i < end; i++ )
    nucl_count += ( tolower( seq[ i ] ) == 'n' ) ? (1) : (0);
    
  return nucl_count;

}









int char_is_nucl( char c )
{

  switch ( tolower( c )) {
  
  case 'a' :
  case 'c' :
  case 'g' :
  case 't' :
  case 'n' :
    return 1;
    
  default :
    return 0;
  
  }

}












int char_is_non_degen_nucl( char c )
{

  switch ( tolower( c )) {
  
  case 'a' :
  case 'c' :
  case 'g' :
  case 't' :
    return 1;
    
  default :
    return 0;
  
  }

}










POS get_col_nucl_offset( char seq[], POS x )
{

  int i, end = strlen( seq );
  
  int count = 0;
  
  for ( i = 0; i < end; i++ ) {
  
    if ( char_is_nucl( seq[ i ] )) {
      if ( count++ == x )
        return i;
    }
  
  }
  
  return end - 1;

}











POS get_offset( COORD a, COORD b )
{

  if ( b < a )
    error( 701, "Offset positions reversed.\n" );

  if ( a * b >= 0 )
    return (POS)(b - a);
    
  return b - a - 1;

}














void get_word_w_length( char wd[], char seq[], POS i, char length )
{

  int char_is_nucl( char c );

  int k, j = 0, len = 0, end = strlen( seq );
  
  for ( k = i; k < end; k++ ) {
  
    wd[ j++ ] = seq[ k ];
    
    if ( char_is_nucl( seq[ k ] ))
      ++len;
    
    if ( seq[ k ] == '\0' )
      return;
      
    if ( len == length ) {
      wd[ j ] = '\0';
      return;
    }  
  
  }

}














int strs_match( char a[], char b[] )
{

  if ( strlen( a ) != strlen( b ))
    return 0;
    
  int i;
  
  for ( i = 0; i < strlen( a ); i++ ) {
  
    if ( tolower( a[ i ] ) != tolower( b[ i ] ))
      return 0;
  
  }
  
  return 1;

}















int get_word_and_gaps( char word[], char seq[], char length )
{

  int gaps = 0;

  int i, j = 0, end = strlen( seq );
  
  for ( i = 0; i < end; i++ ) {
  
    if ( char_is_nucl( seq[ i ] ) == 0 )
      ++gaps;
    else
      word[ j++ ] = seq[ i ];
      
    if ( j == length )
      break;
  
  }
  
  word[ j ] == '\0';
  
  return gaps;

}










/*================================================================================*/
/*================================================================================*/
/*================================================================================*/
/*================================================================================*/
/*================================================================================*/













int nucl_matches_bp_degen( char nucl, MF bp )
{

  switch ( tolower( nucl )) {
  
  case 'a' :
    return ( bp == A_NUCL ) ? (1) : (0);    
  
  case 'c' :
    return ( bp == C_NUCL ) ? (1) : (0);    

  case 'g' :
    return ( bp == G_NUCL ) ? (1) : (0);    

  case 't' :
    return ( bp == T_NUCL ) ? (1) : (0);    
    
  case 'k' :
    if (( bp == G_NUCL ) || ( bp == T_NUCL ))
      return 1;
    return 0;

  case 'm' :
    if (( bp == A_NUCL ) || ( bp == C_NUCL ))
      return 1;
    return 0;

  case 'r' :
    if (( bp == A_NUCL ) || ( bp == G_NUCL ))
      return 1;
    return 0;

  case 's' :
    if (( bp == C_NUCL ) || ( bp == G_NUCL ))
      return 1;
    return 0;

  case 'w' :
    if (( bp == A_NUCL ) || ( bp == T_NUCL ))
      return 1;
    return 0;

  case 'y' :
    if (( bp == C_NUCL ) || ( bp == T_NUCL ))
      return 1;
    return 0;
    
  case 'b' :
    if (( bp == C_NUCL ) || ( bp == G_NUCL ) || ( bp == T_NUCL ))
      return 1;
    return 0;

  case 'd' :
    if (( bp == A_NUCL ) || ( bp == G_NUCL ) || ( bp == T_NUCL ))
      return 1;
    return 0;

  case 'h' :
    if (( bp == A_NUCL ) || ( bp == C_NUCL ) || ( bp == T_NUCL ))
      return 1;
    return 0;

  case 'v' :
    if (( bp == A_NUCL ) || ( bp == C_NUCL ) || ( bp == G_NUCL ))
      return 1;
    return 0;
    
  case 'n' :
    return 1;
    
  default :
    error( 705, "Unrecognized degenerate nucleotide for comparison.\n" );
    break;

  }

}













char get_double_degen_char( MF a, MF b )
{

  MF x = MIN( a, b );
  MF y = MAX( a, b );
  
  switch ( x ) {
  
  case A_NUCL :
  
    switch ( y ) {
    
    case C_NUCL :
      return 'M';
      
    case G_NUCL :
      return 'R';
      
    case T_NUCL :
      return 'W';
      
    default :
      error( 730, "Unrecognized nucl.\n" );
      break;
    
    }
    
  case C_NUCL :
  
    switch ( y ) {
    
    case G_NUCL :
      return 'S';
      
    case T_NUCL :
      return 'Y';
      
    default :
      error( 731, "Unrecognized nucl.\n" );
      break;
    
    }
    
  case G_NUCL :
  
    switch ( y ) {
    
    case T_NUCL :
      return 'K';
      
    default :
      error( 732, "Unrecognized nucl.\n" );
      break;
    
    }
    
  default :
    error( 733, "Unrecognized nucl.\n" );
    break;
  
  }

}








