#include "analysis.h"



















void set_pdata_leaves( TREE *tree, PDATA_LIST *pdata )
{

  NODE_LIST *leaf = tree->spec_header;

  int i = 0;
  
  while ( leaf != NULL ) {
    if ( leaf->my_node->leaf )
      set_single_pdata_leaf_and_step( &leaf, pdata, i++ );
    else
      leaf = leaf->next;
  }

}












void set_single_pdata_leaf_and_step( NODE_LIST **node, PDATA_LIST *pdata, int i )
{
  set_single_pdata_leaf( *node, pdata, i );
  *node = (*node)->next;
}











void set_single_pdata_leaf( NODE_LIST *node, PDATA_LIST *pdata, int i )
{

  if ( pdata->my_pdata->use_array[ i ] ) 
    set_my_single_pdata_leaf( node, pdata, i );

}












void set_my_single_pdata_leaf( NODE_LIST *node, PDATA_LIST *pdata, int i )
{

  void set_use_leaf( NODE_LIST *leaf );
  
  /*  In ../birth/set.c:  */
  set_use_leaf( node );
  
  int x = pdata->my_pdata->ct_array[ i ];
  
  node->my_info->x[ x ] = 1.0;
  node->my_info->max = x;
  node->my_info->mean = x;
  
}
















void set_use_leaf( NODE_LIST *leaf )
{

  NODE_LIST *node = leaf;
  
  while (( node != NULL ) && ( node->my_info->use == 0 )) {
  
    node->my_info->use = 1;
    
    node = node->up;
  
  }

}

















/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/












void copy_models( MODEL_LIST *a, MODEL_LIST *b )
{

  *(a->my_model->mf) = *(b->my_model->mf);
  *(a->my_model->my_region) = *(b->my_model->my_region);
  *(a->my_model->stats) = *(b->my_model->stats);
  *(a->my_model->null_stats) = *(b->my_model->null_stats);
  
  a->my_model->m = b->my_model->m;
  a->my_model->part_n = b->my_model->part_n;
  
  copy_part_list( a->my_model->part_header, b->my_model->part_header );

}















void copy_part_list( PART_LIST *a_header, PART_LIST *b_header )
{

  PART_LIST *a = a_header, *b = b_header;
  
  while ( b != NULL )
    copy_single_part_lister_and_step( &a, &b );

}















void copy_single_part_lister_and_step( PART_LIST **a, PART_LIST **b )
{
  copy_single_part_lister( *a, *b );
  *a = (*a)->next;
  *b = (*b)->next;
}
















void copy_single_part_lister( PART_LIST *a, PART_LIST *b )
{

  if ( a == NULL )
    error( 5300, "Unexpected null partition while copying list.\n" );
    
  LIST *list_header = NULL;
  list_header = (LIST *)copy_list( b->my_part->branch_header );
  
  a->my_part->branch_header = list_header;
  
  *(a->my_part->pp) = *(b->my_part->pp);
  
  a->my_part->part = b->my_part->part;

}
















LIST *copy_list( LIST *b_header )
{

  LIST *a_header = NULL;
  LIST *b = b_header;
  
  while ( b != NULL )
    a_header = (LIST *)copy_single_lister_and_step( a_header, &b );
    
  return a_header;

}















LIST *copy_single_lister_and_step( LIST *a_header, LIST **b )
{
  a_header = (LIST *)copy_single_lister( a_header, *b );
  *b = (*b)->next;
  return a_header;
}















LIST *copy_single_lister( LIST *a_header, LIST *b )
{

  LIST *create_lister( LIST *x, char tt );
  
  LIST *new = NULL;
  new = (LIST *)create_lister( new, INT_TYPE );
  
  new->elem = b->elem;
  new->type = b->type;
  
  return put_lister_end_of_list( a_header, new );

}

















LIST *put_lister_end_of_list( LIST *header, LIST *new )
{

  if ( header == NULL )
    return new;
    
  LIST *curr = header;
  
  while ( curr->next != NULL )
    curr = curr->next;
    
  curr->next = new;
  
  return header;

}











void set_pt_cell_from_pt_list( PT_CELL *pt_cell, PART_LIST *part_header )
{

  PART_LIST *part = part_header;

  int i;
  
  for ( i = 0; i < pt_cell->n; i++ )
    set_pt_cell_from_pt_list_single_part_and_step( pt_cell, &part, i ); 

}














void set_pt_cell_from_pt_list_single_part_and_step( PT_CELL *pt_cell, 
  PART_LIST **part, int i )
{
  set_pt_cell_from_pt_list_single_part( pt_cell, *part, i );
  *part = (*part)->next;
}











void set_pt_cell_from_pt_list_single_part( PT_CELL *pt_cell, PART_LIST *part,
  int i )
{

  *(pt_cell->pp[ i ]) = *(part->my_part->pp);

  /*  In this file, above:  */
  pt_cell->branch[ i ] = (LIST *)copy_list( part->my_part->branch_header );
  
}














/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/













void clear_tree( NODE_LIST *root )
{

  if ( root == NULL )
    return;
    
  clear_node_lister( root );
  
  int i;
  
  for ( i = 0; i < BRANCH_BUFFER; i++ ) {
  
    if ( root->down[ i ] == NULL )
      return;
      
    clear_tree( root->down[ i ] );
    
  }

}













void clear_node_lister( NODE_LIST *node )
{
  clear_node_info( node->my_info );
}













void clear_node_info( NODE_INFO *info )
{

  extern int max_occur;
  
  int end = max_occur + 3;

  int i;
  
  for ( i = 0; i < end; i++ )
    info->x[ i ] = 0.0;
    
  info->mean = 0.0;
  info->max = 0;
  info->use = 0;
  info->change = 0;

}

















void clear_node_info_array( STAT x[], int max )
{

  int my_max = MIN( max, OCCUR_BUFFER );

  int i;
  
  for ( i = 0; i < my_max; i++ )
    x[ i ] = 0.0; 

}












void clear_tree_all( NODE_LIST *root )
{

  if ( root == NULL )
    return;
    
  clear_node_lister_all( root );
  
  int i;
  
  for ( i = 0; i < BRANCH_BUFFER; i++ ) {
  
    if ( root->down[ i ] == NULL )
      break;
      
    clear_tree_all( root->down[ i ] );
    
  }

}














void clear_node_lister_all( NODE_LIST *node )
{

  /*  In this file, above:  */
  clear_node_info( node->my_info );
  clear_node_info( node->my_dd );
  
  int i;
  
  for ( i = 0; i < DBUFF; i++ )
    /*  In this file, above:  */
    clear_node_info( node->my_d[ i ] );
  
}














/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
















void copy_tree_data( TREE_DATA *d, TREE_DATA *a )
{

  MODEL_LIST *create_model_lister( MODEL_LIST *x, MOTIF *mf, INDEX m );
  OCCUR_CELL *create_occur_cell( OCCUR_CELL *x, INDEX nn );
//  void copy_occur_cells( OCCUR_CELL *a, OCCUR_CELL *b );
  NODE_LIST *copy_tree( NODE_LIST *root );
  
  d->model = (MODEL_LIST *)create_model_lister( d->model, a->model->my_model->mf,
    a->model->my_model->m );
    
//  d->occur = (OCCUR_CELL *)create_occur_cell( d->occur, a->occur->n );
  
  /*  In this file, above:  */  
  copy_models( d->model, a->model );
  /*  In lhood.c:  */
//  copy_occur_cells( d->occur, a->occur );

  /*  In ../tree/get.c:  */
  d->root = (NODE_LIST *)copy_tree( a->root );
  
}











void kill_all_tree_data( TREE_DATA *d )
{

  void kill_tree_from_root( NODE_LIST *root );

  /*  In ../main/dyn_alloc.c:  */
  kill_model_list( d->model );
  kill( d->occur );
  
  /*  In ../tree/create_tree.c:  */
  kill_tree_from_root( d->root );

}



























