#include "em.h"















#define MAX_ITER 7
#define PARAM_THRESH 1e-5
#define LHOOD_THRESH 1e-7


#define DO_NEWTON 1


#define MIN_BD_RATE 1e-25


#define PRINT_BOOL 0



#define NEG_INF -1e20














MODEL_LIST *get_opt_model_em( MODEL_LIST *new_model, MODEL_LIST *old_model,
  DATA *data )
{

  MODEL_LIST *create_model_lister( MODEL_LIST *x, MOTIF *mf, INDEX m );
  EM_INFO *create_em_info( EM_INFO *x, INDEX n );
  void copy_models( MODEL_LIST *a, MODEL_LIST *b );
  
  EM_INFO *info = NULL;

  new_model = (MODEL_LIST *)create_model_lister( new_model, data->mf, 
    old_model->my_model->m );
  
  info = (EM_INFO *)create_em_info( info, old_model->my_model->part_n );
  
  /*  In ../birth/model_set.c:  */
  copy_models( new_model, old_model );
  
  /*  In this file, below:  */
  set_em_info_from_model( info, old_model, data );
  
  info->data->new_model = new_model;
  info->data->old_model = old_model;
  
  /*  In this file, below:  */
  new_model = (MODEL_LIST *)get_my_opt_model_em( new_model, info );
  
  kill_em_info( info );
  
  return new_model;

}











MODEL_LIST *get_opt_model_em_new( MODEL_LIST *new_model, MODEL_LIST *old_model,
  DATA *data )
{

  MODEL_LIST *create_model_lister( MODEL_LIST *x, MOTIF *mf, INDEX m );
  EM_INFO *create_em_info( EM_INFO *x, INDEX n );
  void copy_models( MODEL_LIST *a, MODEL_LIST *b );
  
  EM_INFO *info = NULL;

  new_model = (MODEL_LIST *)create_model_lister( new_model, data->mf, 
    old_model->my_model->m );
  
  info = (EM_INFO *)create_em_info( info, old_model->my_model->part_n );
  
  /*  In ../birth/model_set.c:  */
  copy_models( new_model, old_model );
  
  /*  In this file, below:  */
  set_em_info_from_model( info, old_model, data );
  
  info->data->new_model = new_model;
  info->data->old_model = old_model;
  
  /*  In this file, below:  */
  new_model = (MODEL_LIST *)get_my_opt_model_em_new( new_model, info );
  
  kill_em_info( info );
  
  return new_model;

}









MODEL_LIST *get_opt_model_em_full( MODEL_LIST *new_model, MODEL_LIST *old_model,
  DATA *data )
{

  MODEL_LIST *create_model_lister( MODEL_LIST *x, MOTIF *mf, INDEX m );
  EM_INFO *create_em_info( EM_INFO *x, INDEX n );
  void copy_models( MODEL_LIST *a, MODEL_LIST *b );
  
  EM_INFO *info = NULL;

  new_model = (MODEL_LIST *)create_model_lister( new_model, data->mf, 
    old_model->my_model->m );
  
  info = (EM_INFO *)create_em_info( info, old_model->my_model->part_n );
  
  /*  In ../birth/model_set.c:  */
  copy_models( new_model, old_model );
  
  /*  In this file, below:  */
  set_em_info_from_model( info, old_model, data );
  
  info->data->new_model = new_model;
  info->data->old_model = old_model;
  
  /*  In this file, below:  */
  new_model = (MODEL_LIST *)get_my_opt_model_em_full( new_model, info );
  
  kill_em_info( info );
  
  return new_model;

}













void set_em_info_from_model( EM_INFO *info, MODEL_LIST *model, DATA *data )
{

  void set_pt_cell_from_pt_list( PT_CELL *pt_cell, PART_LIST *part_header );

  extern int init;
  
  /*  In ../birth/model_set.c:  */
  set_pt_cell_from_pt_list( info->x, model->my_model->part_header );
  set_pt_cell_from_pt_list( info->old_x, model->my_model->part_header );
  
  if ( init ) {
    int i;
    for ( i = 0; i < info->x->n; i++ ) {
      *(info->x->pp[ i ]) = *(info->x->pp[ 0 ]);
      *(info->old_x->pp[ i ]) = *(info->x->pp[ 0 ]);
    }
  }
  
  info->data = data;
  
  info->lhood = NEG_INF;
  info->old_lhood = NEG_INF;
  
  info->part_n = model->my_model->part_n;
  
  info->iter = 0;
  info->conv = 0;
  info->outcome = OUTCOME_GOING;
  
}













MODEL_LIST *get_my_opt_model_em( MODEL_LIST *model, EM_INFO *info )
{

  do {  
  
    if ( PRINT_BOOL )
      printf( "Iteration :  %d\n", info->iter );

    do_opt_model_em_single_iter( info );

  } while ( em_still_going( info ));
  
  return set_model_lister_em_info( model, info );

}












MODEL_LIST *get_my_opt_model_em_new( MODEL_LIST *model, EM_INFO *info )
{

  do {
  
    do_opt_model_em_new_single_iter( info );
  
  } while ( em_still_going( info ));
  
  return set_model_lister_em_info( model, info );

}









MODEL_LIST *get_my_opt_model_em_full( MODEL_LIST *model, EM_INFO *info )
{

  do {
  
    do_opt_model_em_full_single_iter( info );
    
  } while ( em_still_going( info ));
  
  return set_model_lister_em_info( model, info );

}







/*============================================================================*/
/*============================================================================*/
/*============================================================================*/
/*============================================================================*/
/*============================================================================*/











int em_still_going( EM_INFO *info )
{

  if ( info->conv )
    return 0;

  if ( ++(info->iter) >= MAX_ITER ) {
    info->outcome = OUTCOME_MAX_ITER;
    return 0;
  }
    
  if ( em_converged( info )) {
    info->conv = 1;
    info->outcome = OUTCOME_CONVERGED;
    return 0;
  }
  
  return 1;

}











int em_converged( EM_INFO *info )
{

  if ( em_lhood_converged( info ))
    return 1;

  int i;
  
  for ( i = 0; i < info->x->n; i++ ) {
    if ( params_converged( info->x->pp[ i ], info->old_x->pp[ i ] ) == 0 )
      return 0;
  }
  
  return 1;

}












int em_lhood_converged( EM_INFO *info )
{

  STAT ldiff = fabs( info->lhood - info->old_lhood );

  return ( ldiff / fabs( info->lhood ) < LHOOD_THRESH ) ? (1) : (0);

}








int params_converged( PARAMS *a, PARAMS *b )
{

  if ( fabs(a->a - b->a) / (a->a) >= PARAM_THRESH )
    return 0;
    
  return ( fabs(a->b - b->b) / (a->b) < PARAM_THRESH ) ? (1) : (0);

}












/*============================================================================*/
/*============================================================================*/
/*============================================================================*/
/*============================================================================*/
/*============================================================================*/










void do_opt_model_em_single_iter( EM_INFO *info )
{

  TAB_CELL *create_tab_cell( TAB_CELL *x, INDEX n );
  STAT get_max_q_lhood_params_tree( TAB_CELL *tab_cell, DATA *data, 
    PT_CELL *pt_cell );
  TREE *make_and_set_tree( TREE *tree );
  STAT get_double_max_q_lhood_params_tree( TAB *tab1, TAB *tab2, DATA *data1, 
    TREE *tree, PT_CELL *pt_cell );
  
  TAB_CELL *tab_cell = NULL;
  tab_cell = (TAB_CELL *)create_tab_cell( tab_cell, info->part_n );

  info->old_lhood = info->lhood;
  
  /*  In lhood.c:  */
  info->lhood = get_max_q_lhood_params_tree( tab_cell, info->data, info->x );
  
  /*  In this file, below:  */
  update_model_params_birth_death_ct( tab_cell, info );
  
  kill_tab_cell( tab_cell );
  
}










void do_opt_model_em_new_single_iter( EM_INFO *info )
{

  TAB_CELL *create_tab_cell( TAB_CELL *x, INDEX n );
  STAT get_max_q_lhood_params_tree_new( TAB_CELL *tab_cell, DATA *data, 
    PT_CELL *pt_cell );
  TREE *make_and_set_tree( TREE *tree );
  STAT get_double_max_q_lhood_params_tree( TAB *tab1, TAB *tab2, DATA *data1, 
    TREE *tree, PT_CELL *pt_cell );
    
  extern int newton;
  
  newton = DO_NEWTON;
  
  TAB_CELL *tab_cell = NULL;
  tab_cell = (TAB_CELL *)create_tab_cell( tab_cell, info->part_n );

  info->old_lhood = info->lhood;
  
  /*  In lhood.c:  */
  info->lhood = get_max_q_lhood_params_tree_new( tab_cell, info->data, info->x );
  
  printf( "hi lhood :  %f\n", (double)info->lhood );
  
  if (( newton == 1 ) && ( info->part_n <= 3 ) && ( info->lhood < info->old_lhood ))
    /*  In newton.c:  */
    do_newton( info );
  else
    /*  In this file, below:  */
    update_model_params_birth_death_ct( tab_cell, info );

  
  kill_tab_cell( tab_cell );

}













void do_opt_model_em_full_single_iter( EM_INFO *info )
{

  TAB_CELL *create_tab_cell( TAB_CELL *x, INDEX n );
  STAT get_max_q_lhood_params_tree_new( TAB_CELL *tab_cell, DATA *data, 
    PT_CELL *pt_cell );
  STAT get_max_q_lhood_params_tree_full( TAB_CELL *tab_cell, DATA *data, 
    PT_CELL *pt_cell );
  TREE *make_and_set_tree( TREE *tree );
  STAT get_double_max_q_lhood_params_tree( TAB *tab1, TAB *tab2, DATA *data1, 
    TREE *tree, PT_CELL *pt_cell );
    
  extern int newton;
  
  TAB_CELL *tab_cell = NULL;
  tab_cell = (TAB_CELL *)create_tab_cell( tab_cell, info->part_n );

  info->old_lhood = info->lhood;
  
  /*  In lhood.c:  */
  info->lhood = get_max_q_lhood_params_tree_full( tab_cell, info->data, info->x );
  
  update_model_params_birth_death_ct( tab_cell, info );
    
  kill_tab_cell( tab_cell );

}











void update_model_params_birth_death_ct( TAB_CELL *tab_cell, EM_INFO *info )
{

  copy_over_pt_cell_params( info->old_x, info->x );
  
  set_new_model_params_tab_cell_pt_cell( info->x, tab_cell, info->data->reg );
  
}










void set_new_model_params_tab_cell_pt_cell( PT_CELL *pt_cell, TAB_CELL *tab_cell,
  REGION *reg )
{

  int i;
  
  for ( i = 0; i < pt_cell->n; i++ )
    set_new_model_params_pt_cell( pt_cell->pp[ i ], tab_cell->tab[ i ], reg );

}








void copy_over_pt_cell_params( PT_CELL *a, PT_CELL *b )
{

  int i;
  
  for ( i = 0; i < b->n; i++ )
    *(a->pp[ i ]) = *(b->pp[ i ]);

}











void set_new_model_params_pt_cell( PARAMS *pp, TAB *tab, REGION *reg )
{

  COORD get_region_width( REGION *reg );
  
  extern int reg_start, reg_end;

  TIME t = tab->t;
  
  /*  In ../main/reg_defn.c:  */
  COORD wid = reg_end - reg_start;;
  
  STAT x = ( tab->n != 0.0 ) ? ((STAT)tab->siten / (STAT)tab->n) : (1.0);

  pp->a = tab->birth / (t * (wid) * MILLION);
  pp->b = ( x == 0.0 ) ? (0.0) : (tab->death / (t * x * MILLION));
  
  pp->a = ( pp->a < MIN_BD_RATE ) ? (MIN_BD_RATE) : (pp->a);
  pp->b = ( pp->b < MIN_BD_RATE ) ? (MIN_BD_RATE) : (pp->b);
  
  if ( PRINT_BOOL ) {
    printf( "birth = %d\n", (int)tab->birth );
    printf( "death = %d\n", (int)tab->death );
    printf( "time  = %.1f\n", (float)t );
    printf( "siten = %d\n", (int)tab->siten );
    printf( "n     = %d\n", (int)tab->n );
    printf( "a = %e\nb = %e\n\n", (double)pp->a, (double)pp->b );
  }

}















MODEL_LIST *set_model_lister_em_info( MODEL_LIST *model, EM_INFO *info )
{

  PART_LIST *pt = model->my_model->part_header;

  int i;
  
  for ( i = 0; i < info->x->n; i++ )
    set_single_pt_lister_and_step( &pt, info->x->pp[ i ] );
    
  set_model_lister_stats_em_info( model, info );
    
  return model;

}












void set_single_pt_lister_and_step( PART_LIST **pt, PARAMS *pp )
{
  set_single_pt_lister( *pt, pp );
  *pt = (*pt)->next;
}












void set_single_pt_lister( PART_LIST *pt, PARAMS *pp )
{
  *(pt->my_part->pp) = *pp;
}













void set_model_lister_stats_em_info( MODEL_LIST *model, EM_INFO *info )
{

  MODEL_STATS *s = model->my_model->stats;
  
  s->lhood = info->lhood;
  s->conv = info->conv;
  s->outcome = info->outcome;
  
  s->pval = 1.0;

}



















