#include "model.h"























void do_motif_pred( void )
{

  extern char minor_mode;
  
  switch ( minor_mode ) {
  
  case '\0' :
    do_motif_pred_kmers();
    break;
    
  default :
    error( 3000, "Unknown minor mode passed to program.\n" );
    break;
  
  }

}










void do_motif_pred_kmers( void )
{

  check_kmer_ranks();
  
  do_my_motif_pred_kmers();
  
}








void check_kmer_ranks( void )
{

  extern int min_rank, max_rank;
  
  if ( min_rank == max_rank )
    error( 3001, "Min-max ranks not set.  Use '-b <min>:<max>'.\n" );

}









void do_my_motif_pred_kmers( void )
{

  FILE *open_text_file_writing( char a[], char b[] );
  
  extern char input_file[ BUFFER ];
  extern int min_rank, max_rank;
  extern int reg_start, reg_end;
  
  char output_file[ BUFFER ];
  
  sprintf( output_file, "%s.kmerRes_%d:%d_%d:%d", input_file,
    reg_start, reg_end, min_rank, max_rank );
    
  FILE *out_fh = NULL, *sum_fh = NULL;
  out_fh = (FILE *)open_text_file_writing( output_file, "out" );
  sum_fh = (FILE *)open_text_file_writing( output_file, "sum" );
  
  do_motif_pred_kmers_w_fh( out_fh, sum_fh );
  
  close_file( out_fh );
  close_file( sum_fh );
  
}








void do_motif_pred_kmers_w_fh( FILE *out_fh, FILE *sum_fh )
{

  MOTIF_LIST *mf_header = NULL;
  mf_header = (MOTIF_LIST *)get_motif_freq_list_binary( mf_header );
  
  do_motif_pred_w_motif_list( mf_header, out_fh, sum_fh );

}











MOTIF_LIST *get_motif_freq_list_binary( MOTIF_LIST *mf_header )
{

  FILE *open_binary_file_reading( char a[], char b[] );
  void get_string_from_stack( char cons[], MF stack, char length );
  MOTIF_LIST *create_motif_lister( MOTIF_LIST *x );
  
  extern char input_file[ BUFFER ];
  
  FILE *mf_fh = NULL;
  mf_fh = (FILE *)open_binary_file_reading( input_file, "m" );
  
  MF stack;
  
  MOTIF_LIST *last_motif = NULL;
  
  while ( get( &stack, sizeof(MF), mf_fh )) {
  
    MOTIF_LIST *mf = NULL;
    mf = (MOTIF_LIST *)create_motif_lister( mf );
    
    mf->my_mf->my_motif = stack;    
    get_string_from_stack( mf->my_mf->cons, stack, WORD_LENGTH );
    
    if ( last_motif == NULL ) {
      last_motif = mf;
      mf_header = mf;
    }
    else {
      last_motif->next = mf;
      last_motif = mf;
    }
  
  }
  
  close_file( mf_fh );
  
  return mf_header;

}











void do_motif_pred_w_motif_list( MOTIF_LIST *mf_header, FILE *out_fh, 
  FILE *sum_fh )
{

  extern int min_rank, max_rank;

  MOTIF_LIST *mf = mf_header;
  
  int numb = 0;
  
  while ( mf != NULL ) {
   
    if ( numb++ >= min_rank )
      /*  In this file, below:  */
      do_motif_pred_single_motif( mf->my_mf->my_motif, out_fh, sum_fh );
      
    if ( numb == max_rank )
      break;
      
    mf = mf->next;
    
  }

}











void do_motif_pred_single_motif( MF w, FILE *out_fh, FILE *sum_fh )
{

  void get_string_from_stack( char cons[], MF stack, char length );
  DATA_LIST *get_origin_data( DATA_LIST *data, MOTIF *mf );
  RESULT_LIST *get_result_lister_w_data( RESULT_LIST *r, DATA_LIST *data );

  static int i = 0;

  MOTIF mf = { .my_motif = w, .length = WORD_LENGTH };
  RESULT_LIST *res = NULL;
  DATA_LIST *data = NULL;
  
  /*  In ../main/nucl.c:  */
  get_string_from_stack( mf.cons, w, WORD_LENGTH );
  
  printf( "Motif %d  %d :  %s\n", ++i, (int)w, mf.cons );

  /*  In data.c:  */
  data = (DATA_LIST *)get_origin_data( data, &mf );
  
  /*  In res.c:  */
  res = (RESULT_LIST *)get_result_lister_w_data( res, data );
  
  /*  In print.c:  */
  print_result_lister_pred_fh( res, data->data->tree, &mf, out_fh, sum_fh );
  
}












/*=============================================================================*/
/*=============================================================================*/
/*=============================================================================*/
/*=============================================================================*/
/*=============================================================================*/








void do_motif_pred_single( void )
{

  extern MOTIF major_motif;
  
  do_my_motif_pred_single( &major_motif );

}











void do_my_motif_pred_single( MOTIF *mf )
{

  void get_string_from_stack( char cons[], MF stack, char length );
  DATA_LIST *get_origin_data( DATA_LIST *data, MOTIF *mf );
  RESULT_LIST *get_result_lister_w_data( RESULT_LIST *r, DATA_LIST *data );

  RESULT_LIST *res = NULL;
  DATA_LIST *data = NULL;
  
  /*  In data.c:  */
  data = (DATA_LIST *)get_origin_data( data, mf );
  
  /*  In res.c:  */
  res = (RESULT_LIST *)get_result_lister_w_data( res, data );
  
  /*  In print.c:  */
  print_model_list( res->my_result->model_header, data->data->tree, "Model-list" );
  

}





