#include "newton.h"











#define OUTPUT 1
#define IND 0





#define MAX_ITER 5
#define LH_THRESH 1e-7
#define PP_RATIO 1e-4











void do_newton( EM_INFO *info )
{

  BROY bb;
  
  /*  In ../birth/opt.c:  */
  make_broy( &bb, info->part_n * 2 );
  
  /*  In this file, below:  */
  set_broy_from_em_info( &bb, info );
  
  /*  In this file, below:  */
  do_newton_method( &bb, info );
  
  /*  In ../birth/opt.c:  */
  kill_broy( &bb );

}














void set_broy_from_em_info( BROY *bb, EM_INFO *info )
{

  /*  In em.c:  */
  copy_over_pt_cell_params( info->x, info->old_x );
  
  /*  In this file, below:  */
  set_vec_from_pt_cell( bb->old_x, info->old_x );

}













void set_vec_from_pt_cell( VEC *vec, PT_CELL *pt_cell )
{

  int i, j = 0;
  
  for ( i = 0; i < pt_cell->n; i++ ) {
    vec->v[ j++ ] = pt_cell->pp[ i ]->a;
    vec->v[ j++ ] = pt_cell->pp[ i ]->b;
  }

}














void do_newton_method( BROY *bb, EM_INFO *info )
{

  printf( "Starting Newton iterations...\n\n" );
  
  do {
  
    do_newton_single_iter( bb, info );
  
  } while ( newton_is_going( bb ));
  
  info->conv = bb->conv;

}














int newton_is_going( BROY *bb )
{

  if ( bb->loop_iter > MAX_ITER )
    return 0;

  bb->conv = 0;

  if ( bb->outcome == OUTCOME_DIVERGED ) {
    bb->outcome = OUTCOME_GOING;
//    bb->loop_iter = 0;
    return 1;
  }

  if ( newton_conv( bb ))
    return 0;  
    
  /*  
  if ( bb->loop_iter > MAX_ITER )
    return 0;
  */
    
  return 1;

}










int newton_conv( BROY *bb )
{

  if ( fabs( bb->lhood - bb->old_lhood ) < LH_THRESH * fabs( bb->lhood ) ) {
    bb->conv = 1;
    return 1;
  }
  
  int i;
  
  for ( i = 0; i < bb->x->i; i++ ) {
  
    if ( fabs( bb->s->v[ i ] ) >= fabs( bb->x->v[ i ] ) * PP_RATIO )
      return 0;
  
  }
  
  bb->conv = 1;
  
  return 1;

}










void do_newton_single_iter( BROY *bb, EM_INFO *info )
{

  if ( (bb->loop_iter)++ == 0 )
    do_newton_first_loop( bb, info );
    
  else
    do_newtown_subsequent_loop( bb, info );
    
  ++(info->iter);
    
}












void do_newton_first_loop( BROY *bb, EM_INFO *info )
{

  void get_mx_inv( MX *inv, MX *a );
  void add_vec( VEC *c, VEC *a, VEC *b );
  
  /*  In this file, below:  */
  get_hessian_and_grad_w_info( bb->j, bb->f, info );
  
  /*  In ../birth/mx.c:  */
  get_mx_inv( bb->b, bb->j );
  
  /*  In ../birth/opt.c:  */
  get_step_param_vec( bb );
  
  /*  In ../birth/mx.c:  */
  add_vec( bb->x, bb->old_x, bb->s );
  
  printf( "Vec :\n" );
  print_vec( bb->x );
  printf( "Step :\n" );
  print_vec( bb->s );
  
  /*  In this file, below:  */
  update_newton_iter_end( bb, info );

}













void do_newtown_subsequent_loop( BROY *bb, EM_INFO *info )
{

  void get_new_b_mx( BROY *bb, VEC *w );
  void add_vec( VEC *c, VEC *a, VEC *b );

  VEC w = *(bb->f);
  
  *(bb->old_x) = *(bb->x);
  
  /*  In this file, below:  */
  get_grad_vec_newton( bb->f, info );
  
  /*  In ../birth/opt.c:  */
  get_new_b_mx( bb, &w );
  
  /*  In ../birth/opt.c:  */
  get_step_param_vec( bb );
  
  /*  In ../birth/mx.c:  */
  add_vec( bb->x, bb->old_x, bb->s );
  
  /*  In this file, below:  */
  update_newton_iter_end( bb, info );

}








void update_newton_iter_end( BROY *bb, EM_INFO *info )
{

  int vec_is_neg( VEC *v );
  
  bb->old_lhood = bb->lhood;
  
  /*  In ../birth/mx.c:  */
  if ( vec_is_neg( bb->x )) {
  
    /*  In ../birth/mx.c:  */
    get_pos_x_vec( bb->x, bb->s, bb->old_x );
    
    bb->outcome = OUTCOME_DIVERGED;
    
  }
  
  /*  In em.c:  */
  copy_over_pt_cell_params( info->old_x, info->x );
  
  /*  In this file, below:  */
  update_pt_cell_from_vec( info->x, bb->x );
  
  /*  In this file, below:  */
  bb->lhood = info->lhood = get_lhood_from_info( info );
  
  if (( bb->lhood < bb->old_lhood ) && ( bb->loop_iter > 1 ))
    bb->outcome = OUTCOME_DIVERGED;
  
}












void update_pt_cell_from_vec( PT_CELL *pt_cell, VEC *x )
{

  int i, j = 0;
  
  for ( i = 0; i < pt_cell->n; i++ ) {
    pt_cell->pp[ i ]->a = x->v[ j++ ];
    pt_cell->pp[ i ]->b = x->v[ j++ ];
  }

}












STAT get_lhood_from_info( EM_INFO *info )
{

  TAB_CELL *create_tab_cell( TAB_CELL *x, INDEX n );
  STAT get_max_q_lhood_params_tree_new( TAB_CELL *tab_cell, DATA *data, 
    PT_CELL *pt_cell );

  TAB_CELL *tab_cell = NULL;
  tab_cell = (TAB_CELL *)create_tab_cell( tab_cell, info->part_n );
  
  /*  In lhood.c:  */
  STAT lhood = get_max_q_lhood_params_tree_new( tab_cell, info->data, info->x );

  kill_tab_cell( tab_cell );
  
  return lhood;

}











void get_step_param_vec( BROY *bb )
{

  void vec_mult_scalar( VEC *v, STAT c );
  void mx_times_vec( VEC *x, MX *a, VEC *b );

  VEC f_neg = *(bb->f);
  
  /*  In mx.c:  */
  vec_mult_scalar( &f_neg, -1.0 );
  
  /*  In mx.c:  */
  mx_times_vec( bb->s, bb->b, &f_neg );
  
}










void get_new_b_mx( BROY *bb, VEC *w )
{

  void substract_vec( VEC *c, VEC *a, VEC *b );
  void vec_mult_scalar( VEC *v, STAT c );
  void mx_times_vec( VEC *x, MX *a, VEC *b );
  STAT dot_prod( VEC *a, VEC *b );
  void vec_mult_mx( VEC *c, VEC *s, MX *a );
  void add_vec( VEC *c, VEC *a, VEC *b );
  void outer_prod( MX *c, VEC *a, VEC *b );
  void increm_mx( MX *a, MX *b );

  int n = bb->n;

  VEC y = { .i = n }, z = { . i = n }, q = { .i = n }, u = { .i = n };
  MX a = { .i = n, .j = n };
  
  /*  In mx.c:  */
  subtract_vec( &y, bb->f, w );
  
  VEC y_neg = y;
  
  /*  In mx.c:  */
  vec_mult_scalar( &y_neg, -1.0 );
  
  /*  In mx.c:  */
  mx_times_vec( &z, bb->b, &y_neg );
  
  /*  In mx.c:  */
  STAT p = -1.0 * dot_prod( bb->s, &z );
  
  /*  In mx.c:  */
  vec_mult_mx( &u, bb->s, bb->b );
  
  /*  In mx.c:  */
  add_vec( &q, bb->s, &z );
  
  vec_mult_scalar( &q, 1.0 / p );
  
  outer_prod( &a, &q, &u );
  
  increm_mx( bb->b, &a );
  
}










/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/










void get_hessian_and_grad_w_info( MX *j, VEC *f, EM_INFO *info )
{

  void set_trans_probs_pt_cell( TREE *tree, PT_CELL *pt_cell, REGION *reg );
  
  /*  In lhood.c:  */
  set_trans_probs_pt_cell( info->data->tree, info->x, info->data->reg );
  
  /*  In this file, below:  */
  set_trans_probs_derivs( info->data->tree, info->x, info->data->reg );
  
  printf( "Creating hessian matrix...\n\n" );
  
  /*  In this file, below:  */
  get_my_hessian_and_grad_w_info( j, f, info );
  
  printf( "Hessian matrix J :\n" );
  print_mx( j );
  printf( "Gradient vector F :\n" );
  print_vec( f );

}













void get_my_hessian_and_grad_w_info( MX *mx, VEC *f, EM_INFO *info )
{

  int pi, bi;
  int n = info->part_n;
  
  for ( pi = 0; pi < n; pi++ ) {
    for ( bi = 0; bi < BD; bi++ )
      /*  In this file, below:  */
      get_hessian_and_grad_single_row_w_info( mx, f, info, pi, bi );
  }
  
  /*  In ../birth/mx.c:  */
  sym_mx( mx );

}













void get_hessian_and_grad_single_row_w_info( MX *mx, VEC *f, EM_INFO *info, int pi, 
  int bi )
{

  int i = (BD * pi) + bi;
  
  int pj, bj, j;
  
  int n = info->part_n;
  
  for ( pj = 0; pj < n; pj++ ) {
    for ( bj = 0; bj < BD; bj++ ) {
    
      j = (BD * pj) + bj;
      
      if ( i > j )
        continue;
        
      mx->mx[ i ][ j ] = get_hessian_elem_w_info( info, f, i, j, pi, bi, pj, bj );
      
      if ( OUTPUT )
        printf( "j[ %d ][ %d ] = %15.5e\n", i, j, (double)mx->mx[ i ][ j ] );
  
    }
  }

}













STAT get_hessian_elem_w_info( EM_INFO *info, VEC *f, int i, int j, int pi, int bi,
  int pj, int bj )
{

  if ( i == 0 )
    f->v[ j ] = 0.0;

  STAT x = 0.0;
  
  PDATA_LIST *pdata = info->data->pdata;
  
  while ( pdata != NULL )
    x += get_hessian_elem_single_pdata_and_step( &pdata, info, f, i, j, pi, bi, 
      pj, bj );
      
  return x;

}











STAT get_hessian_elem_single_pdata_and_step( PDATA_LIST **pdata, EM_INFO *info,
  VEC *f, int i, int j, int pi, int bi, int pj, int bj )
{

  STAT x = get_hessian_elem_single_pdata( *pdata, info, f, i, j, pi, bi, pj, bj );
  
  *pdata = (*pdata)->next;
  
  return x;

}










STAT get_hessian_elem_single_pdata( PDATA_LIST *pdata, EM_INFO *info, VEC *f, int i,
  int j, int pi, int bi, int pj, int bj )
{

  void clear_tree_all( NODE_LIST *root );
  
  /*  In ../birth/set.c:  */
  clear_tree_all( info->data->tree->root );
  
  /*  In data.c:  */
  set_pdata_leaves( info->data->tree, pdata );
  
  /*  In this file, below:  */
  set_use_vals_tree( info->data->tree );
  
  /*  In this file, below:  */
  return get_my_hessian_elem_single_pdata( info->data->tree, f, i, j, pi, bi, 
    pj, bj );

}












void set_use_vals_tree( TREE *tree )
{

  NODE_LIST *node = tree->spec_header;
  
  while ( node->my_node->leaf )
    set_use_vals_single_leaf_and_step( &node );

}










void set_use_vals_single_leaf_and_step( NODE_LIST **node )
{
  set_use_vals_single_leaf( *node );
  *node = (*node)->next;
}











void set_use_vals_single_leaf( NODE_LIST *node )
{
  if ( node->my_info->use )
    /*  In ../birth/set.c:  */
    set_use_leaf( node );
}







STAT get_my_hessian_elem_single_pdata( TREE *tree, VEC *f, int i, int j, int pi, 
  int bi, int pj, int bj )
{

  void get_hessian_elem_ct_d_dd_single_node_and_step( NODE_LIST **node, int pi, 
    int bi, int pj, int bj );

  NODE_LIST *node = tree->spec_header;
  
  while ( node != NULL )
    /*  In this file, below:  */
    get_hessian_elem_pdata_single_node_and_step( &node, pi, bi, pj, bj );
    
  /*  In this file, below:  */
  STAT a = get_node_info_w_prior( tree->root->my_info, tree->root->prior );
  STAT b = get_node_info_w_prior( tree->root->my_dd, tree->root->prior );
  STAT c1 = get_node_info_w_prior( tree->root->my_d[ 0 ], tree->root->prior );
  STAT c2 = get_node_info_w_prior( tree->root->my_d[ 1 ], tree->root->prior );
  
  if ( i == 0 )
    f->v[ j ] += c2 / a;
    
  return ((a * b) - (c1 * c2)) / (a * a);

}








void get_hessian_elem_pdata_single_node_and_step( NODE_LIST **node, int pi, int bi,
  int pj, int bj )
{
  get_hessian_elem_pdata_single_node( *node, pi, bi, pj, bj );
  *node = (*node)->next;
}










void get_hessian_elem_pdata_single_node( NODE_LIST *node, int pi, int bi, int pj,
  int bj )
{

  if (( node->my_node->leaf ) || ( node->my_info->use == 0 ))
    return;
    
  get_my_hessian_elem_pdata_single_node( node, pi, bi, pj, bj );
  
}








void get_my_hessian_elem_pdata_single_node( NODE_LIST *node, int pi, int bi, int pj,
  int bj )
{

  int get_max_ct_daughter_nodes( NODE_LIST *node );
  STAT get_lhood_pdata_single_node_ct( NODE_LIST *node, int i );
  STAT get_dd_single_ct( NODE_LIST *node, int i, int pi, int bi, int pj, int bj );

  /*  In lhood.c:  */
  int max = node->my_info->max = get_max_ct_daughter_nodes( node );
  
  int i;
  
  for ( i = 0; i <= max; i++ ) {
  
    /*  In lhood.c:  */
    node->my_info->x[ i ] = get_lhood_pdata_single_node_ct( node, i );
    
    /*  In ../birth/grad_vec.c:  */
    node->my_d[ 0 ]->x[ i ] = get_deriv_single_ct( node, 0, i, pi, bi );
    
    if (( pi == pj ) && ( bi == bj ))
      node->my_d[ 1 ]->x[ i ] = node->my_d[ 0 ]->x[ i ];
    else
      /*  In ../birth/grad_vec.c:  */
      node->my_d[ 1 ]->x[ i ] = get_deriv_single_ct( node, 1, i, pj, bj );
      
    /*  In ../birth/hess_mx.c:  */
    node->my_dd->x[ i ] = get_dd_single_ct( node, i, pi, bi, pj, bj );
  
  }  

}







STAT get_node_info_w_prior( NODE_INFO *x, NODE_INFO *prior )
{

  extern int max_occur;
  
  int end = max_occur + 3;
  
  STAT lhood = 0.0;
  
  int i;
  
  for ( i = 0; i < end; i++ ) 
    lhood += x->x[ i ] * prior->x[ i ];  
  
  return lhood;

}













STAT get_deriv_single_ct( NODE_LIST *node, int d_index, int i, int p, int b )
{

  void check_nan( STAT x, char func[], char file[] );

  if ( node->my_info->x[ i ] == 0.0 )
    return 0.0;
    
  STAT sum = 0.0;
  
  int k;
  
  for ( k = 0; k < BRANCH_BUFFER; k++ ) {
  
    NODE_LIST *d = node->down[ k ];
  
    if ( d == NULL )
      break;
  
    sum += get_deriv_single_ct_single_daughter_node( d, d_index, i, p, b );
    
  }
    
  check_nan( sum, "get_deriv_single_ct", "grad_vec.c" );
  
  return node->my_info->x[ i ] * sum;

}














STAT get_deriv_single_ct_single_daughter_node( NODE_LIST *d, int d_index, int i, 
  int p, int b )
{
  
  int ct_done( STAT x, STAT total, STAT *last );

  if ( d->my_info->use == 0 )
    return 0.0;
  
  STAT numer = 0.0, denom = 0.0;

  int j;
    
  for ( j = 0; j <= d->my_info->max; j++ ) {
    denom += get_deriv_single_daughter_denom( d, i, j );    
    numer += get_deriv_single_daughter_numer( d, d_index, i, j, p, b );  
  }
  
  return ( denom == 0.0 ) ? (0.0) : (numer / denom);

}












STAT get_deriv_single_daughter_denom( NODE_LIST *d, int i, int j )
{
  return d->my_trans->p[ i ][ j ] * d->my_info->x[ j ];
}












STAT get_deriv_single_daughter_numer( NODE_LIST *d, int d_index, int i, int j, 
  int p, int b )
{

  STAT x = d->my_trans->p[ i ][ j ] * d->my_d[ d_index ]->x[ j ];
  
  if ( d->my_trans->part != p )
    return x;
    
  return x + (d->my_trans->p_d[ b ][ i ][ j ] * d->my_info->x[ j ]);

}














STAT get_dd_single_ct( NODE_LIST *node, int i, int pi, int bi, int pj, int bj )
{

  STAT x = node->my_info->x[ i ];

  if ( x == 0.0 )
    return 0.0;
    
  STAT da = node->my_d[ 0 ]->x[ i ];
  STAT db = node->my_d[ 1 ]->x[ i ];
    
  STAT dd = get_dd_term_single_ct( node, i, pi, bi, pj, bj );
  
  return ( x * dd ) + ( da * db / x );

}















STAT get_dd_term_single_ct( NODE_LIST *node, int i, int pi, int bi, int pj,
  int bj )
{

  STAT sum = 0.0;
  
  int k;
  
  for ( k = 0; k < BRANCH_BUFFER; k++ ) {
  
    NODE_LIST *d = node->down[ k ];
    
    if ( d == NULL )
      break;
      
    sum += get_dd_term_single_daughter_node( d, i, pi, bi, pj, bj );
  
  }
  
  return sum;

}












STAT get_dd_term_single_daughter_node( NODE_LIST *d, int i, int pi, int bi,
  int pj, int bj )
{

  STAT get_node_lhood_single_ct_pair( NODE_LIST *d, int i, int j );
  STAT get_deriv_single_daughter_numer( NODE_LIST *d, int d_index, int i, int j, 
    int p, int b );

  if ( d->my_info->use == 0 )
    return 0.0;
    
  STAT a = 0.0, b = 0.0, c[ DBUFF ] = { 0.0, 0.0 };

  int j;
  
  for ( j = 0; j <= d->my_info->max; j++ ) {
  
    /*  In lhood.c:  */
    a += get_node_lhood_single_ct_pair( d, i, j );
    
    /*  In this file, below:  */
    b += get_second_deriv_dd_single_ct( d, i, j, pi, bi, pj, bj );
    
    /*  In grad_vec.c:  */
    c[ 0 ] += get_deriv_single_daughter_numer( d, 0, i, j, pi, bi );
    c[ 1 ] += get_deriv_single_daughter_numer( d, 1, i, j, pj, bj );
    
  }
  
  /*  In this file, below:  */
  return get_second_deriv_term( a, b, c );

}













STAT get_second_deriv_term( STAT a, STAT b, STAT c[] )
{
  return (b / a) - ((c[ 0 ] * c[ 1 ]) / (a * a));
}









STAT get_second_deriv_s_single_ct( NODE_LIST *d, int i, int j )
{
  return d->my_trans->p[ i ][ j ] * d->my_info->x[ j ];
}












STAT get_second_deriv_d_single_ct( NODE_LIST *d, int d_index, int i, int j, 
  int p, int b )
{

  STAT get_deriv_single_daughter_numer( NODE_LIST *d, int d_index, int i, int j, 
    int p, int b );

  /*  In grad_vec.c:  */
  return get_deriv_single_daughter_numer( d, d_index, i, j, p, b );

}










STAT get_second_deriv_dd_single_ct( NODE_LIST *d, int i, int j, int pi, int bi, 
  int pj, int bj )
{

  STAT x = d->my_trans->p[ i ][ j ] * d->my_dd->x[ j ];

  if ( d->my_trans->part == pi )
    x += d->my_trans->p_d[ bi ][ i ][ j ] * d->my_d[ 1 ]->x[ j ];
    
  if ( d->my_trans->part == pj )
    x += d->my_trans->p_d[ bj ][ i ][ j ] * d->my_d[ 0 ]->x[ j ];
        
  if (( d->my_trans->part == pi ) && ( pi == pj ))
    x += d->my_trans->p_dd[ bi ][ bj ][ i ][ j ] * d->my_info->x[ j ];

  return x;
  
}
















void set_trans_probs_derivs( TREE *tree, PT_CELL *pt_cell, REGION *reg )
{

  NODE_LIST *node = tree->spec_header;
  
  while ( node != NULL )
    set_trans_probs_deriv_single_node_and_step( &node, pt_cell, reg );

}











void set_trans_probs_deriv_single_node_and_step( NODE_LIST **node, PT_CELL *pt_cell,
  REGION *reg )
{
  set_trans_probs_deriv_single_node( *node, pt_cell, reg );
  *node = (*node)->next;
}











void set_trans_probs_deriv_single_node( NODE_LIST *node, PT_CELL *pt_cell,
  REGION *reg )
{

  PARAMS *get_params_pt_cell_by_numb( NODE_LIST *node, PT_CELL *pt_cell );
  void set_node_u_deriv( STAT u[], PARAMS *pp, TIME t );
  void set_node_v_deriv( STAT v[], PARAMS *pp, TIME t );
  void set_node_u_double_deriv( STAT u[ BD ][ BD ], PARAMS *pp, TIME t );
  void set_node_v_double_deriv( STAT v[ BD ][ BD ], PARAMS *pp, TIME t );
  void set_p_deriv( NODE_LIST *node, PARAMS *pp, COUNT n );
  
  extern int max_occur;

  PARAMS *pp = NULL;
  /*  In lhood.c:  */
  pp = (PARAMS *)get_params_pt_cell_by_numb( node, pt_cell );
  
  /*  In this file, below:  */
  node->my_trans->part = get_part_numb_pt_cell( node, pt_cell );

  /*  In this file, below:  */
  set_node_u_deriv( node->my_trans->u_d, pp, node->my_node->t );
  set_node_v_deriv( node->my_trans->v_d, pp, node->my_node->t );
  
  /*  In this file, below:  */
  set_node_u_double_deriv( node->my_trans->u_dd, pp, node->my_node->t );
  set_node_v_double_deriv( node->my_trans->v_dd, pp, node->my_node->t );
  
  /*  In this file, below:  */
  set_p_deriv( node, pp, max_occur + 3 );
  
}












int get_part_numb_pt_cell( NODE_LIST *node, PT_CELL *pt_cell )
{

  int node_in_list_numb( NODE_LIST *node, LIST *header );
  
  if ( pt_cell->n == 1 )
    return 0;
    
  int i;
  
  for ( i = 0; i < pt_cell->n; i++ ) {
    /*  In lhood.c:  */
    if ( node_in_list_numb( node, pt_cell->branch[ i ] ))
      return i;
  }
  
  return 0;

}














/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/
/*==============================================================================*/

















void get_grad_vec_newton( VEC *f, EM_INFO *info )
{

  void set_trans_probs_pt_cell( TREE *tree, PT_CELL *pt_cell, REGION *reg );
  
  /*  In lhood.c:  */
  set_trans_probs_pt_cell( info->data->tree, info->x, info->data->reg );
  
  /*  In this file, above:  */
  set_trans_probs_derivs( info->data->tree, info->x, info->data->reg );
  
  printf( "Creating gradient vector...\n\n" );
  
  /*  In this file, below:  */
  get_my_grad_vec_newton( f, info );

}














void get_my_grad_vec_newton( VEC *f, EM_INFO *info )
{

  int i = 0, p, b;
  
  for ( p = 0; p < info->part_n; p++ ) {
  
    for ( b = 0; b < BD; b++ ) {

      f->v[ i ] = get_grad_vec_elem_newton( info, p, b );
      
      if ( OUTPUT )
        printf( "f[ %d ] = %15.5e\n", i, (double)f->v[ i ] );
        
      ++i;
        
    }  
  
  }

}














STAT get_grad_vec_elem_newton( EM_INFO *info, int p, int b )
{

  STAT x = 0.0;

  PDATA_LIST *pdata = info->data->pdata;
  
  while ( pdata != NULL )
    x += get_grad_vec_elem_single_pdata_and_step( &pdata, info, p, b );
    
  return x;

}













STAT get_grad_vec_elem_single_pdata_and_step( PDATA_LIST **pdata, EM_INFO *info,
  int p, int b )
{

  STAT x = get_grad_vec_elem_single_pdata( *pdata, info, p, b );
  
  *pdata = (*pdata)->next;
  
  return x;

}














STAT get_grad_vec_elem_single_pdata( PDATA_LIST *pdata, EM_INFO *info, int p, 
  int b )
{

  void clear_tree_all( NODE_LIST *root );
  
  /*  In ../birth/set.c:  */
  clear_tree_all( info->data->tree->root );
  
  /*  In data.c:  */
  set_pdata_leaves( info->data->tree, pdata );
  
  /*  In this file, below:  */
  set_use_vals_tree( info->data->tree );
  
  /*  In this file, below:  */
  return get_my_grad_vec_elem_single_pdata( info->data->tree, p, b );

}












STAT get_my_grad_vec_elem_single_pdata( TREE *tree, int p, int b )
{

  NODE_LIST *node = tree->spec_header;
  
  while ( node != NULL )
    /*  In this file, below:  */
    set_grad_deriv_vals_single_node_pdata_and_step( &node, p, b );
    
  /*  In this file, above:  */  
  STAT x = get_node_info_w_prior( tree->root->my_d[ IND ], tree->root->prior );
  STAT y = get_node_info_w_prior( tree->root->my_info, tree->root->prior );  
  
  return x / y;

}














void set_grad_deriv_vals_single_node_pdata_and_step( NODE_LIST **node, int p,
  int b )
{
  set_grad_deriv_vals_single_node_pdata( *node, p, b );
  *node = (*node)->next;
}












void set_grad_deriv_vals_single_node_pdata( NODE_LIST *node, int p, int b )
{

  if (( node->my_node->leaf ) || ( node->my_info->use == 0 ))
    return;
    
  set_my_grad_deriv_vals_single_node_pdata( node, p, b );
  
}










void set_my_grad_deriv_vals_single_node_pdata( NODE_LIST *node, int p, int b )
{

  int get_max_ct_daughter_nodes( NODE_LIST *node );
  STAT get_lhood_pdata_single_node_ct( NODE_LIST *node, int i );
  STAT get_deriv_single_ct( NODE_LIST *node, int d_index, int i, int p, int b );
  STAT get_dd_single_ct( NODE_LIST *node, int i, int pi, int bi, int pj, int bj );

  /*  In lhood.c:  */
  int max = node->my_info->max = get_max_ct_daughter_nodes( node );
  
  int i;
  
  for ( i = 0; i <= max; i++ ) {
  
    /*  In lhood.c:  */
    node->my_info->x[ i ] = get_lhood_pdata_single_node_ct( node, i );
    
    /*  In ../birth/grad_vec.c:  */
    node->my_d[ IND ]->x[ i ] = get_deriv_single_ct( node, IND, i, p, b );
    
  }  

}











/*=============================================================================*/
/*=============================================================================*/
/*=============================================================================*/
/*=============================================================================*/
/*=============================================================================*/















void make_broy( BROY *b, INDEX n )
{

  VEC *create_vec( VEC *v, INDEX n );
  MX *create_mx( MX *mx, INDEX i, INDEX j );
  TREE_DATA *create_tree_data( TREE_DATA *d );
  
  b->x = (VEC *)create_vec( b->x, n );
  b->s = (VEC *)create_vec( b->s, n );
  b->f = (VEC *)create_vec( b->f, n );
  b->old_x = (VEC *)create_vec( b->old_x, n );
  
  b->j = (MX *)create_mx( b->j, n, n );
  b->b = (MX *)create_mx( b->b, n, n );
  
  b->d = (TREE_DATA *)create_tree_data( b->d );
  
  b->n = n;
  
  b->lhood = 0.0;
  
  b->major_iter = 0;
  b->loop_iter = 0;
  b->total_iter = 0;
  
  b->conv = BROY_GOING;
  b->outcome = BROY_GOING;

}















void set_broy( BROY *bb, VEC *param, MODEL_LIST *model, NODE_LIST *root, 
  OCCUR_CELL *occur_cell )
{

//  STAT get_lhood_tree_data( TREE_DATA *d );

  *(bb->old_x) = *param;

  bb->d->model = model;
  bb->d->root = root;
  bb->d->occur = occur_cell;
  
  /*  In lhood.c:  */
//  bb->lhood = get_lhood_tree_data( bb->d );

  bb->lhood = 0.0;
  
}













void kill_broy( BROY *bb )
{

  kill_vec( bb->x );
  kill_vec( bb->old_x );
  kill_vec( bb->s );
  kill_vec( bb->f );
  
  kill_mx( bb->j );
  kill_mx( bb->b );
  
  kill( bb->d );

}
















