#include "print.h"































void print_model_lister_tree( MODEL_LIST *model, TREE *tree )
{
  print_model_lister_tree_fh( model, tree, stdout );
}











void print_model_lister_tree_fh( MODEL_LIST *model, TREE *tree, FILE *fh )
{

  print_model_lister_header( model, fh );

  print_model_stats( model->my_model->stats, "Next", fh );
  print_model_stats( model->my_model->null_stats, "Null", fh );

  print_model_lister_tree_part_list( model, tree, fh );
  
  print_dot_sep( fh );

}











void print_model_lister_header( MODEL_LIST *model, FILE *fh )
{

  print_sep( fh );
  
  fprintf( fh, ">  Motif:    %10s   %ld\n", model->my_model->mf->cons,
    (long)model->my_model->mf->my_motif );
    
  fprintf( fh, "M:           %d\n", (int)model->my_model->m );
  fprintf( fh, "Part-N:      %d\n", (int)model->my_model->part_n );

}











void print_model_stats( MODEL_STATS *s, char str[], FILE *fh )
{

  print_dot_sep( fh );
  
  fprintf( fh, "Model_Stats:  %s\n", str );
  fprintf( fh, "Lhood:        %.6f\n", (double)s->lhood );
  fprintf( fh, "Old-Lhood:    %.6f\n", (double)s->old_lhood );
  fprintf( fh, "Best-Lhood:   %.6f\n", (double)s->best_lhood );
  fprintf( fh, "Pval:         %.2e\n", (double)s->pval );
  fprintf( fh, "DOF-numer:    %d\n", (int)s->dof_numer );
  fprintf( fh, "DOF-denom:    %d\n", (int)s->dof_denom );
  fprintf( fh, "Node:         %s\n", s->node_label );
  fprintf( fh, "Best-node:    %s\n", s->best_label );

}








void print_model_lister_tree_part_list( MODEL_LIST *model, TREE *tree, FILE *fh )
{

  print_dot_sep( fh );
  
  fprintf( fh, "Part-N:   %d\n", (int)model->my_model->part_n );
  
  print_part_list_fh( model->my_model->part_header, tree, fh );

}












void print_part_list( PART_LIST *part_header, TREE *tree )
{
  print_part_list_fh( part_header, tree, stdout );
}













void print_part_list_fh( PART_LIST *part_header, TREE *tree, FILE *fh )
{

  int i = 0;

  PART_LIST *part = part_header;
  
  while ( part != NULL )
    print_single_part_lister_fh_and_step( &part, tree, ++i, fh );

  fflush( fh );

}














void print_single_part_lister_fh_and_step( PART_LIST **part, TREE *tree, int i,
  FILE *fh )
{
  print_single_part_lister_fh( *part, tree, i, fh );
  *part = (*part)->next;
}














void print_single_part_lister_fh( PART_LIST *part, TREE *tree, int i, FILE *fh )
{

  char label[ BUFFER ];
  
  get_label_part_lister( label, part, tree );
  
  fprintf( fh, "%3d\t%20s\t%15.6e\t%15.6e\n",
    i,
    label,
    (double)part->my_part->pp->a,
    (double)part->my_part->pp->b );

}













void get_label_part_lister( char label[], PART_LIST *part, TREE *tree )
{

  void get_label_from_numb( char label[], TREE *tree, int numb );

  if ( part->my_part->branch_header->next != NULL )
    strcpy( label, "All" );
  else
    /*  In ../tree/get.c:  */
    get_label_from_numb( label, tree, part->my_part->branch_header->elem.d );

}












/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/













void print_model_list_down( MODEL_LIST *model_header, TREE *tree )
{
  print_model_list_down_fh( model_header, tree, stdout );
}













void print_model_list_down_fh( MODEL_LIST *model_header, TREE *tree, FILE *fh )
{

  MODEL_LIST *model = model_header;
  
  while ( model != NULL ) {
  
    print_model_lister_tree_fh( model, tree, fh );
    
    print_sep( fh );
    
    model = model->down;
  
  }

}











void print_model_list( MODEL_LIST *model_header, TREE *tree, char str[] )
{
  print_model_list_fh( model_header, tree, str, stdout );
}













void print_model_list_fh( MODEL_LIST *model_header, TREE *tree, char str[],
  FILE *fh )
{

  int i = 0;
  
  MODEL_LIST *model = model_header;
  
  while ( model != NULL ) {
  
    print_sep( fh );
      
    fprintf( fh, "%s:  %d\n", str, ++i );
  
    print_dot_sep( fh );
    
    print_model_lister_tree_fh( model, tree, fh );
    
    model = model->next;
  
  }

}














void print_model_list_all( MODEL_LIST *model_header, TREE *tree, char str[] )
{
  print_model_list_all_fh( model_header, tree, str, stdout );
}











void print_model_list_all_fh( MODEL_LIST *model_header, TREE *tree, char str[],
  FILE *fh )
{

  int i = 0;
  
  MODEL_LIST *model = model_header;
  
  while ( model != NULL ) {
  
    print_sep( fh );
      
    fprintf( fh, "%s:  %d\n", str, ++i );
  
    print_dot_sep( fh );
    
    print_model_list_down_fh( model, tree, fh );
    
    model = model->next;
  
  }

}








/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/
/*===============================================================================*/












void print_result_lister_pred_fh( RESULT_LIST *r, TREE *tree, MOTIF *mf,
  FILE *out_fh, FILE *sum_fh )
{

  void print_null_model_single_mf_line( MODEL_LIST *model, MOTIF *mf, FILE *fh );

  /*  In this file:  */
  print_result_lister_summary( r, sum_fh );
  
  RESULT *rr = r->my_result;
  
  print_model_list( rr->model_header, tree, "Model-list" );
  
  print_model_lister_header( rr->model_header, out_fh );
  
  print_model_list_fh( rr->null_model, tree, "Null-model", out_fh );
  print_model_list_all_fh( rr->model_header, tree, "Total-Models", out_fh );
  
  fflush( out_fh );
  fflush( sum_fh );

}











void print_result_lister_model_fh( RESULT_LIST *result, TREE *tree, 
  MODEL_FH *my_fh )
{

  print_result_lister_summary( result, my_fh->sum_fh );
  
  RESULT *rr = result->my_result;
  
  print_model_lister_header( rr->model_header, my_fh->model_fh );
  
  print_model_list_fh( rr->model_header, tree, "Model-list", my_fh->model_fh );
  
  print_model_list( rr->model_header, tree, "Model-list" );
  
  print_model_lister_header( rr->model_header, my_fh->all_fh );

  print_model_list_fh( rr->null_model, tree, "Null-model", my_fh->all_fh );
  print_model_list_all_fh( rr->model_header, tree, "Total-Models", my_fh->all_fh );
  
  fflush( my_fh->sum_fh );
  fflush( my_fh->model_fh );
  fflush( my_fh->all_fh );

}













void print_result_lister_summary( RESULT_LIST *result, FILE *fh )
{

  RESULT *rr = result->my_result;

  fprintf( fh, "%10s\t%15.4e\t%20s\t%20s\n",
    rr->mf->cons,
    (double)rr->pval,
    rr->my_model->my_model->stats->best_label,
    rr->my_model->my_model->stats->beta_label );

  printf( "%10s\t%15.4e\t%20s\t%20s\n",
    rr->mf->cons,
    (double)rr->pval,
    rr->my_model->my_model->stats->best_label,
    rr->my_model->my_model->stats->beta_label );

}












void print_null_model_single_mf_line( MODEL_LIST *model, MOTIF *mf, FILE *fh )
{

  PARAMS *pp = model->my_model->part_header->my_part->pp;
  
  fprintf( fh, "%-9d\t%-10s\t%-10.4e\t%-10.4e\t%-10d\t%-15.6lf\n",
    (int)mf->my_motif,
    mf->cons,
    (double)pp->a,
    (double)pp->b, 
    model->my_model->stats->conv,
    (double)model->my_model->stats->lhood );
    
  fflush( fh );

}

















