#include "sig.h"









#define MAXIT 100
#define EPS 3.0e-7
#define FPMIN 1.0e-30
















void set_model_lister_sig( MODEL_LIST *model, MODEL_STATS *s, MODEL_LIST *old_model, 
  MODEL_LIST *satd_model, DATA *data, char label[], MOTIF *mf )
{

  STAT get_lhood_model_w_node_label( MODEL_LIST *model, DATA *data, 
    char node_label[], COUNT *nn );
  INDEX get_numb_of_nodes_in_tree( TREE *tree );
  STAT get_best_lhood_model_w_node_label( MODEL_LIST *model, DATA *data, 
    char node_label[] );
  STAT get_best_lhood_model_single_prob( MODEL_LIST *model, DATA *data, 
    char node_label[] );
  int get_numb_of_data_pts_data( DATA *data );

  strcpy( s->node_label, label );
  
  /*  In ../tree/get.c:  */
  INDEX numb = get_numb_of_nodes_in_tree( data->tree );
  
  /*  In this file, below:  */
  get_best_node_label( s->best_label, model, data->tree );
  get_beta_node_label( s->beta_label, model, data->tree );
  
  s->lhood = model->my_model->stats->lhood;
  s->old_lhood = old_model->my_model->stats->lhood;

  /*  In lhood.c:  */
//  s->best_lhood = get_best_lhood_model_w_node_label( model, data, s->node_label );
  
  s->best_lhood = get_best_lhood_model_single_prob( model, data, s->node_label );
  
  printf( "best lhood :   %.6f\n", (double)s->best_lhood );
  
  int dm = model->my_model->part_n * 2;
  int dn = old_model->my_model->part_n * 2;
  
  /*  In lhood.c:  */
  int n = get_numb_of_data_pts_data( data );
  
  s->dof_numer = dm - dn;
  s->dof_denom = n - dm;
  
  printf( "dof numer = %d\n", (int)s->dof_numer );
  printf( "dof denom = %d\n", (int)s->dof_denom );
  printf( "lhood     = %.6f\n", (float)s->lhood );
  printf( "old-lhood = %.6f\n", (float)s->old_lhood );
  printf( "b-lhood   = %.6f\n", (float)s->best_lhood );
  printf( "\n" );
  
  /*  In this file, below:  */
  s->pval = get_pval_f_test_stats_struct( s );
  
  printf( "pval :  %.3e\n", (double)s->pval );
  
  print_part_list( model->my_model->part_header, data->tree );  

}













void get_best_node_label( char label[], MODEL_LIST *model, TREE *tree )
{

  void get_label_part_lister( char label[], PART_LIST *part, TREE *tree );

  PART_LIST *part = model->my_model->part_header;
  
  STAT ratio = 0.0;
  
  while ( part != NULL ) {
  
    PARAMS *pp = part->my_part->pp;
  
    STAT new_ratio = pp->a / pp->b;
    
    if ( new_ratio > ratio ) {
      /*  In print.c:  */
      get_label_part_lister( label, part, tree );
      ratio = new_ratio;
    }
    
    part = part->next;
  
  }

}











void get_beta_node_label( char label[], MODEL_LIST *model, TREE *tree )
{

  void get_label_part_lister( char label[], PART_LIST *part, TREE *tree );

  PART_LIST *part = model->my_model->part_header;
  
  STAT ratio = 0.0;
  
  while ( part != NULL ) {
  
    PARAMS *pp = part->my_part->pp;
  
    STAT new_ratio = pp->b / pp->a;
    
    if ( new_ratio > ratio ) {
      /*  In print.c:  */
      get_label_part_lister( label, part, tree );
      ratio = new_ratio;
    }
    
    part = part->next;
  
  }

}












MODEL_LIST *get_model_lister_sig( MODEL_LIST *model, MODEL_STATS *stats, DATA *data, 
  char node_label[] )
{

  set_model_stats( model, stats, data, node_label );

  return model;

}















void set_model_stats( MODEL_LIST *model, MODEL_STATS *stats, DATA *data,
  char node_label[] )
{

  STAT get_lhood_model_w_node_label( MODEL_LIST *model, DATA *data, 
    char node_label[], COUNT *nn );
  STAT get_best_lhood_model_w_node_label( MODEL_LIST *model, DATA *data, 
    char node_label[] );

  MODEL_STATS *s = model->my_model->stats;
  
  COUNT n = 0;
  
  /*  In lhood.c:  */
  s->lhood = get_lhood_model_w_node_label( model, data, node_label, &n );
  
  error( 100, "Hello\n" );

  s->old_lhood = stats->old_lhood;
  s->best_lhood = stats->best_lhood;
  s->best_lhood = 0.0;

  s->n = n;
  
  NODE_LIST *node = data->tree->spec_header;
  int x = 0;
  
  while ( node != NULL ) {
    x += 2;
    node = node->next;
  }
  
  x -= 2;
  
  s->dof_numer = 2;
  s->dof_denom = n - (2 * model->my_model->part_n);
  
  printf( "lhood :  %.6f\n", (double)s->lhood );
  printf( "old   :  %.6f\n", (double)s->old_lhood );
  printf( "best  :  %.6f\n", (double)s->best_lhood );
  printf( "n     :  %d\n\n", (int)n );

  s->pval = get_pval_f_test_stats_struct( s );
  
  printf( "Pval  :  %.3e\n\n", (double)s->pval );
  
}












PVAL get_pval_f_test_stats_struct( MODEL_STATS *s )
{

  STAT v1 = (STAT)s->dof_numer;
  STAT v2 = (STAT)s->dof_denom;
  
  STAT numer = (s->lhood - s->old_lhood) / (STAT)(s->dof_numer);
  STAT denom = (s->best_lhood - s->lhood) / (STAT)(s->dof_denom);
  
  STAT f = numer / denom;
  
  printf( "F = %f\n\n", (float)f );
  
  STAT x = v2 / (v2 + (v1 * f));
  
  return betai( v2 / 2.0, v1 / 2.0, x );
  
}












PVAL cumulative_binom_prob( STAT p, int k, int n )
{
  return betai( (PVAL)k, (PVAL)(n - k + 1.0), (PVAL)p );
}











PVAL get_pval_chi_sq( MODEL_STATS *s )
{

  STAT v = s->dof_numer;
  STAT x = 2.0 * (s->lhood - s->old_lhood);
  
  return gammq( v / 2.0, x / 2.0 );

}











PVAL gammq( STAT a, STAT x )
{

  PVAL gamser, gammcf, gln;

  if ( x < 0.0 || a <= 0.0 )
    return 1.0;

  if ( x < (a + 1.0)) {
    gser( &gamser, a, x, &gln );
    return 1.0 - gamser;
  }
  else {
    gcf( &gammcf, a, x, &gln );
    return gammcf;
  }

}












void gser( STAT *gamser, STAT a, STAT x, STAT *gln )
{

  int n;
  STAT sum, del, ap;
  
  *gln = gammln( a );
  
  if ( x <= 0.0 ) {
    if ( x < 0.0 )
      error( 1, "See sig.c.\n" );
    *gamser = 0.0;
    return;
  }
  else {
  
    ap = a;
    del = sum = 1.0 / a;
    
    for ( n = 1; n <= MAXIT; n++ ) {
    
      ++ap;
      del *= x / ap;
      sum += del;
      if ( fabs( del ) < fabs( sum ) * EPS ) {
        *gamser = sum * exp( (-1.0 * x) + (a * log( x )) - (*gln));
        return;
      }
    
    }
  
  }
  
  error( 2, "Iter > MAXIT.\n" );

}











void gcf( STAT *gammcf, STAT a, STAT x, STAT *gln )
{

  int i;
  STAT an, b, c, d, del, h;
  
  *gln = gammln( a );
  
  b = x + 1.0 - a;
  c = 1.0 / FPMIN;
  d = 1.0 / b;
  h = d;
  
  for ( i = 1; i <= MAXIT; i++ ) {
  
    an = (-1.0 * i) * (i - a);
    b += 2.0;
    d = (an * d) + b;
    if ( fabs( d ) < FPMIN )
      d = FPMIN;
    c = b + (an / c);
    if ( fabs( c ) < FPMIN )
      c = FPMIN;
    d = 1.0 / d;
    del = d * c;
    h *= del;
    if ( fabs( del - 1.0 ) < EPS ) 
      break;
  
  }
  
  *gammcf = exp( (-1.0 * x) + (a * log( x )) - (*gln)) * h;

}












PVAL betai( PVAL a, PVAL b, PVAL x )
{

  if ( x < 0.0 || x > 1.0 )
    return 1.0;
  
  PVAL bt;
    
  if ( x == 0.0 || x == 1.0 )
    bt = 0.0;
  else
    bt = exp( gammln( a + b ) - gammln( a ) - gammln( b ) + (a * log( x ))
      + (b * log( 1.0 - x )) );
      
  if ( x < (a + 1.0) / (a + b + 2.0) )
    return bt * betacf( a, b, x ) / a;

  return 1.0 - (bt * betacf( b, a, 1.0 - x ) / b);

}














PVAL betacf( PVAL a, PVAL b, PVAL x )
{

  int m, m2;
  PVAL aa, c, d, del, h, qab, qam, qap;
  
  qab = a + b;
  qap = a + 1.0;
  qam = a - 1.0;
  c = 1.0;
  d = 1.0 - (qab * x / qap);
  
  if ( fabs( d ) < FPMIN )
    d = FPMIN;
    
  d = 1.0 / d;
  h = d;
  
  for ( m = 1; m <= MAXIT; m++ ) {
  
    m2 = 2 * m;
    aa = (m * (b - m) * x) / ((qam + m2) * (a + m2));
    d = 1.0 + (aa * d);

    if ( fabs( d ) < FPMIN )
      d = FPMIN;

    c = 1.0 + (aa / c);

    if ( fabs( c ) < FPMIN )
      c = FPMIN;

    d = 1.0 / d;
    h *= d * c;
    aa = (-1.0 * (a + m) * (qab + m) * x) / ((a + m2) * (qap + m2));

    d = 1.0 + (aa * d);

    if ( fabs( d ) < FPMIN )
      d = FPMIN;

    c = 1.0 + (aa / c);

    if ( fabs( c ) < FPMIN )
      c = FPMIN;

    d = 1.0 / d;
    del = d * c;
    h *= del;

    if ( fabs( del - 1.0 ) < EPS )
      break;
  
  }
  
  if ( m > MAXIT )
    error( 6300, "Maximum iterations for P-value generation.\n" );
    
  return h;

}










PVAL gammln( PVAL xx )
{

  PVAL x, y, tmp, ser;
  
  static PVAL cof[ 6 ] = { 76.18009172947146, -86.50532032941677, 24.01409824083091,
    -1.231739572450155, 0.1208650973866179e-2, -0.5395239384953e-5 };

  int j;
  
  y = x = xx;
  tmp = x + 5.5;
  tmp -= (x + 0.5) * log( tmp );
  ser = 1.000000000190015;
  
  for ( j = 0; j <= 5; j++ )
    ser += cof[ j ] / ++y;
    
  return (-1.0 * tmp) + log( 2.5066282746310005 * ser / x );

}












































