#include "rel.h"












#define COMMON_ANC "Catarrhini"











void get_rel_origin_stats_for_rel_lister( REL_ORIGIN_LIST *rel, 
  ORIGIN_LIST *major_origin, ORIGIN_LIST *minor_origin, DATA_LIST *major_data,
  DATA_LIST *minor_data )
{

  extern char major_spec[ BUFFER ];
  
  initialize_rel_hist( rel->my_rel->rel_hist, major_data->data->tree, major_spec );

  get_my_rel_origin_stats_for_rel_lister( rel, major_origin, minor_origin,
    major_data, minor_data );
  
}














void initialize_rel_hist( REL_HIST *rel_hist[ STRAND_BUFFER ][ STRAND_BUFFER ], 
  TREE *tree, char spec[] )
{

  NODE_LIST *get_node_from_label( NODE_LIST *node_header, char label[] );
  
  NODE_LIST *node = NULL;
  /*  In ../tree/get.c:  */
  node = (NODE_LIST *)get_node_from_label( tree->spec_header, spec );
  
  int i = 0, j, a, b;
  
  while ( node != NULL ) {
  
    for ( a = 0; a < STRAND_BUFFER; a++ ) {
      for ( b = 0; b < STRAND_BUFFER; b++ ) {
        strcpy( rel_hist[ a ][ b ]->node[ i ], node->my_node->label );
        rel_hist[ a ][ b ]->time[ i ] = node->my_node->t;
      }
    }
    
    ++i;
    
    if ( ! strcmp( COMMON_ANC, node->my_node->label ))
      break;
      
    node = node->up;
  
  }
  
  int n = i;
  
  for ( a = 0; a < STRAND_BUFFER; a++ ) {
    for ( b = 0; b < STRAND_BUFFER; b++ ) {
      for ( i = 0; i < n; i++ ) {
        for ( j = 0; j < n; j++ ) {
          rel_hist[ a ][ b ]->ct[ i ][ j ] = 0;
          rel_hist[ a ][ b ]->seq_ct[ i ][ j ] = 0;
        }
      }
      rel_hist[ a ][ b ]->n = n;
      rel_hist[ a ][ b ]->total_seqs = 0;
    }
  }
  

}














void get_my_rel_origin_stats_for_rel_lister( REL_ORIGIN_LIST *rel, 
  ORIGIN_LIST *major_origin, ORIGIN_LIST *minor_origin,
  DATA_LIST *major_data, DATA_LIST *minor_data )
{

  int major_strand, minor_strand;
  
  for ( major_strand = 0; major_strand < STRAND_BUFFER; major_strand++ ) {
    for ( minor_strand = 0; minor_strand < STRAND_BUFFER; minor_strand++ )
      get_rel_origin_cooccur_single_strand_pair( rel, major_origin, minor_origin,
        major_data, minor_data, major_strand, minor_strand );
  }

  set_rel_origin_statistical_sig( rel );

}












void get_rel_origin_cooccur_single_strand_pair( REL_ORIGIN_LIST *rel,
  ORIGIN_LIST *major_origin, ORIGIN_LIST *minor_origin, DATA_LIST *major_data,
  DATA_LIST *minor_data, int major_strand, int minor_strand )
{

  int a = major_strand, b = minor_strand;

  SITE_LIST *major_site = major_origin->my_origin->site_header[ a ];
  SITE_LIST *minor_site = minor_origin->my_origin->site_header[ b ];
  
  while (( major_site != NULL ) && ( minor_site != NULL ))
    get_rel_origin_cooccur_single_site_and_step( rel->my_rel->rel_hist[ a ][ b ],
      &major_site, &minor_site );

}











void get_rel_origin_cooccur_single_site_and_step( REL_HIST *my_hist,
  SITE_LIST **major_site, SITE_LIST **minor_site )
{

  SEQ_INFO *major = (*major_site)->site->loc;
  SEQ_INFO *minor = (*minor_site)->site->loc;
  
  step_site_listers_nonmatching_chrs( major_site, minor_site );
  
  if (( *major_site == NULL ) || ( *minor_site == NULL ))
    return;
    
  step_site_listers_to_same_seq( major_site, minor_site );
  
  if (( *major_site == NULL ) || ( *minor_site == NULL ))
    return;

  get_my_rel_origin_cooccur_single_site_and_step( my_hist, major_site, minor_site );

}











void step_site_listers_nonmatching_chrs( SITE_LIST **major_site, 
  SITE_LIST **minor_site )
{

  while ( *major_site != NULL ) {
  
    SEQ_INFO *info = (*minor_site)->site->loc;
    
    if ( ! strcmp( info->chr, (*major_site)->site->loc->chr ))
      break;
  
    if ( chr_in_site_list( info->chr, *major_site ))
      *major_site = (*major_site)->next;
    else
      break;
  
  }
  
  if ( *major_site == NULL )
    return;
  
  while ( *minor_site != NULL ) {
  
    SEQ_INFO *info = (*major_site)->site->loc;
    SEQ_INFO *minor_info = (*minor_site)->site->loc;
    
    if ( ! strcmp( info->chr, minor_info->chr ))
      break;

    if ( chr_in_site_list( info->chr, *minor_site ))
      *minor_site = (*minor_site)->next;
    else if ( chr_in_site_list( minor_info->chr, *major_site )) {
      *minor_site = NULL;
      *major_site = NULL;
      break;
    }
  
  }
  
}










int chr_in_site_list( char chr[], SITE_LIST *my_site )
{

  SITE_LIST *site = my_site;

  while ( site != NULL ) {

    if ( ! strcmp( site->site->loc->chr, chr ))
      return 1;

    site = site->next;

  }
  
  return 0;

}








void step_site_listers_to_same_seq( SITE_LIST **major_site, SITE_LIST **minor_site )
{

  if (( *major_site == NULL ) || ( *minor_site == NULL ))
    return;

  while (( *major_site != NULL ) && 
    ( (*major_site)->site->seq_numb < (*minor_site)->site->seq_numb ))
  {
    *major_site = (*major_site)->next;
  }
  
  if ( *major_site == NULL )
    return;

  while (( *minor_site != NULL ) &&
    ( (*minor_site)->site->seq_numb < (*major_site)->site->seq_numb ))
  {
    *minor_site = (*minor_site)->next;
  }

}









void get_my_rel_origin_cooccur_single_site_and_step( REL_HIST *my_hist, 
  SITE_LIST **major_site, SITE_LIST **minor_site )
{

  if ( (*major_site)->site->seq_numb != (*minor_site)->site->seq_numb )
    return;
    
  printf( "SeqNumb = %d      %s    %s\n", (int)(*major_site)->site->seq_numb,
    (*major_site)->site->loc->gene,
    (*minor_site)->site->loc->gene );
    
  char major_node[ BUFFER ], minor_node[ BUFFER ];
  
  if ( ! get_node_from_site_lister_and_step( major_node, major_site ))
    return;
    
  if ( ! get_node_from_site_lister_and_step( minor_node, minor_site ))
    return;
    
  update_rel_origin_hist_w_nodes( my_hist, major_node, minor_node );

}














int get_node_from_site_lister_and_step( char node[], SITE_LIST **site )
{

  strcpy( node, (*site)->stats->node );
  
  SEQ seq = (*site)->site->seq_numb;
  
  int all_match = 1;
  
  while (( *site != NULL ) && ( (*site)->site->seq_numb == seq )) {

    if (( strcmp( (*site)->stats->node, node )) && 
      ( is_lin_spec( (*site)->stats->node )))
    {
      all_match = 0;
    }
    else {
      strcpy( node, (*site)->stats->node );
    }
            
    *site = (*site)->next;
    
  }
  
  return all_match;

}











int is_lin_spec( char node[] )
{

  if ( ! strcmp( node, "hg19" ))
    return 1;
    
  if ( ! strcmp( node, "Homininae2" ))
    return 1;
    
  if ( ! strcmp( node, "Homininae" ))
    return 1;
    
  if ( ! strcmp( node, "Hominidae" ))
    return 1;

  if ( ! strcmp( node, "Catarrhini" ))
    return 1;

  /*
  if ( ! strcmp( node, "Simiiformes" ))
    return 1;
  
  if ( ! strcmp( node, "Haplorrhini" ))
    return 1;
    
  if ( ! strcmp( node, "Primates" ))
    return 1;
  */
    
  return 0;
  
}









void update_rel_origin_hist_w_nodes( REL_HIST *my_hist, char major_node[], 
  char minor_node[] )
{

  int major_index = get_node_index_from_rel_hist( my_hist, major_node );
  int minor_index = get_node_index_from_rel_hist( my_hist, minor_node );
  
  if (( major_index != IMPOSSIBLE ) && ( minor_index != IMPOSSIBLE ))
    ++(my_hist->seq_ct[ major_index ][ minor_index ]);

}














int get_node_index_from_rel_hist( REL_HIST *my_hist, char node[] )
{

  if ( ! strcmp( node, "Eutheria" ))
    return IMPOSSIBLE;

  int i;
  
  for ( i = 0; i < my_hist->n; i++ ) {
  
    if ( ! strcmp( my_hist->node[ i ], node )) 
      return i;
  
  }
  
  return IMPOSSIBLE;

}













/*============================================================================*/
/*============================================================================*/
/*============================================================================*/
/*============================================================================*/
/*============================================================================*/












void set_rel_origin_statistical_sig( REL_ORIGIN_LIST *rel )
{

  int a, b;
  
  for ( a = 0; a < STRAND_BUFFER; a++ ) {
  
    for ( b = 0; b < STRAND_BUFFER; b++ )
      set_rel_origin_hist_stat_sig( rel->my_rel->rel_hist[ a ][ b ],
        rel->my_rel->rel_stats[ a ][ b ] );
  
  }
  
  int x, y;
  
  get_best_rel_origin_indices( &x, &y, rel );
  
  set_rel_origin_best_stats( rel, x, y );

}














void set_rel_origin_hist_stat_sig( REL_HIST *my_hist, OREL_STATS *my_stats )
{

  PVAL get_same_branch_pval_rel_hist( REL_HIST *my_hist );
  
  /*  In this file, below:  */
  set_rel_origin_hist_exp_vals( my_hist );
  
  /*  In rel_sig.c:  */
  my_hist->pval = get_same_branch_pval_rel_hist( my_hist );

}












void set_rel_origin_hist_exp_vals( REL_HIST *my_hist )
{

  int major_total[ BUFFER ], minor_total[ BUFFER ];
  
  int total = get_rel_origin_hist_arrays( my_hist, major_total, minor_total );
  
  set_my_rel_origin_hist_exp_vals( my_hist, major_total, minor_total, total );

}










int get_rel_origin_hist_arrays( REL_HIST *my_hist, int major_total[],
  int minor_total[] )
{

  my_hist->x = 0;

  int i, j;
  
  for ( i = 0; i < my_hist->n; i++ ) {
    major_total[ i ] = 0;
    minor_total[ i ] = 0;
  }
  
  int total = 0, ct = 0;
  
  for ( i = 0; i < my_hist->n; i++ ) {
    for ( j = 0; j < my_hist->n; j++ ) {
    
      total += my_hist->seq_ct[ i ][ j ];
      
      if ( i == j )
        my_hist->x += my_hist->seq_ct[ i ][ j ];
        
      minor_total[ j ] += my_hist->seq_ct[ i ][ j ];
      major_total[ i ] += my_hist->seq_ct[ i ][ j ];
      
    }
    
  }
  
  return total;
  
}








void set_my_rel_origin_hist_exp_vals( REL_HIST *my_hist, int major_total[], 
  int minor_total[], int total )
{

  int i, j;
  
  for ( i = 0; i < my_hist->n; i++ ) {
    for ( j = 0; j < my_hist->n; j++ )
      my_hist->exp[ i ][ j ] = 
        (float)(major_total[ i ] * minor_total[ j ]) / (float)total;
  }

}











void print_rel_origin_hist( REL_HIST *my_hist )
{
  print_rel_origin_hist_fh( my_hist, stdout );
}











void print_rel_origin_hist_fh( REL_HIST *my_hist, FILE *fh )
{

  int i, j;
  
  print_dot_sep( fh );
  
  for ( i = 0; i < my_hist->n; i++ ) {
  
    fprintf( fh, "%15s\t\t", my_hist->node[ i ] );
    
    for ( j = 0; j < my_hist->n; j++ ) 
      fprintf( fh, "%5d  (%.1f)    ", 
        my_hist->seq_ct[ i ][ j ], my_hist->exp[ i ][ j ] );
    
    fprintf( fh, "\n" );
  
  }
  
  fprintf( fh, "\n" );
  
  printf( "Pval = %.3e\n", (double)my_hist->pval );
  
  printf( "\n\n" );

}














void get_best_rel_origin_indices( int *x, int *y, REL_ORIGIN_LIST *rel )
{

  PVAL pval = 2.0;
  
  int a, b;
  
  for ( a = 0; a < STRAND_BUFFER; a++ ) {
    for ( b = 0; b < STRAND_BUFFER; b++ ) {
    
      if ( rel->my_rel->rel_hist[ a ][ b ]->pval < pval ) {
        pval = rel->my_rel->rel_hist[ a ][ b ]->pval;
        *x = a;
        *y = b;
      }
      
    }
  }

}














void set_rel_origin_best_stats( REL_ORIGIN_LIST *rel, int x, int y )
{

  PVAL get_same_branch_pval_rel_hist( REL_HIST *my_hist );

  *(rel->my_rel->best_hist) = *(rel->my_rel->rel_hist[ x ][ y ]);
  
  /*  In this file, below:  */
  int a = opposite_strand_index( x );
  int b = opposite_strand_index( y );
  
  /*  In this file, below:  */
  update_rel_origin_hist( rel->my_rel->best_hist, 
    rel->my_rel->rel_hist[ a ][ b ] );
    
  /*  In rel_sig.c:  */
  rel->my_rel->best_hist->pval = 
    get_same_branch_pval_rel_hist( rel->my_rel->best_hist );

  rel->my_rel->best_hist->major_index[ 0 ] = x;
  rel->my_rel->best_hist->major_index[ 1 ] = y;
  rel->my_rel->best_hist->minor_index[ 0 ] = a;
  rel->my_rel->best_hist->minor_index[ 1 ] = b;
  
  *(rel->my_rel->rev_hist) = *(rel->my_rel->rel_hist[ x ][ b ]);
  
  /*  In this file, below:  */
  update_rel_origin_hist( rel->my_rel->rev_hist,
    rel->my_rel->rel_hist[ a ][ y ] );
    
  /*  In rel_sig.c:  */
  rel->my_rel->rev_hist->pval = 
    get_same_branch_pval_rel_hist( rel->my_rel->rev_hist );
    
  rel->my_rel->rev_hist->major_index[ 0 ] = a;
  rel->my_rel->rev_hist->major_index[ 1 ] = y;
  rel->my_rel->rev_hist->minor_index[ 0 ] = x;
  rel->my_rel->rev_hist->minor_index[ 1 ] = b;
  
  *(rel->my_rel->all_hist) = *(rel->my_rel->best_hist);
  /*  In this file, below:  */
  update_rel_origin_hist( rel->my_rel->all_hist, rel->my_rel->rev_hist );
  
  rel->my_rel->all_hist->pval = 
    get_same_branch_pval_rel_hist( rel->my_rel->all_hist );
  
  /*  In this file, above:  */
//  print_rel_origin_hist( rel->my_rel->rev_hist );
  //print_rel_origin_hist( rel->my_rel->best_hist );

  print_rel_origin_hist( rel->my_rel->all_hist );

}











int opposite_strand_index( int x )
{
  return ( x == PLUS_STRAND ) ? (MINUS_STRAND) : (PLUS_STRAND);
}









void update_rel_origin_hist( REL_HIST *hist, REL_HIST *h )
{

  int i, j;
  
  for ( i = 0; i < hist->n; i++ ) {
  
    for ( j = 0; j < hist->n; j++ ) {
    
      hist->seq_ct[ i ][ j ] += h->seq_ct[ i ][ j ];
      hist->exp[ i ][ j ] += h->seq_ct[ i ][ j ];
    
    }
  
  }
  
  hist->x += h->x;

}
















/*============================================================================*/
/*============================================================================*/
/*============================================================================*/
/*============================================================================*/
/*============================================================================*/














void get_rel_origin_stats_for_after_rel_lister( REL_ORIGIN_LIST *rel, 
  ORIGIN_LIST *major_origin, ORIGIN_LIST *minor_origin, DATA_LIST *major_data, 
  DATA_LIST *minor_data )
{

  extern char major_spec[ BUFFER ];
  
  initialize_rel_hist( rel->my_rel->rel_hist, major_data->data->tree, major_spec );

  get_my_rel_origin_stats_for_after_rel_lister( rel, major_origin, minor_origin,
    major_data, minor_data );

}














void get_my_rel_origin_stats_for_after_rel_lister( REL_ORIGIN_LIST *rel,
  ORIGIN_LIST *major_origin, ORIGIN_LIST *minor_origin, DATA_LIST *major_data,
  DATA_LIST *minor_data )
{

  int major_strand, minor_strand;
  
  for ( major_strand = 0; major_strand < STRAND_BUFFER; major_strand++ ) {
    for ( minor_strand = 0; minor_strand < STRAND_BUFFER; minor_strand++ )
      get_rel_origin_cooccur_single_after_strand_pair( rel, major_origin, 
        minor_origin, major_data, minor_data, major_strand, minor_strand );
  }

  set_rel_origin_statistical_sig( rel );

}










void get_rel_origin_cooccur_single_after_strand_pair( REL_ORIGIN_LIST *rel,
  ORIGIN_LIST *major_origin, ORIGIN_LIST *minor_origin, DATA_LIST *major_data,
  DATA_LIST *minor_data, int major_strand, int minor_strand )
{

  int a = major_strand, b = minor_strand;

  SITE_LIST *major_site = major_origin->my_origin->site_header[ a ];
  SITE_LIST *minor_site = minor_origin->my_origin->site_header[ b ];
  
  while (( major_site != NULL ) && ( minor_site != NULL ))
    get_rel_origin_cooccur_after_site_and_step( rel->my_rel->rel_hist[ a ][ b ],
      &major_site, &minor_site );

}












void get_rel_origin_cooccur_after_site_and_step( REL_HIST *my_hist,
  SITE_LIST **major_site, SITE_LIST **minor_site )
{

  SEQ_INFO *major = (*major_site)->site->loc;
  SEQ_INFO *minor = (*minor_site)->site->loc;
  
  step_site_listers_nonmatching_chrs( major_site, minor_site );
  
  if (( *major_site == NULL ) || ( *minor_site == NULL ))
    return;
    
  step_site_listers_to_same_seq( major_site, minor_site );
  
  if (( *major_site == NULL ) || ( *minor_site == NULL ))
    return;
    
  if ( site_listers_overlap( *major_site, *minor_site ))
  {
    *major_site = (*major_site)->next;
    *minor_site = (*minor_site)->next;
    return;    
  }
  
  get_my_rel_origin_cooccur_single_site_and_step( my_hist, major_site, 
    minor_site );

}















int site_listers_overlap( SITE_LIST *major_site, SITE_LIST *minor_site )
{

  if ( major_site->site->seq_numb != minor_site->site->seq_numb )
    return 0;
    
  if ( strcmp( major_site->site->loc->chr, minor_site->site->loc->chr ))
    return 0;
    
  if ( major_site->site->loc->start > minor_site->site->loc->end )
    return 0;
    
  if ( major_site->site->loc->end < minor_site->site->loc->start )
    return 0;
    
  return 1;

}








    
    
    
    
    
    