#include "pwm.h"













/*
#define PLUS_STRAND 1
#define MINUS_STRAND -1
*/








/*  Defined in pwm.h:  
#define MAX_KMERS 250
#define MY_MAX_KMERS (MAX_KMERS * 2)
#define KMER_LEN 10
*/

#define NUMB_OF_SEQ_BUFFER 70000

#define OVER_ITER 3

#define CT_RATIO 0.80
#define SEQ_CT_RATIO 0.75







void do_pwm( void )
{

  extern MOTIF major_motif;
  extern char input_file[ BUFFER ], seq_align_file[ BUFFER ];
  
  if ( ! major_motif.length )
    error( 4000, "Motif required.\n" );
    
  strcpy( input_file, seq_align_file );
    
  do_my_pwm( &major_motif );

}











void do_my_pwm( MOTIF *mf )
{

  FILE *open_text_file_writing( char a[], char b[] );
  
  extern char input_file[ BUFFER ];
  extern int reg_start, reg_end;
  
  char output_fname[ BUFFER ], cons[ BUFFER ];
  
  upper( cons, mf->cons );
  
  sprintf( output_fname, "%s.%d:%d_motif_%s.txt", input_file,
    reg_start, reg_end, cons );
  //sprintf( output_fname, "%s.%d:%d_motif_%s_%.2f_%.2f.txt", input_file, reg_start, reg_end, cons, CT_RATIO, SEQ_CT_RATIO);

  FILE *out = NULL, *pwm_fh = NULL;
  out = (FILE *)open_text_file_writing( output_fname, "" );
  pwm_fh = (FILE *)open_text_file_writing( output_fname, "pwm" );
  
  do_pwm_w_fh( mf, out, pwm_fh );
  
  close_file( out );
  close_file( pwm_fh );

}












void do_pwm_w_fh( MOTIF *mf, FILE *out, FILE *pwm_fh )
{

  PWM pwm;
  
  get_pwm_from_motif( &pwm, mf, out, pwm_fh );
  
}










void get_pwm_from_motif( PWM *pwm, MOTIF *mf, FILE *out, FILE *pwm_fh )
{

  FILE *open_binary_file_reading( char a[], char b[] );

  extern char input_file[ BUFFER ];

  FILE *ft_fh = NULL, *ind_fh = NULL, *fpos_fh = NULL, *seq_fh = NULL;
  ft_fh = (FILE *)open_binary_file_reading( input_file, "f" );
  ind_fh = (FILE *)open_binary_file_reading( input_file, "i" );
  seq_fh = (FILE *)open_binary_file_reading( input_file, "s" );
  fpos_fh = (FILE *)open_binary_file_reading( input_file, "p" );
  
  get_pwm_from_motif_w_fhs( pwm, mf, ft_fh, ind_fh, seq_fh, fpos_fh, out, pwm_fh );

  close_file( ft_fh );
  close_file( ind_fh );
  close_file( seq_fh );
  close_file( fpos_fh );

}












void get_pwm_from_motif_w_fhs( PWM *pwm, MOTIF *mf, FILE *ft_fh, FILE *ind_fh,
  FILE *seq_fh, FILE *fpos_fh, FILE *out, FILE *pwm_fh )
{
  
  SPEC_CELL *get_spec_cell( void );
  
  SPEC_CELL *spec_cell = NULL;
  /*  In ../model/data.c:  */
  spec_cell = (SPEC_CELL *)get_spec_cell();
  
  /*  In this file, below:  */
  get_my_pwm_from_motif( pwm, mf, spec_cell, ft_fh, ind_fh, seq_fh, fpos_fh,
    out, pwm_fh );

}












void get_my_pwm_from_motif( PWM *pwm, MOTIF *mf, SPEC_CELL *spec_cell, FILE *ft_fh,
  FILE *ind_fh, FILE *seq_fh, FILE *fpos_fh, FILE *out, FILE *pwm_fh )
{

  PWM old_pwm, bg_pwm;
  
  initialize_pwm( &old_pwm, mf->length );
  initialize_pwm( &bg_pwm, mf->length );
  
  set_initial_pwm_w_motif( &old_pwm, mf, spec_cell, ft_fh, ind_fh, seq_fh, fpos_fh );

  set_bg_pwm( &bg_pwm, mf, spec_cell, ft_fh, ind_fh, seq_fh, fpos_fh );
  
  set_pwm_freq( &old_pwm );
  set_pwm_freq( &bg_pwm );
  
  print_pwm( &old_pwm );
  print_pwm( &bg_pwm );
  
  float score = get_min_init_kmer_pwm_score( &old_pwm, &bg_pwm, mf );
  float new_score;
  
  int iter_numb = 0;
  
  do {
  
    printf( "score = %.2f\n", score );
  
    new_score = pwm_iteration( pwm, &old_pwm, &bg_pwm, mf, spec_cell, 
      score, ft_fh, ind_fh, seq_fh, fpos_fh ); 
      
    print_pwm( pwm );
      
    if ( ! pwms_differ( pwm, &old_pwm )) 
      break;
    
    score = new_score;
    old_pwm = *pwm;
  
  } while ( ++iter_numb <= 5 );
  
  printf( "score = %.2f\n", score );
  
  float score_thresh = get_pwm_score_thresh( pwm, &bg_pwm, score, spec_cell,
    ft_fh, ind_fh, seq_fh, fpos_fh );
    
  printf( "Final score = %.4f\n", score_thresh );
  
  print_pwm_kmer_list( pwm, &bg_pwm, mf, score_thresh, out );
  print_pwm_file_fh( pwm, &bg_pwm, score_thresh, pwm_fh );
  
}












void initialize_pwm( PWM *pwm, char length )
{

  int i, j;
  
  for ( i = 0; i < length; i++ ) {
    for ( j = 0; j < NUMB_OF_NUCL; j++ ) {
      pwm->pwm[ i ][ j ] = 0;
      pwm->freq[ i ][ j ] = 0.0;
    }
  }
  
  pwm->length = length;

}










void set_pwm_freq( PWM *pwm )
{

  int i, j;
  
  for ( i = 0; i < pwm->length; i++ ) {
  
    long total = 0;
    
    for ( j = 0; j < NUMB_OF_NUCL; j++ )
      total += pwm->pwm[ i ][ j ];
      
    for ( j = 0; j < NUMB_OF_NUCL; j++ )
      pwm->freq[ i ][ j ] = (float)pwm->pwm[ i ][ j ] / (float)total;
  
  }

}










float get_min_init_kmer_pwm_score( PWM *pwm, PWM *bg_pwm, MOTIF *mf )
{

  int get_motif_stack_mismatches( MOTIF *mf, MF stack );
    
  float min = 10000.0;
  
  MF i;
  
  for ( i = 0; i < NUMB_OF_WORDS; i++ ) {
    /*  In ../main/nucl.c:  */
    if ( get_motif_stack_mismatches( mf, i ) <= 1 ) {
      /*  In this file, below:  */
      float score = get_kmer_pwm_score( i, pwm, bg_pwm, mf->length );
      min = ( isinf( score ) || isnan( score )) ? (min) : (MIN( min, score ));
    }
  }
  
  return min;

}














float get_kmer_pwm_score( MF stack, PWM *pwm, PWM *bg_pwm, char length )
{

  MF kmer = stack;
  
  float score = 0.0;

  int i, j;
  
  for ( i = length - 1; i >= 0; i-- ) {
    int bp = kmer % NUMB_OF_NUCL;
    score += log( pwm->freq[ i ][ bp ] ) - log( bg_pwm->freq[ i ][ bp ] );
    kmer = kmer >> BITS_PER_NUCL;
  }
  
  return ( isinf(score) || isnan(score) ) ? (0.0) : (score);

}














int pwms_differ( PWM *pwm, PWM *old_pwm )
{

  int i, j;
  
  for ( i = 0; i < pwm->length; i++ ) {
    for ( j = 0; j < NUMB_OF_NUCL; j++ ) {
      if ( pwm->pwm[ i ][ j ] != old_pwm->pwm[ i ][ j ] )
        return 1;
    }
  }
  
  return 0;

}













void get_rev_compl_pwm( PWM *rc_pwm, PWM *pwm )
{

  int i, j;
  
  for ( i = 0; i < pwm->length; i++ ) {
    for ( j = 0; j < NUMB_OF_NUCL; j++ ) {
      
      rc_pwm->pwm[ pwm->length - i - 1 ][ NUMB_OF_NUCL - j - 1 ] = 
        pwm->pwm[ i ][ j ];
      
      rc_pwm->freq[ pwm->length - i - 1 ][ NUMB_OF_NUCL - j - 1 ] = 
        pwm->freq[ i ][ j ];
    
    }
  }

}












void print_pwm_file_fh( PWM *pwm, PWM *bg_pwm, float score_thresh, 
  FILE *pwm_fh )
{

  print_pwm_fh( pwm, pwm_fh );
  
  fprintf( pwm_fh, "\n" );
  
  print_pwm_fh( bg_pwm, pwm_fh );
  
  fprintf( pwm_fh, "\n" );
  fprintf( pwm_fh, "ScoreThresh:  %.4f\n", score_thresh );

}








void print_pwm( PWM *pwm )
{

  print_sep( stdout );
  
  int i, j;
  
  for ( j = 0; j < NUMB_OF_NUCL; j++ ) {
    for ( i = 0; i < pwm->length; i++ ) 
      printf( "%10ld   ", pwm->pwm[ i ][ j ] );
    printf( "\n" );
  }
  
  printf( "\n\n" );
  
  for ( j = 0; j < NUMB_OF_NUCL; j++ ) {
  
    for ( i = 0; i < pwm->length; i++ )
      printf( "%10.4f   ", pwm->freq[ i ][ j ] );
      
    printf( "\n" );
  
  }
  
  print_sep( stdout );

}










void print_pwm_fh( PWM *pwm, FILE *fh )
{

  print_sep( fh );
  
  int i, j;
  
  for ( j = 0; j < NUMB_OF_NUCL; j++ ) {
    for ( i = 0; i < pwm->length; i++ ) 
      fprintf( fh, "%10ld   ", pwm->pwm[ i ][ j ] );
    fprintf( fh, "\n" );
  }
  
  fprintf( fh, "\n\n" );
  
  for ( j = 0; j < NUMB_OF_NUCL; j++ ) {
  
    for ( i = 0; i < pwm->length; i++ )
      fprintf( fh, "%10.4f   ", pwm->freq[ i ][ j ] );
      
    fprintf( fh, "\n" );
  
  }
  
  print_sep( fh );

}










void print_pwm_kmer_list( PWM *pwm, PWM *bg_pwm, MOTIF *mf, float score, FILE *out )
{

  int get_motif_stack_mismatches( MOTIF *mf, MF stack );
  void get_rev_complement_seq( char rev_compl[], char seq[] );
  void get_string_from_stack( char cons[], MF stack, char length );
  int get_motif_stack_mismatches( MOTIF *mf, MF stack );
  MF get_stack_from_string( char cons[] );
  
  float min_score = 10000.0;
  
  PWM rc_pwm = { .length = pwm->length }, rc_bg = rc_pwm;
  
  char cons_array[ MY_MAX_KMERS ][ KMER_LEN ];
  int buffer = 0;
  
  get_rev_compl_pwm( &rc_pwm, pwm );
  get_rev_compl_pwm( &rc_bg, bg_pwm );
    
  MOTIF rc_mf = { .length = mf->length };
  
  /*  In ../main/nucl.c:  */
  get_rev_complement_seq( rc_mf.cons, mf->cons );
  
  char cons[ BUFFER ];
    
  MF w;
  
  for ( w = 0; w < NUMB_OF_WORDS; w++ ) {

    /*  In this file, above:  */
    float curr_score = get_kmer_pwm_score( w, pwm, bg_pwm, mf->length );
    
    /*  In ../main/nucl.c:  */
    int mm = get_motif_stack_mismatches( mf, w );
    
    if ((( curr_score >= score ) && ( mm < 3 )) || ( mm == 0 )) {

      /*  In ../main/nucl.c:  */
      get_string_from_stack( cons, w, pwm->length );
      
      /*  In this file, below:  */
      int index = cons_in_cons_array( cons, cons_array, buffer );

      if ( index < 0 ) {
        /*
        fprintf( out, "%s\n", cons );
        printf( "%d  %s\n", buffer + 1, cons );
        */
        strcpy( cons_array[ buffer++ ], cons );
      }
      else {

        MF new_w = get_stack_from_string( cons_array[ index ] );

        float new_a = get_kmer_pwm_score( new_w, pwm, bg_pwm, mf->length );
        float new_b = get_kmer_pwm_score( new_w, &rc_pwm, &rc_bg, mf->length );
        float new_score = MAX( new_a, new_b );
        
        if ( new_score >= curr_score )
          printf( "\t\t\t\t\tFiltered kmer :  %s\n", cons );
        else {
          printf( "\t\t\t\t\tFiltered kmer :  %s\n", cons_array[ index ] );
          strcpy( cons_array[ index ], cons );
        }

      }

    }

  }
  
  for ( w = 0; w < NUMB_OF_WORDS; w++ ) {
  
    float rc_score = get_kmer_pwm_score( w, &rc_pwm, &rc_bg, mf->length );
    
    int rc_mm = get_motif_stack_mismatches( &rc_mf, w );

    if (( rc_score >= score ) && ( rc_mm < 3 )) {

      /*  In ../main/nucl.c:  */
      get_string_from_stack( cons, w, pwm->length );
      
      /*  In this file, below:  */
      int index = cons_in_cons_array( cons, cons_array, buffer );

      if ( index < 0 ) {
        /*
        fprintf( out, "%s\n", cons );
        printf( "%d  %s\n", buffer + 1, cons );
        */
        strcpy( cons_array[ buffer++ ], cons );
      }
      else {
      
        /*
        MF new_w = get_stack_from_string( cons_array[ index ] );

        float new_a = get_kmer_pwm_score( new_w, pwm, bg_pwm, mf->length );
        float new_b = get_kmer_pwm_score( new_w, &rc_pwm, &rc_bg, mf->length );
        float new_score = MAX( new_a, new_b );
        
        if ( new_score >= rc_score )
        */
          printf( "\t\t\t\t\tFiltered kmer :  %s\n", cons );
        /*
        else {
          printf( "\t\t\t\t\tFiltered kmer :  %s\n", cons_array[ index ] );
          
          strcpy( cons_array[ index ], cons );
        }
        */

      }

    }  
  
  }
  
  int i;
  
  for ( i = 0; i < buffer; i++ ) {
    fprintf( out, "%s\n", cons_array[ i ] );
    printf( "%3d  %s\n", i + 1, cons_array[ i ] );
  }
  
}











int cons_in_cons_array( char cons[], char cons_array[ MY_MAX_KMERS ][ KMER_LEN ],
  int buffer )
{

  int i;
  
  for ( i = 0; i < buffer; i++ ) {
    if ( cons_strings_overlap( cons, cons_array[ i ] ))
      return i;
  }
  
  return -1;

}













int cons_strings_overlap( char a[], char b[] )
{

  int offset;
  
  for ( offset = -2; offset <= 2; offset++ ) {

    int j;
    for ( j = 0; j <= WORD_LENGTH; j++ ) {

      if ( j + offset < 0 )
        continue;

      if (( a[ j ] == '\0' ) || ( b[ j + offset ] == '\0' ))
        return 1;

      if ( a[ j ] != b[ j + offset ] )
        break;

    }

  }
  
  return 0;

}







void set_initial_pwm_w_motif( PWM *pwm, MOTIF *mf, SPEC_CELL *spec_cell, FILE *ft_fh, 
  FILE *ind_fh, FILE *seq_fh, FILE *fpos_fh )
{

  int get_motif_stack_mismatches( MOTIF *mf, MF stack );
  void get_rev_complement_seq( char rev_compl[], char seq[] );
  
  MOTIF rc_mf = { .length = mf->length };
  
  /*  In ../main/nucl.c:  */
  get_rev_complement_seq( rc_mf.cons, mf->cons );
    
  MF w;
  
  for ( w = 0; w < NUMB_OF_WORDS; w++ ) {
    /*  In ../main/nucl.c:  */
    if ( get_motif_stack_mismatches( mf, w ) <= 1 ) {
      /*  In this file, below:  */
      update_pwm_single_kmer( pwm, w, mf, spec_cell, PLUS_STRAND, ft_fh, ind_fh, 
        seq_fh, fpos_fh );
    }
    else if ( get_motif_stack_mismatches( &rc_mf, w ) <= 1 ) {
      /*  In this file, below:  */
      update_pwm_single_kmer( pwm, w, mf, spec_cell, MINUS_STRAND, ft_fh, ind_fh, 
        seq_fh, fpos_fh );
    }
  }
  
}















/*================================================================================*/
/*================================================================================*/
/*================================================================================*/
/*================================================================================*/
/*================================================================================*/









float pwm_iteration( PWM *pwm, PWM *old_pwm, PWM *bg_pwm, MOTIF *mf, 
  SPEC_CELL *spec_cell, float score, FILE *ft_fh, FILE *ind_fh, FILE *seq_fh, 
  FILE *fpos_fh )
{

  int get_motif_stack_mismatches( MOTIF *mf, MF stack );
  void get_rev_complement_seq( char rev_compl[], char seq[] );
  
  float min_score = 10000.0;
  
  PWM rc_pwm = { .length = pwm->length }, rc_bg = rc_pwm;
  
  initialize_pwm( pwm, old_pwm->length );
  
  get_rev_compl_pwm( &rc_pwm, old_pwm );
  get_rev_compl_pwm( &rc_bg, bg_pwm );
  
  MOTIF rc_mf = { .length = mf->length };
  
  get_rev_complement_seq( rc_mf.cons, mf->cons );
  
  MF w;
  
  for ( w = 0; w < NUMB_OF_WORDS; w++ ) {
  
    /*  In this file, above:  */
    float curr_score = get_kmer_pwm_score( w, old_pwm, bg_pwm, mf->length );
    
    /*  In ../main/nucl.c:  */
    int mm = get_motif_stack_mismatches( mf, w );
    
    float rc_score = get_kmer_pwm_score( w, &rc_pwm, &rc_bg, mf->length );
    
    int rc_mm = get_motif_stack_mismatches( &rc_mf, w );

    if ((( curr_score >= score ) && ( mm < 3 )) || ( mm == 0 )) {
      /*  In this file, below:  */
      update_pwm_single_kmer( pwm, w, mf, spec_cell, PLUS_STRAND, ft_fh, ind_fh, 
        seq_fh, fpos_fh );
      min_score = MIN( curr_score, min_score );
    }
    else if ((( rc_score >= score ) && ( rc_mm < 3 )) || ( rc_mm == 0 )) {
      /*  In this file, below:  */
      update_pwm_single_kmer( pwm, w, mf, spec_cell, MINUS_STRAND, ft_fh, ind_fh, 
        seq_fh, fpos_fh );
      min_score = MIN( rc_score, min_score );    
    }

  }
  
  set_pwm_freq( pwm );
  
  return min_score;

}






/*================================================================================*/
/*================================================================================*/
/*================================================================================*/
/*================================================================================*/
/*================================================================================*/











void update_pwm_single_kmer( PWM *pwm, MF w, MOTIF *mf, SPEC_CELL *spec_cell, 
  char strand, FILE *ft_fh, FILE *ind_fh, FILE *seq_fh, FILE *fpos_fh )
{

  void initialize_first_ft_array( FT ft_array[], int n );
  void set_position( FPOS fpos, FILE *ff );
  
  FT ft_array[ NUMB_OF_SPEC ];
  FT_IND ind_array[ NUMB_OF_SPEC ];  
  FPOS fpos;
  
  /*  In ../model/data.c:  */
  initialize_first_ft_array( ft_array, NUMB_OF_SPEC );
  
  set_position( sizeof(FPOS) * w, fpos_fh );  
  get( &fpos, sizeof(FPOS), fpos_fh );
  
  if ( fpos < 0 ) {
    return;
    error( 3500, "Unexpected kmer: Occurrences not recorded.\n" );
  }
  
  set_position( fpos, ind_fh );
  get( ind_array, sizeof(FT_IND) * spec_cell->n, ind_fh );
  
  /*  In this file, below:  */
  update_my_pwm_single_kmer( pwm, w, mf, ft_array, ind_array, spec_cell, strand,
    ft_fh, ind_fh, seq_fh, fpos_fh ); 

}












void update_my_pwm_single_kmer( PWM *pwm, MF w, MOTIF *mf, FT ft_array[],
  FT_IND ind_array[], SPEC_CELL *spec_cell, char strand, FILE *ft_fh, FILE *ind_fh, 
  FILE *seq_fh, FILE *fpos_fh )
{

  int update_ft_array_from_binary_fh( FT ft_array[], FT_IND ind_array[], int i,
    FILE *ft_fh );
  int get_new_seq_region_fh( SEQ_INFO *seq_info, int length[], int numb_of_spec,
    FILE *seq_fh );
  void set_position( FPOS fpos, FILE *ff );
  
  SEQ_INFO seq_info;
  int length[ NUMB_OF_SPEC ];  
    
  int i;
  for ( i = 0; i < spec_cell->n; i++ ) 
    /*  In ../model/data.c:  */
    update_ft_array_from_binary_fh( ft_array, ind_array, i, ft_fh );
    
  set_position( 0, seq_fh );

  int seq_numb = 0;
    
  /*  In ../model/data.c:  */
  while ( get_new_seq_region_fh( &seq_info, length, spec_cell->n, seq_fh )) 
    /*  In this file, below:  */
    update_pwm_single_kmer_single_seq( pwm, w, mf, ft_array, ind_array, spec_cell,
      &seq_info, length, strand, ++seq_numb, ft_fh );

}










void update_pwm_single_kmer_single_seq( PWM *pwm, MF w, MOTIF *mf, FT ft_array[],
  FT_IND ind_array[], SPEC_CELL *spec_cell, SEQ_INFO *seq_info, int length[], 
  char strand, SEQ seq_numb, FILE *ft_fh )
{

  void set_use_array_from_length_array( int use_array[], int length[], int n );
  int array_sum( int array[], int n );

  extern int score_thresh;

  int use_array[ NUMB_OF_SPEC ];
  
  if ( seq_info->score < score_thresh ) 
    return;
    
  /*  In ../model/data.c:  */
  set_use_array_from_length_array( use_array, length, spec_cell->n );
  
  /*  In ../model/data.c:  */
  if ( array_sum( use_array, spec_cell->n ) <= 2 ) 
    return;
    
  /*  In this file, below:  */
  update_my_pwm_single_kmer_single_seq( pwm, w, mf, ft_array, ind_array, use_array,
    seq_info, strand, seq_numb, ft_fh );

}










void update_my_pwm_single_kmer_single_seq( PWM *pwm, MF w, MOTIF *mf, 
  FT ft_array[], FT_IND ind_array[], int use_array[], SEQ_INFO *seq_info, 
  char strand, SEQ seq_numb, FILE *ft_fh )
{

  int update_ft_array_from_binary_fh( FT ft_array[], FT_IND ind_array[], int i,
    FILE *ft_fh );
    
  extern int reg_start, reg_end;
    
  int spec_numb = 0;
  
  if ( ! use_array[ spec_numb ] )
    return;
    
  FT *ft = ft_array + spec_numb;
  
  int last = -1;
  
  while ( ft->seq <= seq_numb ) {
  
    if ( ft->seq == seq_numb ) {
      int center = seq_info->n / 2;
      int x = ft->pos[ 0 ];
      if (( center + reg_start <= x ) && ( center + reg_end > x )) {
        if ( last < x ) {
          /*  In this file, below:  */
          update_pwm_single_ft_struct( pwm, ft, mf->length, strand );
          last = ft->pos[ pwm->length - 1 ];
        }
      }
    }
    else {
      last = -1;
    }
    
    /*  In ../model/data.c:  */
    if ( ! update_ft_array_from_binary_fh( ft_array, ind_array, spec_numb, ft_fh ))
      break;
      
    ft = ft_array + spec_numb;
  
  }

}











void update_pwm_single_ft_struct( PWM *pwm, FT *ft, char length, char strand )
{

  void get_rev_complement_seq( char rev_compl[], char seq[] );

  char rc[ BUFFER ];

  if ( strand == PLUS_STRAND ) 
    update_pwm_single_string( pwm, ft->site, length );
  else {
    get_rev_complement_seq( rc, ft->site );
    update_pwm_single_string( pwm, rc, length );
  }
  
}











void update_pwm_single_string( PWM *pwm, char string[], char length )
{

  MF get_bp_from_nucl( char nucl );

  int i;
  
  for ( i = 0; i < length; i++ ) 
    ++(pwm->pwm[ i ][ get_bp_from_nucl( string[ i ] ) ]);
  
  if ( string[ length ] != '\0' )
    error( 4500, "Error in kmer string length.\n" );

}

















/*================================================================================*/
/*================================================================================*/
/*================================================================================*/
/*================================================================================*/
/*================================================================================*/










void set_bg_pwm( PWM *pwm, MOTIF *mf, SPEC_CELL *spec_cell, FILE *ft_fh, 
  FILE *ind_fh, FILE *seq_fh, FILE *fpos_fh )
{

  FILE *open_text_file_reading( char a[], char b[] );
  FILE *open_binary_file_reading( char a[], char b[] );
  
  extern char input_file[ BUFFER ], region_file[ BUFFER ];
  
  FILE *in = NULL, *reg_fh = NULL;
  in = (FILE *)open_text_file_reading( input_file, "" );
  
  set_bg_pwm_w_fh( pwm, mf, spec_cell, in, reg_fh );
  
  close_file( in );

}









void set_bg_pwm_w_fh( PWM *pwm, MOTIF *mf, SPEC_CELL *spec_cell, FILE *in,
  FILE *reg_fh )
{

  char line[ LINE_BUFFER ];
  
  while ( fgets( line, LINE_BUFFER, in ) != NULL )
    get_bg_pwm_single_maf_line( pwm, mf, line );
    
}










void get_bg_pwm_single_maf_line( PWM *pwm, MOTIF *mf, char line[] )
{

  extern char major_spec[ BUFFER ];
  
  if ( line[ 0 ] != 's' )
    return;
  
  char spec[ BUFFER ];

  get_spec_from_maf_line( spec, line );
  
  if ( strcmp( spec, major_spec ))
    return;
    
  char seq[ LINE_BUFFER ];  
    
  sscanf( line, "%*s%*s%*d%*d%*s%*d%s", seq );
  
  get_bg_pwm_single_seq( pwm, mf, seq );

}
  
  










void get_bg_pwm_single_seq( PWM *pwm, MOTIF *mf, char seq[] )
{

  MF get_bp_from_nucl( char nucl );

  int i, j, end = strlen( seq );
  
  for ( i = 0; i < end; i++ ) {
  
    char c = tolower( seq[ i ] );
  
    switch ( c ) {
    
    case '-' :
    case 'n' :
      break;
  
    default :
      ++(pwm->pwm[ 0 ][ get_bp_from_nucl( c ) ]);        
      break;
      
    }
  
  }
  
  for ( j = 1; j < mf->length; j++ ) {
    for ( i = 0; i < NUMB_OF_NUCL; i++ )
      pwm->pwm[ j ][ i ] = pwm->pwm[ j - 1 ][ i ];
  }

}  
  
  
  
  
  
  
  
  
  
  
  
/*================================================================================*/
/*================================================================================*/
/*================================================================================*/
/*================================================================================*/
/*================================================================================*/
















float get_pwm_score_thresh( PWM *pwm, PWM *bg_pwm, float init_score, 
  SPEC_CELL *spec_cell, FILE *ft_fh, FILE *ind_fh, FILE *seq_fh, FILE *fpos_fh )
{

  float score_array[ MAX_KMERS ];
  MF kmer_array[ MAX_KMERS ];
  
  int i;
  for ( i = 0; i < MAX_KMERS; i++ ) {
    score_array[ i ] = init_score;
    kmer_array[ i ] = 0;
  }
    
  return get_my_pwm_score_thresh( pwm, bg_pwm, score_array, spec_cell, kmer_array, 
    ft_fh, ind_fh, seq_fh, fpos_fh );

}












float get_my_pwm_score_thresh( PWM *pwm, PWM *bg_pwm, float score_array[],
  SPEC_CELL *spec_cell, MF kmer_array[], FILE *ft_fh, FILE *ind_fh, FILE *seq_fh, 
  FILE *fpos_fh )
{

  int buffer = get_sorted_pwm_kmer_array( kmer_array, score_array, pwm, bg_pwm );

  return get_pwm_score_thresh_w_kmer_array( pwm, bg_pwm, kmer_array, score_array,
    buffer, spec_cell, ft_fh, ind_fh, seq_fh, fpos_fh );

}













int get_sorted_pwm_kmer_array( MF kmer_array[], float score_array[], PWM *pwm,
  PWM *bg_pwm )
{

  int buffer = 0;

  MF w;
  
  for ( w = 0; w < NUMB_OF_WORDS; w++ ) {
  
    /*  In this file, above:  */
    float score = get_kmer_pwm_score( w, pwm, bg_pwm, pwm->length );
    
    if ( score < score_array[ MAX_KMERS - 1 ] )
      continue;

    ++buffer;
      
    int i, j;
    
    for ( i = 0; i < MAX_KMERS; i++ ) {
      if ( score >= score_array[ i ] ) {
        for ( j = MAX_KMERS - 1; j > i; j-- ) {
          score_array[ j ] = score_array[ j - 1 ];
          kmer_array[ j ] = kmer_array[ j - 1 ];
        }
        kmer_array[ i ] = w;
        score_array[ i ] = score;
        break;
      }
    }
    
  }
  
  return MIN( buffer, MAX_KMERS );

}











float get_pwm_score_thresh_w_kmer_array( PWM *pwm, PWM *bg_pwm, 
  MF kmer_array[], float score_array[], int buffer, SPEC_CELL *spec_cell, 
  FILE *ft_fh, FILE *ind_fh, FILE *seq_fh, FILE *fpos_fh )
{

  int old_ct = 0;
  int new_ct;
  int total_seqs;
  int iter = 0;
  
  int index = 0;
  
  float exp_ct;
  
  do {
  
    exp_ct = get_kmer_expected_ct( kmer_array[ index ], bg_pwm, spec_cell, 
      &total_seqs, seq_fh );
  
    new_ct = get_kmer_ct_w_kmer_array_and_index( pwm, bg_pwm, kmer_array,
      index++, spec_cell, ft_fh, ind_fh, seq_fh, fpos_fh );
      
    printf( "Seq w/ occurs :  %d   (%d)  Exp: %.1f\n", new_ct, old_ct,
      exp_ct );
      
    if ( index >= buffer )
      break;
      
    if ( (float)(new_ct - old_ct) >= exp_ct * CT_RATIO ) {
      old_ct = new_ct;
      iter = 0;
      continue;
    }
    else 
      old_ct = new_ct;
  
  } while (( iter++ <= OVER_ITER ) && ( new_ct < total_seqs * SEQ_CT_RATIO ));
  
  printf( "Final index :  %d   %d\n", index, buffer );
  
  return score_array[ index - 1 ];

}












float get_kmer_expected_ct( MF w, PWM *bg_pwm, SPEC_CELL *spec_cell, 
  int *total_seqs, FILE *seq_fh )
{

  void set_position( FPOS fpos, FILE *ff );
  int get_new_seq_region_fh( SEQ_INFO *seq_info, int length[], int numb_of_spec,
    FILE *seq_fh );
  void set_use_array_from_length_array( int use_array[], int length[], int n );
    
  extern int score_thresh, reg_start, reg_end;

  set_position( 0, seq_fh );
  
  int length[ NUMB_OF_SPEC ];
  long seq_ct = 0;
  
  SEQ_INFO seq_info;
  
  *total_seqs = 0;
  
  int seq_numb = 0;
  
  while ( get_new_seq_region_fh( &seq_info, length, spec_cell->n, seq_fh )) {

    int use_array[ NUMB_OF_SPEC ];
    
    if ( seq_info.score < score_thresh ) 
      continue;
      
    /*  In ../model/data.c:  */
    set_use_array_from_length_array( use_array, length, spec_cell->n );
  
    /*  In ../model/data.c:  */
    if ( array_sum( use_array, spec_cell->n ) > 2 ) {
      ++seq_ct;
      ++(*total_seqs);
    }
      
    printf( "seq = %6d   %ld\n", ++seq_numb, seq_ct );
    
  }
  
  double freq = 1.0;
    
  MF stack = w;  
    
  int i;
  for ( i = bg_pwm->length - 1; i >= 0; i-- ) {
    freq *= bg_pwm->freq[ i ][ stack % NUMB_OF_NUCL ];
    stack = stack >> BITS_PER_NUCL;
  }
  
  double seq_freq = 1.0 - powl( 1.0 - freq, reg_end - reg_start );
  
  return (float)(seq_freq * seq_ct);

}







int get_kmer_ct_w_kmer_array_and_index( PWM *pwm, PWM *bg_pwm, MF kmer_array[],
  int index, SPEC_CELL *spec_cell, FILE *ft_fh, FILE *ind_fh, FILE *seq_fh, 
  FILE *fpos_fh )
{

  char ct_array[ NUMB_OF_SEQ_BUFFER ];
  
  int i;
  for ( i = 0; i < NUMB_OF_SEQ_BUFFER; i++ )
    ct_array[ i ] = 0;
    
  return get_my_kmer_ct_w_kmer_array_and_index( pwm, bg_pwm, ct_array, kmer_array,
    index, spec_cell, ft_fh, ind_fh, seq_fh, fpos_fh );
    
}










int get_my_kmer_ct_w_kmer_array_and_index( PWM *pwm, PWM *bg_pwm, char ct_array[],
  MF kmer_array[], int index, SPEC_CELL *spec_cell, FILE *ft_fh, FILE *ind_fh, 
  FILE *seq_fh, FILE *fpos_fh )
{

  void get_rev_complement_seq( char rev_compl[], char seq[] );
  void get_string_from_stack( char cons[], MF stack, char length );
  MF get_stack_from_string( char cons[] );
    
  PWM rc_pwm = { .length = pwm->length }, rc_bg = rc_pwm;
  
  get_rev_compl_pwm( &rc_pwm, pwm );
  get_rev_compl_pwm( &rc_bg, bg_pwm );
  
  char rc_cons[ BUFFER ];
  
  MOTIF mf = { .length = pwm->length };
  
  printf( "kmer = %ld\n", (long)kmer_array[ index ] );
  
  int i;
  
  for ( i = 0; i <= index; i++ ) {
  
    /*  In ../main/nucl.c:  */
    get_string_from_stack( mf.cons, kmer_array[ i ], pwm->length );
    
    /*  In this file, below:  */
    update_seq_ct_array_single_kmer( ct_array, kmer_array[ i ], &mf, spec_cell,
      ft_fh, ind_fh, seq_fh, fpos_fh );
    
    /*  In ../main/nucl.c:  */    
    get_rev_complement_seq( rc_cons, mf.cons );
    MF rc_w = get_stack_from_string( rc_cons );    
    
    /*  In this file, below:  */
    update_seq_ct_array_single_kmer( ct_array, rc_w, &mf, spec_cell, ft_fh, ind_fh,
      seq_fh, fpos_fh );
        
  }
  
  return array_char_sum( ct_array, NUMB_OF_SEQ_BUFFER );
  
}












int array_char_sum( char ct_array[], int n )
{

  int total = 0;
  
  int i;
  
  for ( i = 0; i < n; i++ )
    total += ct_array[ i ];
    
  return total;

}










void update_seq_ct_array_single_kmer( char ct_array[], MF w, MOTIF *mf, 
  SPEC_CELL *spec_cell, FILE *ft_fh, FILE *ind_fh, FILE *seq_fh, FILE *fpos_fh )
{

  void initialize_first_ft_array( FT ft_array[], int n );
  void set_position( FPOS fpos, FILE *ff );
  
  FT ft_array[ NUMB_OF_SPEC ];
  FT_IND ind_array[ NUMB_OF_SPEC ];  
  FPOS fpos;
  
  /*  In ../model/data.c:  */
  initialize_first_ft_array( ft_array, NUMB_OF_SPEC );
  
  set_position( sizeof(FPOS) * w, fpos_fh );  
  get( &fpos, sizeof(FPOS), fpos_fh );
  
  if ( fpos < 0 ) {
    return;
    error( 3500, "Unexpected kmer: Occurrences not recorded.\n" );
  }
  
  set_position( fpos, ind_fh );
  get( ind_array, sizeof(FT_IND) * spec_cell->n, ind_fh );
  
  /*  In this file, below:  */
  update_my_seq_ct_array_single_kmer( ct_array, w, mf, ft_array, ind_array, 
    spec_cell, ft_fh, ind_fh, seq_fh, fpos_fh ); 

}












void update_my_seq_ct_array_single_kmer( char ct_array[], MF w, MOTIF *mf, 
  FT ft_array[], FT_IND ind_array[], SPEC_CELL *spec_cell, FILE *ft_fh, 
  FILE *ind_fh, FILE *seq_fh, FILE *fpos_fh )
{

  int update_ft_array_from_binary_fh( FT ft_array[], FT_IND ind_array[], int i,
    FILE *ft_fh );
  int get_new_seq_region_fh( SEQ_INFO *seq_info, int length[], int numb_of_spec,
    FILE *seq_fh );
  void set_position( FPOS fpos, FILE *ff );
  
  SEQ_INFO seq_info;
  int length[ NUMB_OF_SPEC ];  
    
  int i;
  for ( i = 0; i < spec_cell->n; i++ ) 
    /*  In ../model/data.c:  */
    update_ft_array_from_binary_fh( ft_array, ind_array, i, ft_fh );
    
  set_position( 0, seq_fh );

  int seq_numb = 0;
    
  /*  In ../model/data.c:  */
  while ( get_new_seq_region_fh( &seq_info, length, spec_cell->n, seq_fh )) 
    /*  In this file, below:  */
    update_seq_ct_array_single_kmer_single_seq( ct_array, w, mf, ft_array, ind_array, 
      spec_cell, &seq_info, length, ++seq_numb, ft_fh );

}










void update_seq_ct_array_single_kmer_single_seq( char ct_array[], MF w, MOTIF *mf, 
  FT ft_array[], FT_IND ind_array[], SPEC_CELL *spec_cell, SEQ_INFO *seq_info, 
  int length[], SEQ seq_numb, FILE *ft_fh )
{

  void set_use_array_from_length_array( int use_array[], int length[], int n );
  int array_sum( int array[], int n );

  extern int score_thresh;

  int use_array[ NUMB_OF_SPEC ];
  
  if ( seq_info->score < score_thresh ) 
    return;
    
  /*  In ../model/data.c:  */
  set_use_array_from_length_array( use_array, length, spec_cell->n );
  
  /*  In ../model/data.c:  */
  if ( array_sum( use_array, spec_cell->n ) <= 2 ) 
    return;
    
  /*  In this file, below:  */
  update_my_seq_ct_array_single_kmer_single_seq( ct_array, w, mf, ft_array, 
    ind_array, use_array, seq_info, seq_numb, ft_fh );

}










void update_my_seq_ct_array_single_kmer_single_seq( char ct_array[], MF w, 
  MOTIF *mf, FT ft_array[], FT_IND ind_array[], int use_array[], SEQ_INFO *seq_info, 
  SEQ seq_numb, FILE *ft_fh )
{

  int update_ft_array_from_binary_fh( FT ft_array[], FT_IND ind_array[], int i,
    FILE *ft_fh );
    
  extern int reg_start, reg_end;
    
  int spec_numb = 0;
  
  if ( ! use_array[ spec_numb ] )
    return;
    
  FT *ft = ft_array + spec_numb;
  
  int last = -1;
  
  while ( ft->seq <= seq_numb ) {
  
    if ( ft->seq >= NUMB_OF_SEQ_BUFFER )
      break;
  
    if ( ft->seq == seq_numb ) {
      int center = seq_info->n / 2;
      int x = ft->pos[ 0 ];
      if (( center + reg_start <= x ) && ( center + reg_end > x )) {
        if ( last < x ) {
          /*  In this file, below:  */
          ct_array[ ft->seq ] = 1;
          last = ft->pos[ mf->length - 1 ];
        }
      }
    }
    else {
      last = -1;
    }
    
    /*  In ../model/data.c:  */
    if ( ! update_ft_array_from_binary_fh( ft_array, ind_array, spec_numb, ft_fh ))
      break;
      
    ft = ft_array + spec_numb;
  
  }

}



































