#include "../ext/struct.h"








#define MAX_KMERS 250
#define MY_MAX_KMERS (MAX_KMERS * 2)
#define KMER_LEN 10






void do_pwm( void );

void do_my_pwm( MOTIF *mf );

void do_pwm_w_fh( MOTIF *mf, FILE *out, FILE *pwm_fh );

void get_pwm_from_motif( PWM *pwm, MOTIF *mf, FILE *out, FILE *pwm_fh );

void get_pwm_from_motif_w_fhs( PWM *pwm, MOTIF *mf, FILE *ft_fh, FILE *ind_fh,
  FILE *seq_fh, FILE *fpos_fh, FILE *out, FILE *pwm_fh );

void get_my_pwm_from_motif( PWM *pwm, MOTIF *mf, SPEC_CELL *spec_cell, FILE *ft_fh,
  FILE *ind_fh, FILE *seq_fh, FILE *fpos_fh, FILE *out, FILE *pwm_fh );

void initialize_pwm( PWM *pwm, char length );

void set_pwm_freq( PWM *pwm );

float get_min_init_kmer_pwm_score( PWM *pwm, PWM *bg_pwm, MOTIF *mf );

float get_kmer_pwm_score( MF stack, PWM *pwm, PWM *bg_pwm, char length );

int pwms_differ( PWM *pwm, PWM *old_pwm );

void get_rev_compl_pwm( PWM *rc_pwm, PWM *pwm );

void print_pwm_file_fh( PWM *pwm, PWM *bg_pwm, float score_thresh, 
  FILE *pwm_fh );

void print_pwm( PWM *pwm );

void print_pwm_fh( PWM *pwm, FILE *fh );

void print_pwm_kmer_list( PWM *pwm, PWM *bg_pwm, MOTIF *mf, float score, FILE *out );

int cons_in_cons_array( char cons[], char cons_array[ MY_MAX_KMERS ][ KMER_LEN ],
  int buffer );

int cons_strings_overlap( char a[], char b[] );

void set_initial_pwm_w_motif( PWM *pwm, MOTIF *mf, SPEC_CELL *spec_cell, FILE *ft_fh, 
  FILE *ind_fh, FILE *seq_fh, FILE *fpos_fh );



/*================================================================================*/
/*================================================================================*/
/*================================================================================*/



float pwm_iteration( PWM *pwm, PWM *old_pwm, PWM *bg_pwm, MOTIF *mf, 
  SPEC_CELL *spec_cell, float score, FILE *ft_fh, FILE *ind_fh, FILE *seq_fh, 
  FILE *fpos_fh );



/*================================================================================*/
/*================================================================================*/
/*================================================================================*/



void update_pwm_single_kmer( PWM *pwm, MF w, MOTIF *mf, SPEC_CELL *spec_cell, 
  char strand, FILE *ft_fh, FILE *ind_fh, FILE *seq_fh, FILE *fpos_fh );

void update_my_pwm_single_kmer( PWM *pwm, MF w, MOTIF *mf, FT ft_array[],
  FT_IND ind_array[], SPEC_CELL *spec_cell, char strand, FILE *ft_fh, FILE *ind_fh, 
  FILE *seq_fh, FILE *fpos_fh );

void update_pwm_single_kmer_single_seq( PWM *pwm, MF w, MOTIF *mf, FT ft_array[],
  FT_IND ind_array[], SPEC_CELL *spec_cell, SEQ_INFO *seq_info, int length[], 
  char strand, SEQ seq_numb, FILE *ft_fh );

void update_my_pwm_single_kmer_single_seq( PWM *pwm, MF w, MOTIF *mf, 
  FT ft_array[], FT_IND ind_array[], int use_array[], SEQ_INFO *seq_info, 
  char strand, SEQ seq_numb, FILE *ft_fh );

void update_pwm_single_ft_struct( PWM *pwm, FT *ft, char length, char strand );

void update_pwm_single_string( PWM *pwm, char string[], char length );



/*================================================================================*/
/*================================================================================*/
/*================================================================================*/



void set_bg_pwm( PWM *pwm, MOTIF *mf, SPEC_CELL *spec_cell, FILE *ft_fh, 
  FILE *ind_fh, FILE *seq_fh, FILE *fpos_fh );

void set_bg_pwm_w_fh( PWM *pwm, MOTIF *mf, SPEC_CELL *spec_cell, FILE *in,
  FILE *reg_fh );

void get_bg_pwm_single_maf_line( PWM *pwm, MOTIF *mf, char line[] );

void get_bg_pwm_single_seq( PWM *pwm, MOTIF *mf, char seq[] );



/*================================================================================*/
/*================================================================================*/
/*================================================================================*/



float get_pwm_score_thresh( PWM *pwm, PWM *bg_pwm, float init_score, 
  SPEC_CELL *spec_cell, FILE *ft_fh, FILE *ind_fh, FILE *seq_fh, FILE *fpos_fh );

float get_my_pwm_score_thresh( PWM *pwm, PWM *bg_pwm, float score_array[],
  SPEC_CELL *spec_cell, MF kmer_array[], FILE *ft_fh, FILE *ind_fh, FILE *seq_fh, 
  FILE *fpos_fh );

int get_sorted_pwm_kmer_array( MF kmer_array[], float score_array[], PWM *pwm,
  PWM *bg_pwm );

float get_pwm_score_thresh_w_kmer_array( PWM *pwm, PWM *bg_pwm, 
  MF kmer_array[], float score_array[], int buffer, SPEC_CELL *spec_cell, 
  FILE *ft_fh, FILE *ind_fh, FILE *seq_fh, FILE *fpos_fh );

float get_kmer_expected_ct( MF w, PWM *bg_pwm, SPEC_CELL *spec_cell, 
  int *total_seqs, FILE *seq_fh );

int get_kmer_ct_w_kmer_array_and_index( PWM *pwm, PWM *bg_pwm, MF kmer_array[],
  int index, SPEC_CELL *spec_cell, FILE *ft_fh, FILE *ind_fh, FILE *seq_fh, 
  FILE *fpos_fh );

int get_my_kmer_ct_w_kmer_array_and_index( PWM *pwm, PWM *bg_pwm, char ct_array[],
  MF kmer_array[], int index, SPEC_CELL *spec_cell, FILE *ft_fh, FILE *ind_fh, 
  FILE *seq_fh, FILE *fpos_fh );

int array_char_sum( char ct_array[], int n );

void update_seq_ct_array_single_kmer( char ct_array[], MF w, MOTIF *mf, 
  SPEC_CELL *spec_cell, FILE *ft_fh, FILE *ind_fh, FILE *seq_fh, FILE *fpos_fh );

void update_my_seq_ct_array_single_kmer( char ct_array[], MF w, MOTIF *mf, 
  FT ft_array[], FT_IND ind_array[], SPEC_CELL *spec_cell, FILE *ft_fh, 
  FILE *ind_fh, FILE *seq_fh, FILE *fpos_fh );

void update_seq_ct_array_single_kmer_single_seq( char ct_array[], MF w, MOTIF *mf, 
  FT ft_array[], FT_IND ind_array[], SPEC_CELL *spec_cell, SEQ_INFO *seq_info, 
  int length[], SEQ seq_numb, FILE *ft_fh );

void update_my_seq_ct_array_single_kmer_single_seq( char ct_array[], MF w, 
  MOTIF *mf, FT ft_array[], FT_IND ind_array[], int use_array[], SEQ_INFO *seq_info, 
  SEQ seq_numb, FILE *ft_fh );

















