#include "pwm_func.h"












#define DEFAULT_PWM_LENGTH 7
#define DEFAULT_SCORE 3.86








void try_pwm( FILE *pwm_fh, FILE *kmer_fh )
{

  PWM pwm = { .length = DEFAULT_PWM_LENGTH };
  PWM bg_pwm = { .length = DEFAULT_PWM_LENGTH };
  
  pwm.pwm[ 0 ][ A_NUCL ] = 63585;
  pwm.pwm[ 0 ][ C_NUCL ] = 0;
  pwm.pwm[ 0 ][ G_NUCL ] = 0;
  pwm.pwm[ 0 ][ T_NUCL ] = 16988;
  
  pwm.pwm[ 1 ][ A_NUCL ] = 0;
  pwm.pwm[ 1 ][ C_NUCL ] = 0;
  pwm.pwm[ 1 ][ G_NUCL ] = 80573;
  pwm.pwm[ 1 ][ T_NUCL ] = 0;
  
  pwm.pwm[ 2 ][ A_NUCL ] = 80573;
  pwm.pwm[ 2 ][ C_NUCL ] = 0;
  pwm.pwm[ 2 ][ G_NUCL ] = 0;
  pwm.pwm[ 2 ][ T_NUCL ] = 0;

  pwm.pwm[ 3 ][ A_NUCL ] = 15302;
  pwm.pwm[ 3 ][ C_NUCL ] = 7347;
  pwm.pwm[ 3 ][ G_NUCL ] = 12568;
  pwm.pwm[ 3 ][ T_NUCL ] = 45356;

  pwm.pwm[ 4 ][ A_NUCL ] = 76996;
  pwm.pwm[ 4 ][ C_NUCL ] = 0;
  pwm.pwm[ 4 ][ G_NUCL ] = 1765;
  pwm.pwm[ 4 ][ T_NUCL ] = 1812;
  
  pwm.pwm[ 5 ][ A_NUCL ] = 65169;
  pwm.pwm[ 5 ][ C_NUCL ] = 0;
  pwm.pwm[ 5 ][ G_NUCL ] = 13736;
  pwm.pwm[ 5 ][ T_NUCL ] = 1668;

  pwm.pwm[ 6 ][ A_NUCL ] = 19328;
  pwm.pwm[ 6 ][ C_NUCL ] = 11497;
  pwm.pwm[ 6 ][ G_NUCL ] = 39005;
  pwm.pwm[ 6 ][ T_NUCL ] = 10743;
  
  bg_pwm.pwm[ 0 ][ A_NUCL ] = 7362492;
  bg_pwm.pwm[ 0 ][ C_NUCL ] = 6368611;
  bg_pwm.pwm[ 0 ][ G_NUCL ] = 6375813;
  bg_pwm.pwm[ 0 ][ T_NUCL ] = 7372436;
  
  int i, j;
  for ( i = 1; i < bg_pwm.length; i++ ) {
    for ( j = 0; j < NUMB_OF_NUCL; j++ ) 
      bg_pwm.pwm[ i ][ j ] = bg_pwm[ 0 ][ j ];
  }
  
  get_and_print_pwm( &pwm, &bg_pwm, DEFAULT_SCORE, pwm_fh, kmer_fh );

}







void get_and_print_pwm( PWM *pwm, PWM *bg_pwm, float score_thresh, FILE *pwm_fh, 
  FILE *kmer_fh )
{

  set_pwm_freq( pwm );
  
  print_pwm_fh( pwm, pwm_fh );
  print_pwm_kmer_list( pwm, bg_pwm, score_thresh, kmer_fh );

}










void set_pwm_freq( PWM *pwm )
{

  int i, j;
  
  for ( i = 0; i < pwm->length; i++ ) {
  
    long total = 0;
    
    for ( j = 0; j < NUMB_OF_NUCL; j++ )
      total += pwm->pwm[ i ][ j ];
      
    for ( j = 0; j < NUMB_OF_NUCL; j++ )
      pwm->freq[ i ][ j ] = (float)pwm->pwm[ i ][ j ] / (float)total;
  
  }

}









void print_pwm_fh( PWM *pwm, FILE *fh )
{

  print_sep( fh );
  
  int i, j;
  
  for ( j = 0; j < NUMB_OF_NUCL; j++ ) {
    for ( i = 0; i < pwm->length; i++ ) 
      fprintf( fh, "%10ld   ", pwm->pwm[ i ][ j ] );
    fprintf( fh, "\n" );
  }
  
  fprintf( fh, "\n\n" );
  
  for ( j = 0; j < NUMB_OF_NUCL; j++ ) {
  
    for ( i = 0; i < pwm->length; i++ )
      fprintf( fh, "%10.4f   ", pwm->freq[ i ][ j ] );
      
    fprintf( fh, "\n" );
  
  }
  
  print_sep( fh );

}














void print_pwm_kmer_list( PWM *pwm, PWM *bg_pwm, float score_thresh, FILE *fh )
{

  int get_motif_stack_mismatches( MOTIF *mf, MF stack );
  void get_rev_complement_seq( char rev_compl[], char seq[] );
  void get_string_from_stack( char cons[], MF stack, char length );
  int get_motif_stack_mismatches( MOTIF *mf, MF stack );
  void get_rev_complement_seq( char rev_compl[], char seq[] );
  
  float min_score = 10000.0;
  
  PWM rc_pwm = { .length = pwm->length }, rc_bg = rc_pwm;
  
  get_rev_compl_pwm( &rc_pwm, pwm );
  get_rev_compl_pwm( &rc_bg, bg_pwm );
    
  char cons[ BUFFER ];
    
  MF w;
  
  for ( w = 0; w < NUMB_OF_WORDS; w++ ) {

    /*  In this file, below:  */
    float curr_score = get_kmer_pwm_score( w, pwm, bg_pwm, pwm->length );
    float rc_score = get_kmer_pwm_score( w, &rc_pwm, &rc_bg, pwm->length );
    
    if ( curr_score >= score_thresh ) {
      /*  In ../main/nucl.c:  */
      get_string_from_stack( cons, w, pwm->length );
      fprintf( out, "%s\n", cons );
    }
    else if ( rc_score >= score_thresh ) {
      /*  In ../main/nucl.c:  */
      get_string_from_stack( cons, w, pwm->length );      
      fprintf( out, "%s\n", cons );
    }

  }

}












void get_rev_compl_pwm( PWM *rc_pwm, PWM *pwm )
{

  int i, j;
  
  for ( i = 0; i < pwm->length; i++ ) {
    for ( j = 0; j < NUMB_OF_NUCL; j++ ) {
      
      rc_pwm->pwm[ pwm->length - i - 1 ][ NUMB_OF_NUCL - j - 1 ] = 
        pwm->pwm[ i ][ j ];
      
      rc_pwm->freq[ pwm->length - i - 1 ][ NUMB_OF_NUCL - j - 1 ] = 
        pwm->freq[ i ][ j ];
    
    }
  }

}













float get_kmer_pwm_score( MF stack, PWM *pwm, PWM *bg_pwm, char length )
{

  MF kmer = stack;
  
  float score = 0.0;

  int i, j;
  
  for ( i = length - 1; i >= 0; i-- ) {
    int bp = kmer % NUMB_OF_NUCL;
    score += log( pwm->freq[ i ][ bp ] ) - log( bg_pwm->freq[ i ][ bp ] );
    kmer = kmer >> BITS_PER_NUCL;
  }
  
  return ( isinf( score ) || isnan( score )) ? (-10.0) : (score);

}






























