#include "print_tree.h"




























void print_tree( NODE_LIST *root )
{
  print_tree_fh( root, stdout );
}
















void print_tree_fh( NODE_LIST *node, FILE *fh )
{

  if ( node == NULL )
    return;

  print_node( node, fh );
  
  if ( node->my_node->leaf )
    return;
  
  int i;
  
  for ( i = 0; i < BRANCH_BUFFER; i++ )
    print_tree_fh( node->down[ i ], fh );

}














void print_node( NODE_LIST *node, FILE *fh )
{

  if ( node->my_node->leaf )
    print_leaf( node, fh );
    
  else
    print_non_leaf( node, fh );
    
  print_sep( fh );
    
}










void print_leaf( NODE_LIST *node, FILE *fh )
{
  fprintf( fh, "%-16s   (%s)\n", node->my_node->label, node->my_node->spec );
  
  if ( node->my_node->t >= 1.0 )
    fprintf( fh, "\t\t%7.1f branch length\n", node->my_node->t );
  else
    fprintf( fh, "\t\t%7.5f branch length\n", node->my_node->t );
}











void print_non_leaf( NODE_LIST *node, FILE *fh )
{

  STAT get_time_to_node( NODE_LIST *node );
  
  /*  In get.c:  */
  STAT t = get_time_to_node( node );

  if ( t >= 1.0 )
    fprintf( fh, "%-16s\t\t   %7.1f branch length\n", node->my_node->label, 
      (float)t );
  else
    fprintf( fh, "%-16s\t\t   %7.5f branch length\n", node->my_node->label,
      (float)t );
    
  fprintf( fh, "\n" );
  
  if ( node->up != NULL ) {
    fprintf( fh, "\tUpstream node :\n" );
    if ( t >= 1.0 )
      fprintf( fh, "\t\t%-16s   %7.1f branch length\n",
        node->up->my_node->label, node->up->my_node->t );
    else
      fprintf( fh, "\t\t%-16s   %7.5f branch length\n",
        node->up->my_node->label, node->up->my_node->t );
  }

  fprintf( fh, "\tDownstream nodes :\n" );
  
  int i;
  for ( i = 0; i < BRANCH_BUFFER; i++ ) {
    if ( node->down[ i ] != NULL ) {
      if ( t >= 1.0 )
        fprintf( fh, "\t\t%-16s   %7.1f branch length\n",
          node->down[ i ]->my_node->label, node->down[ i ]->my_node->t ); 
      else
        fprintf( fh, "\t\t%-16s   %7.5f branch length\n",
          node->down[ i ]->my_node->label, node->down[ i ]->my_node->t );
    }
  }
  
  fprintf( fh, "\tLeaves :\n" );
  
  print_downstream_leaves( node, fh );
  
}















void print_downstream_leaves( NODE_LIST *node, FILE *fh )
{

  if ( node == NULL )
    return;

  if ( node->my_node->leaf ) {
    fprintf( fh, "\t\t%-16s   (%s)\n", node->my_node->label, node->my_node->spec );
    return;
  }
  
  int i;
  
  for ( i = 0; i < BRANCH_BUFFER; i++ )
    print_downstream_leaves( node->down[ i ], fh );

}













